/**********************************************************************************************************************************/
/* Project      : TAPER                                                                                                           */
/* Module       : TPAUDIO.C                                                                                                       */
/* Description  : Audio functions (read/write VOC/direct samples)                                                                 */
/* Version type : Program module                                                                                                  */
/* Last changed : 19-07-1999  5:50                                                                                                */
/* Update count : 69                                                                                                              */
/* OS type      : PC (DOS)                                                                                                        */
/* Compiler     : Watcom C32 Optimizing Compiler Version 10.x                                                                     */
/* Linker       : Watcom C Linker                                                                                                 */
/* Notes        : Uses gotos __SKIPNOISE, __PILOTAGAIN and __NEXTSUBBLOCK.                                                        */
/*                The initial decoder was co-written by Tomaz Kac (and was tested in VOC2TZX).                                    */
/*                                                                                                                                */
/*                Copyleft (C) 1997-2001 ThunderWare Research Center, written by Martijn van der Heide.                           */
/*                                                                                                                                */
/*                This program is free software; you can redistribute it and/or                                                   */
/*                modify it under the terms of the GNU General Public License                                                     */
/*                as published by the Free Software Foundation; either version 2                                                  */
/*                of the License, or (at your option) any later version.                                                          */
/*                                                                                                                                */
/*                This program is distributed in the hope that it will be useful,                                                 */
/*                but WITHOUT ANY WARRANTY; without even the implied warranty of                                                  */
/*                MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                                                   */
/*                GNU General Public License for more details.                                                                    */
/*                                                                                                                                */
/*                You should have received a copy of the GNU General Public License                                               */
/*                along with this program; if not, write to the Free Software                                                     */
/*                Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                                     */
/*                                                                                                                                */
/**********************************************************************************************************************************/

#define __TPAUDIO_MAIN__

#include "wbench32/wbench32.h"
#include <math.h>
#include <dos.h>
#include <fcntl.h>
#include <io.h>
#include "tpfile.h"
#include "tpscreen.h"
#include "tpmenu.h"
#include "tpmain.h"
#include "tptape.h"
#include "tpfilter.h"
#include "tpserpar.h"
#include "tpaudio.h"

#pragma pack                ()

#define _Average(_Pulse1,_Pulse2)  ((long)(((double)(_Pulse1 + _Pulse2) / 2) + 0.5))                       /* Average of 2 pulses */
#define _Min(_Pulse1,_Pulse2)      (_Pulse1 < _Pulse2 ? _Pulse1 : _Pulse2)                                /* Shortest of 2 pulses */
#define _Max(_Pulse1,_Pulse2)      (_Pulse1 > _Pulse2 ? _Pulse1 : _Pulse2)                                 /* Longest of 2 pulses */
#define _NumSamples(_Pulse,_TRate) ((long)(((double)_Pulse / _TRate) + 0.5))                           /* T-states to num samples */

/**********************************************************************************************************************************/
/* Define the static variables: Spectrum                                                                                          */
/**********************************************************************************************************************************/

static char *_SchemeNames[DS_MAX] = { "(none)",
                                      "Force ROM Timing",
                                      "Firebird BleepLoad",
                                      "Alkatraz",
                                      "SpeedLock 1",
                                      "SpeedLock 2",
                                      "SpeedLock 3",
                                      "SpeedLock 4",
                                      "SpeedLock 5",
                                      "SpeedLock 6",
                                      "SpeedLock 7",
                                      "Digital Integration",
                                      "Dinamic Loader",
                                      "Sinclair User",
                                      "Your Sinclair",
                                      "Players",
                                      "Paul Owens Protection System" };

long   MinPilotPulseLength = 650;                                     /* These are needed to be able to skip noise between blocks */
long   MaxPilotPulseLength = 3200;
long   MinNumPilotPulses = 50;                                                /* At least this number to be recognized as a pilot */
long   UsedMinNumPilotPulses;                                                      /* Possibly extended for some decoding schemes */
long   ThresholdPilotPulse = 30;                                                     /* May be this much off (%) to be acceptable */
long   NumSyncPulses = 2;                                                                                                /* [0-3] */
long   FMinBitPulseLength = 100;                                                             /* Minimal T-states in any bit pulse */
long   ThresholdBitPulse = 25;                                                /* Makes the difference between a '0' and a '1' (%) */
double ThresholdEndMarker = 1.4;                                                   /* End reached if this times LengthOneBitPulse */
double ThresholdSingleEndMarker = 1.6;
long   EndBitPulseLength = 2400;                         /* If a pulse has this many T-states, it's the pause right after a block */
bool   TestDoubleBitPulse = FALSE;                                                  /* If TRUE, check ThresholdDoubleBitPulse too */
long   ThresholdDoubleBitPulse = 35;                                   /* 2 consequetive pulses for a bit should be identical (%) */
long   ThresholdStandardLoading = 15;        /* If the timing values are this close (%) to the NS_xxx values, it's Normal Loading */
bool   DropPartialEndByte = FALSE;                                      /* TRUE if any bits after the last byte are to be dropped */
bool   ForceStandardSyncs = FALSE;                                                                /* Force standard sync values ? */
short  DecodingScheme = 0xFF;                                                                       /* No special decoding scheme */
                                                                       /* 0xFF marks TAPER startup - set to TPConfig.DefDecodeROM */
bool   LossyDetection = FALSE;            /* If FALSE, use (fixed) averaging. If TRUE, see what's most likely for each pulse read */
bool   ApproximateBitTimers = TRUE;                                                                  /* Use '1' = 2 * '0' if TRUE */

bool   ApplyFilter = 0xFF;                                                    /* TRUE if filtering is needed for all read samples */
                                                                       /* 0xFF marks TAPER startup - set to TPConfig.DefUseFilter */
bool   ShowSoundStreamInfo = FALSE;                                                     /* Show graph after decoding each block ? */

bool   HandlingZ80EmRAW = FALSE;
bool   HandlingCSW = FALSE;

dword  WriteSampleRate = NS_SAMPRATE;                                                     /* The sample rate to write a VOC block */

static bool        _SkipVOCSoundContinue;                                             /* Step over these VOC block ID's when TRUE */

static struct      _ReadBuffer_s                                                                          /* The file read buffer */
{                                                                         /* Variables order is optimized rather than logical : ) */
  byte            *_Buffer;
  dword            _DBufferEnd;                                                          /* BufferEnd as a dword (for XMS access) */
  dword            _FileLength;                                                                                /* Length in bytes */
  dword            _FileCount;                                                       /* Length counter in bits (for DRB decoding) */
  dword            _FileEnd;                                                                          /* Maps end-of-block to EOF */
  dword            _FileIndex;                                                                     /* Index to next block to load */
  dword            _InitialFileIndex;                                                                        /* (Used to restore) */
  dword            _XMSIndex;                                                                   /* (Only needed for DRB decoding) */
  word             _BufferLength;                                                                        /* Length of file buffer */
  word             _BufferEnd;                                                                  /* Used length of the file buffer */
  word             _BufferIndex;                                                                                    /* Read index */
  word             _BaseLine;                                                                         /* Signal baseline (filter) */
  word             _InitialBaseLine;                                                                         /* (Used to restore) */
  short            _TapeWindow;                                                                 /* (Only needed for DRB decoding) */
  int              _Fh;                                                /* Either the file handle, or < 0 to indicate DRB decoding */
  byte             _SkipLength;                                            /* To step over VOC soundcontinue blocks in the filter */
  byte             _InitialSkipLength;                                                                       /* (Used to restore) */
  byte             _Bit;                                                                        /* (Only needed for DRB decoding) */
  bool             _LastBlock;                                               /* TRUE when the last block was read into the buffer */
  bool             _FirstByte;                                                            /* TRUE only just after _OpenReadBuffer */
  double           _TRate;                                                            /* Translation value SampleRate -> T-states */
}                  _ReadBuffer = { NULL, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0x8000, 0x8000, 0, 0, 0, 0, 0, FALSE, FALSE, 0.0 };

static struct      _WriteBuffer_s                                                                        /* The file write buffer */
{
  int              _Fh;
  byte            *_Buffer;
  byte            *_SBBuffer[2];
  word             _BufferLength;
  word             _BufferIndex;
  dword            _InitialFileIndex;
  dword            _FileIndex;
  double           _TRate;                                                            /* Translation value SampleRate -> T-states */
  byte             _ID01;                                                                    /* TRUE for VOC ID 1, FALSE for ID 2 */
  byte             _State;
  bool             _Dirty;
}                  _WriteBuffer = { 0, NULL, NULL, NULL, 0, 0, 0, 0, 0.0, 0, 0xF0, FALSE };

static struct _ReadBuffer_s _DataBackupReadBuffer;
static struct _ReadBuffer_s _EndBackupReadBuffer;

static char                  _SBWHeader[30];
static struct WindowInfo     _SBWindow = {30, 130, 220, 54, 220, 54, 220, 54, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                          WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER | WINF_NOMOVE, NULL,
                                          _SBWHeader};
static bool                  _KeepSBWindow = FALSE;

static short                 _SelectBlockX = 215;
static short                 _SelectBlockY = 125;

static bool _DEBUGSBDMA = FALSE;
static byte                 *_AudioBuffer = NULL;
static word                  _AudioBufferSelector;
static bool                  _AudioBufferAllocated = FALSE;

static struct _VOCID9Header_s
{
  byte  _ID;
  byte  _Length[3];
  dword _SampleRate;
  dword _Info;
  dword _Reserved;
} _VOCID9Header = { 0x09, 0x00, 0x00, 0x00, 0x00000000, 0x00000108, 0x00000000 };

/**********************************************************************************************************************************/
/* Define the SpeedLock decoding variables                                                                                        */
/**********************************************************************************************************************************/

static long  PilotBuffer[50];                                                                            /* Store clicking pilots */
static short PilotIndex = 0;
static short SLBlockCnt = 0;
static long  MidSyncBuffer[21];                                                      /* 21 bits midsync in SpeedLocks 2/3/4/5/6/7 */
static long  MidSyncMatch[21] =
             { 875, 875, 875, 875, 875, 1750, 1750, 1750, 1750, 1750, 1750, 1750, 1750, 1750, 1750, 1750, 875, 875, 875, 875, 875 };
static bool  MidSyncFound = FALSE;
static dword TmpXMSIndex;
static long  ComposedSync[16] =
             { 667, 735, 714, 714, 1428, 1428, 2144, 2144, 2144, 2144, 2144, 2144, 1428, 1428, 714, 714 };

/**********************************************************************************************************************************/
/* Define the look-ahead bit repair stack.                                                                                        */
/* Bits can get corrupted in each of the two composed pulses. If either of the read pulses is way too short, that pulses is in    */
/* fact a distortion, and the next 2 pulses should be added up to the read pulses.                                                */
/* If the first of the 2 initially read pulses is too short, the corruption distorted the previous bit, and a look-ahead stack is */
/* set up in the analyse loop so that the decode loop knows it needs to add up the next 2 pulses as well.                         */
/* If the second of the 2 initially read pulses is too short, the corruption distorted the current bit, in which case we can      */
/* simply join the next 2 pulses and no stack is needed for this.                                                                 */
/**********************************************************************************************************************************/

#define MAXREPAIRBITS    200                                                                   /* Max number of look-ahead pulses */
static struct BitRepairStack_s
{
  byte           NumUsed;                                                                           /* Number of locations filled */
  byte           Index;                                                                                        /* Read-back index */
  dword          Locations[MAXREPAIRBITS];                                       /* Bit-locations that need the look-ahead repair */
}                BitRepairStack;

/**********************************************************************************************************************************/
/* Define the sound stream decode graph info                                                                                      */
/**********************************************************************************************************************************/

#define SA_MAXPULSELENGTH 4000                                                                         /* Max logged pulse length */
static struct SA_Pulses_s                                                                                               /* Pulses */
{
  dword          Counts0;                                                                            /* Pulse that led to a 0-bit */
  dword          Counts1;                                                                            /* Pulse that led to a 1-bit */
}                SA_Pulses[SA_MAXPULSELENGTH];
static dword     SA_Joined[SA_MAXPULSELENGTH];                                                                      /* Bit-spread */
static dword     SA_BadBits;                                                 /* Number of corrupted (and hopefully repaired) bits */

/**********************************************************************************************************************************/
/* Define the variables used for time elapse when playing through SoundBlaster                                                    */
/**********************************************************************************************************************************/

static void (_interrupt *_OldINT1CHandler)();
static bool  _PlayClockRunning = FALSE;
static dword _PlayLeftTicks    = 0;
static word  _PlayLeftMinutes  = 0;
static byte  _PlayLeftSeconds  = 0;
static bool  _CheckTimeOut     = TRUE;
static byte  _OldSB_DMAComplete = 0xFF;
static byte  _SB_DMACompleteCount;
static byte  _SB_DMACompleteTimeOut;

/**********************************************************************************************************************************/
/* Define the 'record' picture                                                                                                    */
/**********************************************************************************************************************************/

#define R_WIDTH  14
#define R_HEIGHT 14
static byte _RecordPicture[R_WIDTH * R_HEIGHT + 2] = {
  R_WIDTH, R_HEIGHT,
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x08, 0x08, 0x08, 0x08, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
  0xFF, 0xFF, 0xFF, 0x08, 0x08, 0x04, 0x04, 0x04, 0x04, 0x08, 0x08, 0xFF, 0xFF, 0xFF,
  0xFF, 0xFF, 0x08, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x0F, 0xFF, 0xFF,
  0xFF, 0x08, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x0F, 0xFF,
  0xFF, 0x08, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x0F, 0xFF,
  0x08, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x0F,
  0x08, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x0F,
  0x08, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x0F,
  0x08, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x0F,
  0xFF, 0x08, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x0F, 0xFF,
  0xFF, 0x08, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x0F, 0xFF,
  0xFF, 0xFF, 0x08, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x0F, 0xFF, 0xFF,
  0xFF, 0xFF, 0xFF, 0x0F, 0x0F, 0x04, 0x04, 0x04, 0x04, 0x0F, 0x0F, 0xFF, 0xFF, 0xFF,
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x0F, 0x0F, 0x0F, 0x0F, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF };

/**********************************************************************************************************************************/
/* The path info used to save out the intermediate sample from the Sampler.                                                       */
/**********************************************************************************************************************************/

static char _SavePath[MAXDIR + 1] = {"\0"};
static char _SaveFile[13]         = {"\0"};
static char _SaveMask[13]         = "*.VOC";

/**********************************************************************************************************************************/
/* Define the static routines                                                                                                     */
/**********************************************************************************************************************************/

static void  _interrupt _PlayClockHandler   (void);
static void  _StartPlayClock                (bool WantTime);
static void  _StopPlayClock                 (void);

static void  _interrupt _RecordClockHandler (void);
static void  _StartRecordClock              (struct WindowInfo *Window);
static void  _StopRecordClock               (void);

static void  _SpecFilter                    (byte *Buffer, word BufLength, word TRate, word *BaseLine, byte *SkipLength);

static void  _AddEntryToTape                (short WhichWindow, struct TapeEntry_s *Entry);
static void  _AddSpaceToTape                (short WhichWindow, dword AddedSpace);

static bool  _OpenReadBuffer                (short WhichWindow, int Fh, dword FileIndex, dword FileLength, dword SampleRate,
                                             dword XMSIndex);
static bool  _OpenWriteBuffer               (int Fh, dword FileIndex, bool Force);
static void  _CloseFileBuffer               (bool WhichBuffer);
static void  _BackupReadState               (struct _ReadBuffer_s *RBuffer);
static bool  _RestoreReadState              (struct _ReadBuffer_s *RBuffer);

static bool  _ReadState                     (bool *State);
static long  _ReadPulse                     (bool *LastState);
static long  _PulseDiff                     (long ReferencePulse, long NewPulse);
static dword _GetCurrentFileIndex           (void);

static bool  _WriteStateSB                  (byte State);
static bool  _WriteStateFile                (byte State);
static bool  _WritePulse                    (byte *LastState, long PulseLength);

static void  _ProcessInterBlockPause        (short WhichWindow, dword PauseLength, dword SampleRate);
static void  _InsertDigitalIntegrationPilot (short WhichWindow, struct TapeEntry_s *Entry);
static void  _InsertClickingPilot           (short WhichWindow);
static void  _InsertSpeedLock4567Pilot      (short WhichWindow, long PilotLength);
static void  _InsertSpeedLock234567MidSync  (short WhichWindow, struct TapeEntry_s *Entry);
static void  _InsertGroupEnd                (short WhichWindow, struct TapeEntry_s *Entry);
static void  _GetBitTimingValues            (word *Zero, word *One, long *Average);
static long  _GetSLSyncMatchValue           (long PulseLength);
static comboent _HandleSelectBlock          (comboent CurrentBlockNo);
static void  _PrintSoundAnalysis            (struct TapeEntry_s *Entry);

/**********************************************************************************************************************************/
/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> START OF THE INTERNAL LIBRARY ROUTINES <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< */
/**********************************************************************************************************************************/

static void _interrupt _PlayClockHandler (void)

/**********************************************************************************************************************************/
/* Pre   : The variables must be set up (This is the INT 1C handler).                                                             */
/* Post  : If a second passed, the remaining play time string is adjusted and the poll-flag `PlayTimeUpdated' is set.             */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  static byte _TimeIndex;
  static word _TimeTemp;
  static byte _NewSB_DMAComplete;
  static byte _SB_DMACompleteInterval = 0;

  if (_PlayClockRunning)
  {
    _PlayLeftTicks += (dword)65536;
    if (_PlayLeftTicks >= (dword)0x1234DC)
    {
      _PlayLeftTicks -= (dword)0x1234DC;
      if (_PlayLeftSeconds == 0)
      {
        if (_PlayLeftMinutes != 0)                                                       /* We don't want negative time displayed */
        {
          _PlayLeftSeconds = 59;
          _PlayLeftMinutes --;
        }
      }
      else
        _PlayLeftSeconds --;
      _TimeIndex = 5;
      _TimeTemp = _PlayLeftMinutes;                                                    /* Fill remaining time string, but quickly */
      if (_TimeTemp >= 1000)
      { PlayTimeString[_TimeIndex ++] = _TimeTemp / 1000 + '0'; _TimeTemp %= 1000; }
      if (_TimeTemp >= 100)
      { PlayTimeString[_TimeIndex ++] = _TimeTemp / 100 + '0'; _TimeTemp %= 100; }
      if (_TimeTemp >= 10)
      { PlayTimeString[_TimeIndex ++] = _TimeTemp / 10 + '0'; _TimeTemp %= 10; }
      PlayTimeString[_TimeIndex ++] = _TimeTemp + '0';
      PlayTimeString[_TimeIndex ++] = ':';
      _TimeTemp = _PlayLeftSeconds;
      PlayTimeString[_TimeIndex ++] = _TimeTemp / 10 + '0';
      if (_TimeTemp >= 10)
        _TimeTemp %= 10;
      PlayTimeString[_TimeIndex ++] = _TimeTemp + '0';
      PlayTimeString[_TimeIndex ++] = ' ';
      PlayTimeString[_TimeIndex ++] = '\0';
      PlayTimeUpdated = TRUE;                                                                /* Signal: time to update the window */
    }
  }
  _NewSB_DMAComplete = (*SB_DMAComplete);                                                      /* Read current DMA_Complete state */
  if (_OldSB_DMAComplete == 0xFF)                                                             /* Need to initialize the counter ? */
  {
    _OldSB_DMAComplete = _NewSB_DMAComplete;
    _SB_DMACompleteCount = 0;
    _SB_DMACompleteInterval = 0;
  }
  else if (_SB_DMACompleteCount < 4)                                                             /* (4 are enough to find errors) */
  {
    if (_OldSB_DMAComplete != _NewSB_DMAComplete)                                                         /* An IRQ has occured ? */
    {
      _OldSB_DMAComplete = _NewSB_DMAComplete;
      _SB_DMACompleteCount ++;
      _SB_DMACompleteInterval = 0;
    }
    else                                                                                                      /* (No IRQ occured) */
      if (++ _SB_DMACompleteInterval == _SB_DMACompleteTimeOut)
        SB_DMANoIRQTimeOut = TRUE;
  }
  _chain_intr (_OldINT1CHandler);
}

static void _StartPlayClock (bool WantTime)

/**********************************************************************************************************************************/
/* Pre   : We are about to start playing, and we want a remaining time indication.                                                */
/* Post  : The timer is set up and started.                                                                                       */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  dword FullSeconds;

  _PlayClockRunning = FALSE;
  SB_DMANoIRQTimeOut = FALSE;
  if (WantTime)
  {
    FullSeconds = (dword)(PlayTimeLength / 3500000.0);                                                   /* (T-states to seconds) */
    _PlayLeftMinutes = (word)(FullSeconds / 60);
    _PlayLeftSeconds = (byte)(FullSeconds % 60);
    _PlayLeftTicks = (dword)(((PlayTimeLength - 3500000.0 * FullSeconds) / 3500000.0) * (double)0x1234DC);
    PlayTimeUpdated = FALSE;
    strcpy (PlayTimeString, " ETA ");
    Wprintf (&_SBWindow, 1, TEXT_CENTER, "ETA %d:%02d", _PlayLeftMinutes, _PlayLeftSeconds);           /* Write initial time left */
  }
  _OldSB_DMAComplete = 0xFF;                                                                /* Signal: need to init this variable */
  _SB_DMACompleteTimeOut = (((dword)_WriteBuffer._BufferLength * 3) / WriteSampleRate + 1) * 50;
  SB_DMANoIRQTimeOut = FALSE;
  DisableInterrupts ();                                                                                       /* Attach the timer */
  _OldINT1CHandler = _dos_getvect (0x1C);
  _dos_setvect (0x1C, _PlayClockHandler);
  EnableInterrupts ();
  if (WantTime)
    _PlayClockRunning = TRUE;                                                                                 /* Start the timer! */
}

static void _StopPlayClock (void)

/**********************************************************************************************************************************/
/* Pre   : We just stopped playing, while we had a remaining time indication.                                                     */
/* Post  : The timer is stopped.                                                                                                  */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  _PlayClockRunning = FALSE;
  DisableInterrupts ();
  _dos_setvect (0x1C, _OldINT1CHandler);
  EnableInterrupts ();
  PlayTimeUpdated = FALSE;
}

static void _interrupt _RecordClockHandler (void)

/**********************************************************************************************************************************/
/* Pre   : The variables must be set up (This is the INT 1C handler).                                                             */
/* Post  : If a second passed, the recording time string is adjusted and the poll-flag `PlayTimeUpdated' is set.                  */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  static byte _TimeIndex;
  static word _TimeTemp;

  if (_PlayClockRunning)
  {
    _PlayLeftTicks += (dword)65536;
    if (_PlayLeftTicks >= (dword)0x1234DC)
    {
      _PlayLeftTicks -= (dword)0x1234DC;
      if (_PlayLeftSeconds == 59)
      {
        _PlayLeftSeconds = 0;
        _PlayLeftMinutes ++;
      }
      else
        _PlayLeftSeconds ++;
      _TimeIndex = 13;
      _TimeTemp = _PlayLeftMinutes;                                                    /* Fill recording time string, but quickly */
      if (_TimeTemp >= 1000)
      { PlayTimeString[_TimeIndex ++] = _TimeTemp / 1000 + '0'; _TimeTemp %= 1000; }
      if (_TimeTemp >= 100)
      { PlayTimeString[_TimeIndex ++] = _TimeTemp / 100 + '0'; _TimeTemp %= 100; }
      if (_TimeTemp >= 10)
      { PlayTimeString[_TimeIndex ++] = _TimeTemp / 10 + '0'; _TimeTemp %= 10; }
      PlayTimeString[_TimeIndex ++] = _TimeTemp + '0';
      PlayTimeString[_TimeIndex ++] = ':';
      _TimeTemp = _PlayLeftSeconds;
      PlayTimeString[_TimeIndex ++] = _TimeTemp / 10 + '0';
      if (_TimeTemp >= 10)
        _TimeTemp %= 10;
      PlayTimeString[_TimeIndex ++] = _TimeTemp + '0';
      PlayTimeString[_TimeIndex ++] = ' ';
      PlayTimeString[_TimeIndex ++] = '\0';
      PlayTimeUpdated = TRUE;                                                                /* Signal: time to update the window */
    }
  }
  _chain_intr (_OldINT1CHandler);
}

static void _StartRecordClock (struct WindowInfo *Window)

/**********************************************************************************************************************************/
/* Pre   : We are about to start recording, and we want a time indication.                                                        */
/* Post  : The timer is set up and started.                                                                                       */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  _PlayLeftMinutes = _PlayLeftSeconds = _PlayLeftTicks = 0;
  PlayTimeUpdated = FALSE;
  _PlayClockRunning = FALSE;
  strcpy (PlayTimeString, " Record time ");
  Wprintf (Window, 4, TEXT_CENTER, "Record time 0:00");                                              /* Write initial record time */
  DisableInterrupts ();                                                                                       /* Attach the timer */
  _OldINT1CHandler = _dos_getvect (0x1C);
  _dos_setvect (0x1C, _RecordClockHandler);
  EnableInterrupts ();
  _PlayClockRunning = TRUE;                                                                                   /* Start the timer! */
}

static void _StopRecordClock (void)

/**********************************************************************************************************************************/
/* Pre   : We just stopped recording, while we had a time indication.                                                             */
/* Post  : The timer is stopped.                                                                                                  */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  _PlayClockRunning = FALSE;
  DisableInterrupts ();
  _dos_setvect (0x1C, _OldINT1CHandler);
  EnableInterrupts ();
  PlayTimeUpdated = FALSE;
}

static void _SpecFilter (byte *Buffer, word BufLength, word TRate, word *BaseLine, byte *SkipLength)

/**********************************************************************************************************************************/
/* Pre   : `Buffer' points to the sound buffer, `BufLength' is the length to go, `TRate' is the sample rate in T-states,          */
/*         `BaseLine' holds the current base line value, `SkipLength' holds the number of bytes in a VOC header left to skip.     */
/* Post  : The signal has been filtered. `BaseLine' and `SkipLength' hold the end values.                                         */
/*         This filter emulates the capacitor in a Spectrum, ignoring (in this case: adjusting) shifts in the base line.          */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  word register      Alpha;
  dword              NewBase;
  byte register     *SoundStream;
  byte               NewSoundByte;
  byte               OldSoundByte;
  dword              FileIndexNow;
  bool               DoSkipHeader;
  byte               TmpSkipLength;
  word               VOCHeaderRun;

  if (TRate < 900)                                                                                      /* 900 T's, approx. 4 kHz */
    return;                                                                            /* Don't filter with such low sample rates */
  SoundStream = Buffer;
  Alpha = (TRate + 64) >> 7;                                                                               /* This is 256 * Alpha */
  if (Alpha == 0)
    Alpha = 1;
  Alpha = (256 - Alpha);
  FileIndexNow = _ReadBuffer._FileIndex;

  while (BufLength > 0)
  {
    __SKIPLOOP: while (*SkipLength > 0 && BufLength > 0)
                { SoundStream ++; (*SkipLength) --; BufLength --; FileIndexNow ++; }
    if (BufLength == 0)
      return;
    if (*SoundStream == 0x01)                                                                         /* VOC Sound Start Header ? */
    { TmpSkipLength = 6; goto __MAYBESKIP; }                                                                   /* 6 bytes to skip */
    else if (*SoundStream == 0x02)                                                                 /* VOC Sound Continue Header ? */
    { TmpSkipLength = 4; goto __MAYBESKIP; }                                                                   /* 4 bytes to skip */
    else if (*SoundStream == 0x09)                                                                  /* GoldWave Extended Header ? */
    { TmpSkipLength = 16;                                                                                     /* 16 bytes to skip */
      __MAYBESKIP: DoSkipHeader = FALSE;                  /* Only skip if it indeed is a header - it could be just a sample value */
                   for (VOCHeaderRun = 0 ; !DoSkipHeader && VOCHeaderRun < VOCHeaderLocationIndex ; VOCHeaderRun ++)
                     if (VOCHeaderLocations[VOCHeaderRun] == FileIndexNow)
                       DoSkipHeader = TRUE;
                   if (DoSkipHeader)
                   {                                                                            /* It's indeed a header; skip it! */
                     *SkipLength = TmpSkipLength;
                     goto __SKIPLOOP;
                   }
    }
    NewSoundByte = *SoundStream * Alpha;                                                   /* New value is 'Alpha * Signal / 256' */
    OldSoundByte = *SoundStream - 5;                                                                               /* Tiny offset */
    if (OldSoundByte > (byte)(*BaseLine & 0x00FF))
      *SoundStream = 0xFF;
    else
      *SoundStream = 0x00;
    NewBase = *BaseLine * NewSoundByte;                                               /* New base is '(1 - Alpha) * current base' */
    *BaseLine += (word)(NewBase >> 8);                                                                   /* Update baseline value */
    if (NewBase & 0x00000080)
      *BaseLine += 1;
    SoundStream ++;
    BufLength --;
    FileIndexNow ++;
  }
}

static void _AddEntryToTape (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the entry to be added.                                    */
/* Post  : The entry has been added to the end of the tape.                                                                       */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *DirP;

  if ((DirP = TapeWindow[WhichWindow].Directory) == NULL)                                                        /* First entry ? */
    TapeWindow[WhichWindow].Directory = Entry;
  else
  {
    DirP = TapeWindow[WhichWindow].Directory;
    while (DirP->Next != NULL)
      DirP = DirP->Next;
    DirP->Next = Entry;
  }
}

static void _AddSpaceToTape (short WhichWindow, dword AddedSpace)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `AddedSpace' holds the number of bytes (multiple of 4) to be added.         */
/* Post  : The requested space has been added to the end of the tape XMS block.                                                   */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  if (TapeWindow[WhichWindow].Tape.XMSHandle == -1)                                                              /* First entry ? */
    XMSAlloc (&(TapeWindow[WhichWindow].Tape.XMSHandle), AddedSpace);
  else
    XMSRealloc (TapeWindow[WhichWindow].Tape.XMSHandle, TapeWindow[WhichWindow].Tape.Size + AddedSpace, TRUE);
}

static bool _OpenReadBuffer (short WhichWindow, int Fh, dword FileIndex, dword FileLength, dword SampleRate, dword XMSIndex)

/**********************************************************************************************************************************/
/* Pre   : `FileIndex' holds the current position in the file `Fh', `FileLength' holds the length, which is not necessarily the   */
/*         actual end of the file (but will be treated as such), `SampleRate' is the sample rate in Hz.                           */
/*         If `Fh' is < 0, we're decoding a Direct Recording block.                                                               */
/*         In that case, `FileIndex' starts a 0 (XMS index) and `FileLength' holds the number of BITS.                            */
/* Post  : The file cache buffer has been set up and initialized as being empty. If the buffer was already allocated, only the    */
/*         pointers have been reset.                                                                                              */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  if (_ReadBuffer._Buffer == NULL)                                                             /* Not set up yet (for re-inits) ? */
  {
    _ReadBuffer._BufferLength = 16384;                                                                  /* Assume we can get 16Kb */
    if (GlobalFileCopyBuffer != NULL)                                                                        /* Globally set up ? */
      _ReadBuffer._Buffer = GlobalFileCopyBuffer;
    else if ((_ReadBuffer._Buffer = (byte *)malloc ((size_t)16384)) == NULL)                                    /* Try 16K buffer */
      if ((_ReadBuffer._Buffer = (byte *)malloc ((size_t)4096)) == NULL)                               /* Retry at 4K if it fails */
      { ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Out of memory"); return (FALSE); }
      else
        _ReadBuffer._BufferLength = 2048;
    _ReadBuffer._BaseLine   = _ReadBuffer._InitialBaseLine   = 0x8000;
    _ReadBuffer._SkipLength = _ReadBuffer._InitialSkipLength = 0;
    _ReadBuffer._Bit        = 0;
  }
  else if (Fh >= 0)
    if (lseek (Fh, FileIndex, SEEK_SET) == -1)                           /* Move the filepointer (back) to the requested position */
      return (FALSE);
  _ReadBuffer._BufferEnd     = 0;
  _ReadBuffer._BufferIndex   = 0;                                                             /* Signal: fill buffer on next read */
  _ReadBuffer._Fh            = Fh;
  _ReadBuffer._FileLength    = FileLength;
  _ReadBuffer._FileCount     = 0;
  _ReadBuffer._FileEnd       = FileIndex + FileLength;
  _ReadBuffer._FileIndex     = _ReadBuffer._InitialFileIndex = FileIndex;
  _ReadBuffer._Bit           = 0;
  _ReadBuffer._LastBlock     = FALSE;
  _ReadBuffer._FirstByte     = TRUE;                               /* Signal: first state byte was NOT read away (for _ReadPulse) */
  _ReadBuffer._TRate         = (double)3500000 / (double)SampleRate;                         /* Determine no. T-states per sample */
  _ReadBuffer._TapeWindow    = WhichWindow;
  _ReadBuffer._XMSIndex      = XMSIndex;
  return (TRUE);
}

static bool _OpenWriteBuffer (int Fh, dword FileIndex, bool Force)

/**********************************************************************************************************************************/
/* Pre   : `Fh' holds the file handle, `FileIndex' is the current fie index.                                                      */
/* Post  : The file write buffer has been allocated.                                                                              */
/*         If `Fh' < 0, output will be sent to the SoundBlaster output instead and the buffering is different!                    */
/*         NOTE: While the SoundBlaster is sending sound, NO window activity can be done that requires the GlobalVideoPlaneBuffer */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  char  BlockHeader[6] = "\x01\x00\x00\x00\x00\x00";
  dword LinearAddress;
  word  BufLen;                                                                                             /* Used buffer length */

  BlockHeader[4] = (byte)(256 - (byte)((double)1000000 / (double)WriteSampleRate));                    /* Determine Time Constant */
  _WriteBuffer._Fh               = Fh;
  if (HandlingZ80EmRAW)
    _WriteBuffer._TRate          = (double)3500000 / (double)Z80EMSAMPRATE;                                  /* Fixed sample rate */
   else
    _WriteBuffer._TRate          = (double)3500000 / (double)WriteSampleRate;                /* Determine no. T-states per sample */
  _WriteBuffer._Dirty            = FALSE;                                                              /* Signal: buffer is empty */

  if (Fh < 0)                                                     /* Setup is requested for the SoundBlaster/Parallel buffering ? */
  {
    if (Force ||
        (TPConfig.PlaySoundBlaster && !SoundBlasterInfo.DoingDMA && _WriteBuffer._BufferIndex == 0) ||
        (!TPConfig.PlaySoundBlaster && !ParallelRunning))                                                    /* Already running ? */
    {
      if (Force || TPConfig.PlaySoundBlaster)
      {
        LinearAddress = (dword)_AudioBuffer;                                                         /* (Always linear in Watcom) */
        BufLen = ~(word)LinearAddress + 1;                                                          /* Calculate to page boundary */
        if (BufLen < (word)32760)                                                               /* Smaller than half the length ? */
        {
          LinearAddress = (LinearAddress + 0x00010000L) & 0x00FF0000L;                                      /* Start at next page */
          BufLen = (word)65520 - BufLen;                                                      /* Use remaining part of the buffer */
        }
        _WriteBuffer._BufferLength = (word)(BufLen >> 1);                                                        /* 64K / 2 = 32K */
        _WriteBuffer._SBBuffer[0] = (byte *)LinearAddress;                                       /* Determine the 2 buffer halves */
        _WriteBuffer._SBBuffer[1] = (byte *)(LinearAddress + _WriteBuffer._BufferLength);
        _WriteBuffer._BufferIndex  = 0;
        _WriteBuffer._ID01         = 0;                                                               /* Keep track of last state */
        if (SoundBlasterInfo.DSPMajorVersion < 4)                                                          /* Not a SB16 and up ? */
          _WriteBuffer._TRate      = (double)3.5 * ((double)256 - (byte)BlockHeader[4]);        /* Adjust no. T-states per sample */
                                                                /* Time Constant rather than samplerate, just like a VOC file ID1 */
      }
      _WriteBuffer._State        = 0xF0;                                                        /* (So goes to 0x10 at the start) */
      
      (*SB_DMAOverrun) = RealSB_DMAOverrun = FALSE;
      if (TPConfig.PlaySoundBlaster)
        SB_Speaker (TRUE);                                                                                  /* Open the line-outs */
    }
    else if (_WriteBuffer._State == 0x80)                                                      /* Last pulse played was silence ? */
      _WriteBuffer._State = 0xF0;                                                               /* Force next pulse to go to 0x10 */
    return (TRUE);                                                                                               /* (Cannot fail) */
  }

  if (_WriteBuffer._Buffer == NULL)                                                            /* Not set up yet (for re-inits) ? */
  {
    _WriteBuffer._BufferLength = 16384;                                                                 /* Assume we can get 16Kb */
    if (GlobalFileCopyBuffer != NULL)                                                                        /* Globally set up ? */
      _WriteBuffer._Buffer = GlobalFileCopyBuffer;
    else if ((_WriteBuffer._Buffer = (byte *)malloc ((size_t)16384)) == NULL)                                   /* Try 16K buffer */
      if ((_WriteBuffer._Buffer = (byte *)malloc ((size_t)4096)) == NULL)                              /* Retry at 4K if it fails */
      { ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Out of memory"); return (FALSE); }
      else
        _WriteBuffer._BufferLength = 4096;
    _WriteBuffer._State          = 0xF0;                                                        /* (So goes to 0x10 at the start) */
    _WriteBuffer._BufferLength -= 1;
  }
  else if (_WriteBuffer._State == 0x80)                                                        /* Last pulse played was silence ? */
    _WriteBuffer._State = 0xF0;                                                                 /* Force next pulse to go to 0x10 */
  _WriteBuffer._InitialFileIndex = FileIndex;
  if (!HandlingZ80EmRAW && !HandlingCSW)
  {
    if (write (Fh, BlockHeader, 6) != 6)                                           /* Write block ID: 'Sound at <SampleRate> Hz.' */
      return (FALSE);
    _WriteBuffer._TRate          = (double)3.5 * ((double)256 - (byte)BlockHeader[4]);          /* Adjust no. T-states per sample */
  }
  _WriteBuffer._FileIndex        = FileIndex + 6;
  _WriteBuffer._ID01             = TRUE;                                                          /* Signal: header is a VOC ID 1 */
  _WriteBuffer._BufferIndex      = 0;
  return (TRUE);
}

static void _CloseFileBuffer (bool WhichBuffer)

/**********************************************************************************************************************************/
/* Pre   : `WhichBuffer' is TRUE for the _ReadBuffer and `FALSE' for the _WriteBuffer.                                            */
/* Post  : The file cache buffer has been released.                                                                               */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  if (WhichBuffer)
  {
    if (_ReadBuffer._Buffer == NULL)                                                                              /* Not set up ? */
      return;
    if (GlobalFileCopyBuffer == NULL)                                                                        /* Globally set up ? */
      free (_ReadBuffer._Buffer);                                                                     /* Nope, release the memory */
    _ReadBuffer._Buffer = NULL;
  }
  else
  {
    if (_WriteBuffer._Buffer == NULL)                                                                             /* Not set up ? */
      return;
    if (_WriteBuffer._Fh >= 0 && GlobalFileCopyBuffer == NULL)                                               /* Globally set up ? */
      free (_WriteBuffer._Buffer);                                                                    /* Nope, release the memory */
    _WriteBuffer._Buffer = NULL;
  }
}

static void _BackupReadState (struct _ReadBuffer_s *RBuffer)

/**********************************************************************************************************************************/
/* Pre   : None.                                                                                                                  */
/* Post  : The entire _ReadBuffer structure has been copied to a safe place.                                                      */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  memcpy (RBuffer, &_ReadBuffer, sizeof (struct _ReadBuffer_s));
}

static bool _RestoreReadState (struct _ReadBuffer_s *RBuffer)

/**********************************************************************************************************************************/
/* Pre   : This is the restore from the last `_BackupReadState'.                                                                  */
/* Post  : The _ReadBuffer has been restored to the backup position and the file segment has been reread amd refiltered.          */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  //if (_ReadBuffer._InitialFileIndex == RBuffer->_InitialFileIndex)         /* Is the currently loaded block the block we want ? */
    //return (TRUE);                                                                          /* Then we won't have to reload it! */
  memcpy (&_ReadBuffer, RBuffer, sizeof (struct _ReadBuffer_s));                            /* Restore buffer location and status */
  if (_ReadBuffer._Fh < 0)                                                                 /* Decoding a Direct Recording block ? */
  {
    CopyFromXMS (_ReadBuffer._DBufferEnd, TapeWindow[_ReadBuffer._TapeWindow].Tape.XMSHandle,
                 _ReadBuffer._XMSIndex + _ReadBuffer._InitialFileIndex, _ReadBuffer._Buffer);
    return (TRUE);
  }
  _ReadBuffer._BaseLine   = _ReadBuffer._InitialBaseLine;                                                /* Restore filter values */
  _ReadBuffer._SkipLength = _ReadBuffer._InitialSkipLength;
  if (lseek (_ReadBuffer._Fh, _ReadBuffer._InitialFileIndex, SEEK_SET) == -1)   /* Return the filepointer to the initial position */
    return (FALSE);
  if (read (_ReadBuffer._Fh, _ReadBuffer._Buffer, _ReadBuffer._BufferEnd) < _ReadBuffer._BufferEnd)          /* Reread the buffer */
    return (FALSE);
  if (!HandlingZ80EmRAW && !HandlingCSW)
    if (ApplyFilter)                                                                                          /* Apply the filter */
      _SpecFilter (_ReadBuffer._Buffer, _ReadBuffer._BufferEnd, (word)_ReadBuffer._TRate, &(_ReadBuffer._BaseLine),
                   &(_ReadBuffer._SkipLength));
  return (TRUE);
}

static bool _ReadState (bool *State)

/**********************************************************************************************************************************/
/* Pre   : None.                                                                                                                  */
/* Post  : The next signal state is returned in `State'. All handling of the read buffer is done here as well. When the buffer is */
/*         fully read, the next block is read from disk, if any. FALSE is returned if the block had already ended.                */
/*         If the next state byte appears to be a VOC 'Sound Continue' block header, this header is skipped (if the global        */
/*         `_SkipVOCSoundContinue' was TRUE).                                                                                     */
/* Import: _ReadState (to skip the VOC header).                                                                                   */
/**********************************************************************************************************************************/

{
  word   LengthToRead;
  word   NRead;
  byte   ReadByte;
  bool   DoSkipHeader;
  dword  CurrentFileIndex;
  word   VOCHeaderRun;

  if (_ReadBuffer._BufferIndex == 0 && _ReadBuffer._Bit == 0)                                         /* Buffer empty or looped ? */
  {
    if (_ReadBuffer._LastBlock)                                                                      /* Last block done already ? */
      return (FALSE);                                                                                              /* Signal: EOF */
    _ReadBuffer._InitialFileIndex  = _ReadBuffer._FileIndex;                                               /* Store current value */
    _ReadBuffer._InitialBaseLine   = _ReadBuffer._BaseLine;
    _ReadBuffer._InitialSkipLength = _ReadBuffer._SkipLength;
    if (_ReadBuffer._FileEnd - _ReadBuffer._FileIndex > (dword)_ReadBuffer._BufferLength)         /* Determine size of next block */
      LengthToRead = _ReadBuffer._BufferLength;
    else
      LengthToRead = (word)(_ReadBuffer._FileEnd - _ReadBuffer._FileIndex);
    _ReadBuffer._BufferEnd = LengthToRead;
    if (_ReadBuffer._Fh < 0)                                                               /* Decoding a Direct Recording block ? */
    {
      _ReadBuffer._DBufferEnd = (_ReadBuffer._BufferEnd & 0xFFFC) + (_ReadBuffer._BufferEnd % 4 ? 4 : 0);
      CopyFromXMS (_ReadBuffer._DBufferEnd, TapeWindow[_ReadBuffer._TapeWindow].Tape.XMSHandle,
                   _ReadBuffer._XMSIndex + _ReadBuffer._InitialFileIndex, _ReadBuffer._Buffer);
      _ReadBuffer._FileIndex += _ReadBuffer._DBufferEnd;
      if (_ReadBuffer._FileIndex >= _ReadBuffer._FileEnd)                              /* If this is the last block, set a marker */
        _ReadBuffer._LastBlock = TRUE;
    }
    else
    {
      if ((NRead = read (_ReadBuffer._Fh, _ReadBuffer._Buffer, LengthToRead)) < LengthToRead)                  /* Premature EOF ? */
      {
        if (NRead == 0)                                                                                         /* Nothing read ? */
          return (FALSE);                                                                                          /* Signal: EOF */
        _ReadBuffer._BufferEnd = NRead;
        _ReadBuffer._LastBlock = TRUE;
        _ReadBuffer._FileIndex = _ReadBuffer._FileEnd;
      }
      else                                                                              /* (The next block was read successfully) */
      {
        _ReadBuffer._FileIndex += _ReadBuffer._BufferEnd;
        if (_ReadBuffer._FileIndex >= _ReadBuffer._FileEnd)                            /* If this is the last block, set a marker */
          _ReadBuffer._LastBlock = TRUE;
      }
      if (!HandlingZ80EmRAW && !HandlingCSW)
        if (ApplyFilter && _ReadBuffer._BufferEnd > 0)                                                        /* Apply the filter */
          _SpecFilter (_ReadBuffer._Buffer, _ReadBuffer._BufferEnd, (word)_ReadBuffer._TRate, &(_ReadBuffer._BaseLine),
                       &(_ReadBuffer._SkipLength));
    }
  }
  ReadByte = *(_ReadBuffer._Buffer + _ReadBuffer._BufferIndex);                                                  /* Read the byte */
  if ((ReadByte == 0x01 || ReadByte == 0x02 || ReadByte == 0x09) && _SkipVOCSoundContinue)   /* VOC Sound Continue block header ? */
  {
    DoSkipHeader = FALSE;                                                           /* It could very well be just a sample value! */
    CurrentFileIndex = _GetCurrentFileIndex ();                                           /* Make sure it's indeed a header block */
    for (VOCHeaderRun = 0 ; !DoSkipHeader && VOCHeaderRun < VOCHeaderLocationIndex ; VOCHeaderRun ++)
      if (VOCHeaderLocations[VOCHeaderRun] == CurrentFileIndex)
        DoSkipHeader = TRUE;
    if (DoSkipHeader)                                                                                   /* It's a header; skip it */
    {
      _SkipVOCSoundContinue = FALSE;                                                  /* (In case a size byte happens to be 0x02) */
      if (++ _ReadBuffer._BufferIndex == _ReadBuffer._BufferEnd)                                                  /* Adjust index */
        _ReadBuffer._BufferIndex = 0;
      if (!_ReadState (State) || !_ReadState (State) || !_ReadState (State))
      { _SkipVOCSoundContinue = TRUE; return (FALSE); }                                         /* Skip all 4 bytes of the header */
      if (ReadByte == 0x01)                        /* If 'Sound Start' rather than 'Sound Continue', the header is 2 bytes longer */
      {
        if (!_ReadState (State) || !_ReadState (State))                                          /* (Ignore change in samplerate) */
        { _SkipVOCSoundContinue = TRUE; return (FALSE); }
      }
      else if (ReadByte == 0x09)                                         /* If 'GoldWave Ext', the header is even 12 bytes longer */
      {
        if (!_ReadState (State) || !_ReadState (State) || !_ReadState (State) || !_ReadState (State) ||
            !_ReadState (State) || !_ReadState (State) || !_ReadState (State) || !_ReadState (State) ||
            !_ReadState (State) || !_ReadState (State) || !_ReadState (State) || !_ReadState (State))
        { _SkipVOCSoundContinue = TRUE; return (FALSE); }
      }
      _SkipVOCSoundContinue = TRUE;
      return (_ReadState (State));
    }
  }
  if (_ReadBuffer._Fh < 0)
  {
    if (_ReadBuffer._Bit == 0)
      *State = (ReadByte & 0x80 ? TRUE : FALSE);
    else
      *State = (ReadByte & (0x80 >> _ReadBuffer._Bit) ? TRUE : FALSE);
    if (++ _ReadBuffer._Bit == 8)
    {
      if (++ _ReadBuffer._BufferIndex == _ReadBuffer._BufferEnd)                                            /* Update the pointer */
        _ReadBuffer._BufferIndex = 0;                                  /* Loop back, the buffer is auto-refilled on the next call */
      _ReadBuffer._Bit = 0;
    }
    if (++ _ReadBuffer._FileCount == _ReadBuffer._FileLength)                                                  /* Done last bit ? */
    {
      _ReadBuffer._BufferIndex = 0;
      _ReadBuffer._Bit = 0;
      _ReadBuffer._LastBlock = TRUE;
    }
  }
  else
  {
    if (HandlingZ80EmRAW || HandlingCSW)
      *State = (bool)ReadByte;                                                                        /* Export the byte directly */
    else
      *State = (ReadByte >= 0x80 ? TRUE : FALSE);                                                         /* Set the signal state */
    if (++ _ReadBuffer._BufferIndex == _ReadBuffer._BufferEnd)                                              /* Update the pointer */
      _ReadBuffer._BufferIndex = 0;                                    /* Loop back, the buffer is auto-refilled on the next call */
  }
  return (TRUE);                                                                                            /* Signal: valid byte */
}

static long _ReadPulse (bool *LastState)

/**********************************************************************************************************************************/
/* Pre   : `LastState' is the last signal state read (either previous call or with _ReadState).                                   */
/* Post  : The number of signal samples in the next bytes is determined. This number if returned as number of T states.           */
/*         Note that the first byte of the next pulse has been read away (to determine the end), so the file pointer is left at   */
/*         1 byte after the start of the next pulse (or: end-of-pulse + 2).                                                       */
/* Import: _ReadState.                                                                                                            */
/**********************************************************************************************************************************/

{
  dword DPLen = 0;
  long PulseLength = 0;                            /* Last call read the first entry away! (unless _ReadBuffer._FirstByte = TRUE) */
                                                                                           /* The 'wrong' is solved to begin with */
  bool CurrentState;

  if (!_ReadState (&CurrentState))                                                                         /* Fetch current state */
    return (-1L);                                                                                                  /* Signal: EOF */
  if (HandlingZ80EmRAW || HandlingCSW)
  {
    if (_ReadBuffer._FirstByte)                                                                      /* There is no 'last' call ? */
      _ReadBuffer._FirstByte = FALSE;
    if ((byte)CurrentState == 0x00)                                                                         /* Encoded as dword ? */
    {
      if (!_ReadState (&CurrentState))  return (-1L);                                                           /* Get next state */
      DPLen = ((dword)CurrentState & 0x0000FF);
      if (!_ReadState (&CurrentState))  return (-1L);                                                           /* Get next state */
      DPLen |= (((dword)CurrentState & 0x0000FF) << 8);
      if (!_ReadState (&CurrentState))  return (-1L);                                                           /* Get next state */
      DPLen |= (((dword)CurrentState & 0x0000FF) << 16);
      if (!_ReadState (&CurrentState))  return (-1L);                                                           /* Get next state */
      DPLen |= (((dword)CurrentState & 0x0000FF) << 24);
    }
    else                                                                                                     /* (Encoded as byte) */
      DPLen = ((dword)CurrentState & 0x0000FF);
    PulseLength = (long)DPLen;
  }
  else                                                                                                                   /* (VOC) */
  {
    if (_ReadBuffer._FirstByte)                                                                      /* There is no 'last' call ? */
      _ReadBuffer._FirstByte = FALSE;
    else
      PulseLength ++;                                                                         /* (Include this byte in the count) */
    while (CurrentState != *LastState)                                                        /* Until the next signal transition */
    {
      PulseLength ++;
      if (!_ReadState (&CurrentState))                                                                          /* Get next state */
        CurrentState = *LastState;                                                                     /* Map EOF as pulse change */
    }
  }
  *LastState = !*LastState;                                                                                     /* Flip the state */
  return ((long)(PulseLength * _ReadBuffer._TRate));                               /* Return the number of T states of this pulse */
}

static long _PulseDiff (long ReferencePulse, long NewPulse)

/**********************************************************************************************************************************/
/* Pre   : `ReferencePulse' is the pulse average, `NewPulse' is the pulse to match against this average.                          */
/* Post  : The difference of `NewPulse' to the `ReferencePulse' is returned as a percentage.                                      */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  double Diff;
  double Shorter;
  double Longer;

  if (ReferencePulse >= NewPulse)  { Longer = (double)ReferencePulse; Shorter = (double)NewPulse; }
  else                             { Longer = (double)NewPulse; Shorter = (double)ReferencePulse; }
  Diff = (double)(Longer - Shorter);
  return ((long)(Diff / (Diff + Longer) * 100));
}

static dword _GetCurrentFileIndex (void)

/**********************************************************************************************************************************/
/* Pre   : None.                                                                                                                  */
/* Post  : The current file index has been returned.                                                                              */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  if (_ReadBuffer._Fh < 0)
    if (_ReadBuffer._BufferIndex == 0 && _ReadBuffer._Bit == 0)
      return (_ReadBuffer._FileCount + (_ReadBuffer._LastBlock ? 1 : 0) + (_ReadBuffer._FirstByte ? 1 : 0));
    else
      return (_ReadBuffer._FileCount);
  else
    if (_ReadBuffer._BufferIndex == 0)                                                                                /* (looped) */
      return (_ReadBuffer._FileIndex + (_ReadBuffer._LastBlock ? 1 : 0) + (_ReadBuffer._FirstByte ? 1 : 0));    /* Adjust for EOF */
    else
      return (_ReadBuffer._FileIndex - _ReadBuffer._BufferEnd + _ReadBuffer._BufferIndex);                    /* Current location */
}

static bool _WriteStateSB (byte State)

/**********************************************************************************************************************************/
/* Pre   : `State' is the current signal state. It should be one of 0x10, 0xF0 or 0x80 (silence).                                 */
/* Post  : The signal state is written to the SoundBlaster DMA buffer.                                                            */
/*         While the DMA buffer is full, the keyboard is scanned. If [ESC] was pressed, FALSE is returned.                        */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct ButtonInfo     *ButtonP;
#ifdef __WATCOMC__
  int                    NoCacheLoop;
#endif

  if (_WriteBuffer._BufferIndex == _WriteBuffer._BufferLength)                                          /* Buffer full allready ? */
  {
    (*SB_BufferReady) = _WriteBuffer._ID01;                                                                /* Signal: buffer full */
    if (TPConfig.PlaySoundBlaster && !SoundBlasterInfo.DoingDMA)                        /* Just wrote first buffer - no DMA yet ? */
      SB_PlayAutoInitDMA (_WriteBuffer._SBBuffer[0], (word)WriteSampleRate, _WriteBuffer._BufferLength << 1);   /* Start playing! */
    else
    {
      while ((*SB_DMAComplete) != _WriteBuffer._ID01)                                   /* Wait for the other buffer to come free */
      {
        HandleWindow (&_SBWindow, &ButtonP);                                                   /* (Only to scan the keyboard etc) */
        if (GlobalKey == K_ESC)                                                                                /* [ESC] pressed ? */
        {
          GlobalKey = 0x0000;
          RealSB_DMAOverrun = FALSE;                                                      /* Report the error to FlushWriteBuffer */
          return (FALSE);                                                                                      /* Return: 'error' */
        }
        else if (SB_DMANoIRQTimeOut)
          return (FALSE);
        else if (GlobalKey == K_F1 && (GlobalShiftStatus & KS_ACTRL))                                             /* [CTRL][F1] ? */
        { _DEBUGSBDMA = TRUE; GlobalKey = 0x0000; return (FALSE); }
        GlobalKey = 0x0000;                                                                             /* (Drop all keys anyway) */
        if (PlayTimeUpdated)  { PlayTimeUpdated = FALSE; Wprintf (&_SBWindow, 1, TEXT_CENTER, PlayTimeString); }
#ifdef __WATCOMC__
        NoCacheLoop = kbhit ();
#endif
      }
    }
    if (*SB_DMAOverrun)                                                                              /* Did a DMA overrun occur ? */
    {
      RealSB_DMAOverrun = TRUE;                                                           /* Report the error to FlushWriteBuffer */
      return (FALSE);                                                                                    /* Then return the error */
    }
    _WriteBuffer._ID01 ^= 0x01;
    _WriteBuffer._BufferIndex = 0;
  }
  *(_WriteBuffer._SBBuffer[_WriteBuffer._ID01] + (_WriteBuffer._BufferIndex ++)) = State;             /* 'Write' the signal state */
  if (PlayTimeUpdated)  { PlayTimeUpdated = FALSE; Wprintf (&_SBWindow, 1, TEXT_CENTER, PlayTimeString); }
  return (TRUE);
}

static bool _WriteStateFile (byte State)

/**********************************************************************************************************************************/
/* Pre   : `State' is the current signal state. It should be one of 0x10, 0xF0 or 0x80 (silence).                                 */
/* Post  : The signal state is added to the file. The buffer is automatically flushed when needed.                                */
/*         Also, new VOC Sound block headers are automatically inserted when the total sample length reaches 16Mb.                */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  dword SampleLength;

  if (_WriteBuffer._BufferIndex == _WriteBuffer._BufferLength)                                           /* Buffer full already ? */
  {
    if (write (_WriteBuffer._Fh, _WriteBuffer._Buffer, _WriteBuffer._BufferLength) != _WriteBuffer._BufferLength)
      return (FALSE);                                                                        /* Write the full buffer to the file */
    _WriteBuffer._FileIndex += _WriteBuffer._BufferLength;
    SampleLength = _WriteBuffer._FileIndex - _WriteBuffer._InitialFileIndex - 4;         /* Total size of the sample block so far */
    if (!HandlingZ80EmRAW && !HandlingCSW && SampleLength >= 0x00FFFFF0)                                  /* More than 16777200 ? */
    {
      if (lseek (_WriteBuffer._Fh, _WriteBuffer._InitialFileIndex + 1, SEEK_SET) == -1)         /* Go back to the sample's header */
        return (FALSE);
      if (write (_WriteBuffer._Fh, &SampleLength, 3) != 3)                                             /* Write the sample length */
        return (FALSE);
      if (lseek (_WriteBuffer._Fh, _WriteBuffer._FileIndex, SEEK_SET) == -1)                /* And go back to the end of the file */
        return (FALSE);
      _WriteBuffer._InitialFileIndex = _WriteBuffer._FileIndex;
      if (write (_WriteBuffer._Fh, "\x02\x00\x00\x00", 4) != 4)                          /* Write next block ID: 'Sound Continue' */
        return (FALSE);
      _WriteBuffer._FileIndex += 4;
      _WriteBuffer._ID01 = FALSE;                                                                   /* Signal: VOC header is ID 2 */
    }
    _WriteBuffer._BufferIndex = 0;
  }
  *(_WriteBuffer._Buffer + (_WriteBuffer._BufferIndex ++)) = State;                                   /* 'Write' the signal state */
  _WriteBuffer._Dirty = TRUE;
  return (TRUE);
}

static bool _WritePulse (byte *LastState, long PulseLength)

/**********************************************************************************************************************************/
/* Pre   : `LastState' is the previous signal state, `PulseLength' is the length of the pulse to write in T-states.               */
/* Post  : The pulse has been added to the file.                                                                                  */
/*         If `LastState' was 0x80, this value is kept, otherwise the amplitude value has been toggled.                           */
/* Import: _WriteStateFile, _WriteStateSB.                                                                                        */
/**********************************************************************************************************************************/

{
  long  CntSamples;
  long  NumSamples;
  byte *PNumSamples;
  byte  NullByte = 0;

  if (*LastState != 0x80)                                                                               /* Not a 'silent' state ? */
    *LastState = (*LastState == 0x10 ? 0xF0 : 0x10);
  if (ParallelRunning)
    return (Par_PlayPulse (*LastState, PulseLength));
  NumSamples = _NumSamples (PulseLength, _WriteBuffer._TRate);
  if (HandlingZ80EmRAW || HandlingCSW)
  {
    if (NumSamples < 256)                                                                           /* Can be encoded as 1 byte ? */
    { if (!_WriteStateFile ((byte)NumSamples))  return (FALSE); }
    else
    {
      PNumSamples = (byte *)&NumSamples;                                                                     /* (Quick and dirty) */
      { if (!_WriteStateFile (NullByte))  return (FALSE); }
      for (CntSamples = 0 ; CntSamples < 4 ; CntSamples ++)
      { if (!_WriteStateFile (*(PNumSamples ++)))  return (FALSE); }
    }
  }
  else
    for (CntSamples = 0 ; CntSamples < NumSamples ; CntSamples ++)
      if (_WriteBuffer._Fh < 0)  { if (!_WriteStateSB (*LastState))  return (FALSE); }
      else                       { if (!_WriteStateFile (*LastState))  return (FALSE); }
  return (TRUE);
}

static void _InsertSpeedLock4567Pilot (short WhichWindow, long PilotLength)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `PilotLength' is the NumPilotPulses value for the pilot itself.             */
/* Post  : The pilot, composed sync and 6-bit flag values have been inserted in the virtual tape.                                 */
/* Import: _AddEntryToTape, _AddSpaceToTape.                                                                                      */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *DirP;
  struct TapeEntry_s *NewDirP;
  word      register  ByteIndex;                                /* Index into WorkBuffer (where the decoded bytes will be stored) */
  dword               XMSIndex;
  word                Zero;
  word                One;

  sprintf (WorkBuffer, "SpeedLock %d block %d", DecodingScheme - DS_SLOCK1 + 1, ++ SLBlockCnt);
  for (ByteIndex = 0 ; ByteIndex < 16 ; ByteIndex ++)                                                  /* Insert full sync signal */
    *((word *)WorkBuffer + ByteIndex + 10) = (word)ComposedSync[ByteIndex];
  Zero = (word)((DecodingScheme == DS_SLOCK6 || DecodingScheme == DS_SLOCK7) ? 753 : 551);
  One  = (word)((DecodingScheme == DS_SLOCK6 || DecodingScheme == DS_SLOCK7) ? 1508 : 1102);
  *((word *)WorkBuffer + (ByteIndex ++) + 10) = One;                                    /* SpeedLock 4-7 have a fixed flag 111010 */
  *((word *)WorkBuffer + (ByteIndex ++) + 10) = One;
  *((word *)WorkBuffer + (ByteIndex ++) + 10) = One;
  *((word *)WorkBuffer + (ByteIndex ++) + 10) = One;
  *((word *)WorkBuffer + (ByteIndex ++) + 10) = One;
  *((word *)WorkBuffer + (ByteIndex ++) + 10) = One;
  *((word *)WorkBuffer + (ByteIndex ++) + 10) = Zero;
  *((word *)WorkBuffer + (ByteIndex ++) + 10) = Zero;
  *((word *)WorkBuffer + (ByteIndex ++) + 10) = One;
  *((word *)WorkBuffer + (ByteIndex ++) + 10) = One;
  *((word *)WorkBuffer + (ByteIndex ++) + 10) = Zero;
  *((word *)WorkBuffer + (ByteIndex ++) + 10) = Zero;
  XMSIndex = ByteIndex * 2 + 20;
  _AddSpaceToTape (WhichWindow, XMSIndex);
  CopyToXMS (XMSIndex, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer, TapeWindow[WhichWindow].Tape.Size);
  TapeWindow[WhichWindow].Tape.Size += XMSIndex;
  if ((NewDirP = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
    FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
  memset (NewDirP, 0, sizeof (struct TapeEntry_s));
  NewDirP->ComboSelected = 0x00;
  NewDirP->BlockTypeID = 0x21;                                                                                     /* Group Start */
  GetBlockTypeInfo (0x21, &(NewDirP->IsTapeData), &(NewDirP->IsIdentifiedData), NULL, NULL, NULL);
  NewDirP->BlockLength = 19;                                                                 /* (19 characters in the group name) */
  NewDirP->AllocLength = 20;
  NewDirP->Next = NULL;
  _AddEntryToTape (WhichWindow, NewDirP);
  DirP = NewDirP;
  if ((DirP->Next = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)                                 /* Pilot */
    FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
  memset (DirP->Next, 0, sizeof (struct TapeEntry_s));
  DirP->Next->ComboSelected = 0x00;
  DirP->Next->BlockTypeID = 0x12;                                                                           /* Store as Pure Tone */
  GetBlockTypeInfo (0x12, &(DirP->Next->IsTapeData), &(DirP->Next->IsIdentifiedData), NULL, NULL, NULL);
  DirP->Next->Timing.LengthPilotPulse = 2100;
  DirP->Next->Timing.NumPilotPulses = PilotLength;
  DirP->Next->Next = NULL;
  DirP = DirP->Next;
  if ((DirP->Next = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)                           /* Sync + flag */
    FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
  memset (DirP->Next, 0, sizeof (struct TapeEntry_s));
  DirP->Next->ComboSelected = 0x00;
  DirP->Next->BlockTypeID = 0x13;                                                                       /* Store as Row of Pulses */
  GetBlockTypeInfo (0x13, &(DirP->Next->IsTapeData), &(DirP->Next->IsIdentifiedData), NULL, NULL, NULL);
  DirP->Next->BlockLength = 56;                                                                          /* (28 halfpulses total) */
  DirP->Next->AllocLength = 56;
  DirP->Next->Next = NULL;
  DirP = DirP->Next;
  TapeWindow[WhichWindow].NumEntries += 3;
}

static void _InsertDigitalIntegrationPilot (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the currently stored information.                         */
/* Post  : The composed pilot and single sync pulse has been inserted into the virtual tape, preceeded by a Group Start block.    */
/* Import: _AddEntryToTape, _AddSpaceToTape.                                                                                      */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *DirP;
  struct TapeEntry_s *NewDirP;
  dword               XMSIndex;

  sprintf (WorkBuffer, "DI block %d", ++ SLBlockCnt);
  *((word *)(WorkBuffer + 12)) = 725;                                                                    /* The single sync pulse */
  XMSIndex = 16;
  _AddSpaceToTape (WhichWindow, XMSIndex);
  CopyToXMS (XMSIndex, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer, TapeWindow[WhichWindow].Tape.Size);
  TapeWindow[WhichWindow].Tape.Size += XMSIndex;
  if ((NewDirP = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
    FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
  memset (NewDirP, 0, sizeof (struct TapeEntry_s));
  NewDirP->ComboSelected = 0x00;
  NewDirP->BlockTypeID = 0x21;                                                                                     /* Group Start */
  GetBlockTypeInfo (0x21, &(NewDirP->IsTapeData), &(NewDirP->IsIdentifiedData), NULL, NULL, NULL);
  NewDirP->BlockLength = 10;                                                                 /* (10 characters in the group name) */
  NewDirP->AllocLength = 12;
  NewDirP->Next = NULL;
  _AddEntryToTape (WhichWindow, NewDirP);
  DirP = NewDirP;
  if ((DirP->Next = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)                               /* Pilot 1 */
    FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
  memset (DirP->Next, 0, sizeof (struct TapeEntry_s));
  DirP->Next->ComboSelected = 0x00;
  DirP->Next->BlockTypeID = 0x12;                                                                           /* Store as Pure Tone */
  GetBlockTypeInfo (0x12, &(DirP->Next->IsTapeData), &(DirP->Next->IsIdentifiedData), NULL, NULL, NULL);
  DirP->Next->Timing.LengthPilotPulse = (word)PilotBuffer[0];
  DirP->Next->Timing.NumPilotPulses = (word)PilotBuffer[1];
  DirP->Next->Next = NULL;
  DirP = DirP->Next;
  if ((DirP->Next = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)                               /* Pilot 2 */
    FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
  memset (DirP->Next, 0, sizeof (struct TapeEntry_s));
  DirP->Next->ComboSelected = 0x00;
  DirP->Next->BlockTypeID = 0x12;
  GetBlockTypeInfo (0x12, &(DirP->Next->IsTapeData), &(DirP->Next->IsIdentifiedData), NULL, NULL, NULL);
  DirP->Next->Timing.LengthPilotPulse = Entry->Timing.LengthPilotPulse;
  DirP->Next->Timing.NumPilotPulses = (word)256;
  DirP->Next->Next = NULL;
  DirP = DirP->Next;
  if ((DirP->Next = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)                                  /* Sync */
    FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
  memset (DirP->Next, 0, sizeof (struct TapeEntry_s));
  DirP->Next->ComboSelected = 0x00;
  DirP->Next->BlockTypeID = 0x13;                                                                       /* Store as Row of Pulses */
  GetBlockTypeInfo (0x13, &(DirP->Next->IsTapeData), &(DirP->Next->IsIdentifiedData), NULL, NULL, NULL);
  DirP->Next->BlockLength = 2;                                                                                   /* (1 halfpulse) */
  DirP->Next->AllocLength = 4;
  DirP->Next->Next = NULL;
  DirP = DirP->Next;
  TapeWindow[WhichWindow].NumEntries += 4;
}

static void _InsertClickingPilot (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT; a clicking pilot and sync has been read and stored.                         */
/* Post  : The clicking pilot, sync and 6-bit flag values have been inserted into the virtual tape.                               */
/*         To reduce the number of needed entries, a Loop is used according to the repetition scheme of this pilot.               */
/* Import: _AddEntryToTape, _AddSpaceToTape.                                                                                      */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *DirP;
  struct TapeEntry_s *NewDirP;
  word      register  ByteIndex;                                /* Index into WorkBuffer (where the decoded bytes will be stored) */
  dword               XMSIndex;
  struct PilotTest_s
  {
    long              NumPulses;
    byte              Occurences;
    byte              Location[50];
  }                   PilotTest[51];                                                                 /* (Add 1 for noise removal) */
  byte                SinglePulses;
  byte                Repeats;
  byte                RepeatLength;
  bool                Found = FALSE;
  byte                RemoveAt;

  sprintf (WorkBuffer, "SpeedLock %d block %d", DecodingScheme - DS_SLOCK1 + 1, ++ SLBlockCnt);

  /* Create a normalized pilot graph */
  PilotTest[0].NumPulses = PilotBuffer[PilotIndex - 1];
  PilotTest[0].Location[0] = PilotIndex - 1;
  PilotTest[0].Occurences = 1;
  SinglePulses = 1;
  for (Repeats = PilotIndex - 2 ; Repeats > 0 ; Repeats --)                                     /* (Ignore the first pilot pulse) */
  {
    Found = FALSE;
    for (RepeatLength = 0 ; !Found && RepeatLength < SinglePulses ; RepeatLength ++)
      if (PilotBuffer[Repeats] == PilotTest[RepeatLength].NumPulses)
      {
        PilotTest[RepeatLength].Location[PilotTest[RepeatLength].Occurences ++] = Repeats;
        Found = TRUE;
      }
    if (!Found)
    {
      PilotTest[SinglePulses].NumPulses = PilotBuffer[Repeats];
      PilotTest[SinglePulses].Location[0] = Repeats;
      PilotTest[SinglePulses ++].Occurences = 1;
    }
  }

  /* Remove corrupted pilots from the list */
  RemoveAt = PilotIndex / 8;                                                              /* Noise is considered at each 8 pilots */
  for (Repeats = 0 ; Repeats < SinglePulses ; )                                                                 /* Sift out noise */
    if (PilotTest[Repeats].Occurences <= RemoveAt)
    {
      memcpy (&PilotTest[Repeats], &PilotTest[Repeats + 1], sizeof (struct PilotTest_s) * (SinglePulses - Repeats));
      SinglePulses --;
    }
    else
      Repeats ++;

  /* Now set the repeat length */
  RepeatLength = SinglePulses;

  /* And rearrange the pilot lengths according to the repeat length */
  if (RepeatLength == 1)
    for (Repeats = 0 ; Repeats < PilotIndex ; Repeats ++)                                                       /* Set all pilots */
      PilotBuffer[Repeats] = PilotTest[0].NumPulses;
  else
  {
    memset (&PilotBuffer[0], 0, sizeof (long) * 50);                                                            /* Rebuild pilots */
    for (Repeats = 0 ; Repeats < RepeatLength ; Repeats ++)
      for (SinglePulses = 0 ; SinglePulses < PilotTest[Repeats].Occurences ; SinglePulses ++)
         PilotBuffer[PilotTest[Repeats].Location[SinglePulses]] = PilotTest[Repeats].NumPulses;
    for (Repeats = 1 ; Repeats < PilotIndex ; Repeats ++)
      if (PilotBuffer[Repeats] == 0)                                                               /* This was a corruped pilot ? */
      {                                                                                                /* Try to fill up the hole */
        Found = FALSE;
        if (Repeats >= 4)
          if (PilotBuffer[Repeats - 4] != 0)
          {
            PilotBuffer[Repeats] = PilotBuffer[Repeats - 4];
            Found = TRUE;
          }
          else if (Repeats >= 8)
            if (PilotBuffer[Repeats - 8] != 0)
            {
              PilotBuffer[Repeats] = PilotBuffer[Repeats - 8];
              Found = TRUE;
            }
        if (!Found)
          if (Repeats < PilotIndex - 4)
            if (PilotBuffer[Repeats + 4] != 0)
            {
              PilotBuffer[Repeats] = PilotBuffer[Repeats + 4];
              Found = TRUE;
            }
          else if (Repeats < PilotIndex - 8)
            if (PilotBuffer[Repeats + 8] != 0)
            {
              PilotBuffer[Repeats] = PilotBuffer[Repeats + 8];
              Found = TRUE;
            }
      }
    if (RepeatLength == 2)                                                                   /* Only 2 pilot lengths being used ? */
      if (PilotBuffer[PilotIndex - 1] != PilotBuffer[PilotIndex - 3] ||
          PilotBuffer[PilotIndex - 2] != PilotBuffer[PilotIndex - 4])                                           /* Pattern 0101 ? */
        RepeatLength = 4;                                                                                     /* No, pattern 0110 */
    PilotBuffer[0] = PilotBuffer[4];                                                            /* (Fill the ignored first pilot) */
  }

  /* All set; insert the clicking pilot! */
  SinglePulses = PilotIndex % RepeatLength;
  Repeats = PilotIndex / RepeatLength;
  if (SinglePulses == 0)
  { SinglePulses = RepeatLength; Repeats --; }
  for (ByteIndex = 0 ; ByteIndex < 2 * (SinglePulses + RepeatLength - 1) ; ByteIndex ++)                    /* Prepare the buffer */
    *((word *)WorkBuffer + ByteIndex + 10) = 715;                                                           /* 2 pulses per click */
  *((word *)WorkBuffer + (ByteIndex ++) + 10) = 3150;                                                        /* Insert sync bytes */
  *((word *)WorkBuffer + (ByteIndex ++) + 10) = 3150;
  if (DecodingScheme == DS_SLOCK1)
  {
    TmpXMSIndex = TapeWindow[WhichWindow].Tape.Size + (ByteIndex * 2) + 20;              /* Keep XMS location of the flag (words) */
    XMSIndex = ByteIndex * 2 + 44;                                              /* Total size of all data - clicks, sync and flag */
  }
  else
  {
    *((word *)WorkBuffer + (ByteIndex ++) + 10) = 1130;                                 /* SpeedLock 2/3 have a fixed flag 111010 */
    *((word *)WorkBuffer + (ByteIndex ++) + 10) = 1130;
    *((word *)WorkBuffer + (ByteIndex ++) + 10) = 1130;
    *((word *)WorkBuffer + (ByteIndex ++) + 10) = 1130;
    *((word *)WorkBuffer + (ByteIndex ++) + 10) = 1130;
    *((word *)WorkBuffer + (ByteIndex ++) + 10) = 1130;
    *((word *)WorkBuffer + (ByteIndex ++) + 10) = 565;
    *((word *)WorkBuffer + (ByteIndex ++) + 10) = 565;
    *((word *)WorkBuffer + (ByteIndex ++) + 10) = 1130;
    *((word *)WorkBuffer + (ByteIndex ++) + 10) = 1130;
    *((word *)WorkBuffer + (ByteIndex ++) + 10) = 565;
    *((word *)WorkBuffer + (ByteIndex ++) + 10) = 565;
    XMSIndex = ByteIndex * 2 + 20;
  }
  _AddSpaceToTape (WhichWindow, XMSIndex);
  CopyToXMS (XMSIndex, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer, TapeWindow[WhichWindow].Tape.Size);
  TapeWindow[WhichWindow].Tape.Size += XMSIndex;
  if ((NewDirP = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
    FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
  memset (NewDirP, 0, sizeof (struct TapeEntry_s));
  NewDirP->ComboSelected = 0x00;
  NewDirP->BlockTypeID = 0x21;                                                                                     /* Group Start */
  GetBlockTypeInfo (0x21, &(NewDirP->IsTapeData), &(NewDirP->IsIdentifiedData), NULL, NULL, NULL);
  NewDirP->BlockLength = 19;                                                                 /* (19 characters in the group name) */
  NewDirP->AllocLength = 20;
  NewDirP->Next = NULL;
  _AddEntryToTape (WhichWindow, NewDirP);
  DirP = NewDirP;
  for (ByteIndex = 0 ; ByteIndex < SinglePulses ; ByteIndex ++)                                          /* Store all pilot parts */
  {
    if ((DirP->Next = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
      FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
    memset (DirP->Next, 0, sizeof (struct TapeEntry_s));
    DirP->Next->ComboSelected = 0x00;
    DirP->Next->BlockTypeID = 0x12;                                                                         /* Store as Pure Tone */
    GetBlockTypeInfo (0x12, &(DirP->Next->IsTapeData), &(DirP->Next->IsIdentifiedData), NULL, NULL, NULL);
    DirP->Next->Timing.LengthPilotPulse = 2100;
    DirP->Next->Timing.NumPilotPulses = PilotBuffer[ByteIndex];
    DirP->Next->Next = NULL;
    DirP = DirP->Next;
    if (ByteIndex < SinglePulses - 1)
    {
      if ((DirP->Next = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)                             /* Click */
        FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
      memset (DirP->Next, 0, sizeof (struct TapeEntry_s));
      DirP->Next->ComboSelected = 0x00;
      DirP->Next->BlockTypeID = 0x13;                                                                   /* Store as Row of Pulses */
      GetBlockTypeInfo (0x13, &(DirP->Next->IsTapeData), &(DirP->Next->IsIdentifiedData), NULL, NULL, NULL);
      DirP->Next->BlockLength = 4;                                                                                  /* (2 pulses) */
      DirP->Next->AllocLength = 4;
      DirP->Next->Next = NULL;
      DirP = DirP->Next;
    }
  }
  /* Loop Start */
  if ((DirP->Next = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
    FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
  memset (DirP->Next, 0, sizeof (struct TapeEntry_s));
  DirP->Next->ComboSelected = 0x00;
  DirP->Next->BlockTypeID = 0x24;                                                                                  /* Loop Start */
  GetBlockTypeInfo (0x21, &(DirP->Next->IsTapeData), &(DirP->Next->IsIdentifiedData), NULL, NULL, NULL);
  DirP->Next->PauseLength = Repeats;
  DirP->Next->Next = NULL;
  DirP = DirP->Next;
  for (ByteIndex = 0 ; ByteIndex < RepeatLength ; ByteIndex ++)                                          /* Store all pilot parts */
  {
    if ((DirP->Next = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)                               /* Click */
      FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
    memset (DirP->Next, 0, sizeof (struct TapeEntry_s));
    DirP->Next->ComboSelected = 0x00;
    DirP->Next->BlockTypeID = 0x13;                                                                     /* Store as Row of Pulses */
    GetBlockTypeInfo (0x13, &(DirP->Next->IsTapeData), &(DirP->Next->IsIdentifiedData), NULL, NULL, NULL);
    DirP->Next->BlockLength = 4;                                                                                    /* (2 pulses) */
    DirP->Next->AllocLength = 4;
    DirP->Next->Next = NULL;
    DirP = DirP->Next;
    if ((DirP->Next = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
      FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
    memset (DirP->Next, 0, sizeof (struct TapeEntry_s));
    DirP->Next->ComboSelected = 0x00;
    DirP->Next->BlockTypeID = 0x12;                                                                         /* Store as Pure Tone */
    GetBlockTypeInfo (0x12, &(DirP->Next->IsTapeData), &(DirP->Next->IsIdentifiedData), NULL, NULL, NULL);
    DirP->Next->Timing.LengthPilotPulse = 2100;
    DirP->Next->Timing.NumPilotPulses = PilotBuffer[PilotIndex - RepeatLength + ByteIndex];
    DirP->Next->Next = NULL;
    DirP = DirP->Next;
  }
  /* Loop End */
  if ((DirP->Next = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
    FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
  memset (DirP->Next, 0, sizeof (struct TapeEntry_s));
  DirP->Next->ComboSelected = 0x00;
  DirP->Next->BlockTypeID = 0x25;                                                                                     /* Loop End */
  GetBlockTypeInfo (0x21, &(DirP->Next->IsTapeData), &(DirP->Next->IsIdentifiedData), NULL, NULL, NULL);
  DirP->Next->Next = NULL;
  DirP = DirP->Next;
  if ((DirP->Next = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)                                  /* Flag */
    FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
  memset (DirP->Next, 0, sizeof (struct TapeEntry_s));
  DirP->Next->ComboSelected = 0x00;
  DirP->Next->BlockTypeID = 0x13;                                                                       /* Store as Row of Pulses */
  GetBlockTypeInfo (0x13, &(DirP->Next->IsTapeData), &(DirP->Next->IsIdentifiedData), NULL, NULL, NULL);
  DirP->Next->BlockLength = 28;                                                            /* (6 bits fullpulses + 2 sync pulses) */
  DirP->Next->AllocLength = 28;
  DirP->Next->Next = NULL;
  DirP = DirP->Next;
  TapeWindow[WhichWindow].NumEntries += (SinglePulses + RepeatLength) * 2 + 3;
}

static void _ProcessInterBlockPause (short WhichWindow, dword PauseLength, dword SampleRate)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `PauseLength' is the length in T-states, `SampleRate' the samplerate.       */
/* Post  : The tape length between the current block and the end of the previous has been added as pause length to the last tape  */
/*         entry.                                                                                                                 */
/*         If the decoding scheme was SpeedLock 5 or 6, the pause could be required noise. This depends on the length.            */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *DirP;
  struct TapeEntry_s *PrevDirP;
  struct TapeEntry_s *UsedDirP;
  struct TapeEntry_s *NewDirP;
  bool                UsePrevious = FALSE;                                               /* TRUE if the last entry is a Group End */
  dword               ConvPauseLength;                                                            /* Pause length in milliseconds */

  DirP = TapeWindow[WhichWindow].Directory;                                                       /* Find last block in the chain */
  while (DirP->Next != NULL)
  {
    PrevDirP = DirP;
    DirP = DirP->Next;
  }
  ConvPauseLength = (PauseLength * 1000) / SampleRate;
  if (DecodingScheme == DS_DINAMIC && ConvPauseLength < 7)         /* Dinamic loaders must have pauses of atleast 7 milliseconds! */
    ConvPauseLength = 7;
  else if (DecodingScheme == DS_SLOCK5 || DecodingScheme == DS_SLOCK6)                                      /* SpeedLock 5 or 6 ? */
    if (PauseLength / SampleRate > 10)                                                                  /* More than 10 seconds ? */
    {
      if ((DirP->Next = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
        FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
      memset (DirP->Next, 0, sizeof (struct TapeEntry_s));
      DirP->Next->ComboSelected = 0x00;
      DirP->Next->Timing.SampleRate = SampleRate;
      DirP->Next->BlockTypeID = 0x12;                                                                      /* Save as 'Pure Tone' */
      GetBlockTypeInfo (0x12, &(DirP->Next->IsTapeData), &(DirP->Next->IsIdentifiedData), NULL, NULL, NULL);
      DirP->Next->Timing.LengthPilotPulse = 3000;                                     /* The pulse length must be at least 3000 T */
      DirP->Next->Timing.NumPilotPulses = (word)(((double)PauseLength * 3500) / ((double)SampleRate * 3));
      DirP->Next->Next = NULL;
      return;
    }
  if (DirP->BlockTypeID == 0x22)                                                                   /* Last entry is a Group End ? */
  {
    UsePrevious = TRUE;
    UsedDirP = PrevDirP;                                                                                /* Use the one-last entry */
  }
  else
    UsedDirP = DirP;
  if (UsedDirP->BlockTypeID == 0x10 || UsedDirP->BlockTypeID == 0x11 || UsedDirP->BlockTypeID == 0x14)
    UsedDirP->PauseLength += ConvPauseLength;
  else if (UsedDirP->BlockTypeID == 0x12 &&                                                  /* Pause after a 'Pure Tone Block' ? */
           ConvPauseLength > 0)                                                                         /* Anything substantial ? */
  {
    if ((NewDirP = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
      FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
    memset (NewDirP, 0, sizeof (struct TapeEntry_s));
    NewDirP->ComboSelected = 0x00;
    NewDirP->Timing.SampleRate = SampleRate;
    NewDirP->BlockTypeID = 0x20;                                                                    /* Insert a pause block first */
    GetBlockTypeInfo (0x20, &(NewDirP->IsTapeData), &(NewDirP->IsIdentifiedData), NULL, NULL, NULL);
    NewDirP->PauseLength = ConvPauseLength;
    if (UsePrevious)                                                                         /* Now link the new entry into place */
    {
      NewDirP->Next = DirP;
      PrevDirP->Next = NewDirP;                                          /* (PrevDirP cannot be NULL - Group End cannot be first) */
    }
    else
    {
      DirP->Next = NewDirP;
      NewDirP->Next = NULL;
    }
  }
}

static void _InsertSpeedLock234567MidSync (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/* Post  : The midblock sync signal has been inserted into the virtual tape.                                                      */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  word register Cnt;
  word register SyncIndex;

  if ((Entry->Next = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
    FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
  memset (Entry->Next, 0, sizeof (struct TapeEntry_s));
  Entry->Next->ComboSelected = 0x00;
  Entry->Next->BlockTypeID = 0x13;                                                                      /* Store as Row of Pulses */
  GetBlockTypeInfo (0x13, &(Entry->Next->IsTapeData), &(Entry->Next->IsIdentifiedData), NULL, NULL, NULL);
  Entry->Next->BlockLength = 84;                                                                                     /* (21 bits) */
  Entry->Next->AllocLength = 84;
  Entry->Next->Next = NULL;
  for (Cnt = 0, SyncIndex = 0 ; SyncIndex < 21 ; SyncIndex ++)
  {
    *((word *)WorkBuffer + (Cnt ++)) = (word)MidSyncMatch[SyncIndex];                                       /* (2 pulses per bit) */
    *((word *)WorkBuffer + (Cnt ++)) = (word)MidSyncMatch[SyncIndex];
  }
  XMSRealloc (TapeWindow[WhichWindow].Tape.XMSHandle, TapeWindow[WhichWindow].Tape.Size + 84, TRUE);
  CopyToXMS (84, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer, TapeWindow[WhichWindow].Tape.Size);
  TapeWindow[WhichWindow].Tape.Size += 84;
  TapeWindow[WhichWindow].NumEntries ++;
}

static void _InsertGroupEnd (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/* Post  : A 'Group End' block has been appended to the virtual tape.                                                             */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  if ((Entry->Next = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
    FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
  memset (Entry->Next, 0, sizeof (struct TapeEntry_s));
  Entry->Next->ComboSelected = 0x00;
  Entry->Next->BlockTypeID = 0x22;                                                                                 /* (Group End) */
  GetBlockTypeInfo (0x22, &(Entry->Next->IsTapeData), &(Entry->Next->IsIdentifiedData), NULL, NULL, NULL);
  Entry->Next->Next = NULL;
  TapeWindow[WhichWindow].NumEntries ++;
}

static void _GetBitTimingValues (word *Zero, word *One, long *Average)

/**********************************************************************************************************************************/
/* Pre   : `DecodingScheme' has been set up.                                                                                      */
/* Post  : The bit timing values for that scheme have been returned.                                                              */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  switch (DecodingScheme)
  {
    case DS_ROM     :
    case DS_BLEEP   :
    case DS_DINAMIC : *Zero = NS_LEN0BIT;
                      *One = NS_LEN1BIT;
                      break;
    case DS_PLAYERS : *Zero = 560;
                      *One = 1120;
                      break;
    case DS_SULOAD  :
    case DS_YSLOAD  : *Zero = 565;
                      *One = 1130;
                      break;
    case DS_POPS    : *Zero = 735;
                      *One = 1590;
                      break;
    case DS_SLOCK1  :
    case DS_SLOCK2  :
    case DS_SLOCK3  : *Zero = 565;
                      *One = 1130;
                      break;
    case DS_SLOCK4  :
    case DS_SLOCK5  : *Zero = 551;
                      *One = 1102;
                      break;
    case DS_SLOCK6  :
    case DS_SLOCK7  : *Zero = 753;
                      *One = 1508;
                      break;
  }
  *Average = _Average (*Zero, *One);
}

static long _GetSLSyncMatchValue (long PulseLength)

/**********************************************************************************************************************************/
/* Pre   : `PulseLength' holds the read (averaged) pulse length.                                                                  */
/* Post  : The SpeedLock 4/5/6/7 composed sync value according to this pulse length has been returned.                            */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  if (PulseLength >= 1786)                                                                             /* Middle of 1428 and 2144 */
    return (2144L);
  if (PulseLength >= 1071)                                                                              /* Middle of 714 and 1428 */
    return (1428L);
  if (PulseLength < 357)
    return (0L);
  return (714L);
}

static comboent _HandleSelectBlock (comboent CurrentBlockNo)

/**********************************************************************************************************************************/
/* Pre   : `CurrentBlockNo' holds the block number of this Select Block.                                                          */
/* Post  : The Select menu has been presented. The return value is the next block to be played (selected from the menu).          */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct ButtonInfo          *ButtonP;
  struct WindowInfo           ViewWindow;
  struct ComboInfo            TextWindow;
  struct ComboEntry           TEntry;
  bool                        More         = TRUE;
  bool                        Cancelled    = FALSE;
  byte              register  Cnt;
  word              register  Cnt2;
  byte                        Bt;
  comboent                    SelectedLine = -1;
  char                        MenuOption;
  byte                        MaxOption;

  SelectFont (NULL, FONT_NORMAL);
  memset (&ViewWindow, 0, sizeof (struct WindowInfo));                                           /* Initialize the interface part */
  ViewWindow.StartX      = _SelectBlockX;
  ViewWindow.StartY      = _SelectBlockY;
  ViewWindow.Width       = ViewWindow.MinWidth = ViewWindow.MaxWidth = CurrentFont->Width * 34 + 7;
  ViewWindow.Height      = ViewWindow.MinHeight = ViewWindow.MaxHeight = (CurrentFont->Height + 2) * 22 -2;
  ViewWindow.BorderFlags = WINB_OUT | WINB_SURROUND;
  ViewWindow.BorderSize  = 1;
  ViewWindow.APen        = SystemColors.ReqAPen;
  ViewWindow.BPen        = SystemColors.ReqOPen;
  ViewWindow.Header      = "Select Next Block";
  ViewWindow.Flags       = WINF_ACTIVATE | WINF_HUSE | WINF_HCENTER | WINF_XMSBUFFER;

  DrawWindow (&ViewWindow, NULL, WING_CLOSE | WING_RESIZEMAX, TRUE);
  MakeComboWindow (&TextWindow, &ViewWindow, COMT_SELECT, COMF_NOBUFFER | COMF_VSLIDER | COMF_SYNCSLIDER | COMF_KEYSELECT,
                   ViewWindow.A_StartX + 2, ViewWindow.StartY - 2, 0, 1024, 40, 34, 22, CurrentFont, FALSE, 0, NULL,
                   SystemColors.ReqAPen, SystemColors.ReqBPen, 0, 0);
  TEntry.Selected = ENT_NOSELECT;
  strcpy (TEntry.Text, "{+1Select which part must be played}N");
  if (!InsertComboEntry (&TextWindow, &TEntry, NULL, -1))
    FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
  TEntry.Text[0] = '\0';
  if (!InsertComboEntry (&TextWindow, &TEntry, NULL, -1))
    FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
  TEntry.Selected = 0;
  MaxOption = *WorkBuffer <= 10 ? *WorkBuffer : 10;
  for (Cnt = 0, Cnt2 = 1 ; Cnt < MaxOption ; Cnt ++)                                /* Insert the selection menu (max 10 options) */
  {
    Bt = *(WorkBuffer + Cnt2 + 2);                                                                /* (Length of description text) */
    TEntry.Identifier = (dword)(CurrentBlockNo + *((short *)(WorkBuffer + Cnt2)));
    if (Cnt < 10)
      MenuOption = '0' + Cnt + 1;
    else if (Cnt == 10)
      MenuOption = '0';
    sprintf (TEntry.Text, "  %c ", MenuOption);
    strncpy (TEntry.Text + 4, WorkBuffer + Cnt2 + 3, Bt);
    TEntry.Text[Bt + 4] = '\0';
    if (!InsertComboEntry (&TextWindow, &TEntry, NULL, -1))
      FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
    Cnt2 += (word)Bt + 3;
  }
  PrintComboEntries (&TextWindow);                                                                          /* Print the contents */
  SelectEntry (&TextWindow, 2);                                                                      /* Appoint first menu option */
  if (!SimulatePlayBlock)
    SoundErrorBeeper (BEEP_MEDIUM);                                                                       /* Get user's attention */
  while (More)                                                                                    /* Handle until close requested */
  {
    if (HandleWindow (&ViewWindow, &ButtonP) == WINH_CLOSE)                                              /* Read keyboard as well */
    { More = FALSE; Cancelled = TRUE; }
    else if (HandleComboWindow (&TextWindow, &SelectedLine) == WINH_LINEFIRST)      /* User choose an entry (1 click is enough) ? */
      More = FALSE;                                                                                                      /* Done! */
    else if (GlobalKey == K_ENTER)                                                            /* [RETURN] selects appointed block */
      More = FALSE;
    else if ((GlobalKey & 0x00FF) >= '1' && (GlobalKey & 0x00FF) <= '9')         /* They can also directly type the option number */
    {
      if ((GlobalKey & 0x00FF) - '0' <= MaxOption)
      { TextWindow.LastSelected = (GlobalKey & 0x00FF) - '0' + 1; More = FALSE; }
    }
    else if ((GlobalKey & 0x00FF) == '0' && MaxOption == 10)
    { TextWindow.LastSelected = 11 ; More = FALSE; }
    else if (GlobalKey == K_ESC)
    { More = FALSE; Cancelled = TRUE; }
    else if (More && GlobalKey == K_ESC)
    { More = FALSE; Cancelled = TRUE; }
    MouseStatus ();
    GlobalKey = 0x0000;
  }
  if (!Cancelled)
    GetComboEntry (&TextWindow, &TEntry, TextWindow.LastSelected);
  DestroyComboWindow (&TextWindow, 0);
  DestroyWindow (&ViewWindow);
  if (Cancelled)
    return (CurrentBlockNo + 1);
  return ((short)(TEntry.Identifier));
}

static void _PrintSoundAnalysis (struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `Entry' points to the currently decoded tape block.                                                                    */
/* Post  : A graph has been shown of the pulse and bit spread plus some info is printed, in a seperate window. The user closed it */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct WindowInfo CWindow   = {125, 22, 0, 449, 0, 449, 0, 449, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                 WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER, NULL, "Sound Block Analysis"};
  struct ButtonInfo *ButtonP;
  bool   Done     = FALSE;
  dword  Cnt;
  dword  Cnt2;
  short  Cnt3;
  dword  Value;
  dword  Num0Bits = 0;
  dword  Num1Bits = 0;
  short  OldValue;
  double Limiter;
  dword  TopValue = 0;
  short  Width    = 0;
  short  XOffset  = 42;                                                                                                /* 7 chars */
  short  YOffset  = 5;
  short  BitColor = DBLUE;

  if (Entry->BlockLength == 0)                                                                                   /* Empty block ? */
    return;                                                                                                       /* Don't bother */
  TopValue = 0;
  for (Cnt = 0 ; Cnt < SA_MAXPULSELENGTH ; Cnt ++)                                      /* Determine largest (X- and Y-co) values */
  {
    if (Width < Cnt && SA_Joined[Cnt] != 0)
      Width = Cnt;
    if (SA_Joined[Cnt] > TopValue)
      TopValue = SA_Joined[Cnt];
  }
  for (Cnt = 0 ; Cnt < SA_MAXPULSELENGTH ; Cnt ++)
  {
    Num0Bits += SA_Pulses[Cnt].Counts0;
    Num1Bits += SA_Pulses[Cnt].Counts1;
    if (Width < Cnt && (SA_Pulses[Cnt].Counts0 != 0 || SA_Pulses[Cnt].Counts1 != 0))
      Width = Cnt;
    if (SA_Pulses[Cnt].Counts0 + SA_Pulses[Cnt].Counts1 > TopValue)
      TopValue = SA_Pulses[Cnt].Counts0 + SA_Pulses[Cnt].Counts1;
  }
  Limiter = 400.0 / TopValue;                                          /* Create divider to maintain a graph height of 400 pixels */
  Width += 400;                                                             /* Add an empty pulse piece to the right of the graph */
  if (Width < 3080)                                                                                        /* (We need a minimum) */
    Width = 3080;
  else if (Width > SA_MAXPULSELENGTH)
    Width = SA_MAXPULSELENGTH;
  CWindow.Width = CWindow.MinWidth = CWindow.MaxWidth = Width / 10 + XOffset;       /* Determine window size and keep it centered */
  CWindow.StartX = (Screen.Width - CWindow.Width) / 2;
  DrawWindow (&CWindow, NULL, WING_CLOSE, TRUE);
  WDrawLine (&CWindow, DBLACK, XOffset, YOffset, XOffset, YOffset + 400);                                          /* Draw Y-axis */
  WDrawLine (&CWindow, DBLACK, XOffset - 1, YOffset, XOffset, YOffset);
  Wprintf (&CWindow, 0, 0, "%6ld", TopValue);                                                             /* Print vertical scale */
  Wprintf (&CWindow, 3, 4, "P");
  Wprintf (&CWindow, 4, 4, "u");
  Wprintf (&CWindow, 5, 4, "l");
  Wprintf (&CWindow, 6, 4, "s");
  Wprintf (&CWindow, 7, 4, "e");
  Wprintf (&CWindow, 8, 4, "s");
  WDrawLine (&CWindow, DBLACK, XOffset, YOffset + 401, Width / 10 + XOffset, YOffset + 401);                       /* Draw X-axis */
  for (Cnt = 0 ; Cnt < Width ; Cnt += 100)
    WDrawLine (&CWindow, DBLACK, Cnt / 10 + XOffset, YOffset + 401, Cnt / 10 + XOffset, YOffset + 402);
  for (Cnt = 0, Cnt3 = 0 ; Cnt < Width - 1 ; Cnt += 10, Cnt3 ++)                                    /* Draw pulse variation graph */
  {
    for (Cnt2 = Cnt ; Cnt2 < Cnt + 10 ; Cnt2 ++)
    {
      OldValue = 401;
      if (SA_Pulses[Cnt2].Counts0 != 0)
      {
        Value = SA_Pulses[Cnt2].Counts0;
        OldValue = 400 - (short)((double)Value * Limiter);
        WDrawBox (&CWindow, TRUE, DRED, Cnt3 + XOffset, YOffset + 400, Cnt3 + XOffset + 2, YOffset + OldValue);
      }
      if (SA_Pulses[Cnt2].Counts1 != 0)
      {
        Value = SA_Pulses[Cnt2].Counts1;
        WDrawBox (&CWindow, TRUE, DGREEN, Cnt3 + XOffset, YOffset + OldValue - 1, Cnt3 + XOffset + 2,
                  YOffset + OldValue - (short)((double)Value * Limiter) - 1);
      }
    }
  }
  for (Cnt = 0, Cnt3 = 0 ; Cnt < Width - 1 ; Cnt += 10, Cnt3 ++)                                 /* Super-impose bit-spread graph */
  {
    for (Cnt2 = Cnt ; Cnt2 < Cnt + 10 ; Cnt2 ++)
      if (SA_Joined[Cnt2] > 0)
      {
        BitColor = (Cnt < _Average (Entry->Timing.LengthZeroBitPulse, Entry->Timing.LengthOneBitPulse) ? DBLUE : LBLUE);
        Value = SA_Joined[Cnt2];
        WDrawBox (&CWindow, TRUE, BitColor, Cnt3 + XOffset - 1, YOffset + 400, Cnt3 + XOffset + 1,
                  YOffset + 400 - (short)((double)Value * Limiter));
      }
    }
  Num0Bits /= 2;
  Num1Bits /= 2;
  Wprintf (&CWindow, 41, Width / 60 - 8, "T-states * 100");
  Wprintf (&CWindow, 42,  1, "{+1Num 0-bits}N %ld", Num0Bits);
  Wprintf (&CWindow, 43,  1, "{+1Num 1-bits}N %ld", Num1Bits);
  Wprintf (&CWindow, 42, 21, "{+1Total bits}N %ld (%ld:%ld bytes)", Num0Bits + Num1Bits,
           (Num0Bits + Num1Bits) / 8, (Num0Bits + Num1Bits) % 8);
  Wprintf (&CWindow, 43, 21, "{+1Noise bits}N %ld (%ld)", SA_BadBits, BitRepairStack.NumUsed);
  XOffset = Width / 60 - 13;
  YOffset = Width / 10 - 92;
  WDrawBox (&CWindow, FALSE, DBLACK, YOffset - 5, 7, YOffset + 128, 55);                                           /* Draw legend */
  WDrawBox (&CWindow, TRUE, DRED,   YOffset, 15, YOffset + 9, 18);
  WDrawBox (&CWindow, TRUE, DGREEN, YOffset, 25, YOffset + 9, 28);
  WDrawBox (&CWindow, TRUE, DBLUE,  YOffset, 35, YOffset + 9, 38);
  WDrawBox (&CWindow, TRUE, LBLUE,  YOffset, 45, YOffset + 9, 48);
  Wprintf (&CWindow, 1, XOffset, "Pulse led to 0-bit");
  Wprintf (&CWindow, 2, XOffset, "Pulse led to 1-bit");
  Wprintf (&CWindow, 3, XOffset, "0-bit spread");
  Wprintf (&CWindow, 4, XOffset, "1-bit spread");
  while (!Done)
  {
    if (HandleWindow (&CWindow, &ButtonP) == WINH_CLOSE ||
        GlobalKey == K_ESC || GlobalKey == K_ENTER)
      Done = TRUE;
    MouseStatus ();
  }
  DestroyWindow (&CWindow);
  GlobalKey = 0x0000;
}

/**********************************************************************************************************************************/
/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> START OF THE LIBRARY ROUTINES <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< */
/**********************************************************************************************************************************/

bool DecodeDirectSample (short WhichWindow, int Fh, dword SampleRate, dword BlockLength, dword *FileIndex, bool SkipSoundContinue,
                         bool *Cancelled, dword TPXMSIndex)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/*         `Fh' points to the open tape file, `SampleRate' is the found rate in Hz, `BlockLength' is the length of the sample.    */
/*         `FileIndex' holds the current file index.                                                                              */
/* Post  : The sample has been decoded into readable blocks where possible, which have been inserted in the tape directory.       */
/*         The `FileIndex' has been adjusted to after the block.                                                                  */
/*         If the user hit the `Cancel' button in the percentage requester, `Cancelled' is returned TRUE.                         */
/* Import: _OpenReadBuffer, _ReadState, _ReadPulse, _PulseDiff, _GetCurrentFileIndex, _InsertClickingPilot,                       */
/*         _InsertSpeedLock4567Pilot, _ProcessInterBlockPause, _InsertSpeedLock234567MidSync, _InsertGroupEnd,                    */
/*         _GetBitTimingValues, _GetSLSyncMatchValue, _InsertDigitalIntegrationPilot, _AddEntryToTape, _AddSpaceToTape,           */
/*         _PrintSoundAnalysis.                                                                                                   */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *Entry;
  struct TapeEntry_s *DirP;
  bool                AllocateTapeBlock = TRUE;
  bool                CurrentState = TRUE;                                                            /* First pulse must go down */
  bool                OldState;
  dword               NumPulses;                                                                   /* The actual number of pulses */
  dword               Num0Bits;
  dword               Num1Bits;
  long                TotalTStates;                                                              /* The actual number of T-states */
  long                Total0TStates;                                                                            /* The number / 2 */
  long                Total1TStates;
  dword               CntBits;
  dword               DBlLength;
  long                PulseLength;
  long                PulseLength2;
  long                JoinedPulseLength;
  long                PulseAverage;
  long                MinBitPulseLength;
  long                MinSinglePulseLength;
  short               MinBitPulseDivider = 3;
  long                MaxBitPulseLength;
  long                EndMarkerLength;
  bool                EndFound;                                                            /* Marks the change of a loading state */
  bool                EndMarkerFound;                                                  /* If TRUE, the entire block has been done */
  bool                PilotAgain = TRUE;
  bool                FirstPilot = TRUE;
  dword               DataBlockIndex;
  dword               NextBlock;
  dword               PilotStartIndex;                                        /* File position of the first accepted pilot sample */
  dword               PilotSeekIndex;                                                    /* A copy of the index used for the PReq */
  word                ByteIndex;                                /* Index into WorkBuffer (where the decoded bytes will be stored) */
  byte                BitCounter;                                                      /* Used to rotate the read bits into place */
  byte                BitShift;
  bool                DoingFlagByte;
  bool                DoingParityByte;
  long                PauseLength = 0;                                                            /* Length of endmarker pulse(s) */
  dword               XMSIndex;
  bool                ClickingPilot = FALSE;
  bool                PilotWritten = FALSE;
  bool                DoingNextSubBlock = FALSE;
  bool                SLTurbo = FALSE;
  bool                DIFirstPilot = TRUE;
  bool                PartialEndByteDropped = FALSE;
  dword               CntS;
  double              Weighed0;
  double              Weighed1;

  _SkipVOCSoundContinue = SkipSoundContinue;
  HandlingZ80EmRAW = FALSE;
  HandlingCSW = FALSE;
  if (TapeWindow[WhichWindow].FileType == FT_ACORNRAW)
    HandlingZ80EmRAW = TRUE;
  else if (TapeWindow[WhichWindow].FileType == FT_CSW)
    HandlingCSW = TRUE;
  SLBlockCnt = 0;
  _OpenReadBuffer (WhichWindow, Fh, *FileIndex, BlockLength, SampleRate, TPXMSIndex);               /* Setup the file read buffer */
  if (SampleRate <= 22100)                                                                 /* Watch out for too low sample rates! */
    MinBitPulseDivider = 4;
  *Cancelled = FALSE;
  while (PilotAgain)
  {
    BitRepairStack.NumUsed = 0;                                                              /* Clear look-ahead bit repair stack */
    BitRepairStack.Index   = 0;
    if (ShowSoundStreamInfo)
    {
      SA_BadBits = 0;
      for (CntS = 0 ; CntS < SA_MAXPULSELENGTH ; CntS ++)
      {
        SA_Joined[CntS] = 0;
        SA_Pulses[CntS].Counts0 = 0;
        SA_Pulses[CntS].Counts1 = 0;
      }
    }
    if (AllocateTapeBlock)
    {
      if (!FirstPilot)
        _RestoreReadState (&_EndBackupReadBuffer);
      if ((Entry = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
        FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
      memset (Entry, 0, sizeof (struct TapeEntry_s));
      Entry->ComboSelected = 0x00;
      Entry->BlockTypeID = 0x11;                                                                          /* Assume Turbo Loading */
      Entry->Timing.SampleRate = SampleRate;
      Entry->Timing.NumBitsInFirstByte = 8;
      AllocateTapeBlock = FALSE;
      if (!UpdatePReq (*FileIndex, "Status: Searching for pilot tone ..."))
      { *Cancelled = TRUE; free (Entry); return (TRUE); }
      ClickingPilot = FALSE;
      PilotWritten = FALSE;
      DoingNextSubBlock = FALSE;
      PilotIndex = 0;
      SLTurbo = FALSE;
      DIFirstPilot = TRUE;
    }

    __PILOTAGAIN:

    PilotSeekIndex = _GetCurrentFileIndex () - 1;                                                        /* Tell where we are now */
    if (!UpdatePReq (PilotSeekIndex, NULL))                                /* Just check Cancel (if decoding doesn't work at all) */
    { *Cancelled = TRUE; free (Entry); return (TRUE); }

    if (!ClickingPilot)
    {
      if (!HandlingZ80EmRAW && !HandlingCSW)
      if (!_ReadState (&CurrentState))
      { __SKIPNOISE: *FileIndex += BlockLength;
                     free (Entry);
                     if (lseek (Fh, *FileIndex, SEEK_SET) == -1)
                       return (FALSE);
                     if (!FirstPilot && TapeWindow[WhichWindow].Directory != NULL)          /* Noise was at the end of the tape ? */
                     {
                       DirP = TapeWindow[WhichWindow].Directory;
                       while (DirP->Next != NULL)                                      /* Add the end of the tape as pause length */
                         DirP = DirP->Next;                                          /* to the last block, if it uses pauseblocks */
                       if (DirP->BlockTypeID == 0x10 || DirP->BlockTypeID == 0x11 || DirP->BlockTypeID == 0x14)
                         DirP->PauseLength += ((*FileIndex - NextBlock) * 1000) / Entry->Timing.SampleRate;
                     }
                     return (TRUE);
      }
      CurrentState = !CurrentState;
      if (_ReadPulse (&CurrentState) < 0)  goto __SKIPNOISE;                                                /* Skip initial pulse */
      PilotStartIndex = _GetCurrentFileIndex () - 1;                                         /* Keep the start index of the pilot */
    }
    if ((TotalTStates = _ReadPulse (&CurrentState)) < 0)  goto __SKIPNOISE;         /* Read 2 pulses, to find the initial average */
    if ((PulseLength = _ReadPulse (&CurrentState)) < 0)  goto __SKIPNOISE;
    NumPulses = 2;                                                                                    /* Start decoding the pilot */
    TotalTStates += PulseLength;
    PulseAverage = TotalTStates / 2;
    EndFound = FALSE;
    while (!EndFound)
    {
      if ((PulseLength = _ReadPulse (&CurrentState)) < 0)  goto __SKIPNOISE;                                      /* Read a pulse */
      if (ApplyFilter && PulseLength < 300)                                                           /* Noise spike in the pilot */
      {
        if (NumPulses < 50)                                                               /* We're not actually in a good pilot ? */
          goto __PILOTAGAIN;
        TotalTStates += PulseLength;
        if ((PulseLength = _ReadPulse (&CurrentState)) < 0)  goto __SKIPNOISE;                                 /* Join next pulse */
        TotalTStates += PulseLength;
        PulseAverage = (short)(TotalTStates / ++ NumPulses);                                                    /* Update average */
      }
      else if ((DecodingScheme == DS_SLOCK1 || DecodingScheme == DS_SLOCK2 || DecodingScheme == DS_SLOCK3) &&
          (_PulseDiff (PulseAverage, PulseLength) > 20))                                  /* The pilot pulse and sync are 25% off */
        EndFound = TRUE;
      else if (PulseLength < MinPilotPulseLength || PulseLength > MaxPilotPulseLength)                         /* Skip some noise */
        EndFound = TRUE;
      else if (_PulseDiff (PulseAverage, PulseLength) > ThresholdPilotPulse)                                    /* Too much off ? */
        EndFound = TRUE;                                                                               /* Yep, maybe end of pilot */
      else
      {
        TotalTStates += PulseLength;
        PulseAverage = (short)(TotalTStates / ++ NumPulses);                                                    /* Update average */
      }
    }
    if (NumPulses >= UsedMinNumPilotPulses ||                                                      /* Enough pulses for a pilot ? */
        (NumPulses > 200 &&                                            /* Alkatraz and Bleepload have pilots of around 240 pulses */
         (DecodingScheme == DS_ALKATRAZ || DecodingScheme == DS_BLEEP)) ||
        (NumPulses > 10 && DecodingScheme == DS_DINAMIC) ||                             /* Dinamic has pilots of around 13 pulses */
        ((NumPulses >= 180 && NumPulses <= 250) &&
         (DecodingScheme == DS_SLOCK1 || DecodingScheme == DS_SLOCK2 || DecodingScheme == DS_SLOCK3)) ||    /* (Clicking pilots ) */
        (NumPulses >= 200 && DecodingScheme == DS_DIGITAL && !DIFirstPilot))   /* Digital Integration has 2nd pilot of 256 pulses */
    {
      if (PulseLength > PulseAverage && DecodingScheme == DS_BLEEP)                          /* BleepLoad has 2 pilot-only blocks */
      {
        if (!FirstPilot && TapeWindow[WhichWindow].Directory != NULL)
          _ProcessInterBlockPause (WhichWindow, PilotStartIndex - NextBlock, Entry->Timing.SampleRate);
        Entry->BlockTypeID = 0x12;                                                /* Store pilot-only blocks as 'Pure Tone Block' */
        Entry->Timing.LengthPilotPulse = PulseAverage;
        Entry->Timing.NumPilotPulses = NumPulses;
        GetBlockTypeInfo (Entry->BlockTypeID, &(Entry->IsTapeData), &(Entry->IsIdentifiedData), NULL, NULL, NULL);
        Entry->Next = NULL;
        _AddEntryToTape (WhichWindow, Entry);
        BlockLength -= (_GetCurrentFileIndex () - 1 - *FileIndex);
        *FileIndex = _GetCurrentFileIndex () - 1;
        _BackupReadState (&_EndBackupReadBuffer);                                               /* Store position as end-of-block */
        if (BlockLength > 0)                                                                                       /* Clean end ? */
        {
          AllocateTapeBlock = TRUE;                                                                     /* Then there may be more */
          PilotAgain = TRUE;
          FirstPilot = FALSE;
        }
      }
      else if (PulseLength > PulseAverage && DecodingScheme == DS_DIGITAL && DIFirstPilot)       /* DI 2nd pilot has larger pulse */
      {
        PilotBuffer[0] = PulseAverage;                                                      /* Store the 1st pilot somewhere safe */
        PilotBuffer[1] = NumPulses;
        DIFirstPilot = FALSE;                                                                           /* Signal: read 2nd pilot */
        SLTurbo = TRUE;                                                                                    /* Signal: turbo block */
        PilotAgain = TRUE;
        FirstPilot = FALSE;
      }
      else if ((DecodingScheme == DS_SLOCK1 || DecodingScheme == DS_SLOCK2|| DecodingScheme == DS_SLOCK3) &&/* 'Click' in pilot ? */
               PulseLength < PulseAverage && NumPulses < 1000)                                  /* (Not in the first BASIC parts) */
      {
        ClickingPilot = TRUE;
        if (PilotIndex == 49)                                                                            /* More than 50 clicks ? */
        {
          memcpy (&PilotBuffer[0], &PilotBuffer[1], 49 * sizeof (long));                                 /* Shift the buffer back */
          PilotIndex = 48;
        }
        PilotBuffer[PilotIndex ++] = NumPulses;
        if (_ReadPulse (&CurrentState) < 0)  goto __SKIPNOISE;                                            /* Drop 2nd click pulse */
      }
      else if (PulseLength < PulseAverage ||                                           /* Shorter ? Than it's the first sync half */
               DecodingScheme == DS_SLOCK1 || DecodingScheme == DS_SLOCK2 || DecodingScheme == DS_SLOCK3)
      {                                                                           /* SpeedLock 1 has a sync larger than the pilot */
        if ((DecodingScheme == DS_SLOCK1 || DecodingScheme == DS_SLOCK2 || DecodingScheme == DS_SLOCK3) &&
            ((ClickingPilot && PilotIndex < 16) ||                                                  /* We need at least 16 clicks */
             (!ClickingPilot && NumPulses < 1000)))                                          /* The first click can contain noise */
          goto __PILOTAGAIN;
        if (!UpdatePReq (*FileIndex, "Status: Analysing data ..."))                              /* Update the progress indicator */
        { *Cancelled = TRUE; free (Entry); return (TRUE); }
        PilotAgain = FALSE;
        if (!FirstPilot && TapeWindow[WhichWindow].Directory != NULL)                   /* Handle all space in between the blocks */
          _ProcessInterBlockPause (WhichWindow, PilotStartIndex - NextBlock, Entry->Timing.SampleRate);
        if ((DecodingScheme == DS_SLOCK1 || DecodingScheme == DS_SLOCK2 || DecodingScheme == DS_SLOCK3) && ClickingPilot)
        {
          if (PilotIndex == 49)                                                                        /* Include this last pilot */
          {
            memcpy (&PilotBuffer[0], &PilotBuffer[1], 49 * sizeof (long));
            PilotIndex = 48;
          }
          PilotBuffer[PilotIndex ++] = NumPulses;
          if (_ReadPulse (&CurrentState) < 0)  goto __SKIPNOISE;                                /* Read (and drop) 2nd sync pulse */
        }
        Entry->Timing.LengthPilotPulse = PulseAverage;
        if (NumPulses < 3000 && DecodingScheme == DS_ALKATRAZ)                            /* Alkatraz pilots must be forced a bit */
          Entry->Timing.NumPilotPulses = 3000;
        else
          Entry->Timing.NumPilotPulses = NumPulses;
        if (DecodingScheme == DS_ROM || DecodingScheme == DS_BLEEP || DecodingScheme == DS_DINAMIC ||
            DecodingScheme == DS_SULOAD || DecodingScheme == DS_YSLOAD || DecodingScheme == DS_PLAYERS ||
            DecodingScheme == DS_POPS)
          Entry->Timing.NumSyncPulses = 2;
        else if (((DecodingScheme == DS_SLOCK1 || DecodingScheme == DS_SLOCK2 || DecodingScheme == DS_SLOCK3) && ClickingPilot) ||
                 (DecodingScheme == DS_DIGITAL && SLTurbo))
          Entry->Timing.NumSyncPulses = 0;                                              /* No more syncs when doing SpeedLock 1/2 */
        else
          Entry->Timing.NumSyncPulses = NumSyncPulses;                                                  /* (Only possibility now) */
        Entry->Timing.LengthSyncPulse[0] = PulseLength;                                                           /* (Just read!) */
        NumPulses = 1;                                                                /* Read all syncs after the 1st (fixed) one */
        while (NumPulses < Entry->Timing.NumSyncPulses)
          Entry->Timing.LengthSyncPulse[NumPulses ++] = _ReadPulse (&CurrentState);                         /* Finish sync signal */
        TotalTStates = 0;
        for (NumPulses = 0 ; NumPulses < Entry->Timing.NumSyncPulses ; NumPulses ++)              /* Calculate average sync value */
          TotalTStates += Entry->Timing.LengthSyncPulse[NumPulses];
        if (Entry->Timing.NumSyncPulses > 0)                                     /* Make all sync pulses the average pulse length */
        {
          Entry->Timing.LengthSyncPulse[0] = (word)((double)TotalTStates / Entry->Timing.NumSyncPulses + 0.5);
          for (NumPulses = 1 ; NumPulses < Entry->Timing.NumSyncPulses ; NumPulses ++)
            Entry->Timing.LengthSyncPulse[NumPulses] = Entry->Timing.LengthSyncPulse[0];
        }

        __NEXTSUBBLOCK:

        OldState = CurrentState;                                                                           /* Mark current state */
        DataBlockIndex = _GetCurrentFileIndex () - 1;                                                    /* Mark current location */
        _BackupReadState (&_DataBackupReadBuffer);
        TotalTStates = _ReadPulse (&CurrentState);
        PulseLength = _ReadPulse (&CurrentState);
        NumPulses = 2;                                          /* Start figuring out which ones are 0 and which are 1 bit pulses */
        Num0Bits = 1;
        Num1Bits = 0;
        TotalTStates += PulseLength;
        MinSinglePulseLength = MinBitPulseLength = FMinBitPulseLength;                   /* Start with the provided minimum value */
        if (DecodingScheme == DS_ROM || DecodingScheme == DS_BLEEP ||                 /* Do we already 'know' the timing values ? */
            DecodingScheme == DS_DINAMIC)
        {
          _GetBitTimingValues (&(Entry->Timing.LengthZeroBitPulse), &(Entry->Timing.LengthOneBitPulse), &PulseAverage);
          EndMarkerLength = (long)(ThresholdEndMarker * Entry->Timing.LengthOneBitPulse);
          MinBitPulseLength = Entry->Timing.LengthZeroBitPulse;
          MinSinglePulseLength = _Max (FMinBitPulseLength, MinBitPulseLength / MinBitPulseDivider);
          MaxBitPulseLength = _Max (EndBitPulseLength, (long)(ThresholdSingleEndMarker * Entry->Timing.LengthOneBitPulse));
        }
        else if (((DecodingScheme == DS_SLOCK1 || DecodingScheme == DS_SLOCK2 || DecodingScheme == DS_SLOCK3) && ClickingPilot) ||
                 ((DecodingScheme == DS_SLOCK4 || DecodingScheme == DS_SLOCK5 || DecodingScheme == DS_SLOCK6 ||
                   DecodingScheme == DS_SLOCK7) && DoingNextSubBlock))
        {
          _GetBitTimingValues (&(Entry->Timing.LengthZeroBitPulse), &(Entry->Timing.LengthOneBitPulse), &PulseAverage);
          if (DecodingScheme == DS_SLOCK1)
          {
            EndMarkerLength = (long)(ThresholdEndMarker * Entry->Timing.LengthOneBitPulse);
            MaxBitPulseLength = _Max (EndBitPulseLength, (long)(ThresholdSingleEndMarker * Entry->Timing.LengthOneBitPulse));
          }
          else
          {
            EndMarkerLength = (long)(ThresholdEndMarker * 1750);
            MaxBitPulseLength = _Max (EndBitPulseLength, (long)(ThresholdSingleEndMarker * 1750));          /* '1' in middle sync */
          }
          MinBitPulseLength = Entry->Timing.LengthZeroBitPulse;
          MinSinglePulseLength = _Max (FMinBitPulseLength, MinBitPulseLength / MinBitPulseDivider);
        }
        else if ((DecodingScheme == DS_SLOCK4 || DecodingScheme == DS_SLOCK5 || DecodingScheme == DS_SLOCK6 ||
                  DecodingScheme == DS_SLOCK7) && !SLTurbo)
        {
          Entry->Timing.LengthZeroBitPulse = PulseAverage = TotalTStates / 2;
          Entry->Timing.LengthOneBitPulse = 0;
          Total0TStates = Entry->Timing.LengthZeroBitPulse;
          Total1TStates = 0;
          EndMarkerLength = MaxBitPulseLength = (long)(ThresholdEndMarker * 2144);                    /* Max in the composed sync */
          MidSyncBuffer[0] = MidSyncBuffer[1] = _GetSLSyncMatchValue (PulseAverage);
        }
        else
        {
          Entry->Timing.LengthZeroBitPulse = PulseAverage = TotalTStates / 2;
          Entry->Timing.LengthOneBitPulse = 0;
          Total0TStates = Entry->Timing.LengthZeroBitPulse;
          Total1TStates = 0;
          EndMarkerLength = MaxBitPulseLength = EndBitPulseLength;
        }
        EndMarkerFound = EndFound = FALSE;
        PauseLength = 0;
        while (!EndFound)
        {
          if ((PulseLength = _ReadPulse (&CurrentState)) < 0)                                                    /* Read 2 pulses */
          { PauseLength = 0; EndMarkerFound = EndFound = TRUE; }
          else if ((PulseLength2 = _ReadPulse (&CurrentState)) < 0)
          { PauseLength = PulseLength; EndMarkerFound = EndFound = TRUE; }
          else if (PulseLength > MaxBitPulseLength || PulseLength2 > MaxBitPulseLength)               /* End-of-block 'timeout' ? */
          { PauseLength = PulseLength + PulseLength2; EndMarkerFound = EndFound = TRUE; }
          else if (ApplyFilter && ((PulseLength + PulseLength2) < MinBitPulseLength ||                           /* Way too few ? */
                   PulseLength < MinSinglePulseLength || PulseLength2 < MinSinglePulseLength))
          {
            if (PulseLength < MinSinglePulseLength && PulseLength2 > MinSinglePulseLength)
              if (BitRepairStack.NumUsed < MAXREPAIRBITS)                   /* It's ONLY the first of the two pulses that fails ? */
                BitRepairStack.Locations[BitRepairStack.NumUsed ++] = NumPulses / 2 + 1;  /* Then it belongs to the PREVIOUS bit! */
            PulseLength += PulseLength2;
            if ((PulseLength2 = _ReadPulse (&CurrentState)) < 0)                       /* Skip the noise by reading 2 more pulses */
            { PauseLength = PulseLength; EndMarkerFound = EndFound = TRUE; }                               /* And add them all up */
            else
            {
              PulseLength += PulseLength2;
              if ((PulseLength2 = _ReadPulse (&CurrentState)) < 0)
              { PauseLength = PulseLength; EndMarkerFound = EndFound = TRUE; }
            }
          }
          else if (_Average (PulseLength, PulseLength2) > EndMarkerLength)
          { PauseLength = PulseLength + PulseLength2; EndMarkerFound = EndFound = TRUE; }
          else if (TestDoubleBitPulse)                                                            /* Test the individual pulses ? */
            if (_PulseDiff (PulseLength, PulseLength2) > ThresholdDoubleBitPulse)
              EndFound = TRUE;                                                                      /* Should be almost identical */

          /* If the timing values are fixed (DS_ROM or DS_BLEEP), this routine is used only to determine the end of the block     */
          /* SpeedLocks 4-7 are checked for an initial composed sync pulse; SpeedLocks 2-7 are checked for a middle sync signal   */
          /* Otherwise, this routine is used to determine these values so we need to continue.                                    */

          if (!EndFound && (DecodingScheme == DS_ROM || DecodingScheme == DS_BLEEP || DecodingScheme == DS_DINAMIC ||
              (DecodingScheme == DS_SLOCK1 && ClickingPilot)))
          {
            JoinedPulseLength = _Average (PulseLength, PulseLength2);                                  /* Find the joined average */
            NumPulses += 2;
            if ((LossyDetection  && (labs (JoinedPulseLength - Entry->Timing.LengthZeroBitPulse) >
                                     labs (JoinedPulseLength - Entry->Timing.LengthOneBitPulse))) ||
                (!LossyDetection && (JoinedPulseLength > PulseAverage)))
              Num1Bits ++;
            else
              Num0Bits ++;
          }
          else if (!EndFound &&
                   ((DecodingScheme == DS_SLOCK2 || DecodingScheme == DS_SLOCK3) && ClickingPilot) ||
                   ((DecodingScheme == DS_SLOCK4 || DecodingScheme == DS_SLOCK5 || DecodingScheme == DS_SLOCK6 ||
                     DecodingScheme == DS_SLOCK7) && SLTurbo))
          {
            JoinedPulseLength = _Average (PulseLength, PulseLength2);                                  /* Find the joined average */
            NumPulses += 2;
            if ((LossyDetection  && (labs (JoinedPulseLength - Entry->Timing.LengthZeroBitPulse) >
                                     labs (JoinedPulseLength - Entry->Timing.LengthOneBitPulse))) ||
                (!LossyDetection && (JoinedPulseLength > PulseAverage)))
              Num1Bits ++;
            else
              Num0Bits ++;
            if (NumPulses < 42)                                                 /* First 21 bits just fill up the mid sync buffer */
              MidSyncBuffer[NumPulses / 2] = JoinedPulseLength > 1500 ? 1750 : 875;
            else
            {
              memcpy (&MidSyncBuffer[0], &MidSyncBuffer[1], 20 * sizeof (long));                         /* Shift the buffer back */
              MidSyncBuffer[20] = JoinedPulseLength > 1500 ? 1750 : 875;
              if (!memcmp (&MidSyncBuffer[0], &MidSyncMatch[0], 21 * sizeof (long)))                      /* Found the mid sync ? */
              {
                EndFound = TRUE;
                MidSyncFound = TRUE;
                if (Num0Bits > 20)                                                              /* Remove the sync from the count */
                  Num0Bits -= 21;
                else
                { Num1Bits -= (21 - Num0Bits); Num0Bits = 0; }
              }
            }
          }
          else if (!EndFound)                                                                                       /* Still ok ? */
          {
            JoinedPulseLength = _Average (PulseLength, PulseLength2);                                  /* Find the joined average */
            TotalTStates += PulseLength + PulseLength2;
            NumPulses += 2;
            if (Entry->Timing.LengthOneBitPulse == 0)                                                          /* Still unknown ? */
            {
              if (_PulseDiff (Entry->Timing.LengthZeroBitPulse, JoinedPulseLength) > ThresholdBitPulse)  /* Found the other timer */
              {
                if (JoinedPulseLength > Entry->Timing.LengthZeroBitPulse)                                    /* Found the '1' now */
                {
                  Total1TStates = Entry->Timing.LengthOneBitPulse = JoinedPulseLength;
                  Num1Bits = 1;
                }
                else                                                                                         /* Found the '0' now */
                {
                  Entry->Timing.LengthOneBitPulse = Entry->Timing.LengthZeroBitPulse;
                  Total1TStates = Total0TStates;
                  Total0TStates = Entry->Timing.LengthZeroBitPulse = JoinedPulseLength;
                  Num1Bits = Num0Bits;
                  Num0Bits = 1;
                }
                PulseAverage = _Average (Entry->Timing.LengthZeroBitPulse, Entry->Timing.LengthOneBitPulse); /* Keep true average */
              }
              else
              {
                Total0TStates += JoinedPulseLength;
                Entry->Timing.LengthZeroBitPulse = (word)(Total0TStates / ++ Num0Bits);
                PulseAverage = (long)(TotalTStates / NumPulses);                                                /* Update average */
              }
            }
            else
            {
              if ((LossyDetection  && (labs (JoinedPulseLength - Entry->Timing.LengthZeroBitPulse) >
                                       labs (JoinedPulseLength - Entry->Timing.LengthOneBitPulse))) ||
                  (!LossyDetection && (JoinedPulseLength > PulseAverage)))                                 /* Found 'another' 1 ? */
              {
                Total1TStates += JoinedPulseLength;
                Entry->Timing.LengthOneBitPulse = (word)(Total1TStates / ++ Num1Bits);
                if ((DecodingScheme == DS_PLAYERS || DecodingScheme == DS_SULOAD || DecodingScheme == DS_YSLOAD ||
                     DecodingScheme == DS_POPS) && Num0Bits > 40 && Num1Bits > 40)
                {
                  if (_PulseDiff (NS_LEN0BIT, Entry->Timing.LengthZeroBitPulse) <= ThresholdStandardLoading &&
                      _PulseDiff (NS_LEN1BIT, Entry->Timing.LengthOneBitPulse) <= ThresholdStandardLoading)
                  { Entry->Timing.LengthZeroBitPulse = NS_LEN0BIT ; Entry->Timing.LengthOneBitPulse = NS_LEN1BIT; }
                  else
                  {
                    _GetBitTimingValues (&(Entry->Timing.LengthZeroBitPulse), &(Entry->Timing.LengthOneBitPulse), &PulseAverage);
                    SLTurbo = TRUE;
                  }
                  EndMarkerLength = _Min (EndMarkerLength, (long)(1.2 * Entry->Timing.LengthOneBitPulse));
                  MaxBitPulseLength = _Max (EndBitPulseLength, (long)(ThresholdSingleEndMarker * Entry->Timing.LengthOneBitPulse));
                }
                else if (Num1Bits > 4)                                        /* A pretty good average is found after only 4 bits */
                {
                  EndMarkerLength = (long)(ThresholdEndMarker * Entry->Timing.LengthOneBitPulse);
                  if (DecodingScheme == DS_SULOAD || DecodingScheme == DS_YSLOAD)      /* SU Loader has all blocks glued together */
                    EndMarkerLength = _Min (EndMarkerLength, (long)(1.2 * Entry->Timing.LengthOneBitPulse));
                  MaxBitPulseLength = _Max (EndBitPulseLength, (long)(ThresholdSingleEndMarker * Entry->Timing.LengthOneBitPulse));
                }
              }
              else
              {
                Total0TStates += JoinedPulseLength;
                Entry->Timing.LengthZeroBitPulse = (word)(Total0TStates / ++ Num0Bits);
                if ((DecodingScheme == DS_PLAYERS || DecodingScheme == DS_SULOAD || DecodingScheme == DS_YSLOAD ||
                     DecodingScheme == DS_POPS) && Num0Bits > 40 && Num1Bits > 40)
                {
                  if (_PulseDiff (NS_LEN0BIT, Entry->Timing.LengthZeroBitPulse) <= ThresholdStandardLoading &&
                      _PulseDiff (NS_LEN1BIT, Entry->Timing.LengthOneBitPulse) <= ThresholdStandardLoading)
                  { Entry->Timing.LengthZeroBitPulse = NS_LEN0BIT ; Entry->Timing.LengthOneBitPulse = NS_LEN1BIT; }
                  else
                  {
                    _GetBitTimingValues (&(Entry->Timing.LengthZeroBitPulse), &(Entry->Timing.LengthOneBitPulse), &PulseAverage);
                    SLTurbo = TRUE;
                  }
                  MinBitPulseLength = Entry->Timing.LengthZeroBitPulse;
                  MinSinglePulseLength = _Max (FMinBitPulseLength, MinBitPulseLength / MinBitPulseDivider);
                }
                else if (Num0Bits > 4)
                {
                  MinBitPulseLength = Entry->Timing.LengthZeroBitPulse;
                  MinSinglePulseLength = _Max (FMinBitPulseLength, MinBitPulseLength / MinBitPulseDivider);
                }
              }
              PulseAverage = _Average (Entry->Timing.LengthZeroBitPulse, Entry->Timing.LengthOneBitPulse);   /* Keep true average */
            }
            if (((DecodingScheme == DS_SLOCK4 || DecodingScheme == DS_SLOCK5 || DecodingScheme == DS_SLOCK6 ||
                  DecodingScheme == DS_SLOCK7) && !SLTurbo) &&
                NumPulses <= 14)                                                       /* Composed sync signal has 14 half-pulses */
            {
              MidSyncBuffer[NumPulses - 2] = MidSyncBuffer[NumPulses - 1] = _GetSLSyncMatchValue (JoinedPulseLength);
              if (NumPulses == 14)                                                                        /* Read all 14 pulses ? */
              {
                if (!memcmp (&MidSyncBuffer[0], &ComposedSync[2], 14 * sizeof (long)))             /* Matches the composed sync ? */
                {
                  SLTurbo = TRUE;                                                                  /* Signal: reading turbo block */
                  DataBlockIndex = _GetCurrentFileIndex () - 1;                                             /* Step over the sync */
                  _BackupReadState (&_DataBackupReadBuffer);
                  TotalTStates = _ReadPulse (&CurrentState);                                            /* Re-start the datablock */
                  PulseLength = _ReadPulse (&CurrentState);
                  TotalTStates += PulseLength;
                  NumPulses = 2;                                /* Start figuring out which ones are 0 and which are 1 bit pulses */
                  Num0Bits = 1;
                  Num1Bits = 0;
                  _GetBitTimingValues (&(Entry->Timing.LengthZeroBitPulse), &(Entry->Timing.LengthOneBitPulse), &PulseAverage);
                  MinBitPulseLength = Entry->Timing.LengthZeroBitPulse;
                  MinSinglePulseLength = _Max (FMinBitPulseLength, MinBitPulseLength / MinBitPulseDivider);
                  EndMarkerLength = (long)(ThresholdEndMarker * 1750);
                  MaxBitPulseLength = _Max (EndBitPulseLength, (long)(ThresholdSingleEndMarker * 1750));    /* '1' in middle sync */
                }
              }
            }
          }
        }

        /* If `NumPulses' is 2 (initial value), the block only had a pilot and sync signal, no data body. This is acceptable.    */
        /* If `Num1Bits' = 0, the block contains only the same bits, either all '0's or '1's, so we don't know the timing values */
        /* In this case, let's assume that it were '0's, unless the block appears to be 'Standard Loading'.                      */
        /* So, at this point, we have the timer averages for a '0' and a '1' bit and the total number of bits in the block       */
        /* Also, the length of the after-block initial pause is 'PauseLength' if the block terminated correctly (EndMarkerFound) */

        NextBlock = _GetCurrentFileIndex () - 1;
        BlockLength -= (DataBlockIndex - *FileIndex);
        *FileIndex = DataBlockIndex;
        _BackupReadState (&_EndBackupReadBuffer);                                               /* Store position of end-of-block */
        _RestoreReadState (&_DataBackupReadBuffer);                                                       /* Go back to the start */
        CurrentState = OldState;
        if (!UpdatePReq (*FileIndex, "Status: Decoding data ..."))                               /* Update the progress indicator */
        { *Cancelled = TRUE; free (Entry); return (TRUE); }
        CntBits = Num0Bits + Num1Bits;                                                   /* The total number of bits in the block */

        /* If we have bit approximation on, we have to find an average value, so that the length of a '1' bit is indeed twice as  */
        /* much as a '0' bit. The true average would simply be                                                                    */
        /*     PulseAverage = _Average (Entry->Timing.LengthZeroBitPulse * 2, Entry->Timing.LengthOneBitPulse / 2)                */
        /* but we want a weighing factor to reduce the sampling rate errors. A '1' bit contained twice as many samples, making it */
        /* a much more solid value, so we will make a '1' bit pulse value twice as important as a '0' bit pulse.                  */
        /* After finding the average, the new bit timing values can easily be taken from it.                                      */

        if (ApproximateBitTimers && CntBits > 0)                                                            /* Use aproximation ? */
        {
          Weighed0 = ((double)Entry->Timing.LengthZeroBitPulse - (Entry->Timing.LengthOneBitPulse / 2)) * Num0Bits;
          Weighed1 = ((double)Entry->Timing.LengthOneBitPulse - (Entry->Timing.LengthZeroBitPulse * 2)) * Num1Bits / 2;
          if (Weighed0 > 0)
            PulseAverage = _Average (Entry->Timing.LengthZeroBitPulse * 2, Entry->Timing.LengthOneBitPulse / 2) -
                           labs ((long)(((Weighed0 + Weighed1) / (double)CntBits) + 0.5));
          else
            PulseAverage = _Average (Entry->Timing.LengthZeroBitPulse * 2, Entry->Timing.LengthOneBitPulse / 2) +
                           labs ((long)(((Weighed0 + Weighed1) / (double)CntBits) + 0.5));
          Entry->Timing.LengthZeroBitPulse = PulseAverage - (long)(((double)PulseAverage / 3 + 0.5));
          Entry->Timing.LengthOneBitPulse = Entry->Timing.LengthZeroBitPulse * 2;
        }

        if ((DecodingScheme == DS_SLOCK1 || DecodingScheme == DS_SLOCK2 || DecodingScheme == DS_SLOCK3) &&
            ClickingPilot && !PilotWritten)                                           /* SpeedLock 1-3 have a flag of only 6 bits */
        {
          CntBits += 2;                                                                          /* 'Add' 2 bits to the flag byte */
          _InsertClickingPilot (WhichWindow);
          PilotWritten = TRUE;
        }
        else if ((DecodingScheme == DS_SLOCK4 || DecodingScheme == DS_SLOCK5 || DecodingScheme == DS_SLOCK6 ||
                  DecodingScheme == DS_SLOCK7) && SLTurbo && !PilotWritten)
        {
          CntBits += 2;
          _InsertSpeedLock4567Pilot (WhichWindow, Entry->Timing.NumPilotPulses);
          PilotWritten = TRUE;
        }
        else if (DecodingScheme == DS_DIGITAL && SLTurbo)                                    /* Digital Integration turbo block ? */
          _InsertDigitalIntegrationPilot (WhichWindow, Entry);
        if (Num1Bits == 0 && NumPulses > 2)                                                         /* Only 1 type of bit found ? */
                                                                                        /* Look if it could be 'Standard Loading' */
          if (_PulseDiff (NS_LENPILOT, Entry->Timing.LengthPilotPulse) <= ThresholdStandardLoading &&
              Entry->Timing.NumSyncPulses == 2 &&
              //_PulseDiff (NS_LENSYNC1, Entry->Timing.LengthSyncPulse[0]) <= ThresholdStandardLoading &&
              //_PulseDiff (NS_LENSYNC2, Entry->Timing.LengthSyncPulse[1]) <= ThresholdStandardLoading &&
              (_PulseDiff (NS_LEN0BIT, Entry->Timing.LengthZeroBitPulse) <= ThresholdStandardLoading ||
               _PulseDiff (NS_LEN1BIT, Entry->Timing.LengthZeroBitPulse) <= ThresholdStandardLoading))
            if (Entry->Timing.LengthZeroBitPulse > (NS_LEN0BIT + NS_LEN1BIT) / 2)                                      /* It is ! */
              PulseAverage = 0;                                                                  /* And they're actually all '1's */
            else
              PulseAverage = 5000;                                                                            /* They're all '0's */
          else                                                                                     /* It's really 'Turbo Loading' */
            PulseAverage = 5000;                                                                         /* Force them to be '0's */

        if (DecodingScheme == DS_ROM || DecodingScheme == DS_BLEEP ||                                 /* Force Standard Loading ? */
            DecodingScheme == DS_DINAMIC ||
            ((DecodingScheme == DS_PLAYERS || DecodingScheme == DS_SULOAD || DecodingScheme == DS_YSLOAD ||
              DecodingScheme == DS_POPS) && !SLTurbo))
        {                                   /* BleepLoad blocks will still be saved as custom loading, when NumPilotPulses is off */
          Entry->Timing.LengthPilotPulse = NS_LENPILOT;
          Entry->Timing.LengthSyncPulse[0] = NS_LENSYNC1;
          Entry->Timing.LengthSyncPulse[1] = NS_LENSYNC2;
          Entry->Timing.LengthZeroBitPulse = NS_LEN0BIT;
          Entry->Timing.LengthOneBitPulse = NS_LEN1BIT;
          Entry->Flags.IsNormalSpeed = TRUE;
          PulseAverage = _Average (Entry->Timing.LengthZeroBitPulse, Entry->Timing.LengthOneBitPulse);
        }
        else if (DecodingScheme == DS_YSLOAD || DecodingScheme == DS_PLAYERS)    /* YS Loader and Players have normal sync pulses */
        {
          Entry->Timing.LengthSyncPulse[0] = NS_LENSYNC1;
          Entry->Timing.LengthSyncPulse[1] = NS_LENSYNC2;
          _GetBitTimingValues (&(Entry->Timing.LengthZeroBitPulse), &(Entry->Timing.LengthOneBitPulse), &PulseAverage);
        }
        else if (DecodingScheme == DS_POPS)                                    /* Paul Owens has normal syns too, but fixed pilot */
        {
          Entry->Timing.LengthPilotPulse = NS_LENPILOT;
          Entry->Timing.NumPilotPulses = 2196;
          Entry->Timing.LengthSyncPulse[0] = NS_LENSYNC1;
          Entry->Timing.LengthSyncPulse[1] = NS_LENSYNC2;
          _GetBitTimingValues (&(Entry->Timing.LengthZeroBitPulse), &(Entry->Timing.LengthOneBitPulse), &PulseAverage);
        }
        else if (DecodingScheme == DS_SULOAD && SLTurbo)                                       /* SU Loader has fixed sync pulses */
        {
          Entry->Timing.LengthSyncPulse[0] = Entry->Timing.LengthSyncPulse[1] = 655;
          _GetBitTimingValues (&(Entry->Timing.LengthZeroBitPulse), &(Entry->Timing.LengthOneBitPulse), &PulseAverage);
        }
        else
        {
          if (ForceStandardSyncs &&
              !(((DecodingScheme == DS_SLOCK1 || DecodingScheme == DS_SLOCK2 || DecodingScheme == DS_SLOCK3) && ClickingPilot) ||
               ((DecodingScheme == DS_SLOCK4 || DecodingScheme == DS_SLOCK5 || DecodingScheme == DS_SLOCK6 ||
                 DecodingScheme == DS_SLOCK7) && SLTurbo) ||
               (DecodingScheme == DS_DIGITAL && SLTurbo)))                          /* Not for special decoding schemes of course */
          {
            Entry->Timing.LengthSyncPulse[0] = NS_LENSYNC1;
            Entry->Timing.LengthSyncPulse[1] = NS_LENSYNC2;
          }
        }
        Entry->PauseLength = EndMarkerFound ? (PauseLength / 3500) : 0;                               /* (Pause T-states -> ms. ) */
        Entry->BlockLength = CntBits / 8 + (CntBits % 8 ? 1 : 0);                       /* The total number of bytes in the block */
        Entry->Timing.NumBitsInFirstByte = Entry->Timing.NumBitsInLastByte = 8;                                /* Enable all bits */
        if (NumPulses == 2)                                                                                     /* No data body ? */
        {
          Entry->BlockLength = 0;
          Entry->Flags.FlagByteUsed = FALSE;
          Entry->Flags.ParityByteUsed = FALSE;
          TapeWindow[WhichWindow].NumEntries ++;
          ByteIndex = 0;
          DBlLength = 0;
        }
        else
        {
          if (((DecodingScheme == DS_SLOCK2 || DecodingScheme == DS_SLOCK3 || DecodingScheme == DS_SLOCK4 ||
                DecodingScheme == DS_SLOCK5 || DecodingScheme == DS_SLOCK6 || DecodingScheme == DS_SLOCK7) && DoingNextSubBlock) ||
               (DecodingScheme == DS_DIGITAL && SLTurbo))
          {
            Entry->Flags.FlagByteUsed = FALSE;
            Entry->Flags.ParityByteUsed = FALSE;
            DBlLength = Entry->BlockLength;
          }
          else if (((DecodingScheme == DS_SLOCK1 || DecodingScheme == DS_SLOCK2 || DecodingScheme == DS_SLOCK3) && ClickingPilot) ||
                   ((DecodingScheme == DS_SLOCK4 || DecodingScheme == DS_SLOCK5 || DecodingScheme == DS_SLOCK6 ||
                     DecodingScheme == DS_SLOCK7) && SLTurbo))
          {
            Entry->Flags.FlagByteUsed = TRUE;                                                       /* (But is cut off when read) */
            Entry->Flags.ParityByteUsed = FALSE;
            Entry->BlockLength -= 1;
            DBlLength = Entry->BlockLength;
          }
          else
          {
            Entry->BlockLength -= 1;                                                               /* Unaccount for the flag byte */
            Entry->Flags.FlagByteUsed = TRUE;
            if (Entry->BlockLength > 0)                                                                     /* More than 1 byte ? */
            {
              Entry->Flags.ParityByteUsed = TRUE;                                              /* Then we also have a parity byte */
              Entry->BlockLength -= 1;                                                               /* The remaining data length */
              DBlLength = Entry->BlockLength + 2;
            }
            else
              DBlLength = Entry->BlockLength + 1;
          }
          Entry->AllocLength = (DBlLength & 0xFFFFFFFC) + (DBlLength % 4 ? 4 : 0);
          if (Entry->AllocLength > 0)
            _AddSpaceToTape (WhichWindow, Entry->AllocLength);
          TapeWindow[WhichWindow].NumEntries ++;
          XMSIndex = 0;                                                                                 /* Start at the beginning */
          ByteIndex = 0;
          BitCounter = 0;
          if (Entry->Flags.FlagByteUsed)
            DoingFlagByte = TRUE;                                                        /* First byte is of course the flag byte */
          DoingParityByte = FALSE;
          Entry->FlagByte = 0;
          Entry->CalculatedParity = 0;
          *WorkBuffer = 0;
          for (TotalTStates = 0 ; TotalTStates < CntBits ; TotalTStates ++)                                        /* Do all bits */
          {
            PulseLength = _ReadPulse (&CurrentState);
            PulseLength2 = _ReadPulse (&CurrentState);
            if (ApplyFilter)
              if (BitRepairStack.Index < BitRepairStack.NumUsed && TotalTStates == BitRepairStack.Locations[BitRepairStack.Index])
              {                                     /* Check look-ahead bit repair stack: next pulse to be read is a corruption ? */
                BitRepairStack.Index ++;                                     /* Then the next 2 pulses belong to this bit as well */
                JoinedPulseLength = _ReadPulse (&CurrentState); PulseLength2 += JoinedPulseLength;
                JoinedPulseLength = _ReadPulse (&CurrentState); PulseLength2 += JoinedPulseLength;
                if (ShowSoundStreamInfo)
                  SA_BadBits ++;
              }
              else if ((PulseLength + PulseLength2) < MinBitPulseLength ||                                          /* Skip noise */
                  PulseLength < MinSinglePulseLength || PulseLength2 < MinSinglePulseLength)
              {                       /* The second read pulse is the corruption, so the next 2 pulses belong to this bit as well */
                JoinedPulseLength = _ReadPulse (&CurrentState); PulseLength2 += JoinedPulseLength;
                JoinedPulseLength = _ReadPulse (&CurrentState); PulseLength2 += JoinedPulseLength;
                if (ShowSoundStreamInfo)
                  SA_BadBits ++;
              }
            JoinedPulseLength = _Average (PulseLength, PulseLength2);
            if (ShowSoundStreamInfo)
            {
              if (JoinedPulseLength < SA_MAXPULSELENGTH)
                SA_Joined[JoinedPulseLength] += 2;
              if (PulseLength < SA_MAXPULSELENGTH && PulseLength2 < SA_MAXPULSELENGTH)
                if ((LossyDetection  && (labs (JoinedPulseLength - Entry->Timing.LengthZeroBitPulse) >
                                         labs (JoinedPulseLength - Entry->Timing.LengthOneBitPulse))) ||
                    (!LossyDetection && (JoinedPulseLength > PulseAverage)))
                {
                  SA_Pulses[PulseLength].Counts1 ++;
                  SA_Pulses[PulseLength2].Counts1 ++;
                }
                else
                {
                  SA_Pulses[PulseLength].Counts0 ++;
                  SA_Pulses[PulseLength2].Counts0 ++;
                }
            }
            if (DoingFlagByte)
            {
              Entry->FlagByte <<= 1;                                                  /* Make room for this bit, at bitlocation 0 */
              if ((LossyDetection  && (labs (JoinedPulseLength - Entry->Timing.LengthZeroBitPulse) >
                                       labs (JoinedPulseLength - Entry->Timing.LengthOneBitPulse))) ||
                  (!LossyDetection && (JoinedPulseLength > PulseAverage)))                                   /* Is it a '1' bit ? */
                Entry->FlagByte |= 0x01;                                                                         /* Enter the '1' */
              BitCounter ++;
              if (BitCounter == 8 ||                                                                        /* Done entire byte ? */
                 ((((DecodingScheme == DS_SLOCK1 || DecodingScheme == DS_SLOCK2 || DecodingScheme == DS_SLOCK3) && ClickingPilot) ||
                   ((DecodingScheme == DS_SLOCK4 || DecodingScheme == DS_SLOCK5 || DecodingScheme == DS_SLOCK6 ||
                     DecodingScheme == DS_SLOCK7) && SLTurbo)) &&
                  BitCounter == 6))
              {
                Entry->CalculatedParity = Entry->FlagByte;                                 /* Initialize the parity matching byte */
                if (((DecodingScheme == DS_SLOCK1 || DecodingScheme == DS_SLOCK2 || DecodingScheme == DS_SLOCK3) && ClickingPilot) ||
                    ((DecodingScheme == DS_SLOCK4 || DecodingScheme == DS_SLOCK5 || DecodingScheme == DS_SLOCK6 ||
                      DecodingScheme == DS_SLOCK7) && SLTurbo))
                {
                  TotalTStates += 2;                                                  /* 'Step over' the missing bits in the flag */
                  if (DecodingScheme == DS_SLOCK1)                                              /* (Fixed flag for SpeedLock 2-7) */
                  {
                    for (ByteIndex = 0, BitCounter = 0, BitShift = 0x20 ; BitCounter < 6 ; BitCounter ++, BitShift >>= 1)
                    {
                      *((word *)WorkBuffer + (ByteIndex ++)) = (word)(Entry->FlagByte & BitShift) ? 1130 : 565;
                      *((word *)WorkBuffer + (ByteIndex ++)) = (word)(Entry->FlagByte & BitShift) ? 1130 : 565;
                    }
                    CopyToXMS (24, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer, TmpXMSIndex);    /* Store the flag pattern */
                  }
                  Entry->Flags.FlagByteUsed = FALSE;                                       /* (Room + entry were created earlier) */
                  Entry->CalculatedParity = 0;                                                     /* Start counting at next byte */
                }
                BitCounter = 0;
                ByteIndex = 0;                                                       /* The next byte will be the first data byte */
                *WorkBuffer = 0;
                DoingFlagByte = FALSE;
                if (Entry->BlockLength == 0 && Entry->Flags.ParityByteUsed)                                 /* Block of 0 bytes ? */
                  DoingParityByte = TRUE;                                               /* Then the next byte is the parity byte! */
              }
            }
            else if (DoingParityByte)
            {
              Entry->ParityByte <<= 1;
              if ((LossyDetection  && (labs (JoinedPulseLength - Entry->Timing.LengthZeroBitPulse) >
                                       labs (JoinedPulseLength - Entry->Timing.LengthOneBitPulse))) ||
                  (!LossyDetection && (JoinedPulseLength > PulseAverage)))
                Entry->ParityByte |= 0x01;
              BitCounter ++;                                                                          /* Cannot loop (last byte!) */
            }
            else
            {
              *(WorkBuffer + ByteIndex) <<= 1;
              if ((LossyDetection  && (labs (JoinedPulseLength - Entry->Timing.LengthZeroBitPulse) >
                                       labs (JoinedPulseLength - Entry->Timing.LengthOneBitPulse))) ||
                  (!LossyDetection && (JoinedPulseLength > PulseAverage)))
                *(WorkBuffer + ByteIndex) |= 0x01;
              if (++ BitCounter == 8)                                                                    /* Done an entire byte ? */
              {                                                                                      /* Prepare for the next byte */
                if (TotalTStates == 15 &&                                                                    /* First data byte ? */
                  (((DecodingScheme == DS_SLOCK1 || DecodingScheme == DS_SLOCK2 || DecodingScheme == DS_SLOCK3) && ClickingPilot) ||
                   ((DecodingScheme == DS_SLOCK4 || DecodingScheme == DS_SLOCK5 || DecodingScheme == DS_SLOCK6 ||
                     DecodingScheme == DS_SLOCK7) && SLTurbo && !DoingNextSubBlock)))
                  Entry->FlagByte = *(WorkBuffer + ByteIndex);
                else if (TotalTStates == 7 && (DecodingScheme == DS_DIGITAL ||
                         ((DecodingScheme == DS_SLOCK4 || DecodingScheme == DS_SLOCK5 || DecodingScheme == DS_SLOCK6 ||
                           DecodingScheme == DS_SLOCK7) && SLTurbo && DoingNextSubBlock)))
                  Entry->FlagByte = *(WorkBuffer + ByteIndex);
                Entry->CalculatedParity ^= *(WorkBuffer + ByteIndex);                         /* Include the byte into the parity */
                BitCounter = 0;
                if (++ ByteIndex == (word)49152)                                                              /* Buffer is full ? */
                {
                  CopyToXMS ((dword)49152, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer,
                             TapeWindow[WhichWindow].Tape.Size + XMSIndex);
                  XMSIndex += (dword)49152;
                  ByteIndex = 0;
                }
                if (XMSIndex + ByteIndex == Entry->BlockLength)                                    /* Now points past the block ? */
                  DoingParityByte = TRUE;                            /* Then the data is done, next byte is the final parity byte */
                else
                  *(WorkBuffer + ByteIndex) = 0;
              }
            }
          }
          if (BitCounter == 0)
            BitCounter = 8;
          Entry->Timing.NumBitsInLastByte = BitCounter;
          PartialEndByteDropped = FALSE;
          if (BitCounter != 8)                                                                         /* Not all bits are used ? */
          {
            Entry->Timing.CustomLastByte = Entry->ParityByte << (8 - BitCounter);                          /* Shift the 'byte' up */
            if (((DecodingScheme == DS_SLOCK1 || DecodingScheme == DS_SLOCK2 || DecodingScheme == DS_SLOCK3) && ClickingPilot) ||
                ((DecodingScheme == DS_SLOCK4 || DecodingScheme == DS_SLOCK5 || DecodingScheme == DS_SLOCK6 ||
                  DecodingScheme == DS_SLOCK7 || DecodingScheme == DS_DIGITAL) && SLTurbo))
            {
              Entry->ParityByte = *(WorkBuffer + ByteIndex - 1);
              Entry->CalculatedParity ^= *(WorkBuffer + ByteIndex - 1);
            }
            else
              if (Entry->BlockLength == 0)                                            /* The last 'whole' byte will be the parity */
                Entry->Flags.ParityByteUsed = FALSE;                                    /* (Unless there are no 'whole' bytes...) */
              else
              {
                Entry->ParityByte = *(WorkBuffer + ByteIndex - 1);
                Entry->CalculatedParity ^= Entry->ParityByte;                      /* The byte doesn't count in the parity itself */
                Entry->BlockLength --;
              }
            if (DropPartialEndByte ||                                                               /* We don't want loose ends ? */
                DecodingScheme == DS_ROM || DecodingScheme == DS_BLEEP || DecodingScheme == DS_DINAMIC ||
                DecodingScheme == DS_SULOAD || DecodingScheme == DS_YSLOAD || DecodingScheme == DS_PLAYERS ||
                DecodingScheme == DS_POPS)
            {
              Entry->Timing.NumBitsInLastByte = 8;                              /* Then signal: there is no partial last byte :-) */
              PartialEndByteDropped = TRUE;
            }
          }
          else if (((DecodingScheme == DS_SLOCK1 || DecodingScheme == DS_SLOCK2 || DecodingScheme == DS_SLOCK3) && ClickingPilot) ||
                   ((DecodingScheme == DS_SLOCK4 || DecodingScheme == DS_SLOCK5 || DecodingScheme == DS_SLOCK6 ||
                     DecodingScheme == DS_SLOCK7 || DecodingScheme == DS_DIGITAL) && SLTurbo))
          {
            Entry->ParityByte = *(WorkBuffer + ByteIndex - 1);
            Entry->CalculatedParity ^= *(WorkBuffer + ByteIndex - 1);
          }
        }

        /* The entire block has been processed. Insert it into our virtual tape.                      */
        /* SpeedLock clicking-pilot-blocks are treated very differently.                              */
        /* Otherwise try to find out if the block appears to be 'Standard Loading' or 'Turbo Loading' */

        if (((DecodingScheme == DS_SLOCK1 || DecodingScheme == DS_SLOCK2 || DecodingScheme == DS_SLOCK3) && ClickingPilot) ||
            ((DecodingScheme == DS_SLOCK4 || DecodingScheme == DS_SLOCK5 || DecodingScheme == DS_SLOCK6 ||
              DecodingScheme == DS_SLOCK7) && SLTurbo) ||
            (DecodingScheme == DS_DIGITAL && SLTurbo))
        {
          Entry->BlockTypeID = 0x14;                                                         /* Store the data block as Pure Data */
          if (PartialEndByteDropped)
            Entry->BlockLength --;
        }
        else if (DecodingScheme == DS_POPS && Entry->Timing.NumPilotPulses < 2400)               /* Paul Owens is hard to detect! */
        {
          Entry->Timing.LengthPilotPulse = NS_LENPILOT;
          Entry->Timing.NumPilotPulses = 2196;
          Entry->Timing.LengthSyncPulse[0] = NS_LENSYNC1;
          Entry->Timing.LengthSyncPulse[1] = NS_LENSYNC2;
          _GetBitTimingValues (&(Entry->Timing.LengthZeroBitPulse), &(Entry->Timing.LengthOneBitPulse), &PulseAverage);
        }
        else if (DecodingScheme == DS_ROM ||                                      /* BleepLoad pilot lengths are still considered */
            (_PulseDiff (NS_LENPILOT, Entry->Timing.LengthPilotPulse) <= ThresholdStandardLoading &&
             Entry->Timing.NumSyncPulses == 2 &&
             //_PulseDiff (NS_LENSYNC1, Entry->Timing.LengthSyncPulse[0]) <= ThresholdStandardLoading &&
             //_PulseDiff (NS_LENSYNC2, Entry->Timing.LengthSyncPulse[1]) <= ThresholdStandardLoading &&
             Entry->Timing.NumPilotPulses > 3000 &&
             //(_PulseDiff (NS_HDPILOT, Entry->Timing.NumPilotPulses) <= ThresholdStandardLoading ||
             // _PulseDiff (NS_DTPILOT, Entry->Timing.NumPilotPulses) <= ThresholdStandardLoading) &&
             (NumPulses == 2 || (NumPulses > 2 &&
              (((Num1Bits == 0) && (_PulseDiff (NS_LEN0BIT, Entry->Timing.LengthZeroBitPulse) <= ThresholdStandardLoading ||
                                    _PulseDiff (NS_LEN1BIT, Entry->Timing.LengthZeroBitPulse) <= ThresholdStandardLoading)) ||
               ((Num1Bits != 0) && _PulseDiff (NS_LEN0BIT, Entry->Timing.LengthZeroBitPulse) <= ThresholdStandardLoading &&
                                   _PulseDiff (NS_LEN1BIT, Entry->Timing.LengthOneBitPulse) <= ThresholdStandardLoading))))))
        {
          Entry->BlockTypeID = 0x10;                                                         /* It's a 'Standard Loading' block ! */
          Entry->Timing.LengthPilotPulse = NS_LENPILOT;                                           /* Use the exact values instead */
          Entry->Timing.LengthSyncPulse[0] = NS_LENSYNC1;
          Entry->Timing.LengthSyncPulse[1] = NS_LENSYNC2;
          Entry->Timing.LengthZeroBitPulse = NS_LEN0BIT;
          Entry->Timing.LengthOneBitPulse = NS_LEN1BIT;
          Entry->Flags.IsNormalSpeed = TRUE;
          //if (_PulseDiff (NS_HDPILOT, Entry->Timing.NumPilotPulses) <= ThresholdStandardLoading)
          if (Entry->Timing.NumPilotPulses > _Average (NS_HDPILOT, NS_DTPILOT))
            Entry->Timing.NumPilotPulses = NS_HDPILOT;
          else
            Entry->Timing.NumPilotPulses = NS_DTPILOT;
        }
        if (ShowSoundStreamInfo)
          _PrintSoundAnalysis (Entry);
        GetBlockTypeInfo (Entry->BlockTypeID, &(Entry->IsTapeData), &(Entry->IsIdentifiedData), NULL, NULL, NULL);
        TotalTStates = (ByteIndex & 0xFFFC) + (ByteIndex % 4 ? 4 : 0);
        if (ByteIndex > 0)                                                             /* Store (the remaining part of) the block */
          CopyToXMS (TotalTStates, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer,
                     TapeWindow[WhichWindow].Tape.Size + XMSIndex);
        TapeWindow[WhichWindow].Tape.Size += Entry->AllocLength;
        Entry->Next = NULL;
        _AddEntryToTape (WhichWindow, Entry);
        if ((DecodingScheme == DS_SLOCK2 || DecodingScheme == DS_SLOCK3 || DecodingScheme == DS_SLOCK4 ||
             DecodingScheme == DS_SLOCK5 || DecodingScheme == DS_SLOCK6 || DecodingScheme == DS_SLOCK7) && MidSyncFound)
                                                                                       /* Datablock continues for SpeedLock 2-7 ? */
          _InsertSpeedLock234567MidSync (WhichWindow, Entry);                                       /* Insert the midsync pattern */

        else if (((((DecodingScheme == DS_SLOCK1 || DecodingScheme == DS_SLOCK2 || DecodingScheme == DS_SLOCK3) && ClickingPilot) ||
                   ((DecodingScheme == DS_SLOCK4 || DecodingScheme == DS_SLOCK5 || DecodingScheme == DS_SLOCK6 ||
                     DecodingScheme == DS_SLOCK7) && SLTurbo)) && !MidSyncFound) ||       /* SpeedLock ? Then we need a Group End */
                 (DecodingScheme == DS_DIGITAL && SLTurbo))
          _InsertGroupEnd (WhichWindow, Entry);

        BlockLength -= (NextBlock - *FileIndex);
        *FileIndex = NextBlock;
        if (Fh >= 0)
          if (lseek (Fh, *FileIndex, SEEK_SET) == -1)                             /* Move the filepointer after the initial pause */
            return (FALSE);
        if ((DecodingScheme == DS_SLOCK2 || DecodingScheme == DS_SLOCK3 || DecodingScheme == DS_SLOCK4 ||
             DecodingScheme == DS_SLOCK5 || DecodingScheme == DS_SLOCK6 || DecodingScheme == DS_SLOCK7) && MidSyncFound)
        {                                                                           /* Just done a subblock followed by midsync ? */
          MidSyncFound = FALSE;
          DoingNextSubBlock = TRUE;                                                              /* Signal: no flag byte in front */
          if ((DirP = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
            FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
          memcpy (DirP, Entry, sizeof (struct TapeEntry_s));                                                    /* Use same stats */
          Entry = DirP;
          _RestoreReadState (&_EndBackupReadBuffer);
          if (!UpdatePReq (*FileIndex, "Status: Analysing data ..."))                            /* Update the progress indicator */
          { *Cancelled = TRUE; free (Entry); return (TRUE); }
          goto __NEXTSUBBLOCK;                                                                          /* Read the next subblock */
        }
        if (BlockLength > 0)                                                                                       /* Clean end ? */
        {
          AllocateTapeBlock = TRUE;                                                                     /* Then there may be more */
          PilotAgain = TRUE;
          FirstPilot = FALSE;
        }
      }
    }
  }
  return (TRUE);                                                                                             /* Signal: no errors */
}

bool ReadDirectRecording (short WhichWindow, int Fh, dword SampleRate, dword BlockLength, dword *FileIndex, bool SkipSoundContinue,
                          bool *Cancelled)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/*         `Fh' points to the open tape file, `SampleRate' is the found rate in Hz, `BlockLength' is the length of the sample.    */
/*         `FileIndex' holds the current file index.                                                                              */
/* Post  : The sample has been entirely inserted in the tape directory as one Direct Recording Block.                             */
/*         The `FileIndex' has been adjusted to after the block.                                                                  */
/*         If the user hit the `Cancel' button in the percentage requester, `Cancelled' is returned TRUE.                         */
/* Import: _OpenReadBuffer, _ReadState, _AddEntryToTape, _AddSpaceToTape.                                                         */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *Entry;
  dword               DBlLength;
  dword               XMSIndex;
  word                ByteIndex;
  dword               DByteIndex;
  byte                BitCounter;
  bool                Done = FALSE;
  bool                NewState;

  _SkipVOCSoundContinue = SkipSoundContinue;
  HandlingZ80EmRAW = FALSE;
  HandlingCSW = FALSE;
  if (TapeWindow[WhichWindow].FileType == FT_ACORNRAW)
    HandlingZ80EmRAW = TRUE;
  else if (TapeWindow[WhichWindow].FileType == FT_CSW)
    HandlingCSW = TRUE;
  _OpenReadBuffer (WhichWindow, Fh, *FileIndex, BlockLength, SampleRate, 0);                        /* Setup the file read buffer */
  if ((Entry = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
    FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
  memset (Entry, 0, sizeof (struct TapeEntry_s));
  Entry->ComboSelected = 0x00;
  Entry->BlockTypeID = 0x15;                                                                                /* (Direct Recording) */
  Entry->Timing.SampleRate = (word)(3500000L / SampleRate);                                        /* (SampleRate is in T-States) */
  Entry->Timing.NumBitsInFirstByte = 8;
  Entry->Timing.NumBitsInLastByte = 8;                                               /* Assume last byte is fully used (unlikely) */
  GetBlockTypeInfo (0x15, &(Entry->IsTapeData), &(Entry->IsIdentifiedData), NULL, NULL, NULL);
  DBlLength = BlockLength / 8 + (BlockLength % 7 ? 1 : 0);                         /* (This counts the VOC block headers as well) */
  Entry->AllocLength = (DBlLength & 0xFFFFFFFC) + (DBlLength % 4 ? 4 : 0);
  _AddSpaceToTape (WhichWindow, Entry->AllocLength);
  TapeWindow[WhichWindow].NumEntries ++;
  DBlLength = 0;                                                                   /* Keep track of the total (used) block length */
  *Cancelled = FALSE;
  XMSIndex = 0;                                                                                         /* Start at the beginning */
  while (!*Cancelled && !Done)
  {
    if (!UpdatePReq (*FileIndex, "Converting to Direct Recording ..."))
    { *Cancelled = TRUE; Done = TRUE; }
    ByteIndex = 0;
    BitCounter = 0;
    while (!Done && ByteIndex < 49152)                                                                    /* Read each bufferfull */
      if (!_ReadState (&NewState))                                                                            /* Read next sample */
        Done = TRUE;
      else
      {
        *(WorkBuffer + ByteIndex) <<= 1;
        if (NewState)
          *(WorkBuffer + ByteIndex) |= 0x01;
        if (++ BitCounter == 8)                                                                             /* Done entire byte ? */
        {
          ByteIndex ++;
          BitCounter = 0;
        }
      }
    if (ByteIndex > 0 || BitCounter > 0)                                                                      /* Read something ? */
    {
       if (BitCounter != 0 && Done)                                                        /* Last block - last byte incomplete ? */
       {
         *(WorkBuffer + ByteIndex) <<= (8 - BitCounter);                                                   /* Shift the 'byte' up */
         Entry->Timing.NumBitsInLastByte = BitCounter;
         ByteIndex ++;                                                                           /* And include it into the count */
       }
       DByteIndex = (dword)(ByteIndex & 0xFFFC) + (ByteIndex % 4 ? 4 : 0);
       CopyToXMS (DByteIndex, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer, TapeWindow[WhichWindow].Tape.Size + XMSIndex);
       XMSIndex += DByteIndex;
    }
    DBlLength += ByteIndex;
    *FileIndex = _GetCurrentFileIndex () - 1;
  }
  Entry->BlockLength = DBlLength;                                                                /* Set the resulting blocklength */
  TapeWindow[WhichWindow].Tape.Size += Entry->AllocLength;
  Entry->Next = NULL;                                                                     /* Insert the block into the tape index */
  _AddEntryToTape (WhichWindow, Entry);
  return (!*Cancelled);
}

bool WriteVOCBlock (short WhichWindow, int Fh, struct TapeEntry_s *DirP, short XMSHandle, dword XMSIndex, dword *FileIndex,
                    comboent *NextBlock)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/*         `DirP' points to the current virtual tape directory entry, `FileIndex' keeps track of the position in the file.        */
/*         `XMSHandle'/'XMSIndex' points to the datablock, `Fh' < 0 for SoundBlaster out, or a valid file handle otherwise.       */
/* Post  : The virtual tape block has been appended to the VOC file. `FileIndex' is adjusted accordingly.                         */
/*         If `Fh' < 0 (SoundBlaster out), the block is inserted into the sound DMA buffer. DMA is still running when the routine */
/*         is left. An explicit call to FlushWriteBuffer is then needed to stop it.                                               */
/*         After the block is played, `NextBlock' holds the next blocknumber to be played (for jumps, loops and calls).           */
/* Import: _WritePulse, _HandleSelectBlock.                                                                                       */
/**********************************************************************************************************************************/

{
  static word     _LoopCounter;                                                        /* Number of loops left in this 0x24 block */
  static comboent _LoopStart;                                                         /* Absolute block number of this 0x24 block */
  static struct   _CallStack_s
  {
    struct _CallStack_s *_Next;                                                         /* Points to next call table (0x26 block) */
    struct _CallStack_s *_Prev;                                                            /* Previous one, used for backtracking */
    comboent      _BlockNo;                                                           /* Absolute block number of this 0x26 block */
    word          _Index;                                                         /* Index into the call table of this 0x26 block */
    word          _CallsLeft;                                                          /* Number of calls left in this call table */
  }              *_CallStack = NULL;
  static struct   _CallStack_s *_CallTail = NULL;                                        /* Points to the last entry in the chain */
  static struct   _CallStack_s *_CallWalk;
  static struct   _CallStack_s *_CallNew;
  static comboent _CallNextBlock = -1;                                                  /* Block right after the first call block */

  struct TapeEntry_s *DirWalk;
  struct ButtonInfo *ButtonP;
  struct Timing_s NormalTiming;
  struct Timing_s UsedTiming;
  word           *WPWorkBuffer;
  byte            LastState    = 0x10;                                                        /* (Used for Direct Recording only) */
  bool            OutData      = FALSE;                                                      /* TRUE if a datablock is to be done */
  byte   register Cnt;
  byte   register BitMask;                                                                     /* Rotation for each bit in a byte */
  byte            ByteToWrite;
  byte            BitsToDo;
  long            PulseLength;
  dword           XMSSubIndex  = 0;
  dword           LengthToGo;
  dword           ReadLength;
  dword           DReadLength;
  bool            MoreToGo     = TRUE;
  comboent        Tmp;

  if (InitVOCWriteVars)                                                                                /* Need to cleanup first ? */
  {
    _LoopCounter = 0;
    InitVOCWriteVars = FALSE;
    if (_CallStack != NULL)
    {
      _CallWalk = _CallStack;
      while (_CallWalk->_Next != NULL)
        _CallWalk = _CallWalk->_Next;
      while (_CallWalk->_Prev != NULL)
      {
        _CallWalk = _CallWalk->_Prev;
        free (_CallWalk->_Next);
      }
      free (_CallStack);
    }
    _CallStack = NULL;
    _CallTail = NULL;
    _CallNextBlock = -1;
  }
  NormalTiming.LengthPilotPulse = NS_LENPILOT;                                                   /* Prepare for normal speed data */
  NormalTiming.LengthSyncPulse[0] = NS_LENSYNC1;
  NormalTiming.LengthSyncPulse[1] = NS_LENSYNC2;
  NormalTiming.LengthZeroBitPulse = NS_LEN0BIT;
  NormalTiming.LengthOneBitPulse = NS_LEN1BIT;
  NormalTiming.NumPilotPulses = NS_DTPILOT;                                                                      /* (Assume data) */
  NormalTiming.NumSyncPulses = 2;
  NormalTiming.NumBitsInFirstByte = NS_LENFIRST;
  NormalTiming.NumBitsInLastByte = DirP->Timing.NumBitsInLastByte;
  NormalTiming.CustomLastByte = DirP->Timing.CustomLastByte;
  if (PlayTimeUpdated)  { PlayTimeUpdated = FALSE; Wprintf (&_SBWindow, 1, TEXT_CENTER, PlayTimeString); }
  if (!SimulatePlayBlock &&
      DirP->Flags.IsWritten &&                                                           /* This block has already been written ? */
      _LoopCounter == 0 &&                                                                     /* And we're not in a Loop group ? */
      _CallStack == NULL)                                                                  /* And it's not part of a Call group ? */
  {
    _InfiniteTapeLoop = TRUE;                                                                      /* Signal: infinite tape loop! */
    if (Fh >= 0)                                                                                         /* Writing to VOC file ? */
      return (TRUE);                                          /* Step out without error - the loop check is already taken care of */
    else                                                                       /* (Playing through SoundBlaster or parallel port) */
      return (FALSE);                                                                            /* Step out signalling the error */
  }
  DirP->Flags.IsWritten = TRUE;
  if (!SimulatePlayBlock && !CountPlayTimeLength)
  {
    HandlingZ80EmRAW = FALSE;
    HandlingCSW = FALSE;
    if (TapeWindow[WhichWindow].FileType == FT_ACORNRAW)
      HandlingZ80EmRAW = TRUE;
    else if (TapeWindow[WhichWindow].FileType == FT_CSW)
      HandlingCSW = TRUE;
    if (!_OpenWriteBuffer (Fh, *FileIndex, FALSE))  return (FALSE);
    if (Fh < 0 && TPConfig.PlaySoundBlaster && !SoundBlasterInfo.DoingDMA && _WriteBuffer._BufferIndex == 0)     /* Just set up ? */
    {
      _WriteBuffer._State = 0x80;                                            /* Write some 'silence' first to suppress the 'plop' */
      if (!_WritePulse (&_WriteBuffer._State, 7000))  return (FALSE);                                                  /* (2 ms.) */
      _WriteBuffer._State = 0xF0;
    }
  }
  switch (DirP->BlockTypeID)
  {
    case 0x10 : if (!SimulatePlayBlock)                                                                  /* (Standard Speed Data) */
                {
                  OutData = TRUE;
                  memcpy (&UsedTiming, &NormalTiming, sizeof (struct Timing_s));
                  if (DirP->Flags.IsHeader)
                    UsedTiming.NumPilotPulses = NS_HDPILOT;
                }
                (*NextBlock) ++;
                break;
    case 0x11 : if (!SimulatePlayBlock)                                                                     /* (Turbo Speed Data) */
                {
                  OutData = TRUE;
                  memcpy (&UsedTiming, &DirP->Timing, sizeof (struct Timing_s));                         /* Prepare timing values */
                }
                (*NextBlock) ++;
                break;
    case 0x12 : if (!SimulatePlayBlock)                                                                            /* (Pure Tone) */
                {
                  if (CountPlayTimeLength)
                    PlayTimeLength += (DirP->Timing.NumPilotPulses * DirP->Timing.LengthPilotPulse);
                  else
                    for (PulseLength = 0 ; PulseLength < DirP->Timing.NumPilotPulses ; PulseLength ++)
                      if (!_WritePulse (&_WriteBuffer._State, DirP->Timing.LengthPilotPulse))  return (FALSE);
                }
                (*NextBlock) ++;
                break;
    case 0x13 : if (!SimulatePlayBlock)                                                                        /* (Row of Pulses) */
                {
                  CopyFromXMS (DirP->AllocLength, XMSHandle, XMSIndex, WorkBuffer);
                  WPWorkBuffer = (word *)WorkBuffer;                                                             /* Read as words */
                  if (CountPlayTimeLength)
                    for (PulseLength = DirP->BlockLength / 2 ; PulseLength > 0 ; PulseLength --)                 /* Do all pulses */
                      PlayTimeLength += *(WPWorkBuffer ++);
                  else
                    for (PulseLength = DirP->BlockLength / 2 ; PulseLength > 0 ; PulseLength --)                 /* Do all pulses */
                      if (!_WritePulse (&_WriteBuffer._State, *(WPWorkBuffer ++)))  return (FALSE);
                }
                (*NextBlock) ++;
                break;
    case 0x14 : if (!SimulatePlayBlock)                                                                      /* (Pure Data Block) */
                {
                  OutData = TRUE;
                  memset (&UsedTiming, 0, sizeof (struct Timing_s));                                     /* Reset all timing info */
                  UsedTiming.LengthZeroBitPulse = DirP->Timing.LengthZeroBitPulse;             /* Set only the appropriate values */
                  UsedTiming.LengthOneBitPulse = DirP->Timing.LengthOneBitPulse;
                  UsedTiming.NumBitsInFirstByte = NS_LENFIRST;
                }
                (*NextBlock) ++;
                break;
    case 0x15 : if (!SimulatePlayBlock)                                                                     /* (Direct Recording) */
                {
                  LengthToGo = DirP->BlockLength;
                  MoreToGo = (bool)(LengthToGo > 0);
                  PulseLength = 0;
                  LastState = 0x10;                                                           /* Assume first byte will be a 0x10 */
                  if (CountPlayTimeLength)
                  {
                    if (MoreToGo)
                      PlayTimeLength += (((DirP->BlockLength - 1) * 8 + DirP->Timing.NumBitsInLastByte) * DirP->Timing.SampleRate);
                  }
                  else
                  {
                    while (MoreToGo)                                                              /* Output the entire data block */
                    {
                      ReadLength = LengthToGo > 49152L ? 49152L : LengthToGo;                             /* (Read in 48K chunks) */
                      LengthToGo -= ReadLength;
                      if (LengthToGo == 0)
                        MoreToGo = FALSE;
                      DReadLength = (ReadLength & 0xFFFFFFFC) + (ReadLength % 4 ? 4 : 0);
                      CopyFromXMS (DReadLength, XMSHandle, XMSIndex + XMSSubIndex, WorkBuffer);                /* (Fetch a block) */
                      XMSSubIndex += DReadLength;
                      for (DReadLength = 0 ; DReadLength < ReadLength ; DReadLength ++)                         /* (Do each byte) */
                      {
                        ByteToWrite = *(WorkBuffer + DReadLength);
                        if (MoreToGo == FALSE && DReadLength == ReadLength - 1)                        /* About to do last byte ? */
                          BitsToDo = DirP->Timing.NumBitsInLastByte;
                        else
                          BitsToDo = 8;
                        for (Cnt = 0, BitMask = 0x80 ; Cnt < BitsToDo ; Cnt ++, BitMask >>= 1)         /* (Do each bit in a byte) */
                        {
                          _WriteBuffer._State = ((ByteToWrite & BitMask) ? 0xF0 : 0x10);
                          if (_WriteBuffer._State == LastState)                                        /* Add up until transision */
                            PulseLength ++;
                          else
                          {
                            if (PulseLength > 0)                                             /* (In case first byte was not 0x10) */
                                                                                                               /* Write the pulse */
                              if (!_WritePulse (&LastState, PulseLength * DirP->Timing.SampleRate))  return (FALSE);
                            LastState = _WriteBuffer._State;
                            PulseLength = 1;                                                               /* Include this state! */
                          }
                        }
                      }
                    }
                    if (PulseLength > 0)                                                        /* Any remaining pulse (likely) ? */
                                                                                                         /* Write finishing pulse */
                      if (!_WritePulse (&LastState, PulseLength * DirP->Timing.SampleRate))  return (FALSE);
                  }
                  if (DirP->PauseLength > 0)                                                /* Is there a pause after the block ? */
                    if (CountPlayTimeLength)
                      PlayTimeLength += (dword)DirP->PauseLength * 3500;
                    else
                    {
                      if (!_WritePulse (&LastState, 3500))  return (FALSE);                              /* Finish the last pulse */
                      _WriteBuffer._State = 0x80;                                                      /* Force to 'silent' state */
                      PulseLength = (long)(DirP->PauseLength - 1) * 3500;
                      if (PulseLength > 0)
                        if (!_WritePulse (&_WriteBuffer._State, PulseLength))  return (FALSE);                /* The actual pause */
                    }
                }
                (*NextBlock) ++;
                break;
    case 0x20 : if (!SimulatePlayBlock)                                                             /* (Pause or 'Stop the tape') */
                  if (CountPlayTimeLength)
                    PlayTimeLength += (dword)DirP->PauseLength * 3500;
                  else
                  {
                    if (!_WritePulse (&_WriteBuffer._State, 3500))  return (FALSE);                      /* Finish the last pulse */
                    _WriteBuffer._State = 0x80;                                                        /* Force to 'silent' state */
                    if (DirP->PauseLength > 0)
                    {
                      PulseLength = (long)(DirP->PauseLength - 1) * 3500;
                      if (DirP->PauseLength > 0)
                        if (!_WritePulse (&_WriteBuffer._State, PulseLength))  return (FALSE);                /* The actual pause */
                    }
                    else if (_WriteBuffer._Fh < 0)                                           /* Infinite pause for SoundBlaster ? */
                    {
                      _PlayClockRunning = FALSE;
                      _KeepSBWindow = TRUE;
                      FlushWriteBuffer (FALSE);                                                   /* Wait for the sound to finish */
                                                             /* (If multiple blocks are done, DMA is automatically re-initialsed) */
                      _KeepSBWindow = FALSE;
                      ClearWindow (&_SBWindow);                                        /* (This will leave the nice border alone) */
                      Wprintf (&_SBWindow, 1, TEXT_CENTER,       "{BPAUSED}B");
                      Wprintf (&_SBWindow, 2, TEXT_CENTER, "Press a key to continue");
                      Wprintf (&_SBWindow, 3, TEXT_CENTER,     "[ESC] to abort");
                      SoundErrorBeeper (BEEP_MEDIUM);                                                            /* Tell the user */
                      while (GlobalKey == 0x0000)
                        HandleWindow (&_SBWindow, &ButtonP);
                      if (GlobalKey == K_ESC)
                        return (FALSE);
                      GlobalKey = 0x0000;                                                                       /* (Drop the key) */
                      ClearWindow (&_SBWindow);                                                           /* 'Restore' the window */
                      if (UsePlayTimeClock)
                      {
                        _PlayClockRunning = TRUE;
                        Wprintf (&_SBWindow, 1, TEXT_CENTER, PlayTimeString);
                        Wprintf (&_SBWindow, 3, TEXT_CENTER, "Press [ESC] to abort");
                      }
                      else
                        Wprintf (&_SBWindow, 2, TEXT_CENTER, "Press [ESC] to abort");
                    }
                  }
                (*NextBlock) ++;
                break;
    case 0x23 : (*NextBlock) += DirP->PauseLength;                                                             /* (Jump to Entry) */
                break;                                                                                           /* 'Do' the jump */
    case 0x24 : _LoopCounter = DirP->PauseLength;                                                                 /* (Loop Start) */
                _LoopStart = (*NextBlock);                                                           /* Initialize loop variables */
                (*NextBlock) ++;
                break;
    case 0x25 : if (_LoopCounter > 1)                                                                               /* (Loop End) */
                {                                                                       /* Go back unless the counter has expired */
                  _LoopCounter --;
                  (*NextBlock) = _LoopStart + 1;
                }
                else
                  (*NextBlock) ++;
                break;
    case 0x26 : if ((DirP->BlockLength / 2) == 1)                                                              /* (Call Sequence) */
                                                                                                    /* Only 1 call in the block ? */
                  _CallNextBlock = (*NextBlock) + 1;                                            /* Next pass will take next block */
                else
                {
                  if ((_CallNew = (struct _CallStack_s *)malloc (sizeof (struct _CallStack_s))) == NULL)
                    FatalError (1, _ErrOutOfMemory, sizeof (struct _CallStack_s));
                  _CallNew->_Next = NULL;
                  _CallNew->_BlockNo = (*NextBlock);
                  _CallNew->_Index = 2;                                                                     /* Point to next call */
                  _CallNew->_CallsLeft = (DirP->BlockLength / 2) - 1;
                  if (_CallStack == NULL)                                                                         /* First call ? */
                  {
                    _CallNew->_Prev = NULL;
                    _CallStack = _CallTail = _CallNew;
                  }
                  else                                                                                   /* Add call to the stack */
                  {
                    _CallWalk = _CallStack;
                    while (_CallWalk->_Next != NULL)
                      _CallWalk = _CallWalk->_Next;
                    _CallWalk->_Next = _CallTail = _CallNew;
                    _CallNew->_Prev = _CallWalk;
                  }
                }
                CopyFromXMS (DirP->AllocLength, XMSHandle, XMSIndex, WorkBuffer);                             /* Fetch call table */
                (*NextBlock) += *((short *)WorkBuffer);                                                       /* Setup first call */
                break;
    case 0x27 : if (_CallTail != NULL)                                                                  /* (Return From Sequence) */
                {                                                                                            /* Stack not empty ? */
                  DirWalk = TapeWindow[WhichWindow].Directory;
                  Tmp = _CallTail->_BlockNo;                                                            /* Find the calling block */
                  XMSSubIndex = 0;
                  while (Tmp > 0 && DirWalk)
                  {
                    XMSSubIndex += DirWalk->AllocLength;
                    DirWalk = DirWalk->Next;
                    Tmp --;
                  }
                  CopyFromXMS (DirWalk->AllocLength, XMSHandle, XMSSubIndex, WorkBuffer);                     /* Fetch call block */
                  (*NextBlock) = _CallTail->_BlockNo + *((short *)(WorkBuffer + _CallTail->_Index));           /* Setup next call */
                  while (MoreToGo && _CallTail != NULL)                          /* Point to next call, backtracking where needed */
                  {
                    if (-- _CallTail->_CallsLeft > 0)                                          /* Any more calls after this one ? */
                    {
                      _CallTail->_Index += 2;                                                               /* Point to next call */
                      MoreToGo = FALSE;                                                              /* Signal: ready to continue */
                    }
                    else
                    {
                      if (_CallTail->_Prev == NULL)                                  /* At first stack entry (first call block) ? */
                      {
                        _CallNextBlock = _CallTail->_BlockNo + 1;                               /* Next pass will take next block */
                        free (_CallTail);
                        _CallStack = _CallTail = NULL;                                                        /* Remove the stack */
                      }
                      else
                      {
                        _CallTail = _CallTail->_Prev;
                        free (_CallTail->_Next);                                                            /* Release done block */
                      }
                    }
                  }
                  MoreToGo = TRUE;
                }
                else if (_CallNextBlock >= 0)                                                                   /* Done last call */
                {
                  (*NextBlock) = _CallNextBlock;
                  _CallNextBlock = -1;
                }
                else                                                                                       /* Return without call */
                  (*NextBlock) ++;                                                                          /* (Ignore the block) */
                break;
    case 0x28 : if (!SimulatePlayBlock)                                                                         /* (Select Block) */
                  if (_WriteBuffer._Fh < 0)                                                               /* Doing SoundBlaster ? */
                  {
                    if (!_WritePulse (&_WriteBuffer._State, 3500))  return (FALSE);                      /* Finish the last pulse */
                    _WriteBuffer._State = 0x80;                                                        /* Force to 'silent' state */
                    _KeepSBWindow = TRUE;
                    FlushWriteBuffer (FALSE);                                                     /* Wait for the sound to finish */
                                                             /* (If multiple blocks are done, DMA is automatically re-initialsed) */
                    _KeepSBWindow = FALSE;
                  }
                CopyFromXMS (DirP->AllocLength, XMSHandle, XMSIndex, WorkBuffer);                           /* Fetch select block */
                (*NextBlock) = _HandleSelectBlock (*NextBlock);                           /* Wait for the user to pick next block */
                break;
    case 0x2A : if (!SimulatePlayBlock && !TapeWindow[WhichWindow].ModeIs128K)                      /* (Stop Tape if in 48K Mode) */
                {
                  if (!_WritePulse (&_WriteBuffer._State, 3500))  return (FALSE);                        /* Finish the last pulse */
                  _WriteBuffer._State = 0x80;                                                          /* Force to 'silent' state */
                  if (_WriteBuffer._Fh < 0)                                                  /* Infinite pause for SoundBlaster ? */
                  {
                    _PlayClockRunning = FALSE;
                    _KeepSBWindow = TRUE;
                    FlushWriteBuffer (FALSE);                                                     /* Wait for the sound to finish */
                                                             /* (If multiple blocks are done, DMA is automatically re-initialsed) */
                    _KeepSBWindow = FALSE;
                    ClearWindow (&_SBWindow);                                          /* (This will leave the nice border alone) */
                    Wprintf (&_SBWindow, 1, TEXT_CENTER,       "{BPAUSED}B");
                    Wprintf (&_SBWindow, 2, TEXT_CENTER, "Press a key to continue");
                    Wprintf (&_SBWindow, 3, TEXT_CENTER,     "[ESC] to abort");
                    SoundErrorBeeper (BEEP_MEDIUM);                                                              /* Tell the user */
                    while (GlobalKey == 0x0000)
                      HandleWindow (&_SBWindow, &ButtonP);
                    if (GlobalKey == K_ESC)
                      return (FALSE);
                    GlobalKey = 0x0000;                                                                         /* (Drop the key) */
                    ClearWindow (&_SBWindow);                                                             /* 'Restore' the window */
                    if (UsePlayTimeClock)
                    {
                      _PlayClockRunning = TRUE;
                      Wprintf (&_SBWindow, 1, TEXT_CENTER, PlayTimeString);
                      Wprintf (&_SBWindow, 3, TEXT_CENTER, "Press [ESC] to abort");
                    }
                    else
                      Wprintf (&_SBWindow, 2, TEXT_CENTER, "Press [ESC] to abort");
                  }
                }
                (*NextBlock) ++;
                break;
    case 0x31 : (*NextBlock) ++;                                                                               /* (Message Block) */
                break;
    default   : (*NextBlock) ++;                                                                     /* Simply skip anything else */
                break;
  }
  if (OutData && !SimulatePlayBlock)                                                      /* We still have to do the data block ? */
  {
    if (CountPlayTimeLength)
      PlayTimeLength += (UsedTiming.NumPilotPulses * UsedTiming.LengthPilotPulse);
    else
      for (PulseLength = 0 ; PulseLength < UsedTiming.NumPilotPulses ; PulseLength ++)                   /* Output the pilot tone */
        if (!_WritePulse (&_WriteBuffer._State, UsedTiming.LengthPilotPulse))  return (FALSE);
    if (CountPlayTimeLength)
      for (Cnt = 0 ; Cnt < UsedTiming.NumSyncPulses ; Cnt ++)                                           /* Output the sync pulses */
        PlayTimeLength += UsedTiming.LengthSyncPulse[Cnt];
    else
      for (Cnt = 0 ; Cnt < UsedTiming.NumSyncPulses ; Cnt ++)
        if (!_WritePulse (&_WriteBuffer._State, UsedTiming.LengthSyncPulse[Cnt]))  return (FALSE);
    if (DirP->Timing.NumBitsInFirstByte != 8)                                                          /* Output any initial bits */
    {
      ByteToWrite = DirP->Timing.CustomFirstByte;
      for (Cnt = 0, BitMask = 0x80 ; Cnt < DirP->Timing.NumBitsInFirstByte ; Cnt ++, BitMask >>= 1)
      {
        if (ByteToWrite & BitMask)  PulseLength = UsedTiming.LengthOneBitPulse;                              /* (Determine timer) */
        else                        PulseLength = UsedTiming.LengthZeroBitPulse;
        if (CountPlayTimeLength)
          PlayTimeLength += PulseLength * 2;
        else
        {
          if (!_WritePulse (&_WriteBuffer._State, PulseLength))  return (FALSE);                               /* (Write the bit) */
          if (!_WritePulse (&_WriteBuffer._State, PulseLength))  return (FALSE);                               /* (2 half-pulses) */
        }
      }
    }
    if (DirP->Flags.FlagByteUsed)                                                                         /* Output any flag byte */
    {
      ByteToWrite = DirP->FlagByte;
      for (Cnt = 0, BitMask = 0x80 ; Cnt < 8 ; Cnt ++, BitMask >>= 1)
      {
        if (ByteToWrite & BitMask)  PulseLength = UsedTiming.LengthOneBitPulse;
        else                        PulseLength = UsedTiming.LengthZeroBitPulse;
        if (CountPlayTimeLength)
          PlayTimeLength += PulseLength * 2;
        else
        {
          if (!_WritePulse (&_WriteBuffer._State, PulseLength))  return (FALSE);
          if (!_WritePulse (&_WriteBuffer._State, PulseLength))  return (FALSE);
        }
      }
    }
    LengthToGo = DirP->BlockLength;
    while (MoreToGo)                                                                              /* Output the entire data block */
    {
      ReadLength = LengthToGo > 49152L ? 49152L : LengthToGo;                                             /* (Read in 48K chunks) */
      LengthToGo -= ReadLength;
      if (LengthToGo == 0)
        MoreToGo = FALSE;
      DReadLength = (ReadLength & 0xFFFFFFFC) + (ReadLength % 4 ? 4 : 0);
      CopyFromXMS (DReadLength, XMSHandle, XMSIndex + XMSSubIndex, WorkBuffer);                                /* (Fetch a block) */
      XMSSubIndex += DReadLength;
      for (DReadLength = 0 ; DReadLength < ReadLength ; DReadLength ++)                                         /* (Do each byte) */
      {
        ByteToWrite = *(WorkBuffer + DReadLength);
        for (Cnt = 0, BitMask = 0x80 ; Cnt < 8 ; Cnt ++, BitMask >>= 1)                                /* (Do each bit in a byte) */
        {
          if (ByteToWrite & BitMask)  PulseLength = UsedTiming.LengthOneBitPulse;
          else                        PulseLength = UsedTiming.LengthZeroBitPulse;
          if (CountPlayTimeLength)
            PlayTimeLength += PulseLength * 2;
          else
          {
            if (!_WritePulse (&_WriteBuffer._State, PulseLength))  return (FALSE);
            if (!_WritePulse (&_WriteBuffer._State, PulseLength))  return (FALSE);
          }
        }
      }
    }
    if (DirP->Flags.ParityByteUsed)                                                                     /* Output any parity byte */
    {
      ByteToWrite = DirP->ParityByte;
      for (Cnt = 0, BitMask = 0x80 ; Cnt < 8 ; Cnt ++, BitMask >>= 1)
      {
        if (ByteToWrite & BitMask)  PulseLength = UsedTiming.LengthOneBitPulse;
        else                        PulseLength = UsedTiming.LengthZeroBitPulse;
        if (CountPlayTimeLength)
          PlayTimeLength += PulseLength * 2;
        else
        {
          if (!_WritePulse (&_WriteBuffer._State, PulseLength))  return (FALSE);
          if (!_WritePulse (&_WriteBuffer._State, PulseLength))  return (FALSE);
        }
      }
    }
    if (DirP->Timing.NumBitsInLastByte != 8)                                                         /* Output any remaining bits */
    {
      ByteToWrite = DirP->Timing.CustomLastByte;
      for (Cnt = 0, BitMask = 0x80 ; Cnt < DirP->Timing.NumBitsInLastByte ; Cnt ++, BitMask >>= 1)
      {
        if (ByteToWrite & BitMask)  PulseLength = UsedTiming.LengthOneBitPulse;
        else                        PulseLength = UsedTiming.LengthZeroBitPulse;
        if (CountPlayTimeLength)
          PlayTimeLength += PulseLength * 2;
        else
        {
          if (!_WritePulse (&_WriteBuffer._State, PulseLength))  return (FALSE);
          if (!_WritePulse (&_WriteBuffer._State, PulseLength))  return (FALSE);
        }
      }
    }
    if (DirP->PauseLength > 0)                                                              /* Is there a pause after the block ? */
      if (CountPlayTimeLength)
        PlayTimeLength += (dword)DirP->PauseLength * 3500;
      else
      {
        if (!_WritePulse (&_WriteBuffer._State, 3500))  return (FALSE);                                  /* Finish the last pulse */
        _WriteBuffer._State = 0x80;                                                                    /* Force to 'silent' state */
        PulseLength = (long)(DirP->PauseLength - 1) * 3500;
        if (PulseLength > 0)
          if (!_WritePulse (&_WriteBuffer._State, PulseLength))  return (FALSE);                              /* The actual pause */
      }
  }
  if (Fh > 0 && !SimulatePlayBlock && !CountPlayTimeLength)
  {
    if (!FlushWriteBuffer (FALSE))  return (FALSE);                                                 /* Write any remaining buffer */
    *FileIndex = _WriteBuffer._FileIndex;                                                                    /* (Update globally) */
  }
  return (TRUE);
}

void CleanupTapeBuffers (void)

/**********************************************************************************************************************************/
/* Pre   : None.                                                                                                                  */
/* Post  : The file cache buffer has been released.                                                                               */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  _CloseFileBuffer (TRUE);                                                                        /* Release any file read buffer */
  _CloseFileBuffer (FALSE);                                                                      /* Release any file write buffer */
}

void ResetVOCTimers (void)

/**********************************************************************************************************************************/
/* Pre   : None.                                                                                                                  */
/* Post  : All VOC decoding values have been reset to their defaults.                                                             */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  MinPilotPulseLength = 650;
  MaxPilotPulseLength = 3200;
  MinNumPilotPulses = 50;
  ThresholdPilotPulse = 30;
  NumSyncPulses = 2;
  FMinBitPulseLength = 100;
  ThresholdBitPulse = 25;
  ThresholdEndMarker = 1.4;
  ThresholdSingleEndMarker = 1.6;
  EndBitPulseLength = 2400;
  TestDoubleBitPulse = FALSE;
  ThresholdDoubleBitPulse = 35;
  ThresholdStandardLoading = 15;
  DropPartialEndByte = FALSE;
  DecodingScheme = TPConfig.DefDecodeROM ? DS_ROM : DS_NONE;
  LossyDetection = FALSE;
  ApproximateBitTimers = TRUE;
  VOCHighSpeedSampleRateAdjust = FALSE;
  ApplyFilter = TPConfig.DefUseFilter;
  InsertDirectRecording = FALSE;
  ForceStandardSyncs = FALSE;
}

bool SetupVOCReadTimers (short FileType)

/**********************************************************************************************************************************/
/* Pre   : (We are about to read a VOC file).                                                                                     */
/* Post  : All VOC decoding values are presented to the user. The values are possibly adjusted by the user.                       */
/*         The return value is TRUE if the user choose 'Ok' (or used the close gadget) or FALSE if the user cancelled.            */
/* Import: ResetVOCTimers.                                                                                                        */
/**********************************************************************************************************************************/

{
  char              ValMinPilotPulseLength[5];
  char              ValMaxPilotPulseLength[5];
  char              ValMinNumPilotPulses[5];
  char              ValThresholdPilotPulse[5];
  char              ValFMinBitPulseLength[5];
  char              ValThresholdBitPulse[5];
  char              ValThresholdEndMarker[8];
  char              ValThresholdSingleEndMarker[8];
  char              ValThresholdStandardLoading[5];
  char              ValDecodingScheme[35];
  char              DocPath[MAXDIR];
  char              CHeader[40] = "VOC Decoding Setup";

  struct WindowInfo CWindow   = {125, 40, 384, 395, 384, 395, 384, 395, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                 WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER, NULL, NULL};
  struct ButtonInfo PCancel   = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 50, 370, 8, DBLACK, DWHITE, NULL, 0, "@Cancel", NULL};
  struct ButtonInfo PDefault  = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 127, 370, 8, DBLACK, DWHITE, NULL, 1, "@Default",
                                 NULL};
  struct ButtonInfo PHelp     = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 203, 370, 8, DBLACK, DWHITE, NULL, 14, "@Help", NULL};
  struct ButtonInfo POk       = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 280, 370, 8, DBLACK, DWHITE, NULL, 2, "@Ok", NULL};
  struct ButtonInfo BDScheme  = {BT_TEXT, BC_TEXT, BF_FWIDTH, 132, 20, 34, DBLACK, DWHITE, NULL, 0, NULL, NULL};
  struct ButtonInfo BDSchmSel = {BT_PUSH, BC_GADGET, BF_FWIDTH | BF_KEYSELECT, 343, 20, 1, DBLACK, DWHITE, NULL, 13, "\012", NULL};
  struct ButtonInfo BPiMin    = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 132, 70, 4, DBLACK, DWHITE, NULL, 3,
                                 NULL, NULL};
  struct ButtonInfo BPiMax    = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 318, 70, 4, DBLACK, DWHITE, NULL, 4,
                                 NULL, NULL};
  struct ButtonInfo BPiPulses = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 132, 90, 4, DBLACK, DWHITE, NULL, 5,
                                 NULL, NULL};
  struct ButtonInfo BPiThres  = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 318, 90, 4, DBLACK, DWHITE, NULL, 6,
                                 NULL, NULL};
  struct ButtonInfo BBtMin    = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 132, 140, 4, DBLACK, DWHITE, NULL, 7,
                                 NULL, NULL};
  struct ButtonInfo BBtThres  = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 318, 140, 4, DBLACK, DWHITE, NULL, 8,
                                 NULL, NULL};
  struct ButtonInfo BLossyDet = {BT_BOOL, BC_GADGET, BF_KEYSELECT | BF_BCLEAR, 132, 160, 1, DBLACK, DWHITE, NULL, 12,
                                 &LossyDetection, NULL};
  struct ButtonInfo BApproxBT = {BT_BOOL, BC_GADGET, BF_KEYSELECT | BF_BCLEAR, 318, 160, 1, DBLACK, DWHITE, NULL, 12,
                                 &ApproximateBitTimers, NULL};
  struct ButtonInfo BEndSing  = {BT_INPUT, BC_FLOAT, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 132, 210, 4, DBLACK, DWHITE, NULL, 9,
                                 NULL, NULL};
  struct ButtonInfo BEndDoub  = {BT_INPUT, BC_FLOAT, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 318, 210, 4, DBLACK, DWHITE, NULL, 10,
                                 NULL, NULL};
  struct ButtonInfo BStdTres  = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 132, 260, 4, DBLACK, DWHITE, NULL, 11,
                                 NULL, NULL};
  struct ButtonInfo BDropEnd  = {BT_BOOL, BC_GADGET, BF_KEYSELECT | BF_BCLEAR, 132, 280, 1, DBLACK, DWHITE, NULL, 12,
                                 &DropPartialEndByte, NULL};
  struct ButtonInfo BFilter   = {BT_BOOL, BC_GADGET, BF_KEYSELECT | BF_BCLEAR, 132, 300, 1, DBLACK, DWHITE, NULL, 12,
                                 &ApplyFilter, NULL};
  struct ButtonInfo BStdSync  = {BT_BOOL, BC_GADGET, BF_KEYSELECT | BF_BCLEAR, 132, 320, 1, DBLACK, DWHITE, NULL, 13,
                                 &ForceStandardSyncs, NULL};
  struct ButtonInfo BDirect   = {BT_BOOL, BC_GADGET, BF_KEYSELECT | BF_BCLEAR, 318, 260, 1, DBLACK, DWHITE, NULL, 13,
                                 &InsertDirectRecording, NULL};
  struct ButtonInfo BShowInfo = {BT_BOOL, BC_GADGET, BF_KEYSELECT | BF_BCLEAR, 318, 280, 1, DBLACK, DWHITE, NULL, 13,
                                 &ShowSoundStreamInfo, NULL};
  struct ButtonInfo BSampRAdj = {BT_BOOL, BC_GADGET, BF_KEYSELECT | BF_BCLEAR, 318, 300, 1, DBLACK, DWHITE, NULL, 13,
                                 &VOCHighSpeedSampleRateAdjust, NULL};
  struct ButtonInfo BFreqShft = {BT_BOOL, BC_GADGET, BF_KEYSELECT | BF_BCLEAR, 318, 320, 1, DBLACK, DWHITE, NULL, 13,
                                 &VOCFrequencyShift, NULL};
  struct ComboInfo  SWindow;
  struct ComboEntry SEntry;
  struct ButtonInfo *ButtonP;
  bool              Finished = FALSE;
  bool              Cancelled = FALSE;
  comboent          SelectedLine;
  short   register  Cnt;

  CWindow.Header = CHeader;
  BDScheme.Contents = ValDecodingScheme;
  BPiMin.Contents = ValMinPilotPulseLength;
  BPiMax.Contents = ValMaxPilotPulseLength;
  BPiPulses.Contents = ValMinNumPilotPulses;
  BPiThres.Contents = ValThresholdPilotPulse;
  BBtMin.Contents = ValFMinBitPulseLength;
  BBtThres.Contents = ValThresholdBitPulse;
  BEndSing.Contents = ValThresholdSingleEndMarker;
  BEndDoub.Contents = ValThresholdEndMarker;
  BStdTres.Contents = ValThresholdStandardLoading;
  if (ApplyFilter == 0xFF)                                                                               /* Not initialized yet ? */
  {
    ApplyFilter = TPConfig.DefUseFilter;                                                            /* Use the user default value */
    VOCHighSpeedSampleRateAdjust = FALSE;
    VOCFrequencyShift = FALSE;
  }
  if (DecodingScheme == 0xFF)                                                                            /* Not initialized yet ? */
    DecodingScheme = TPConfig.DefDecodeROM ? DS_ROM : DS_NONE;
  InsertDirectRecording = FALSE;                                                                   /* We don't usually want this! */
  if (FileType == FT_ACORNRAW)
  {
    strcpy (CHeader, "RAW Decoding Setup");
    BFilter.Flags |= BF_GHOSTED;                                                                           /* (Nothing to filter) */
  }
  if (DecodeDirectRecording)
  {
    strcpy (CHeader, "Direct Recording Decoding Setup");
    BDirect.Flags |= BF_GHOSTED;                                                               /* A bit of nonsense otherwise :-) */
  }
  if (FileType != FT_VOC)
  {
    BSampRAdj.Flags |= BF_GHOSTED;
    BFreqShft.Flags |= BF_GHOSTED;
  }
  sprintf (DocPath, "%s%sSAMPLING.DOC", ProgramDrive, ProgramPath);                                     /* Set the help file name */
  AddButton (&CWindow, &POk, TRUE);
  AddButton (&CWindow, &BDScheme, FALSE);
  AddButton (&CWindow, &BDSchmSel, FALSE);
  AddButton (&CWindow, &BPiMin, FALSE);
  AddButton (&CWindow, &BPiMax, FALSE);
  AddButton (&CWindow, &BPiPulses, FALSE);
  AddButton (&CWindow, &BPiThres, FALSE);
  AddButton (&CWindow, &BBtMin, FALSE);
  AddButton (&CWindow, &BBtThres, FALSE);
  AddButton (&CWindow, &BLossyDet, FALSE);
  AddButton (&CWindow, &BApproxBT, FALSE);
  AddButton (&CWindow, &BEndSing, FALSE);
  AddButton (&CWindow, &BEndDoub, FALSE);
  AddButton (&CWindow, &BStdTres, FALSE);
  AddButton (&CWindow, &BDirect, FALSE);
  AddButton (&CWindow, &BDropEnd, FALSE);
  AddButton (&CWindow, &BShowInfo, FALSE);
  AddButton (&CWindow, &BFilter, FALSE);
  AddButton (&CWindow, &BSampRAdj, FALSE);
  AddButton (&CWindow, &BStdSync, FALSE);
  AddButton (&CWindow, &BFreqShft, FALSE);
  AddButton (&CWindow, &PCancel, FALSE);
  AddButton (&CWindow, &PDefault, FALSE);
  AddButton (&CWindow, &PHelp, FALSE);
  strcpy (ValDecodingScheme, _SchemeNames[DecodingScheme]);
  sprintf (ValMinPilotPulseLength, "%ld", MinPilotPulseLength);
  sprintf (ValMaxPilotPulseLength, "%ld", MaxPilotPulseLength);
  sprintf (ValMinNumPilotPulses, "%ld", MinNumPilotPulses);
  sprintf (ValThresholdPilotPulse, "%ld", ThresholdPilotPulse);
  sprintf (ValFMinBitPulseLength, "%ld", FMinBitPulseLength);
  sprintf (ValThresholdBitPulse, "%ld", ThresholdBitPulse);
  sprintf (ValThresholdEndMarker, "%.2f", ThresholdEndMarker);
  sprintf (ValThresholdSingleEndMarker, "%.2f", ThresholdSingleEndMarker);
  sprintf (ValThresholdStandardLoading, "%ld", ThresholdStandardLoading);
  DrawWindow (&CWindow, NULL, WING_CLOSE, TRUE);
  Wprintf (&CWindow, 2,  2,  "{+1Decoding Scheme}N");
  WDrawShadowBox (&CWindow, 5,  0, 11, 62, "Pilot detection");
  Wprintf (&CWindow, 7,  2,  "{+1Minimum Pulse}N");         Wprintf (&CWindow, 7,  27, "{+1T}N");
  Wprintf (&CWindow, 7,  33, "{+1Maximum Pulse}N");         Wprintf (&CWindow, 7,  58, "{+1T}N");
  Wprintf (&CWindow, 9,  2,  "{+1Min Num Pulses}N");
  Wprintf (&CWindow, 9,  33, "{+1Pulse Threshold}N");       Wprintf (&CWindow, 9,  58, "{+1%%}N");
  WDrawShadowBox (&CWindow, 12, 0, 18, 62, "Bit detection");
  Wprintf (&CWindow, 14, 2,  "{+1Minimum Pulse}N");         Wprintf (&CWindow, 14,  27, "{+1T}N");
  Wprintf (&CWindow, 14, 33, "{+1Threshold \'0\'/\'1\'}N"); Wprintf (&CWindow, 14,  58, "{+1%%}N");
  Wprintf (&CWindow, 16, 2,  "{+1Lossy Detection}N");
  Wprintf (&CWindow, 16, 33, "{+1Use '1' = 2 * '0'}N");
  WDrawShadowBox (&CWindow, 19, 0, 23, 62, "End marker detection");
  Wprintf (&CWindow, 21, 2,  "{+1Single Pulse}N");          Wprintf (&CWindow, 21,  27, "{+1\'1\'}N");
  Wprintf (&CWindow, 21, 33, "{+1Joined Pulse}N");          Wprintf (&CWindow, 21,  58, "{+1\'1\'}N");
  WDrawShadowBox (&CWindow, 24, 0, 34, 62, "Other options");
  Wprintf (&CWindow, 26, 2,  "{+1Threshold Standard}N");    Wprintf (&CWindow, 26,  27, "{+1%%}N");
  Wprintf (&CWindow, 26, 33, "{+1Direct Recording}N");
  Wprintf (&CWindow, 28, 2,  "{+1Drop Partial End}N");
  Wprintf (&CWindow, 28, 33, "{+1Show Decode Graph}N");
  Wprintf (&CWindow, 30, 2,  "{+1Digital Filter}N");
  Wprintf (&CWindow, 30, 33, "{+1CoolEdit Adjust}N");
  Wprintf (&CWindow, 32, 2,  "{+1Standard Syncs}N");
  Wprintf (&CWindow, 32, 33, "{+1Frequency Shift}N");
  while (!Finished)
  {
    switch (HandleWindow (&CWindow, &ButtonP))
    {
      case WINH_CLOSE   : Cancelled = Finished = TRUE;
                          break;
      case WINH_RNOHAND : switch (ButtonP->ReactNum)
                          {
                            case 0  : Cancelled = Finished = TRUE; break;                                             /* (Cancel) */
                            case 1  : ResetVOCTimers ();                                                             /* (Default) */
                                      sprintf (ValMinPilotPulseLength, "%ld", MinPilotPulseLength);
                                      sprintf (ValMaxPilotPulseLength, "%ld", MaxPilotPulseLength);
                                      sprintf (ValMinNumPilotPulses, "%ld", MinNumPilotPulses);
                                      sprintf (ValThresholdPilotPulse, "%ld", ThresholdPilotPulse);
                                      sprintf (ValFMinBitPulseLength, "%ld", FMinBitPulseLength);
                                      sprintf (ValThresholdBitPulse, "%ld", ThresholdBitPulse);
                                      sprintf (ValThresholdEndMarker, "%.2f", ThresholdEndMarker);
                                      sprintf (ValThresholdSingleEndMarker, "%.2f", ThresholdSingleEndMarker);
                                      sprintf (ValThresholdStandardLoading, "%ld", ThresholdStandardLoading);
                                      strcpy (ValDecodingScheme, _SchemeNames[DecodingScheme]);
                                      UpdateButtonText (&CWindow, &BPiMin, BF_FITTXT);
                                      UpdateButtonText (&CWindow, &BPiMax, BF_FITTXT);
                                      UpdateButtonText (&CWindow, &BPiPulses, BF_FITTXT);
                                      UpdateButtonText (&CWindow, &BPiThres, BF_FITTXT);
                                      UpdateButtonText (&CWindow, &BBtMin, BF_FITTXT);
                                      UpdateButtonText (&CWindow, &BBtThres, BF_FITTXT);
                                      UpdateButtonText (&CWindow, &BEndSing, BF_FITTXT);
                                      UpdateButtonText (&CWindow, &BEndDoub, BF_FITTXT);
                                      UpdateButtonText (&CWindow, &BStdTres, BF_FITTXT);
                                      UpdateButtonText (&CWindow, &BFilter, BF_FITTXT);
                                      UpdateButtonText (&CWindow, &BStdSync, BF_FITTXT);
                                      UpdateButtonText (&CWindow, &BDirect, BF_FITTXT);
                                      UpdateButtonText (&CWindow, &BShowInfo, BF_FITTXT);
                                      UpdateButtonText (&CWindow, &BSampRAdj, BF_FITTXT);
                                      UpdateButtonText (&CWindow, &BFreqShft, BF_FITTXT);
                                      DrawButton (&CWindow, &BLossyDet);
                                      DrawButton (&CWindow, &BApproxBT);
                                      DrawButton (&CWindow, &BDropEnd);
                                      UpdateButtonText (&CWindow, &BDScheme, BF_FITTXT);
                                      break;
                            case 2  : Finished = TRUE; break;                                                             /* (Ok) */
                            case 13 : MakeComboWindow (&SWindow, &CWindow, COMT_SELECT, COMF_VSLIDER | COMF_SYNCSLIDER |
                                                       COMF_KEYSELECT | COMF_EXCLUSIVE, CWindow.StartX + BDScheme.StartX + 1,
                                                       CWindow.StartY + BDScheme.StartY + 16, 0, 1024, BDScheme.FWidth + 2,
                                                       BDScheme.FWidth, DS_MAX, NULL, FALSE, 0, NULL, DBLACK, DWHITE, 0, 0);
                                      for (Cnt = 0 ; Cnt < DS_MAX ; Cnt ++)
                                      {
                                        SEntry.Selected = 0;
                                        SEntry.Identifier = Cnt;
                                        strcpy (SEntry.Text, _SchemeNames[Cnt]);
                                        if (!InsertComboEntry (&SWindow, &SEntry, NULL, -1))
                                          FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
                                      }
                                      SWindow.LastSelected = DecodingScheme;                    /* Auto-select the current scheme */
                                      PrintComboEntries (&SWindow);
                                      Finished = FALSE;
                                      while (!Finished)
                                      {
                                        MouseStatus ();
                                        if (HandleComboWindow (&SWindow, &SelectedLine) == WINH_LINEFIRST)
                                        {
                                          Finished = TRUE;
                                          DecodingScheme = SWindow.LastSelected;
                                        }
                                        else
                                        {
                                          HandleWindow (&(SWindow.ComboWindow), &ButtonP);            /* (Just read the keyboard) */
                                         if (Mouse.Xx < SWindow.OverlayWindow.A_StartX || Mouse.Xx > SWindow.OverlayWindow.A_EndX ||
                                              Mouse.Yy < SWindow.OverlayWindow.A_StartY || Mouse.Yy > SWindow.OverlayWindow.A_EndY)
                                            if (Mouse.Left)                                /* Left click outside the select box ? */
                                            {
                                              Finished = TRUE;
                                              DelayLeftMouseButton ();
                                            }
                                        }
                                        if (GlobalKey == K_ESC)
                                          Finished = TRUE;
                                        else if (GlobalKey == K_ENTER)
                                        { DecodingScheme = SWindow.LastSelected; Finished = TRUE; }
                                      }
                                      GlobalKey = 0x0000;
                                      Finished = FALSE;
                                      DestroyComboWindow (&SWindow, 0);
                                      strcpy (ValDecodingScheme, _SchemeNames[DecodingScheme]);
                                      UpdateButtonText (&CWindow, &BDScheme, BF_FITTXT);
                                      break;
                            case 14 : if (RequestFloppy (ProgramDrive[0], ResourceFloppyId))
                                        FileViewRequester (80, 100, 128, 80, 80, 25, DocPath, TRUE, "VOC Decoding Help");
                                      break;
                          }
                          break;
      case WINH_EDITEND : if (ButtonP->Flags & BF_FITINPUT)                                            /* Contents were changed ? */
                            switch (ButtonP->ReactNum)
                            {
                              case 3  : MinPilotPulseLength = atol (ValMinPilotPulseLength);
                                        sprintf (ValMinPilotPulseLength, "%ld", MinPilotPulseLength);
                                        UpdateButtonText (&CWindow, &BPiMin, BF_FITTXT);
                                        break;
                              case 4  : MaxPilotPulseLength = atol (ValMaxPilotPulseLength);
                                        sprintf (ValMaxPilotPulseLength, "%ld", MaxPilotPulseLength);
                                        UpdateButtonText (&CWindow, &BPiMax, BF_FITTXT);
                                        break;
                              case 5  : MinNumPilotPulses = atol (ValMinNumPilotPulses);
                                        sprintf (ValMinNumPilotPulses, "%ld", MinNumPilotPulses);
                                        UpdateButtonText (&CWindow, &BPiPulses, BF_FITTXT);
                                        break;
                              case 6  : ThresholdPilotPulse = atol (ValThresholdPilotPulse);
                                        sprintf (ValThresholdPilotPulse, "%ld", ThresholdPilotPulse);
                                        UpdateButtonText (&CWindow, &BPiThres, BF_FITTXT);
                                        break;
                              case 7  : FMinBitPulseLength = atol (ValFMinBitPulseLength);
                                        sprintf (ValFMinBitPulseLength, "%ld", FMinBitPulseLength);
                                        UpdateButtonText (&CWindow, &BBtMin, BF_FITTXT);
                                        break;
                              case 8  : ThresholdBitPulse = atol (ValThresholdBitPulse);
                                        sprintf (ValThresholdBitPulse, "%ld", ThresholdBitPulse);
                                        UpdateButtonText (&CWindow, &BBtThres, BF_FITTXT);
                                        break;
                              case 9  : ThresholdSingleEndMarker = atof (ValThresholdSingleEndMarker);
                                        sprintf (ValThresholdSingleEndMarker, "%.2f", ThresholdSingleEndMarker);
                                        UpdateButtonText (&CWindow, &BEndSing, BF_FITTXT);
                                        break;
                              case 10 : ThresholdEndMarker = atof (ValThresholdEndMarker);
                                        sprintf (ValThresholdEndMarker, "%.2f", ThresholdEndMarker);
                                        UpdateButtonText (&CWindow, &BEndDoub, BF_FITTXT);
                                        break;
                              case 11 : ThresholdStandardLoading = atol (ValThresholdStandardLoading);
                                        sprintf (ValThresholdStandardLoading, "%ld", ThresholdStandardLoading);
                                        UpdateButtonText (&CWindow, &BStdTres, BF_FITTXT);
                                        break;
                            }
    }
    if (GlobalKey == K_ESC)
      Cancelled = Finished = TRUE;
    MouseStatus ();
  }
  DestroyWindow (&CWindow);
  GlobalKey = 0x0000;
  if (!Cancelled && (DecodingScheme == DS_SLOCK2 || DecodingScheme == DS_SLOCK3 || DecodingScheme == DS_SLOCK4 ||
                     DecodingScheme == DS_SLOCK5 || DecodingScheme == DS_SLOCK6 || DecodingScheme == DS_SLOCK7))
    GlobalDecodingScheme = DecodingScheme;
  else
    GlobalDecodingScheme = DS_NONE;
  if (DecodingScheme == DS_SLOCK4 || DecodingScheme == DS_SLOCK5 || DecodingScheme == DS_SLOCK6 || DecodingScheme == DS_SLOCK7)
    UsedMinNumPilotPulses = _Max (MinNumPilotPulses, 500);                           /* SpeedLock 4-7 pilots must be > 500 pulses */
  else
    UsedMinNumPilotPulses = MinNumPilotPulses;
  return (!Cancelled);
}

bool SetupVOCWriteTimers (short WhichWindow, word *SelectedWriteSampleRate)

/**********************************************************************************************************************************/
/* Pre   : (We are about to write a VOC file).                                                                                    */
/* Post  : The user has been asked the sample rate to be used. This value must be between 4000 and 60000 (Hz).                    */
/*         The return value is TRUE if the user choose 'Ok' or FALSE if the user cancelled.                                       */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct ButtonInfo *ButtonP;
  char MessageBuffer[10];
  bool Accepted  = FALSE;
  long TestVal;

  if (!TPConfig.PlaySoundBlaster && WhichWindow >= 0 && MSWindowsRunning)
  {
    if (QuestionRequester (-1, -1, FALSE, FALSE, NULL, NULL, NULL, NULL, NULL, NULL,
                           "Parallel Port playing may not work correctly in Windows 95.\n"
                           "Do you want to play anyway?"))
      return (FALSE);
  }
  SimulatePlayBlock = FALSE;                                                                    /* Signal: this is the real thing */
  if (TPConfig.PlaySoundBlaster || WhichWindow < 0)                                                /* Not playing parallel port ? */
  {
    sprintf (MessageBuffer, "%lu", WriteSampleRate);
    while (!Accepted)
    {
      if (InputRequester (-1, -1, NULL, NULL, MessageBuffer, 6, FALSE, FALSE, "Set Sample Rate", "{CSelect sample rate}C"))
      {
        if (MessageBuffer[0] == '\0')                                                                           /* Empty buffer ? */
          TestVal = 0;                                                                                             /* Force error */
        else
          TestVal = atol (MessageBuffer);
        if (TestVal >= 4000 && TestVal < 60000L)                                                            /* Max rate is 60 kHz */
          Accepted = TRUE;
        else
          ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - The value is out of range\n        Use values 4000 through 60000");
      }
      else                                                                                                         /* (Cancelled) */
        return (FALSE);
    }
    WriteSampleRate = TestVal;                                                                      /* Set the chosen sample rate */
    *SelectedWriteSampleRate = (word)(WriteSampleRate & 0x0000FFFF);                                             /* And export it */
  }
  if (WhichWindow >= 0)                                                                                          /* Doing audio ? */
  {
    _SBWindow.StartX = (WhichWindow == 0 ? 379 : 30);                                           /* Paint it over the OTHER window */
    strcpy (_SBWHeader, TPConfig.PlaySoundBlaster ? "SoundBlaster Playing" : "Parallel Port Playing");
    DrawWindow (&_SBWindow, NULL, 0, TRUE);
    WDrawBox (&_SBWindow, FALSE, DBLUE, -3, -1, _SBWindow.Width + 2, _SBWindow.Height + 2);               /* Make it more visible */
    WDrawBox (&_SBWindow, FALSE, DBLUE, -4, -2, _SBWindow.Width + 3, _SBWindow.Height + 3);
    __SBWindow = &_SBWindow;                                                            /* Export the SoundBlaster Playing window */
    PlayTimeUpdated = FALSE;
    if (!TPConfig.PlaySoundBlaster)                                                                      /* Doing parallel port ? */
    {                                                                                                        /* Wait for the user */
      Wprintf (&_SBWindow, 2, TEXT_CENTER, "Press any key to start ...");
      while (GlobalKey == 0x0000)
        HandleWindow (&_SBWindow, &ButtonP);
      GlobalKey = 0x0000;                                                                                       /* (Drop the key) */
      ClearWindow (&_SBWindow);                                                                           /* 'Restore' the window */
    }
    if (UsePlayTimeClock)
    {
      _StartPlayClock (TRUE);
      Wprintf (&_SBWindow, 3, TEXT_CENTER, "Press [ESC] to abort");
    }
    else
    {
      _StartPlayClock (FALSE);                                                                   /* (Only check for IRQ time-out) */
      Wprintf (&_SBWindow, 2, TEXT_CENTER, "Press [ESC] to abort");
    }
    _WriteBuffer._Fh = -1;                                                                /* (Just in case everything goes wrong) */
  }
  InitVOCWriteVars = TRUE;                                                            /* Signal: need to start with no stacks etc */
  if (!TPConfig.PlaySoundBlaster && !ParallelRunning && WhichWindow >= 0)
    Par_PlayInit ();                                                                         /* Init the main emulation variables */
  return (TRUE);
}

bool FlushWriteBuffer (bool StopImmediately)

/**********************************************************************************************************************************/
/* Pre   : `StopImmediately' is used for SoundBlaster only. If TRUE, the DMA must stop instantly.                                 */
/* Post  : Any remaining buffer has been written to the file, the Sound header has been updated.                                  */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  dword SampleLength;

  if (_WriteBuffer._Fh < 0)                                                                          /* Flushing a SoundBlaster ? */
  {
    if (!TPConfig.PlaySoundBlaster)                                         /* Parallel port interface rather than SoundBlaster ? */
      _WriteBuffer._State = 0x80;                                                                   /* Force into 'silence' state */
    else if (!StopImmediately)
    {
      if (_WriteBuffer._State != 0x80)                                                      /* No pause done yet (finishing edge) */
        if (!_WritePulse (&_WriteBuffer._State, 3500))                                                   /* Finish the last pulse */
          StopImmediately = TRUE;
      _WriteBuffer._State = 0x80;                                                                   /* Force into 'silence' state */
      if (!StopImmediately)
      {
        while (_WriteBuffer._BufferIndex < _WriteBuffer._BufferLength)
          *(_WriteBuffer._SBBuffer[_WriteBuffer._ID01] + (_WriteBuffer._BufferIndex ++)) = 0x80;        /* Fill up with 'silence' */
        (*SB_BufferReady) = _WriteBuffer._ID01;
        if (TPConfig.PlaySoundBlaster)
        {
          if (!SoundBlasterInfo.DoingDMA)                                               /* Just wrote first buffer - no DMA yet ? */
            SB_PlayAutoInitDMA (_WriteBuffer._SBBuffer[0], (word)WriteSampleRate, _WriteBuffer._BufferLength << 1);    /* Start ! */
        }
      }
      while ((*SB_DMAComplete) != _WriteBuffer._ID01 && !StopImmediately)                   /* Wait for this buffer to get played */
      {
        if (PlayTimeUpdated)  { PlayTimeUpdated = FALSE; Wprintf (&_SBWindow, 1, TEXT_CENTER, PlayTimeString); }
        if (*SB_DMAOverrun)
          StopImmediately = TRUE;
        else if (SB_DMANoIRQTimeOut)
          StopImmediately = TRUE;
        else if (kbhit ())
          if (getch () == 0x1B)
            StopImmediately = TRUE;
      }
      if (!StopImmediately)
      {
        _WriteBuffer._ID01 ^= 0x01;                                                   /* Clear the other buffer to avoid a 'plop' */
        for (_WriteBuffer._BufferIndex = 0 ; _WriteBuffer._BufferIndex < _WriteBuffer._BufferLength ; _WriteBuffer._BufferIndex ++)
          *(_WriteBuffer._SBBuffer[_WriteBuffer._ID01] + _WriteBuffer._BufferIndex) = 0x80;
        (*SB_BufferReady) = _WriteBuffer._ID01;
        while ((*SB_DMAComplete) != _WriteBuffer._ID01 && !StopImmediately)                     /* Wait for this buffer to finish */
        {
          if (PlayTimeUpdated)  { PlayTimeUpdated = FALSE; Wprintf (&_SBWindow, 1, TEXT_CENTER, PlayTimeString); }
          if (*SB_DMAOverrun)
            StopImmediately = TRUE;
          else if (SB_DMANoIRQTimeOut)
            StopImmediately = TRUE;
          else if (kbhit ())
            if (getch () == 0x1B)
              StopImmediately = TRUE;
        }
      }
    }
    if (!_KeepSBWindow)                                                                                   /* (Not just pausing ?) */
    {
      _StopPlayClock ();
      DestroyWindow (&_SBWindow);
    }
    _WriteBuffer._BufferIndex = 0;
    if (TPConfig.PlaySoundBlaster)
    {
      if (SoundBlasterInfo.DoingDMA)
        SB_StopDMA ();                                                                          /* Shut down immediately after it */
      SB_Speaker (FALSE);                                                                                           /* Silence... */
    }
    else
      if (ParallelRunning)
        Par_PlayStop ();
    (*SB_DMAOverrun) = RealSB_DMAOverrun;                                                                /* Reset for key stepout */
    if (_DEBUGSBDMA)
    {
      ErrorRequester (-1, -1, NULL, NULL, "SB DMA Debug", "{BDEBUG INFO}B - Main buffer: %06lX\n"
                                                              "             Buf[0]     : %06lX\n"
                                                              "             Buf[1]     : %06lX\n"
                                                              "             Size       : %u",
                      _AudioBuffer, _WriteBuffer._SBBuffer[0], _WriteBuffer._SBBuffer[1], _WriteBuffer._BufferLength);
      _DEBUGSBDMA = FALSE;
    }
    return (TRUE);
  }

  if (_WriteBuffer._Dirty)                                                                                  /* Buffer not empty ? */
  {
    if (write (_WriteBuffer._Fh, _WriteBuffer._Buffer, _WriteBuffer._BufferIndex) != _WriteBuffer._BufferIndex)
      return (FALSE);
    _WriteBuffer._FileIndex += _WriteBuffer._BufferIndex;
    _WriteBuffer._BufferIndex = 0;
    _WriteBuffer._Dirty = FALSE;
  }
  if (!HandlingZ80EmRAW && !HandlingCSW)
  {
    SampleLength = _WriteBuffer._FileIndex - _WriteBuffer._InitialFileIndex - 4;                /* Total size of the sample block */
                                                                                              /* (4 is the VOC header adjustment) */
    if (lseek (_WriteBuffer._Fh, _WriteBuffer._InitialFileIndex + 1, SEEK_SET) == -1)           /* Go back to the sample's header */
      return (FALSE);
    if (write (_WriteBuffer._Fh, &SampleLength, 3) != 3)                                               /* Write the sample length */
      return (FALSE);
    if (lseek (_WriteBuffer._Fh, _WriteBuffer._FileIndex, SEEK_SET) == -1)                  /* And go back to the end of the file */
      return (FALSE);
  }
  _WriteBuffer._InitialFileIndex = _WriteBuffer._FileIndex;
  return (TRUE);
}

char *GetSB_DMANoIRQTimeOutMessage (void)

/**********************************************************************************************************************************/
/* Pre   : (The play flag SB_DMANoIRQTimeOut was set).                                                                            */
/* Post  : An error message depending on the occured number of IRQs is returned.                                                  */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  if (_SB_DMACompleteCount == 0)
    return ("No IRQ occured. Make sure you set the BLASTER environment variable correctly!");
  if (_SB_DMACompleteCount == 1)
    return ("Your sound card does not appear to support AutoInit DMA!");
  return ("An IRQ time-out occured while waiting for DMA to finish.");
}

void AllocateAudioBuffer (void)

/**********************************************************************************************************************************/
/* Pre   : None.                                                                                                                  */
/* Post  : The audio play DMA buffer has been allocated.                                                                          */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  word AudioSegment;

  DOS_memalloc (0x0FFF, &AudioSegment, &_AudioBufferSelector);                                             /* This is 65520 bytes */
  if (AudioSegment != 0)
  {
    _AudioBufferAllocated = TRUE;
    _AudioBuffer = (byte *)((dword)AudioSegment << 4);
  }
  else
    _AudioBufferAllocated = FALSE;
}

void ReleaseAudioBuffer (void)

/**********************************************************************************************************************************/
/* Pre   : None.                                                                                                                  */
/* Post  : The audio play DMA buffer has been released.                                                                           */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  if (_AudioBufferAllocated)
  {
    DOS_memfree (_AudioBufferSelector);
    _AudioBufferAllocated = FALSE;
  }
}

void SetSoftMixerOutput (void)

/**********************************************************************************************************************************/
/* Pre   : None.                                                                                                                  */
/* Post  : The soundcard soft mixer's output registers have been adjusted by the user.                                            */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  short NMaster;
  short NGain;
  short NDACLevel;
  short NTreble;
  short NBass;
  struct WindowInfo CWindow     = {200, 160, 240, 155, 240, 155, 240, 155, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                   WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER, NULL, "Adjust Sound Volume"};
  struct ButtonInfo POk         = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 50, 130, 8, DBLACK, DWHITE, NULL, 0, "@Ok", NULL};
  struct ButtonInfo PPilot      = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 135, 130, 8, DBLACK, DWHITE, NULL, 1, "@Pilot", NULL};
  struct SliderButton SMaster   = {10,  9, 128, DBLACK, DWHITE, 1, BF_KEYSELECT, 0, NULL};
  struct SliderButton SGain     = {10, 29, 128, DBLACK, DWHITE, 2, BF_KEYSELECT, 0, NULL};
  struct SliderButton SDACLevel = {10, 59, 128, DBLACK, DWHITE, 3, BF_KEYSELECT, 0, NULL};
  struct SliderButton STreble   = {10, 79, 128, DBLACK, DWHITE, 4, BF_KEYSELECT, 0, NULL};
  struct SliderButton SBass     = {10, 99, 128, DBLACK, DWHITE, 5, BF_KEYSELECT, 0, NULL};
  struct ButtonInfo     *ButtonP;
  struct SliderButton   *SliderP;
  bool                   Finished = FALSE;
  bool                   PlayingPilot = FALSE;
  long                   NumSamples;
  long                   CntPulses;
  long                   Halftime;
  byte                  *BufferIndex;

  SMaster.NumSteps   = SoundBlasterInfo.MixerMax.MasterVolume;
  SGain.NumSteps     = SoundBlasterInfo.MixerMax.OutputGain;
  SDACLevel.NumSteps = SoundBlasterInfo.MixerMax.DACLevel;
  STreble.NumSteps   = SoundBlasterInfo.MixerMax.Treble;
  SBass.NumSteps     = SoundBlasterInfo.MixerMax.Bass;
  NMaster            = SoundBlasterInfo.Mixer.MasterVolume;
  NGain              = SoundBlasterInfo.Mixer.OutputGain;
  NDACLevel          = SoundBlasterInfo.Mixer.DACLevel;
  NTreble            = SoundBlasterInfo.Mixer.Treble;
  NBass              = SoundBlasterInfo.Mixer.Bass;
  SMaster.Contents   = &NMaster;
  SGain.Contents     = &NGain;
  SDACLevel.Contents = &NDACLevel;
  STreble.Contents   = &NTreble;
  SBass.Contents     = &NBass;
  AddButton (&CWindow, &POk, TRUE);
  AddButton (&CWindow, &PPilot, FALSE);
  AddSliderButton (&CWindow, &SMaster, FALSE);
  if (SoundBlasterInfo.MixerType == SBMTYPE_SB16)
    AddSliderButton (&CWindow, &SGain, FALSE);
  AddSliderButton (&CWindow, &SDACLevel, FALSE);
  if (SoundBlasterInfo.MixerType == SBMTYPE_SB16)
    AddSliderButton (&CWindow, &STreble, FALSE);
  if (SoundBlasterInfo.MixerType == SBMTYPE_SB16)
    AddSliderButton (&CWindow, &SBass, FALSE);
  DrawWindow (&CWindow, NULL, WING_CLOSE, TRUE);
  Wprintf (&CWindow,    1, 25, "{+1Master volume}N");
  if (SoundBlasterInfo.MixerType == SBMTYPE_SB16)
    Wprintf (&CWindow,  3, 25, "{+1Loudness}N");
  Wprintf (&CWindow,    6, 25, "{+1Play level}N");
  if (SoundBlasterInfo.MixerType == SBMTYPE_SB16)
    Wprintf (&CWindow,  8, 25, "{+1Treble}N");
  if (SoundBlasterInfo.MixerType == SBMTYPE_SB16)
    Wprintf (&CWindow, 10, 25, "{+1Bass}N");
  WriteSampleRate = 22010;                                                   /* Want a samplerate that doesn't slow down too much */
  _OpenWriteBuffer (-1, 0L, TRUE);                                                        /* Make sure the buffers are there etc. */
  NumSamples = _NumSamples (NS_LENPILOT, _WriteBuffer._TRate);                            /* Determine samples/normal speed pilot */
  Halftime = _WriteBuffer._BufferLength / NumSamples;                                                /* Determine pulses to write */
  _WriteBuffer._BufferLength = Halftime * NumSamples;                                                            /* Make loop fit */
  BufferIndex = _WriteBuffer._SBBuffer[0];
  for (CntPulses = 0 ; CntPulses < Halftime * 2 ; CntPulses ++, BufferIndex += NumSamples)            /* Write both buffer halves */
    memset (BufferIndex, (byte)(CntPulses & 0x01) ? 0xF0 : 0x10, NumSamples);
  while (!Finished)
  {
    switch (HandleWindow (&CWindow, &ButtonP))
    {
      case WINH_RNOHAND    : switch (ButtonP->ReactNum)
                             {
                               case 0 : Finished = TRUE; break;                                                           /* (Ok) */
                               case 1 : if (PlayingPilot)                                                              /* (Pilot) */
                                        {
                                          if (SoundBlasterInfo.DoingDMA)
                                            SB_StopDMA ();
                                          SB_Speaker (FALSE);
                                          PlayingPilot = FALSE;
                                        }
                                        else
                                        {
                                          PlayingPilot = TRUE;
                                          SB_PlayAutoInitDMA (_WriteBuffer._SBBuffer[0], (word)WriteSampleRate,
                                                              _WriteBuffer._BufferLength << 1);       /* Start playing the pilot! */
                                        }
                                        break;
                             }
                             break;
      case WINH_CLOSE      : Finished = TRUE; break;
      case WINH_SLIDERMOVE : SliderP = (struct SliderButton *)(ButtonP->Font);                              /* One of the sliders */
                             switch (SliderP->ReactNum)
                             {
                               case 1 : SB_WriteMixer (SBM_MASTERVOL, NMaster); break;                         /* (Master volume) */
                               case 2 : SB_WriteMixer (SBM_OUTPUTGAIN, NGain); break;                            /* (Output gain) */
                               case 3 : SB_WriteMixer (SBM_DACLEVEL, NDACLevel); break;                            /* (DAC level) */
                               case 4 : SB_WriteMixer (SBM_TREBLELEVEL, NTreble); break;                              /* (Treble) */
                               case 5 : SB_WriteMixer (SBM_BASSLEVEL, NBass); break;                                    /* (Bass) */
                             }
                             break;
    }
    if (GlobalKey == K_ESC)
      Finished = TRUE;
    MouseStatus ();
  }
  if (SoundBlasterInfo.DoingDMA)                                                                /* (Still) playing pilot signal ? */
  {
    SB_StopDMA ();
    SB_Speaker (FALSE);
  }
  DestroyWindow (&CWindow);
  _CloseFileBuffer (FALSE);                                                                               /* Release write buffer */
}

bool SetSoftMixerInput (void)

/**********************************************************************************************************************************/
/* Pre   : None.                                                                                                                  */
/* Post  : The soundcard soft mixer's input registers have been adjusted by the user and possibly saved.                          */
/*         The return value is TRUE if `Ok' was selected, FALSE if `Cancel' was selected.                                         */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct WindowInfo CWindow   = {207, 78, 226, 269, 226, 269, 226, 269, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                 WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER | WINF_EXCLUSIVE | WINF_NOMOVE,
                                 NULL, "Audio Input"};
  struct ButtonInfo PCancel   = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 16, 244, 8, DBLACK, DWHITE, NULL, 0, "@Cancel", NULL};
  struct ButtonInfo PSave     = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 87, 244, 8, DBLACK, DWHITE, NULL, 2, "@Save", NULL};
  struct ButtonInfo POk       = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 158, 244, 8, DBLACK, DWHITE, NULL, 1, "@Ok", NULL};
  struct SliderButton SInput  = {10, 150, 128, DBLACK, DWHITE, 10, BF_KEYSELECT, 0, NULL};
  struct SliderButton STreble = {10, 170, 128, DBLACK, DWHITE, 11, BF_KEYSELECT, 0, NULL};
  struct SliderButton SBass   = {10, 190, 128, DBLACK, DWHITE, 12, BF_KEYSELECT, 0, NULL};
  struct SliderButton SGain   = {10, 210, 128, DBLACK, DWHITE, 13, BF_KEYSELECT, 0, NULL};
  struct SliderButton   *SliderP;
  struct ButtonInfo     *ButtonP;
  byte         register  CntX;
  byte         register  XAdd;
  byte         register  ShiftValue;
  byte         register  CntY;
  byte                   PrevSample;
  byte                   Sample;
  byte                   RangeLow;
  byte                   RangeHi;
  byte                  *SWinStart;
  dword        register *SLinStart;
  dword        register *WLinStart;
  short                  NInput;
  short                  NTreble;
  short                  NBass;
  short                  NGain;
  bool                   Finished = FALSE;
  bool                   Cancelled = FALSE;
  byte                   Val;
  byte                   ValNo;
  char                  *Ch;

  if (SoundBlasterInfo.MixerType == SBMTYPE_AD1848)
  {
    ErrorRequester (-1, -1, NULL, NULL, NULL, "Your soundcard's sample/mixer chip (AD1848) is not supported yet. Sorry...");
    return (FALSE);
  }
  else if (SoundBlasterInfo.Detected &&
           SoundBlasterInfo.DSPMajorVersion >= 2 &&
           SoundBlasterInfo.MixerType == SBMTYPE_NONE)
    return (TRUE);                                               /* (Undetermined clone with separate play and record/mixer chip) */
  SInput.NumSteps  = SoundBlasterInfo.MixerMax.LineInLevel;
  STreble.NumSteps = SoundBlasterInfo.MixerMax.Treble;
  SBass.NumSteps   = SoundBlasterInfo.MixerMax.Bass;
  SGain.NumSteps   = SoundBlasterInfo.MixerMax.InputGain;
  NInput           = SoundBlasterInfo.Mixer.LineInLevel;
  NTreble          = SoundBlasterInfo.Mixer.Treble;
  NBass            = SoundBlasterInfo.Mixer.Bass;
  NGain            = SoundBlasterInfo.Mixer.InputGain;
  SInput.Contents  = &NInput;
  STreble.Contents = &NTreble;
  SBass.Contents   = &NBass;
  SGain.Contents   = &NGain;
  Ch = TPConfig.SampleValues;
  Val = 0;
  ValNo = 0;
  while (!Finished)                                           /* First parse the configuration setting to set the register values */
  {
    if (isdigit (*Ch))
      Val = Val * 10 + *Ch - '0';
    else if (*Ch == ',' || *Ch == '\0')
    {
      if (Val != 0)
        switch (ValNo)
        {
          case 0 : if (SoundBlasterInfo.MixerMax.LineInLevel != 0)
                   {
                     Val %= SoundBlasterInfo.MixerMax.LineInLevel;
                     NInput = Val;
                     SB_WriteMixer (SBM_LINEINLEVEL, NInput);
                   }
                   break;
          case 1 : if (SoundBlasterInfo.MixerMax.Treble != 0)
                   {
                     Val %= SoundBlasterInfo.MixerMax.Treble;
                     NTreble = Val;
                     SB_WriteMixer (SBM_TREBLELEVEL, NTreble);
                   }
                   break;
          case 2 : if (SoundBlasterInfo.MixerMax.Bass != 0)
                   {
                     Val %= SoundBlasterInfo.MixerMax.Bass;
                     NBass = Val;
                     SB_WriteMixer (SBM_BASSLEVEL, NBass);
                   }
                   break;
          case 3 : if (SoundBlasterInfo.MixerMax.InputGain != 0)
                   {
                     Val %= SoundBlasterInfo.MixerMax.InputGain;
                     NGain = Val;
                     SB_WriteMixer (SBM_INPUTGAIN, NGain);
                   }
                   break;
        }
      Val = 0;
      ValNo ++;
      if (*Ch == '\0')
        Finished = TRUE;
    }
    Ch ++;
  }
  if (SoundBlasterInfo.MixerType != SBMTYPE_SB16)
  {
    CWindow.StartY += 30;                                                            /* Adjust the window for the missing buttons */
    CWindow.Height -= 60;
    CWindow.MinHeight = CWindow.MaxHeight = CWindow.Height;
    POk.StartY -= 60;
    PCancel.StartY -= 60;
  }
  AddButton (&CWindow, &POk, TRUE);
  AddSliderButton (&CWindow, &SInput, FALSE);
  if (SoundBlasterInfo.MixerType == SBMTYPE_SB16)
  {
    AddSliderButton (&CWindow, &STreble, FALSE);
    AddSliderButton (&CWindow, &SBass, FALSE);
    AddSliderButton (&CWindow, &SGain, FALSE);
  }
  AddButton (&CWindow, &PCancel, FALSE);
  AddButton (&CWindow, &PSave, FALSE);
  DrawWindow (&CWindow, NULL, WING_CLOSE, TRUE);
  WDrawLine (&CWindow, SystemColors.ShadowDark, 47, 4, 178, 4);
  WDrawLine (&CWindow, SystemColors.ShadowDark, 47, 4, 47, 135);
  WDrawLine (&CWindow, SystemColors.ShadowLight, 47, 135, 178, 135);
  WDrawLine (&CWindow, SystemColors.ShadowLight, 178, 4, 178, 135);
  WDrawBox (&CWindow, TRUE, DBLACK, 49, 6, 176, 133);
  Wprintf (&CWindow, 15, 25, "{+1Input level}N");
  if (SoundBlasterInfo.MixerType == SBMTYPE_SB16)
  {
    Wprintf (&CWindow, 17, 25, "{+1Treble}N");
    Wprintf (&CWindow, 19, 25, "{+1Bass}N");
    Wprintf (&CWindow, 21, 25, "{+1Loudness}N");
  }
  SB_WriteMixer (SBM_INPUTCTRL, SBI_MICOFF);                                                               /* Turn all inputs off */
  SB_WriteMixer (SBM_INPUTCTRL, SBI_CDOFF);
  SB_WriteMixer (SBM_INPUTCTRL, SBI_LINEINON);                                                               /* Except the LineIn */
  SB_WriteMixer (SBM_OUTPUTCTRL, SBO_MICOFF);                                                             /* Turn all outputs off */
  SB_WriteMixer (SBM_OUTPUTCTRL, SBO_CDOFF);
  SB_WriteMixer (SBM_OUTPUTCTRL, SBO_FMOFF);
  SB_WriteMixer (SBM_OUTPUTCTRL, SBO_LINEINON);                                                              /* Except the LineIn */
  SWinStart = Screen.DisplayStart + 80 * (CWindow.StartY + 6) + (CWindow.StartX + 49) / 8;
  Finished = FALSE;
  while (!Finished)
  {
    switch (HandleWindow (&CWindow, &ButtonP))
    {
      case WINH_RNOHAND    : switch (ButtonP->ReactNum)
                             {
                               case 0 : Finished = Cancelled = TRUE; break;                                           /* (Cancel) */
                               case 1 : Finished = TRUE; break;                                                           /* (Ok) */
                               case 2 : sprintf (TPConfig.SampleValues, "%d,%d,%d,%d",                                  /* (Save) */
                                                 SoundBlasterInfo.Mixer.LineInLevel,
                                                 SoundBlasterInfo.Mixer.Treble,
                                                 SoundBlasterInfo.Mixer.Bass,
                                                 SoundBlasterInfo.Mixer.InputGain);
                                        WriteConfig ();
                                        SoundErrorBeeper (BEEP_SHORT);
                                        break;
                             }
                             break;
      case WINH_CLOSE      : Finished = TRUE; break;
      case WINH_SLIDERMOVE : SliderP = (struct SliderButton *)(ButtonP->Font);                              /* One of the sliders */
                             switch (SliderP->ReactNum)
                             {
                               case 10 : SB_WriteMixer (SBM_LINEINLEVEL, NInput); break;                        /* (LineIn level) */
                               case 11 : SB_WriteMixer (SBM_TREBLELEVEL, NTreble); break;                             /* (Treble) */
                               case 12 : SB_WriteMixer (SBM_BASSLEVEL, NBass); break;                                   /* (Bass) */
                               case 13 : SB_WriteMixer (SBM_INPUTGAIN, NGain); break;                                   /* (Gain) */
                             }
                             break;
    }

    memset (WorkBuffer, 0, 2048);                                                              /* Clear the sample window storage */
    memset (WorkBuffer + 1024, 255, 16);                                                              /* Except for the null-line */
    PrevSample = SB_SampleByte () / 2;                                    /* Read first sample - this sets the left-most position */
    if (*SB_DMAOverrun)                                                                             /* Did a DSP time-out occur ? */
    {
      __DSPTimeOut: SB_Reset ();
                    ErrorRequester (-1, -1, NULL, NULL, NULL, "Your soundcard's sample/mixer chip does not respond...");
                    return (FALSE);
    }
    *(WorkBuffer + 16 * PrevSample) = 128;
    XAdd = 0;
    ShiftValue = 64;
    for (CntX = 1 ; CntX < 128 ; CntX ++)                                                               /* 127 more samples to go */
    {
      Sample = SB_SampleByte () / 2;                                                                          /* Read next sample */
      if (*SB_DMAOverrun)                                                                           /* Did a DSP time-out occur ? */
        goto __DSPTimeOut;
      if (Sample == PrevSample)
        RangeLow = RangeHi = Sample;
      else if (Sample > PrevSample)
      { RangeLow = PrevSample + 1 ; RangeHi = Sample; }
      else
      { RangeLow = Sample ; RangeHi = PrevSample - 1; }
      for (CntY = 0 ; CntY < 128 ; CntY ++)                                 /* Draw the line between this and the previous sample */
        if (CntY >= RangeLow && CntY <= RangeHi)
          *(WorkBuffer + 16 * CntY + XAdd) |= ShiftValue;
      if (!(ShiftValue >>= 1))
      { XAdd ++; ShiftValue = 128; }
      PrevSample = Sample;
    }

    SelectPlane (2);                                            /* Isn't that nice? One bitplane only, resulting in a red line... */
    SLinStart = (dword *)SWinStart;                                                       /* (And a blue mouse inside the box :-) */
    WLinStart = (dword *)WorkBuffer;
    for (CntY = 128 ; CntY ; CntY --)                                                    /* Quickly draw the entire sample window */
    {
      *(SLinStart ++) = *(WLinStart ++);
      *(SLinStart ++) = *(WLinStart ++);
      *(SLinStart ++) = *(WLinStart ++);
      *(SLinStart ++) = *(WLinStart ++);
      SLinStart += 16;                                                                               /* (20 dwords per scan line) */
    }

    if (GlobalKey == K_ESC)
      Finished = TRUE;
    MouseStatus ();
  }
  DestroyWindow (&CWindow);
  GlobalKey = 0x0000;
  return (!Cancelled);
}

bool Sampler (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/* Post  : A physical tape has been sampled and decoded.                                                                          */
/* Import: SetSoftMixerInput.                                                                                                     */
/**********************************************************************************************************************************/

{
  struct WindowInfo   CWindow = {30, 130, 220, 54, 220, 54, 220, 54, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                 WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER | WINF_EXCLUSIVE | WINF_NOMOVE,
                                 NULL, "SoundBlaster Sampling Tape"};
  struct TapeEntry_s *Tail;
  struct TapeEntry_s *DelEntry;
  char                MessageBuffer[8];
  char                TempTapeFileName[MAXDIR];
  char                OldTapeFileName[MAXDIR];
  char                FullSavePath[MAXDIR];
  int                 OldFileType;
  bool                OldChanged;
  bool                OldNonExist;
  dword               WrittenSize    = 0;
  dword               SampleLength;
  dword               ReadSampleRate = 44100;
  dword               DScratch;
  dword               HeaderIndex;
  dword               TimeOutValue;
  int                 Fh;
  bool                Finished       = FALSE;
  bool                Cancelled      = FALSE;
  bool                MoreToSample   = TRUE;
  bool                Accepted       = FALSE;
  bool                SelectionMade  = FALSE;
  bool                Multiple       = FALSE;
  bool                OverwriteAll   = TRUE;
  bool                TimeOut        = FALSE;
  bool                NewDataRead    = FALSE;
  bool                AddedBlocks    = FALSE;
  int                 NowWhat;
  comboent            OldNum;
  comboent            NewNum;

  (*SB_DMAOverrun) = FALSE;
  if (!SetSoftMixerInput ())                                                            /* Allow the user to set the mixer values */
    return (FALSE);
  sprintf (MessageBuffer, "%lu", ReadSampleRate);
  while (!Finished)
  {
    if (!InputRequester (-1, -1, NULL, NULL, MessageBuffer, 6, FALSE, FALSE, "Set Sample Rate", "{CSelect sample rate}C"))
      Cancelled = Finished = TRUE;
    else
    {
      ReadSampleRate = atol (MessageBuffer);
      if (ReadSampleRate < 11025 || ReadSampleRate > 60000)
        ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - The value is out of range\n        Use values 11025 through 60000");
      else
        Finished = TRUE;
    }
  }
  if (!Cancelled)
  {
    AutoInsertNewTape (WhichWindow);                                                                            /* (Sanity check) */
    _VOCID9Header._SampleRate = ReadSampleRate;
    strcpy (OldTapeFileName, TapeWindow[WhichWindow].TapeName);
    MakeFullPath (TempTapeFileName, GetTmpDirSetting (), "TAPTEMP.VOC");
    strcpy (TapeWindow[WhichWindow].TapeName, TempTapeFileName);
    OldFileType = TapeWindow[WhichWindow].FileType;
    OldChanged = TapeWindow[WhichWindow].TapeChanged;
    OldNonExist = TapeWindow[WhichWindow].NonexistingTape;
    TapeWindow[WhichWindow].FileType = FT_VOC;
    CWindow.StartX = (WhichWindow == 0 ? 379 : 30);
    _OpenWriteBuffer (-1, 0L, TRUE);                                                      /* Make sure the buffers are there etc. */
    while (MoreToSample)
    {
      if ((Fh = open (TempTapeFileName, O_BINARY | O_RDWR | O_CREAT | O_TRUNC, 0777)) == -1)
        FatalError (10, "Cannot create intermediate sample file: [%s] %d %s", "TAPTEMP.VOC", errno, StrError (errno));
      if (write (Fh, "Creative Voice File\x1A\x1A\x00\x0A\x01\x29\x11", 26) != 26)
      {
        __WRITEERROR : ErrorRequester (-1, -1, "@Close", NULL, NULL, "{BERROR}B: Error writing file %s\n"
                                       "       - %s\n"
                                       "       Disk full?", TempTapeFileName, StrError (errno));
                       close (Fh);
                       unlink (TempTapeFileName);
                       goto __SAMPCONT;
      }
      HeaderIndex = 26;
      if (write (Fh, &_VOCID9Header, sizeof (struct _VOCID9Header_s)) != sizeof (struct _VOCID9Header_s))  goto __WRITEERROR;
      SampleLength = sizeof (struct _VOCID9Header_s) - 4;
      WrittenSize = 0;
      if (TPConfig.ShowRemainingPlayTime)                                            /* Are we going to show the recording time ? */
        CWindow.Height = CWindow.MaxHeight = 64;                                                   /* Then make the window larger */
      DrawWindow (&CWindow, NULL, 0, TRUE);
      WDrawBox (&CWindow, FALSE, DBLUE, -3, -1, CWindow.Width + 2, CWindow.Height + 2);                   /* Make it more visible */
      WDrawBox (&CWindow, FALSE, DBLUE, -4, -2, CWindow.Width + 3, CWindow.Height + 3);
      Wprintf (&CWindow, 2, TEXT_CENTER, "Press any key to start ...");
      while (!kbhit ())
        ;
      getch ();
      ClearWindow (&CWindow);
      WDrawPicture (&CWindow, 10, 8, _RecordPicture);
      Wprintf (&CWindow, 1, TEXT_CENTER, "{BRECORDING}B");
      Wprintf (&CWindow, 2, TEXT_CENTER, "Press any key to stop");
      Wprintf (&CWindow, 3, TEXT_CENTER, "Sample length: 0 bytes");
      _WriteBuffer._ID01 = 1;
      (*SB_BufferReady) = _WriteBuffer._ID01;                                                   /* Signal: ready for both buffers */
      if (TPConfig.ShowRemainingPlayTime)
        _StartRecordClock (&CWindow);
      else
        PlayTimeUpdated = FALSE;
      SB_SampleAutoInitDMA (_WriteBuffer._SBBuffer[0], (word)ReadSampleRate, _WriteBuffer._BufferLength << 1); /* Start sampling! */
      Finished = FALSE;
      TimeOutValue = ReadTimer ();
      while (!Finished)
      {
        while (!TimeOut && (*SB_DMAComplete) != _WriteBuffer._ID01)                     /* Wait for the other buffer to come free */
        {
          if (PlayTimeUpdated)  { PlayTimeUpdated = FALSE; Wprintf (&CWindow, 4, TEXT_CENTER, PlayTimeString); }
          if (_CheckTimeOut)
            if (ElapsedTime (TimeOutValue) > 3250LU)              /* More than 3.25 seconds passed without an interrupt occured ? */
              TimeOut = TRUE;
          if (kbhit ())                                                                                          /* Key pressed ? */
          {
            getch ();
            Wprintf (&CWindow, 2, TEXT_CENTER, "  Finalizing sample  ");
            RealSB_DMAOverrun = FALSE;
            Finished = TRUE;
          }
        }
        if (*SB_DMAOverrun)                                                                          /* Did a DMA overrun occur ? */
        {
          RealSB_DMAOverrun = TRUE;
          Finished = TRUE;
        }
        if (TimeOut)
          Finished = TRUE;
        else
        {
          _CheckTimeOut = FALSE;
          _WriteBuffer._ID01 ^= 0x01;
          if (write (Fh, _WriteBuffer._SBBuffer[_WriteBuffer._ID01], _WriteBuffer._BufferLength) != _WriteBuffer._BufferLength)
            goto __WRITEERROR;
          WrittenSize += _WriteBuffer._BufferLength;
          SampleLength += _WriteBuffer._BufferLength;
          if (SampleLength > 0x00FF8000)                                                                         /* (24-bit size) */
          {
            if (lseek (Fh, HeaderIndex + 1, SEEK_SET) == -1)  goto __WRITEERROR;                /* Go back to the sample's header */
            if (write (Fh, &SampleLength, 3) != 3)  goto __WRITEERROR;                                 /* Write the sample length */
            if (lseek (Fh, 0, SEEK_END) == -1)  goto __WRITEERROR;                              /* Go back to the end of the file */
            HeaderIndex += SampleLength + 4;                                                                 /* Write next header */
            if (write (Fh, &_VOCID9Header, sizeof (struct _VOCID9Header_s)) != sizeof (struct _VOCID9Header_s))  goto __WRITEERROR;
            SampleLength = sizeof (struct _VOCID9Header_s) - 4;
          }
          (*SB_BufferReady) = _WriteBuffer._ID01;
          Wprintf (&CWindow, 3, TEXT_CENTER, "Sample length: %lu bytes", WrittenSize);
        }
      }
      SB_StopDMA ();                                                                                             /* Shut down DMA */
      SB_Speaker (FALSE);                                                                                           /* Silence... */
      if (TPConfig.ShowRemainingPlayTime)
        _StopRecordClock ();
      DScratch = 0x00000000;                                                                              /* Write VOC end marker */
      if (write (Fh, &DScratch, 4) != 4)  goto __WRITEERROR;
      if (lseek (Fh, HeaderIndex + 1, SEEK_SET) == -1)  goto __WRITEERROR;                      /* Go back to the sample's header */
      if (write (Fh, &SampleLength, 3) != 3)  goto __WRITEERROR;                                       /* Write the sample length */
      close (Fh);
      DestroyWindow (&CWindow);
      if (TimeOut)
        ErrorRequester (-1, -1, NULL, NULL, NULL, "{BERROR}B: Sampling didn't work!\n"
                        "Your soundcard is not suitable for this program. Sorry...");
      else if (RealSB_DMAOverrun)
        ErrorRequester (-1, -1, NULL, NULL, NULL, "{BERROR}B: DMA Overrun!\n       Choose a lower sample rate");
      else
      {
        OldNum = TapeWindow[WhichWindow].IndexWindow.NumEntries;
        if (TapeWindow[WhichWindow].EmptyTape && OldNum == 0)
          OldNum = -1;
        Tail = TapeWindow[WhichWindow].Directory;
        if (Tail)                                                       /* Let Tail point to the last entry of the tape in memory */
          while (Tail->Next)
            Tail = Tail->Next;
        Finished = FALSE;
        while (!Finished)
        {
          ReadTape (WhichWindow, TRUE);                                                            /* Decode and merge the sample */
          SetupNewData (WhichWindow);                                                                          /* Show the result */
          NewNum = TapeWindow[WhichWindow].IndexWindow.NumEntries;
          if (NewNum >= 0)
            SelectNewTapeEntry (WhichWindow, OldNum + 1);
          AddedBlocks = (bool)((NewNum > OldNum) ? TRUE : FALSE);
          SelectionMade = FALSE;
          while (!SelectionMade)
          {
            switch ((NowWhat = QuestionRequester (-1, -1, TRUE, TRUE, "@Ok", "@Drop", "@Retry", "@Save", NULL, NULL,
                                                  "Was the data decoded correctly?")))
            {
              case 0 : Finished = TRUE;                                                                                   /* (Ok) */
                       SelectionMade = TRUE;
                       if (AddedBlocks)
                       {
                         SignalTapeChanged (WhichWindow, TRUE);
                         NewDataRead = TRUE;                                                  /* Signal: need to unghost the menu */
                         OldChanged = TRUE;
                       }
                       break;
              case 1 :                                                                                                  /* (Drop) */
              case 2 : if (AddedBlocks)                                                                                /* (Retry) */
                       {
                         if (Tail == NULL)                                                             /* Tape was empty before ? */
                           Tail = TapeWindow[WhichWindow].Directory;
                         else
                           Tail = Tail->Next;                                                                /* (First new entry) */
                         while (Tail)
                         {
                           DelEntry = Tail;
                           Tail = Tail->Next;
                           DeleteTapeBlock (WhichWindow, DelEntry, FALSE, FALSE);
                         }
                         SetupNewData (WhichWindow);
                       }
                       SignalTapeChanged (WhichWindow, OldChanged);
                       if (NowWhat == 1)
                         Finished = TRUE;
                       SelectionMade = TRUE;
                       break;
              case 3 : Accepted = FALSE;
                       if (AddedBlocks)
                       {
                         SignalTapeChanged (WhichWindow, TRUE);
                         NewDataRead = TRUE;                                                  /* Signal: need to unghost the menu */
                         OldChanged = TRUE;
                       }
                       while (!Accepted)
                         if (FileRequester (180, 50, 48, 29, 20, _SavePath, _SaveFile, _SaveMask, "Select VOC File For Saving",
                                            FALSE, FALSE))
                         {
                           if (strlen (_SaveFile) < 5 || strcmp (_SaveFile + strlen (_SaveFile) - 4, ".VOC"))
                             ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR: %s is not a VOC file type.\n"
                                             "       Please select a VOC file type", _SaveFile);
                           else
                           {
                             MakeFullPath (FullSavePath, _SavePath, _SaveFile);
                             if (!TestFileOverwrite (FullSavePath))
                             {
                               CopyFile (TempTapeFileName, FullSavePath, &Multiple, &OverwriteAll, "Saving file...");
                               Accepted = TRUE;
                               SelectionMade = TRUE;
                               Finished = TRUE;
                             }
                           }
                         }
                         else
                           Accepted = TRUE;
                       break;
            }
          }
        }
      }
      unlink (TempTapeFileName);                                                                  /* Remove the decoded VOC again */
      TapeWindow[WhichWindow].NonexistingTape = OldNonExist;
      GlobalKey = 0x0000;
      if (TimeOut)
        MoreToSample = FALSE;
      else if (QuestionRequester (-1, -1, FALSE, FALSE, NULL, NULL, NULL, NULL, NULL, NULL, "Do you want to sample more data?"))
        MoreToSample = FALSE;
    }
  }
  __SAMPCONT:
  _CloseFileBuffer (FALSE);                                                                               /* Release write buffer */
  SB_WriteMixer (SBM_INPUTCTRL, SBI_RESTOREALL);                                                    /* Restore the input settings */
  SB_WriteMixer (SBM_OUTPUTCTRL, SBO_RESTOREALL);                                                  /* Restore the output settings */
  strcpy (TapeWindow[WhichWindow].TapeName, OldTapeFileName);
  TapeWindow[WhichWindow].FileType = OldFileType;
  return (NewDataRead);
}
