/**********************************************************************************************************************************/
/* Project      : TAPER                                                                                                           */
/* Module       : TPTAPE.C                                                                                                        */
/* Description  : Intermediate tapefile handler                                                                                   */
/* Version type : Program module                                                                                                  */
/* Last changed : 06-01-1999  6:50                                                                                                */
/* Update count : 85                                                                                                              */
/* OS type      : PC (DOS)                                                                                                        */
/* Compiler     : Watcom C32 Optimizing Compiler Version 10.x                                                                     */
/* Linker       : Watcom C Linker                                                                                                 */
/* Notes        : Uses goto's __READERROR, __INVALIDTAPEFILE, __WRITEERROR and __CONTINUE (saves over 20Kb space)                 */
/*                                                                                                                                */
/*                Copyleft (C) 1997-2001 ThunderWare Research Center, written by Martijn van der Heide.                           */
/*                                                                                                                                */
/*                This program is free software; you can redistribute it and/or                                                   */
/*                modify it under the terms of the GNU General Public License                                                     */
/*                as published by the Free Software Foundation; either version 2                                                  */
/*                of the License, or (at your option) any later version.                                                          */
/*                                                                                                                                */
/*                This program is distributed in the hope that it will be useful,                                                 */
/*                but WITHOUT ANY WARRANTY; without even the implied warranty of                                                  */
/*                MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                                                   */
/*                GNU General Public License for more details.                                                                    */
/*                                                                                                                                */
/*                You should have received a copy of the GNU General Public License                                               */
/*                along with this program; if not, write to the Free Software                                                     */
/*                Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                                     */
/*                                                                                                                                */
/**********************************************************************************************************************************/

#define __TPTAPE_MAIN__

#include "wbench32/wbench32.h"
#include <dos.h>
#include <fcntl.h>
#ifdef __WATCOMC__
#include <io.h>
#endif
#include <bios.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <math.h>
#include "tpconfig.h"
#include "tpmenu.h"
#include "tpfile.h"
#include "tpscreen.h"
#include "tpmain.h"
#include "tpaudio.h"
#include "tpedit.h"
#include "tpcust.h"
#include "tpdisass.h"
#include "tpsnap.h"
#include "tptape.h"

#pragma pack                ()

/**********************************************************************************************************************************/
/* Define the static variables                                                                                                    */
/**********************************************************************************************************************************/

static dword             _PRFileLength;
static struct PReqInfo   _PReq;
static char             *_ErrInvalidTapeFile = "Invalid tape file %s";

/**********************************************************************************************************************************/
/* Name the BlockTypeIDs                                                                                                          */
/**********************************************************************************************************************************/

static struct BlockType_s
{
  byte   TypeID;                                                                            /* ID values as defined in TZX format */
  bool   IsTapeData;                                                                 /* TRUE for tape data, FALSE for info blocks */
                                                                           /* TRUE blocks can be saved to VOC or SoundBlaster out */
  bool   IsIdentifiedData;                                                                      /* TRUE if all bytes are readable */
                                                                   /* TRUE blocks can be saved to TAP and shown with 'View As...' */
  char  *FullName;                                                                  /* As printed in the 'View Block Info' window */
  char  *TapeWindowName;                                                                          /* As printed in the TapeWindow */
  short  NumInfoLines;                                  /* Number of additional info lines needed in the 'View Block Info' window */
                                                                      /* -1 means that the info will be printed in a combo window */
  bool   ReportSize;                                                                /* Report data size in the TapeWindow if TRUE */
  bool (*EditHandler)(short WhichWindow, struct TapeEntry_s *Entry);
}        BlockType[] = {
  { 0x10, TRUE,  TRUE,  "Normal Speed Data",    "",                            11, FALSE, EditTapeDataBlock     },
  { 0x11, TRUE,  TRUE,  "Turbo Speed Data",     "",                            11, FALSE, EditTapeDataBlock     },
  { 0x12, TRUE,  FALSE, "Pure Tone",            "  {{Pure Tone}}            ",  2, FALSE, EditPureToneBlock     },
  { 0x13, TRUE,  FALSE, "Row of Pulses",        "  {{Row of Pulses}}        ", -1, TRUE,  EditRowOfPulsesBlock  },
  { 0x14, TRUE,  FALSE, "Pure Data Block",      "",                            11, TRUE , EditTapeDataBlock     },  /* Special */
  { 0x15, TRUE,  FALSE, "Direct Recording",     "  {{Direct Recording}}     ",  3, TRUE,  NULL                  },  /* INST ONLY */
  { 0x16, TRUE,  TRUE,  "C64 ROM Type Data",    "  {{!C64 ROM Type Block}}  ",  0, TRUE,  NULL                  },
  { 0x17, TRUE,  TRUE,  "C64 Turbo Tape Data",  "  {{!C64 Turbo Tape Data}} ",  0, TRUE,  NULL                  },
  { 0x20, TRUE,  FALSE, "Pause/Stop Tape",      "  {{Pause/Stop}}           ",  1, FALSE, EditPauseBlock        },
  { 0x21, FALSE, FALSE, "Group Start",          "",                             1, FALSE, EditGroupStrtBlock    },  /* Special */
  { 0x22, FALSE, FALSE, "Group End",            "GROUP END",                    0, FALSE, NULL                  },  /* AUTOMATIC */
  { 0x23, TRUE,  FALSE, "Jump to Entry",        "",                             1, FALSE, EditJumpToBlock       },  /* Special */
  { 0x24, TRUE,  FALSE, "Loop Start",           "",                             1, FALSE, EditLoopStrtBlock     },  /* Special */
  { 0x25, TRUE,  FALSE, "Loop End",             "LOOP END",                     0, FALSE, NULL                  },  /* AUTOMATIC */
  { 0x26, TRUE,  FALSE, "Call Sequence",        "  {{Call Sequence}}        ", -1, TRUE,  EditCallSequenceBlock },
  { 0x27, TRUE,  FALSE, "Return from Sequence", "  {{Return from Sequence}} ",  0, FALSE, NULL                  },  /* INST ONLY */
  { 0x28, TRUE,  FALSE, "Select Block",         "  {{Select Block}}         ", -1, TRUE,  EditSelectBlock       },
  { 0x2A, FALSE, FALSE, "Stop if in 48K Mode",  "  {{Stop Tape if 48K Mode}}",  0, FALSE, NULL                  },  /* INST ONLY */
  { 0x30, FALSE, FALSE, "Text Description",     "",                            -1, TRUE,  EditTextDescrBlock    },  /* Special */
  { 0x31, FALSE, FALSE, "Message Block",        "  {{Message Block}}        ", -1, TRUE,  EditMessageBlock      },
  { 0x32, FALSE, FALSE, "Archive Info",         "  {{Archive Info}}         ", -1, TRUE,  EditArchiveInfoBlock  },
  { 0x33, FALSE, FALSE, "Hardware Type",        "  {{Hardware Type}}        ", -1, TRUE,  EditHardwareTypeBlock },
  { 0x34, FALSE, FALSE, "Emulation Info",       "  {{Emulation Info}}       ", 11, FALSE, NULL                  },  /* Emulator */
  { 0x35, FALSE, FALSE, "Custom Info Block",    "",                            -1, TRUE,  EditCustomInfoBlock   },  /* Special */
  { 0x40, FALSE, FALSE, "Snapshot Block",       "  {{Snapshot Block}}       ",  5, TRUE,  NULL                  },  /* Emulator */
  { 0xFF, FALSE, FALSE, "UNKNOWN BLOCK",        "<<< UNKNOWN >>>          ",    0, TRUE,  NULL                  }};

static byte NumBlockTypes = (sizeof (BlockType) / sizeof (struct BlockType_s) - 1);                    /* Last one doesn't count! */

static bool _UsingFatalSave = FALSE;

/**********************************************************************************************************************************/
/* Define the static routines                                                                                                     */
/**********************************************************************************************************************************/

static void  _DestroyPReq         (void);
static void  _CheckGroupBlocks    (short WhichWindow);
static void  _CheckJumpBlocks     (short WhichWindow);
static void  _InitializeJumpCodes (short WhichWindow);
static void  _CheckJumpCodes      (short WhichWindow);

/**********************************************************************************************************************************/
/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> START OF THE INTERNAL LIBRARY ROUTINES <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< */
/**********************************************************************************************************************************/

static void _DestroyPReq (void)

/**********************************************************************************************************************************/
/* Pre   : The percentage requester is initialized.                                                                               */
/* Post  : The percentage requester has been removed.                                                                             */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  if (!_UsingFatalSave)
    DestroyPRequester (&_PReq);
}

static void _CheckGroupBlocks (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/* Post  : The tape directory is checked for bad group blocks and repaired if an error is found. The rules are simple:            */
/*         - There must be an equal number of 'Group Start' and 'Group End' blocks;                                               */
/*         - No nesting of groups is allowed.                                                                                     */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *DirP;
  struct TapeEntry_s *PrevP   = NULL;
  struct TapeEntry_s *NewP;
  bool                InGroupBlock = FALSE;
  bool                InLoopBlock = FALSE;

  DirP = TapeWindow[WhichWindow].Directory;
  while (DirP != NULL)
  {
    if (DirP->BlockTypeID == 0x21)                                                                               /* Group Start ? */
      if (InGroupBlock)                                                                        /* Already in a group -> nesting ? */
      {                                                                   /* Then we need insert a Group End block in front of it */
        if ((NewP = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
          FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
        memset (NewP, 0, sizeof (struct TapeEntry_s));
        NewP->ComboSelected = 0x00;
        NewP->BlockTypeID = 0x22;                                                                                 /* (Group End) */
        GetBlockTypeInfo (0x22, &(NewP->IsTapeData), &(NewP->IsIdentifiedData), NULL, NULL, NULL);
        NewP->Next = DirP;                                                                               /* (Link it into place) */
        PrevP->Next = NewP;                                                   /* (Cannot be first entry, so PrevP is never NULL) */
        TapeWindow[WhichWindow].NumEntries ++;
        SignalTapeChanged (WhichWindow, TRUE);
      }
      else
        InGroupBlock = TRUE;
    else if (DirP->BlockTypeID == 0x22)                                                                            /* Group End ? */
      if (!InGroupBlock)                                                                                 /* End without a Start ? */
      {
        DeleteTapeBlock (WhichWindow, DirP, FALSE, TRUE);                                       /* Remove the redundant Group End */
        DirP = PrevP;                                                                                              /* (Step back) */
      }
      else
        InGroupBlock = FALSE;
    else if (DirP->BlockTypeID == 0x24)                                                                           /* Loop Start ? */
      if (InLoopBlock)                                                                          /* Already in a loop -> nesting ? */
      {                                                                    /* Then we need insert a Loop End block in front of it */
        if ((NewP = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
          FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
        memset (NewP, 0, sizeof (struct TapeEntry_s));
        NewP->ComboSelected = 0x00;
        NewP->BlockTypeID = 0x25;                                                                                  /* (Loop End) */
        GetBlockTypeInfo (0x25, &(NewP->IsTapeData), &(NewP->IsIdentifiedData), NULL, NULL, NULL);
        NewP->Next = DirP;                                                                               /* (Link it into place) */
        PrevP->Next = NewP;                                                   /* (Cannot be first entry, so PrevP is never NULL) */
        TapeWindow[WhichWindow].NumEntries ++;
        SignalTapeChanged (WhichWindow, TRUE);
      }
      else
        InLoopBlock = TRUE;
    else if (DirP->BlockTypeID == 0x25)                                                                             /* Loop End ? */
      if (!InLoopBlock)                                                                                  /* End without a Start ? */
      {
        DeleteTapeBlock (WhichWindow, DirP, FALSE, TRUE);                                        /* Remove the redundant Loop End */
        DirP = PrevP;                                                                                              /* (Step back) */
      }
      else
        InLoopBlock = FALSE;
    PrevP = DirP;
    DirP = DirP->Next;
  }
  if (InGroupBlock)                                                                         /* Missing Group End at end of tape ? */
  {                                                                                                         /* Append a Group End */
    if ((NewP = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
      FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
    memset (NewP, 0, sizeof (struct TapeEntry_s));
    NewP->ComboSelected = 0x00;
    NewP->BlockTypeID = 0x22;                                                                                      /* (Group End) */
    GetBlockTypeInfo (0x22, &(NewP->IsTapeData), &(NewP->IsIdentifiedData), NULL, NULL, NULL);
    NewP->Next = NULL;
    PrevP->Next = NewP;
    TapeWindow[WhichWindow].NumEntries ++;
    SignalTapeChanged (WhichWindow, TRUE);
  }
  if (InLoopBlock)                                                                           /* Missing Loop End at end of tape ? */
  {                                                                                                          /* Append a Loop End */
    if ((NewP = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
      FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
    memset (NewP, 0, sizeof (struct TapeEntry_s));
    NewP->ComboSelected = 0x00;
    NewP->BlockTypeID = 0x25;                                                                                       /* (Loop End) */
    GetBlockTypeInfo (0x25, &(NewP->IsTapeData), &(NewP->IsIdentifiedData), NULL, NULL, NULL);
    NewP->Next = NULL;
    PrevP->Next = NewP;
    TapeWindow[WhichWindow].NumEntries ++;
    SignalTapeChanged (WhichWindow, TRUE);
  }
}

static void _CheckJumpBlocks (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/* Post  : The tape directory is checked for bad jump blocks (jump outside tape) and these are limited.                           */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *DirP;
  comboent            BlockNo = 0;
  bool                Changed = FALSE;
  bool                XMSChanged;
  dword               XMSIndex;
  short              *WTableData;
  byte               *BTableData;
  word                Cnt;

  DirP = TapeWindow[WhichWindow].Directory;
  XMSIndex = 0;
  while (DirP != NULL)
  {
    if (DirP->BlockTypeID == 0x23)                                                                             /* Jump To Entry ? */
    {
      if ((short)DirP->PauseLength + BlockNo < 0)
      {
        DirP->PauseLength = (word)(-BlockNo);                                                /* Cannot jump to before first entry */
        Changed = TRUE;
      }
      else if ((short)DirP->PauseLength + BlockNo > TapeWindow[WhichWindow].NumEntries)
      {
        DirP->PauseLength = (word)(TapeWindow[WhichWindow].NumEntries - BlockNo);              /* Cannot jump to after last entry */
        Changed = TRUE;
      }
    }
    else if (DirP->BlockTypeID == 0x26)                                                                        /* Call Sequence ? */
    {
      CopyFromXMS (DirP->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);
      Cnt = DirP->BlockLength / 2;
      WTableData = (short *)WorkBuffer;
      XMSChanged = FALSE;
      for ( ; Cnt > 0 ; Cnt --, WTableData ++)
      {
        if (*WTableData + BlockNo < 0)
        {
          *WTableData = (word)(-BlockNo);                                                    /* Cannot call to before first entry */
          Changed = TRUE;
          XMSChanged = TRUE;
        }
        else if (*WTableData + BlockNo > TapeWindow[WhichWindow].NumEntries)
        {
          *WTableData = (word)(TapeWindow[WhichWindow].NumEntries - BlockNo);                  /* Cannot call to after last entry */
          Changed = TRUE;
          XMSChanged = TRUE;
        }
      }
      if (XMSChanged)
        CopyToXMS (DirP->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer, XMSIndex);
    }
    else if (DirP->BlockTypeID == 0x28)                                                                         /* Select Block ? */
    {
      CopyFromXMS (DirP->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);
      Cnt = *WorkBuffer;
      BTableData = WorkBuffer + 1;
      Changed = FALSE;
      for ( ; Cnt > 0 ; Cnt --, BTableData += (*(BTableData + 2) + 3))
      {
        if (*((short *)BTableData) + BlockNo < 0)
        {
          *((short *)BTableData) = (short)(-BlockNo);                                        /* Cannot call to before first entry */
          Changed = TRUE;
          XMSChanged = TRUE;
        }
        else if (*((short *)BTableData) + BlockNo > TapeWindow[WhichWindow].NumEntries)
        {
          *((short *)BTableData) = (short)(TapeWindow[WhichWindow].NumEntries - BlockNo);      /* Cannot call to after last entry */
          Changed = TRUE;
          XMSChanged = TRUE;
        }
      }
      if (Changed)
        CopyToXMS (DirP->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer, XMSIndex);
    }
    XMSIndex += DirP->AllocLength;
    DirP = DirP->Next;
    BlockNo ++;
  }
  if (Changed)
    SignalTapeChanged (WhichWindow, TRUE);
}

static void _InitializeJumpCodes (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/* Post  : All jump sources and destinations have codes.                                                                          */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *SrcDirP;
  struct TapeEntry_s *DstDirP;
  comboent            SrcIndex;
  comboent            DstIndex;
  byte               *BTableData;
  short              *WTableData;
  short              *WTableJump;
  word                Cnt;
  dword               XMSIndex = 0;

  _MustInitializeJumpCodes = FALSE;                                                                                /* (Not again) */
  SrcDirP = TapeWindow[WhichWindow].Directory;                                                          /* First cleanup the tape */
  while (SrcDirP != NULL)
  {
    SrcDirP->SrcJumpCode = SrcDirP->DstJumpCode = 0x0000;
    SrcDirP = SrcDirP->Next;
  }
  _CheckJumpBlocks (WhichWindow);                                                /* Test and repair possible Jump inconsistencies */
  SrcDirP = TapeWindow[WhichWindow].Directory;                                                               /* Resolve all jumps */
  SrcIndex = 0;
  TapeWindow[WhichWindow].NextJumpCode = 0x0000;
  while (SrcDirP != NULL)
  {
    if (SrcDirP->BlockTypeID == 0x23)                                                                          /* Jump to Entry ? */
    {
      if (++ TapeWindow[WhichWindow].NextJumpCode == 0x0000)                                /* Overflow - more than 65535 jumps ? */
        ;                                                                                                         /* (ah well...) */
      SrcDirP->SrcJumpCode = TapeWindow[WhichWindow].NextJumpCode;                                          /* Set source jump ID */
      DstIndex = (short)SrcDirP->PauseLength + SrcIndex;                                           /* Determine destination block */
      DstDirP = TapeWindow[WhichWindow].Directory;
      if (DstIndex > 0)                                                                                         /* (Sanity check) */
        while (-- DstIndex >= 0 && DstDirP != NULL)                                                     /* Find destination entry */
          DstDirP = DstDirP->Next;
      if (DstDirP != NULL)                                                                                      /* (Sanity check) */
        if (DstDirP->DstJumpCode != 0x0000)                                                  /* This entry is jumped to already ? */
        {
          SrcDirP->SrcJumpCode = DstDirP->DstJumpCode;                                       /* Use that jump code for the source */
          TapeWindow[WhichWindow].NextJumpCode --;                                                   /* Give the 'next code' back */
        }
        else
          DstDirP->DstJumpCode = TapeWindow[WhichWindow].NextJumpCode;                                 /* Set destination jump ID */
    }
    else if (SrcDirP->BlockTypeID == 0x26)                                                                     /* Call Sequence ? */
    {
      CopyFromXMS (SrcDirP->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);
      Cnt = SrcDirP->BlockLength / 2;
      WTableData = (short *)WorkBuffer;
      WTableJump = (short *)(WorkBuffer + SrcDirP->BlockLength);
      for ( ; Cnt > 0 ; Cnt --, WTableData ++, WTableJump ++)
      {
        if (++ TapeWindow[WhichWindow].NextJumpCode == 0x0000)                              /* Overflow - more than 65535 jumps ? */
          ;                                                                                                       /* (ah well...) */
        *WTableJump = TapeWindow[WhichWindow].NextJumpCode;                                                 /* Set source jump ID */
        DstIndex = (short)(*WTableData) + SrcIndex;                                                /* Determine destination block */
        DstDirP = TapeWindow[WhichWindow].Directory;
        if (DstIndex > 0)                                                                                       /* (Sanity check) */
          while (-- DstIndex >= 0 && DstDirP != NULL)                                                   /* Find destination entry */
            DstDirP = DstDirP->Next;
        if (DstDirP != NULL)                                                                                    /* (Sanity check) */
          if (DstDirP->DstJumpCode != 0x0000)                                                /* This entry is jumped to already ? */
          {
            *WTableJump = DstDirP->DstJumpCode;                                              /* Use that jump code for the source */
            TapeWindow[WhichWindow].NextJumpCode --;                                                 /* Give the 'next code' back */
          }
          else
            DstDirP->DstJumpCode = TapeWindow[WhichWindow].NextJumpCode;                               /* Set destination jump ID */
      }
      CopyToXMS (SrcDirP->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer, XMSIndex);
    }
    else if (SrcDirP->BlockTypeID == 0x28)                                                                      /* Select Block ? */
    {
      CopyFromXMS (SrcDirP->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);
      Cnt = *WorkBuffer;
      BTableData = WorkBuffer + 1;
      WTableJump = (short *)(WorkBuffer + SrcDirP->BlockLength);
      for ( ; Cnt > 0 ; Cnt --, WTableJump ++)
      {
        if (++ TapeWindow[WhichWindow].NextJumpCode == 0x0000)                              /* Overflow - more than 65535 jumps ? */
          ;                                                                                                       /* (ah well...) */
        *WTableJump = TapeWindow[WhichWindow].NextJumpCode;                                                 /* Set source jump ID */
        DstIndex = *((short *)BTableData) + SrcIndex;                                              /* Determine destination block */
        DstDirP = TapeWindow[WhichWindow].Directory;
        if (DstIndex > 0)                                                                                       /* (Sanity check) */
          while (-- DstIndex >= 0 && DstDirP != NULL)                                                   /* Find destination entry */
            DstDirP = DstDirP->Next;
        if (DstDirP != NULL)                                                                                    /* (Sanity check) */
          if (DstDirP->DstJumpCode != 0x0000)                                                /* This entry is jumped to already ? */
          {
            *WTableJump = DstDirP->DstJumpCode;                                              /* Use that jump code for the source */
            TapeWindow[WhichWindow].NextJumpCode --;                                                 /* Give the 'next code' back */
          }
          else
            DstDirP->DstJumpCode = TapeWindow[WhichWindow].NextJumpCode;                               /* Set destination jump ID */
        BTableData += (*(BTableData + 2) + 3);                                                         /* Point to next selection */
      }
      CopyToXMS (SrcDirP->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer, XMSIndex);
    }
    XMSIndex += SrcDirP->AllocLength;
    SrcDirP = SrcDirP->Next;
    SrcIndex ++;
  }
}

static void _CheckJumpCodes (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/* Post  : All jump sources and destinations have been adjusted where needed (after in insert/delete).                            */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *SrcDirP;
  struct TapeEntry_s *DstDirP;
  comboent            SrcIndex;
  comboent            DstIndex = 0;
  bool                FoundJump;
  bool                AtLeastOneJump = FALSE;
  dword               SXMSIndex;
  dword               DXMSIndex;
  byte               *BTableData;
  short              *WTableData;
  short              *WTableJump;
  word                Cnt;
  bool                Changed;

  DstDirP = TapeWindow[WhichWindow].Directory;                                   /* Check if the jump destinations are consistent */
  DXMSIndex = 0;
  while (DstDirP != NULL)
  {
    if (DstDirP->DstJumpCode != 0x0000 &&                                                            /* This entry is jumped to ? */
        DstDirP->SrcJumpCode != 0xFFFF)                                                                      /* And it's in use ? */
    {
      SrcDirP = TapeWindow[WhichWindow].Directory;                                             /* Find the matching Jump block(s) */
      SrcIndex = 0;
      SXMSIndex = 0;
      FoundJump = FALSE;
      while (SrcDirP != NULL)
      {
        if (SrcDirP->BlockTypeID == 0x23)                                                                 /* Found a Jump block ? */
        {
          AtLeastOneJump = TRUE;
          if (SrcDirP->SrcJumpCode == DstDirP->DstJumpCode)                                    /* Found the matching Jump block ? */
          {
            SrcDirP->PauseLength = (word)(DstIndex - SrcIndex);                                    /* Ensure the index is correct */
            FoundJump = TRUE;
          }
        }
        else if (SrcDirP->BlockTypeID == 0x26)
        {
          AtLeastOneJump = TRUE;
          CopyFromXMS (SrcDirP->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, SXMSIndex, WorkBuffer);
          Cnt = SrcDirP->BlockLength / 2;
          WTableData = (short *)WorkBuffer;
          WTableJump = (short *)(WorkBuffer + SrcDirP->BlockLength);
          Changed = FALSE;
          for ( ; Cnt > 0 ; Cnt --, WTableData ++, WTableJump ++)
            if (*WTableJump == DstDirP->DstJumpCode)                                           /* Found the matching Call block ? */
            {
              *WTableData = (short)(DstIndex - SrcIndex);                                          /* Ensure the index is correct */
              Changed = TRUE;
              FoundJump = TRUE;
            }
          if (Changed)
            CopyToXMS (SrcDirP->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer, SXMSIndex);
        }
        else if (SrcDirP->BlockTypeID == 0x28)
        {
          AtLeastOneJump = TRUE;
          CopyFromXMS (SrcDirP->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, SXMSIndex, WorkBuffer);
          Cnt = *WorkBuffer;
          BTableData = WorkBuffer + 1;
          WTableJump = (short *)(WorkBuffer + SrcDirP->BlockLength);
          Changed = FALSE;
          for ( ; Cnt > 0 ; Cnt --, WTableJump ++)
          {
            if (*WTableJump == DstDirP->DstJumpCode)                                           /* Found the matching Call block ? */
            {
              *((short *)BTableData) = (short)(DstIndex - SrcIndex);                               /* Ensure the index is correct */
              Changed = TRUE;
              FoundJump = TRUE;
            }
            BTableData += (*(BTableData + 2) + 3);                                                     /* Point to next selection */
          }
          if (Changed)
            CopyToXMS (SrcDirP->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer, SXMSIndex);
        }
        SXMSIndex += SrcDirP->AllocLength;
        SrcDirP = SrcDirP->Next;
        SrcIndex ++;
      }
      if (!FoundJump)                                                                      /* Jump block itself has disappeared ? */
        DstDirP->DstJumpCode = 0x0000;                                                                          /* Then signal it */
    }
    DXMSIndex += DstDirP->AllocLength;
    DstDirP = DstDirP->Next;
    DstIndex ++;
  }
  if (AtLeastOneJump)
  {
    SrcDirP = TapeWindow[WhichWindow].Directory;                                         /* Now check if the Jumps are consistent */
    SrcIndex = 0;
    SXMSIndex = 0;
    while (SrcDirP != NULL)                                               /* We are doing things double here, but that's the risk */
    {                                                                    /* We're looking for missing (deleted) jump destinations */
      if (SrcDirP->SrcJumpCode != 0xFFFF)                                                                /* Is the entry in use ? */
        if (SrcDirP->BlockTypeID == 0x23)                                                                      /* Jump to Entry ? */
        {
          DstIndex = (short)SrcDirP->PauseLength + SrcIndex;                                       /* Determine destination block */
          DstDirP = TapeWindow[WhichWindow].Directory;
          if (DstIndex > 0)                                                                                     /* (Sanity check) */
            while (-- DstIndex >= 0 && DstDirP != NULL)                                                 /* Find destination entry */
              DstDirP = DstDirP->Next;
          if (DstDirP != NULL)                                                                                  /* (Sanity check) */
            if (DstDirP->DstJumpCode != 0x0000)                                              /* This entry is jumped to already ? */
                                                      /* If it's a different code, than the original destination has been deleted */
              SrcDirP->SrcJumpCode = DstDirP->DstJumpCode;                                   /* Use that jump code for the source */
            else
              DstDirP->DstJumpCode = SrcDirP->SrcJumpCode;                                             /* Set destination jump ID */
        }
        else if (SrcDirP->BlockTypeID == 0x26)
        {
          CopyFromXMS (SrcDirP->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, SXMSIndex, WorkBuffer);
          Cnt = SrcDirP->BlockLength / 2;
          WTableData = (short *)WorkBuffer;
          WTableJump = (short *)(WorkBuffer + SrcDirP->BlockLength);
          Changed = FALSE;
          for ( ; Cnt > 0 ; Cnt --, WTableData ++, WTableJump ++)
          {
            DstIndex = (short)(*WTableData) + SrcIndex;                                            /* Determine destination block */
            DstDirP = TapeWindow[WhichWindow].Directory;
            if (DstIndex > 0)
              while (-- DstIndex >= 0 && DstDirP != NULL)                                               /* Find destination entry */
                DstDirP = DstDirP->Next;
            if (DstDirP != NULL)
              if (DstDirP->DstJumpCode != 0x0000)                                            /* This entry is jumped to already ? */
              {
                *WTableJump = DstDirP->DstJumpCode;                                          /* Use that jump code for the source */
                Changed = TRUE;
              }
              else
                DstDirP->DstJumpCode = *WTableJump;                                                    /* Set destination jump ID */
          }
          if (Changed)
            CopyToXMS (SrcDirP->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer, SXMSIndex);
        }
        else if (SrcDirP->BlockTypeID == 0x28)
        {
          CopyFromXMS (SrcDirP->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, SXMSIndex, WorkBuffer);
          Cnt = *WorkBuffer;
          BTableData = WorkBuffer + 1;
          WTableJump = (short *)(WorkBuffer + SrcDirP->BlockLength);
          Changed = FALSE;
          for ( ; Cnt > 0 ; Cnt --, WTableJump ++)
          {
            DstIndex = *((short *)BTableData) + SrcIndex;                                          /* Determine destination block */
            DstDirP = TapeWindow[WhichWindow].Directory;
            if (DstIndex > 0)
              while (-- DstIndex >= 0 && DstDirP != NULL)                                               /* Find destination entry */
                DstDirP = DstDirP->Next;
            if (DstDirP != NULL)
              if (DstDirP->DstJumpCode != 0x0000)                                            /* This entry is jumped to already ? */
              {
                *WTableJump = DstDirP->DstJumpCode;                                          /* Use that jump code for the source */
                Changed = TRUE;
              }
              else
                DstDirP->DstJumpCode = *WTableJump;                                                    /* Set destination jump ID */
            BTableData += (*(BTableData + 2) + 3);                                                     /* Point to next selection */
          }
          if (Changed)
            CopyToXMS (SrcDirP->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer, SXMSIndex);
        }
      SXMSIndex += SrcDirP->AllocLength;
      SrcDirP = SrcDirP->Next;
      SrcIndex ++;
    }
  }
}

/**********************************************************************************************************************************/
/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> START OF THE EXPORTED LIBRARY ROUTINES <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< */
/**********************************************************************************************************************************/

bool UpdatePReq (dword FileIndex, char *StatusText)

/**********************************************************************************************************************************/
/* Pre   : `FileIndex' holds the curent file index.                                                                               */
/* Post  : The percentage requester has been updated.                                                                             */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  MouseStatus ();
  if (!_UsingFatalSave)
    return (UpdatePRequester (&_PReq, (double)FileIndex * 100 / _PRFileLength, StatusText));
  return (TRUE);
}

bool ReadTape (short WhichWindow, bool Merge)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Merge' is TRUE if the current tape in memory should be kept as well.       */
/* Post  : The tape file has been processed and read into the intermediate tape file format.                                      */
/* Import: UpdatePReq, GetBlockTypeInfo.                                                                                          */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *Entry;
  struct TapeEntry_s *DirP;
  struct Timing_s     Timing;
  struct Timing_s     NormalTiming;                                                            /* Normal speed timing (reference) */
  comboent            NoiseNumEntries;                                            /* Noise inside the tape rather than at the end */
  comboent            NoiseCnt;
  int                 Fh;
  dword               Index       = 0;
  bool                MoreToGo    = FALSE;
  word                NRead;
  word                BlLength;                                                                   /* Length of one TAP file block */
  dword               DBlLength;                                                                             /* Length as a dword */
  word      register  Cnt;
  bool                UseTiming   = FALSE;                                                    /* Assume Timing structure not used */
  word                PauseLength = 1000;                                                   /* Assume 1 sec. pause between blocks */
  byte                BlockTypeID;
  byte                BScratch;
  word                WScratch;
  dword               DScratch;
  dword               XMSIndex;
  dword               LengthToGo;
  dword               DLengthToGo;                                                         /* Length as a multiple of 4 (for XMS) */
  word                ReadLength;
  word                RealReadLength;
  bool                InsertBlock = TRUE;                                                      /* Only adjust internally if FALSE */
  bool                PureData = FALSE;                                                    /* If TRUE, no flag/parity byte splits */
  bool                InsertWorkBuffer = FALSE;                          /* If TRUE, data to be inserted into XMS has been set up */
                                                                                      /* in WorkBuffer, with a length of BlLength */
  bool                InsertLargeBlock = FALSE;            /* If TRUE, the next DBlLength bytes of the tape must be read into XMS */
  dword               AddedSize = 0L;
  bool                FirstPart;
  char                NameIDBuffer[65];
  char                StatusText[80];
  bool                ReadingTapeID = TRUE;
  bool                SkipSoundContinue;
  bool                PRCancel = FALSE;
  bool                TZXTooHighReported = FALSE;
  bool                PreviousWasVOCID8 = FALSE;
  int                 ErrNo;
  bool                LoadingSnapshot = FALSE;
  char               *SnapWinHeader = "Converting Snapshot";
  char               *TapeWinHeader = "Reading Tape";

  DecodeDirectRecording = FALSE;                                              /* We're not decoding a Direct Recording block here */
  VOCHeaderLocationIndex = 0;                                                              /* Clear any VOC header location table */
  _TAPNoise = FALSE;
  errno = 0;
  if (TapeWindow[WhichWindow].FileType == FT_UNKNOWN)                                                       /* Final sanity check */
    if (GetFileType (WhichWindow) == FT_UNKNOWN)                                                  /* Ensure the type was resolved */
    {                                                                               /* Unknown type - something has gone wrong... */
      ErrorRequester (-1, -1, "@Close", NULL, NULL, "%s has unknown Spectrum tape file type", TapeWindow[WhichWindow].TapeName);
      return (FALSE);
    }
  if (IsSnapshot (WhichWindow))
    LoadingSnapshot = TRUE;
  if (MaxFloppyIdChanged)
    WriteConfig ();
  if (TapeWindow[WhichWindow].TapeName[1] == ':' && IsFloppy (TapeWindow[WhichWindow].TapeName[0]))
  {
    ReadFloppyId (TapeWindow[WhichWindow].TapeName[0], TRUE);
    TapeWindow[WhichWindow].TapeFloppyId = CurrentFloppyId;
  }
  if ((Fh = open (TapeWindow[WhichWindow].TapeName, O_BINARY | O_RDONLY)) == -1)
    return (TRUE);
  if (TapeWindow[WhichWindow].FileType == FT_VOC ||                                                 /* About to read a VOC file ? */
      TapeWindow[WhichWindow].FileType == FT_ACORNRAW ||                                            /* About to read a RAW file ? */
      TapeWindow[WhichWindow].FileType == FT_CSW)                                                   /* About to read a CSW file ? */
    if (!SetupVOCReadTimers (TapeWindow[WhichWindow].FileType))                    /* Allow the users to select the timing values */
    {
      close (Fh);
      return (TRUE);                                                                            /* (Previous tape is still there) */
    }
  if (!Merge && !_PreReadingTape)
    ReleaseTape (WhichWindow);                                                             /* Release current tape unless merging */
  TapeWindow[WhichWindow].NonexistingTape = FALSE;
  TapeWindow[WhichWindow].BackupTape = TPConfig.AlwaysBackupTape;
  sprintf (StatusText, "%s %s %s into extended memory ...", Merge ? "Adding" : "Reading",
           WhichWindow == TWLEFT ? "source" : "destination", LoadingSnapshot ? "snapshot" : "tape");
  MakePRequester (&_PReq, -1, -1, (bool)((TapeWindow[WhichWindow].FileType == FT_VOC ||
                                          TapeWindow[WhichWindow].FileType == FT_CSW) ? TRUE : FALSE), NULL,
                  LoadingSnapshot ? SnapWinHeader : TapeWinHeader, StatusText);            /* (VOC decoding gets a cancel button) */
  _PRFileLength = filelength (Fh);
  if (MoreToGo = (bool)(_PRFileLength != 0))
    TapeWindow[WhichWindow].EmptyTape = FALSE;
  NormalTiming.LengthPilotPulse = NS_LENPILOT;                                                   /* Prepare for normal speed data */
  NormalTiming.LengthSyncPulse[0] = NS_LENSYNC1;
  NormalTiming.LengthSyncPulse[1] = NS_LENSYNC2;
  NormalTiming.LengthZeroBitPulse = NS_LEN0BIT;
  NormalTiming.LengthOneBitPulse = NS_LEN1BIT;
  NormalTiming.NumPilotPulses = NS_DTPILOT;                                                                    /* (Assume header) */
  NormalTiming.NumSyncPulses = 2;
  NormalTiming.SampleRate = NS_SAMPRATE;
  NormalTiming.NumBitsInFirstByte = NS_LENFIRST;
  NormalTiming.NumBitsInLastByte = NS_LENLAST;
  GlobalDecodingScheme = DS_NONE;
  while (!PRCancel && MoreToGo)
  {
    switch (TapeWindow[WhichWindow].FileType)
    {
      case FT_TAPZ80 :                                                              /* (TAP and LTP are nearly identical formats) */
      case FT_LTP    : if ((NRead = read (Fh, &BlLength, 2)) == -1)                                     /* Can I read two bytes ? */
                       { __READERROR: ErrNo = errno;
                                      if (ErrNo == 0)                                                           /* Read 0 bytes ? */
                                      {
                                        ErrorRequester (-1, -1, "@Close", NULL, NULL, "Tape ends pre-mature. Last block droppped");
                                        goto __CONTINUE;
                                      }
                                      close (Fh);
                                      _DestroyPReq ();
                                      ErrorRequester (-1, -1, "@Close", NULL, NULL, _FileReadError,
                                                      TapeWindow[WhichWindow].TapeName, StrError (ErrNo));
                                      return (FALSE);
                       }
                       if (NRead == 2)                                                   /* Then there is at least one more block */
                       {
                         if (TapeWindow[WhichWindow].FileType == FT_TAPZ80 && BlLength == 0) /* Flag + block + parity = 0 bytes ? */
                         {                                                                           /* Then it is noise. Skip it */
                           _TAPNoise = TRUE;                                                 /* Signal: skipping some tape blocks */
                           NoiseNumEntries = 0;
                           while (BlLength == 0 && NRead == 2)                                 /* (And all next noise blocks too) */
                           {
                             NoiseNumEntries ++;                                             /* Count all (alledged) noise blocks */
                             Index += 2;
                             if ((NRead = read (Fh, &BlLength, 2)) == -1)  goto __READERROR;
                           }
                           if (NRead < 2)  goto __CONTINUE;                                              /* End of file reached ? */
                         }
                         if (_TAPNoise)                                                      /* Did we just skip 0-sized blocks ? */
                         {                                            /* The tape continues, so we need to insert them after all! */
                           DirP = TapeWindow[WhichWindow].Directory;
                           if (DirP != NULL)                                                      /* Go to last entry in the tape */
                             while (DirP->Next != NULL)
                               DirP = DirP->Next;
                           for (NoiseCnt = 0 ; NoiseCnt < NoiseNumEntries ; NoiseCnt ++)
                           {                                                                    /* Create each (empty) tape entry */
                             if ((Entry = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
                               FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
                             Entry->ComboSelected = 0x00;
                             Entry->Flags.IsNormalSpeed = TRUE;
                             Entry->IsTapeData = TRUE;
                             Entry->IsIdentifiedData = TRUE;
                             Entry->BlockTypeID = 0x10;
                             memcpy (&(Entry->Timing), &NormalTiming, sizeof (struct Timing_s));
                             Entry->Flags.FlagByteUsed = FALSE;
                             Entry->Flags.ParityByteUsed = FALSE;
                             Entry->BlockLength = Entry->AllocLength = 0;
                             Entry->PauseLength = 1000;
                             Entry->Next = NULL;
                             if (TapeWindow[WhichWindow].Directory == NULL)
                               DirP = TapeWindow[WhichWindow].Directory = Entry;
                             else
                             {
                               DirP->Next = Entry;
                               DirP = Entry;
                             }
                             TapeWindow[WhichWindow].NumEntries ++;
                           }
                           _TAPNoise = FALSE;                                                     /* Signal: no noise (after all) */
                         }                                                            /* And go on to insert the newly read entry */
                         if ((Entry = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
                           FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
                         Entry->ComboSelected = 0x00;
                         Entry->Flags.IsNormalSpeed = TRUE;                                      /* (Only possible for TAP files) */
                         Entry->IsTapeData = TRUE;
                         Entry->IsIdentifiedData = TRUE;
                         Entry->PauseLength = 1000;                                                    /* 1 sec. interblock pause */
                         Entry->BlockTypeID = 0x10;                                                          /* Normal speed data */
                         memcpy (&(Entry->Timing), &NormalTiming, sizeof (struct Timing_s));
                         if (TapeWindow[WhichWindow].FileType == FT_LTP || BlLength > 0)                       /* Anything left ? */
                         {
                           Entry->Flags.FlagByteUsed = TRUE;
                           if (read (Fh, &(Entry->FlagByte), 1) <= 0)
                           { free (Entry); goto __READERROR; }
                           if (TapeWindow[WhichWindow].FileType == FT_TAPZ80 && BlLength == 1)
                           {
                             Entry->Flags.ParityByteUsed = FALSE;
                             Entry->BlockLength = Entry->AllocLength = 0;
                           }
                           else
                           {
                             if (TapeWindow[WhichWindow].FileType == FT_LTP || BlLength > 2)
                             {
                               Entry->BlockLength = TapeWindow[WhichWindow].FileType == FT_TAPZ80 ? BlLength - 2 : BlLength;
                               Entry->AllocLength = (BlLength & 0xFFFFFFFC) + (BlLength % 4 ? 4 : 0);
                               Entry->CalculatedParity = Entry->FlagByte;
                               if ((Entry->Flags.IsHeader = (Entry->FlagByte == 0x00 && Entry->BlockLength == 17 ? TRUE : FALSE)))
                                 Entry->Timing.NumPilotPulses = NS_HDPILOT;
                               else
                                 Entry->PauseLength = 2000;                                   /* 2 sec. between non-header blocks */
                               if (TapeWindow[WhichWindow].Tape.XMSHandle == -1)                                 /* First entry ? */
                                 XMSAlloc (&(TapeWindow[WhichWindow].Tape.XMSHandle), Entry->AllocLength);
                               else
                                 XMSRealloc (TapeWindow[WhichWindow].Tape.XMSHandle, TapeWindow[WhichWindow].Tape.Size +
                                             Entry->AllocLength, TRUE);
                               XMSIndex = 0;                                                            /* Start at the beginning */
                               LengthToGo = Entry->BlockLength;
                               while (LengthToGo > 0)                                                  /* Read the data in chunks */
                               {
                                 ReadLength = (word)(LengthToGo > (dword)49152 ? 49152 : LengthToGo);
                                 if ((RealReadLength = read (Fh, WorkBuffer, ReadLength)) < ReadLength)        /* Premature EOF ? */
                                 { if (errno != 0)                                                            /* A 'real' error ? */
                                   { free (Entry); goto __READERROR; }
                                   if (!QuestionRequester (-1, -1, FALSE, FALSE, "@Drop", "@Keep", NULL, NULL, NULL, NULL,
                                                           "The last block in this tape is incomplete.\n"
                                                           "It is defined as %lu bytes, but only %lu bytes were read.\n"
                                                           "Do you want to drop it or keep it?", Entry->BlockLength,
                                                           Entry->BlockLength - LengthToGo + RealReadLength -
                                                           (PureData ? 0 : 1)))
                                   { free (Entry); goto __CONTINUE; }
                                   Entry->Flags.ParityByteUsed = TRUE;
                                   Entry->ParityByte = *(WorkBuffer + RealReadLength - 1);
                                   RealReadLength --;
                                   Entry->BlockLength = Entry->BlockLength - LengthToGo + RealReadLength;
                                   for (Cnt = 0 ; Cnt < RealReadLength ; Cnt ++)
                                     Entry->CalculatedParity ^= *(WorkBuffer + Cnt);
                                   DLengthToGo = (dword)(RealReadLength & 0xFFFC) + (RealReadLength % 4 ? 4 : 0);
                                   CopyToXMS (DLengthToGo, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer,
                                              TapeWindow[WhichWindow].Tape.Size + XMSIndex);
                                   goto __TAPCONT;
                                 }
                                 for (Cnt = 0 ; Cnt < ReadLength ; Cnt ++)                                    /* Calculate parity */
                                   Entry->CalculatedParity ^= *(WorkBuffer + Cnt);
                                 DLengthToGo = (dword)(ReadLength & 0xFFFC) + (ReadLength % 4 ? 4 : 0);
                                 CopyToXMS (DLengthToGo, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer,
                                            TapeWindow[WhichWindow].Tape.Size + XMSIndex);
                                 XMSIndex += RealReadLength;
                                 LengthToGo -= RealReadLength;
                               }
                             }
                             else
                             {
                               Entry->BlockLength = Entry->AllocLength = 0;
                               Entry->CalculatedParity = Entry->FlagByte;
                             }
                             Entry->Flags.ParityByteUsed = TRUE;
                             if (read (Fh, &(Entry->ParityByte), 1) <= 0)
                             { free (Entry); goto __READERROR; }
                           }
                         }
                         __TAPCONT:
                         Index += BlLength + 2;                                                              /* Update file index */
                         TapeWindow[WhichWindow].Tape.Size += Entry->AllocLength;
                         Entry->Next = NULL;
                         if (TapeWindow[WhichWindow].Directory == NULL)
                           TapeWindow[WhichWindow].Directory = Entry;
                         else
                         {
                           DirP = TapeWindow[WhichWindow].Directory;
                           while (DirP->Next != NULL)
                             DirP = DirP->Next;
                           DirP->Next = Entry;
                         }
                         TapeWindow[WhichWindow].NumEntries ++;
                       }
                       else                                                                             /* Less than 2 bytes read */
                       {
                         Index += NRead;                                                                  /* Make the index 'fit' */
                         _TAPNoise = TRUE;
                       }
                       break;
      case FT_ZXS    : if ((NRead = read (Fh, &BlLength, 2)) == -1)  goto __READERROR;                  /* Can I read two bytes ? */
                       if (NRead == 2)                                                   /* Then there is at least one more block */
                       {
                         Index += 2;
                         if (BlLength == 0)                                                        /* Unallocated block (noise) ? */
                         {
                           while (BlLength == 0 && NRead == 2)                                          /* Skip all noise at once */
                           {
                             if ((NRead = read (Fh, &BlLength, 2)) == -1)  goto __READERROR;
                             Index += 2;
                           }
                           if (NRead < 2)  goto __CONTINUE;                                                      /* End of file ? */
                         }
                         if (InsertBlock)                                                        /* Previous block was inserted ? */
                         {
                           if ((Entry = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
                             FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
                           memcpy (&Timing, &NormalTiming, sizeof (struct Timing_s));            /* Prepare for normal speed data */
                         }
                         Entry->ComboSelected = 0x00;
                         Entry->Flags.IsNormalSpeed = !UseTiming;
                         Entry->IsTapeData = TRUE;
                         Entry->IsIdentifiedData = TRUE;
                         Entry->PauseLength = PauseLength;
                         Entry->BlockTypeID = 0x10;                                                          /* Normal speed data */
                         InsertBlock = TRUE;                                                          /* Assume normal data block */
                         switch (BlLength)
                         {
                           case 0x0000 : Entry->Flags.FlagByteUsed = FALSE;                                /* Unallocated (noise) */
                                         Entry->Flags.ParityByteUsed = FALSE;
                                         Entry->BlockLength = Entry->AllocLength = 0;
                                         break;
                           case 0x0001 : InsertBlock = FALSE;                                             /* Direct sampling data */
                                         if ((NRead = read (Fh, &BScratch, 1)) <= 0)                          /* Read sample rate */
                                         { free (Entry); goto __READERROR; }
                                         Entry->Timing.SampleRate = (dword)((double)3.5 * pow (10, 6) /
                                                                           ((((BScratch >> 1) & 0x07) + 8) *
                                                                           pow (2, ((BScratch >> 4) & 0x0F))));  /* ONLY 8-BIT!!! */
                                         if ((NRead = read (Fh, &(Entry->BlockLength), 4)) <= 0)
                                         { free (Entry); goto __READERROR; }
                                         Index += 5;
                                         if (!DecodeDirectSample (WhichWindow, Fh, Entry->Timing.SampleRate, Entry->BlockLength,
                                                                  &Index, FALSE, &PRCancel, 0))
                                         { free (Entry); goto __READERROR; }
                                         free (Entry);                                              /* (Has already been handled) */
                                         break;
                           case 0x0002 : Entry->BlockTypeID = 0x20;                                             /* Infinite Pause */
                                         Entry->Flags.IsNormalSpeed = FALSE;
                                         Entry->PauseLength = 0;
                                         Entry->BlockLength = Entry->AllocLength = 0;
                                         GetBlockTypeInfo (0x20, &(Entry->IsTapeData), &(Entry->IsIdentifiedData), NULL, NULL,
                                                           NULL);
                                         break;
                           case 0x0003 : InsertBlock = FALSE;                                           /* Change speed/no detail */
                                         if ((NRead = read (Fh, &BScratch, 1)) <= 0)                         /* Read speed 'name' */
                                         { free (Entry); goto __READERROR; }
                                         switch (BScratch)
                                         {
                                           case 'l' :
                                           case 'L' : Timing.LengthSyncPulse[0] = NS_LENSYNC1;                    /* Normal speed */
                                                      Timing.LengthSyncPulse[1] = NS_LENSYNC2;
                                                      Timing.LengthZeroBitPulse = NS_LEN0BIT;
                                                      Timing.LengthOneBitPulse = NS_LEN1BIT;
                                                      UseTiming = FALSE;
                                                      break;
                                           case 'h' :
                                           case 'H' : Timing.LengthSyncPulse[0] = HS_LENSYNC1;                      /* High speed */
                                                      Timing.LengthSyncPulse[1] = HS_LENSYNC2;
                                                      Timing.LengthZeroBitPulse = HS_LEN0BIT;
                                                      Timing.LengthOneBitPulse = HS_LEN1BIT;
                                                      UseTiming = TRUE;
                                                      break;
                                           case 't' :
                                           case 'T' :
                                           case 'x' :
                                           case 'X' : Timing.LengthSyncPulse[0] = TS_LENSYNC1;             /* Turbo/Maximum speed */
                                                      Timing.LengthSyncPulse[1] = TS_LENSYNC2;
                                                      Timing.LengthZeroBitPulse = TS_LEN0BIT;
                                                      Timing.LengthOneBitPulse = TS_LEN1BIT;
                                                      UseTiming = TRUE;
                                                      break;
                                         }
                                         Index += 1;
                                         break;
                           case 0x0004 : Entry->BlockTypeID = 0x20;                                            /* Specified Pause */
                                         Entry->Flags.IsNormalSpeed = FALSE;
                                         if ((NRead = read (Fh, &ReadLength, 2)) <= 0)                   /* Read new pause length */
                                         { free (Entry); goto __READERROR; }
                                         Entry->PauseLength = ReadLength;
                                         GetBlockTypeInfo (0x20, &(Entry->IsTapeData), &(Entry->IsIdentifiedData), NULL, NULL,
                                                           NULL);
                                         Index += 2;
                                         break;
                           case 0x0005 : InsertBlock = FALSE;                                         /* Change speed with detail */
                                         if ((NRead = read (Fh, &BScratch, 1)) <= 0)                        /* Read object 'name' */
                                         { free (Entry); goto __READERROR; }
                                         if ((NRead = read (Fh, &ReadLength, 2)) <= 0)                     /* Read object 'value' */
                                         { free (Entry); goto __READERROR; }
                                         switch (BScratch)
                                         {
                                           case '0' : Timing.LengthZeroBitPulse = ReadLength; break;
                                           case '1' : Timing.LengthOneBitPulse = ReadLength; break;
                                           case '2' : Timing.LengthPilotPulse = ReadLength; break;
                                           case 'a' :
                                           case 'A' : Timing.NumPilotPulses = (word)(((dword)3500 * ReadLength) /
                                                                                     (2 * Timing.LengthPilotPulse)); break;
                                           case 'o' :
                                           case 'O' : PauseLength = ReadLength; break;
                                         }
                                         UseTiming = TRUE;
                                         Index += 3;
                                         break;
                           default     : Entry->BlockTypeID = UseTiming ? 0x11 : 0x10;                       /* Actual data block */
                                         GetBlockTypeInfo (Entry->BlockTypeID, &(Entry->IsTapeData), &(Entry->IsIdentifiedData),
                                                           NULL, NULL, NULL);
                                         memcpy (&(Entry->Timing), &Timing, sizeof (struct Timing_s));
                                         if ((NRead = read (Fh, &BScratch, 1)) <= 0)               /* Starting phase (don't care) */
                                         { free (Entry); goto __READERROR; }
                                         if ((NRead = read (Fh, &BScratch, 1)) <= 0)                        /* Custom length byte */
                                         { free (Entry); goto __READERROR; }
                                         Entry->Timing.NumBitsInFirstByte = (BScratch & 0x38) >> 3;
                                         if (Entry->Timing.NumBitsInFirstByte == 0)
                                           Entry->Timing.NumBitsInFirstByte = 8;
                                         Entry->Timing.NumBitsInLastByte = BScratch & 0x07;
                                         if (Entry->Timing.NumBitsInLastByte == 0)
                                           Entry->Timing.NumBitsInLastByte = 8;
                                         if ((NRead = read (Fh, &(Entry->Timing.CustomFirstByte), 1)) <= 0)       /* Leading byte */
                                         { free (Entry); goto __READERROR; }
                                         Index += BlLength - 2;
                                         BlLength -= 5;                                                  /* Exclude this 'header' */
                                         if (Entry->Timing.NumBitsInLastByte != 8)                /* Additional bits at the end ? */
                                           BlLength --;
                                         if (BlLength == 0)                                  /* Flag + block + parity = 0 bytes ? */
                                         {
                                           Entry->Flags.FlagByteUsed = FALSE;
                                           Entry->Flags.ParityByteUsed = FALSE;
                                           Entry->BlockLength = Entry->AllocLength = 0;
                                         }
                                         else
                                         {
                                           Entry->Flags.FlagByteUsed = TRUE;
                                           if (read (Fh, &(Entry->FlagByte), 1) <= 0)
                                           { free (Entry); goto __READERROR; }
                                           if (BlLength == 1)
                                           {
                                             Entry->Flags.ParityByteUsed = FALSE;
                                             Entry->BlockLength = Entry->AllocLength = 0;
                                           }
                                           else
                                           {
                                             if (BlLength > 2)
                                             {
                                               Entry->BlockLength = BlLength - 2;
                                               Entry->AllocLength = (BlLength & 0xFFFFFFFC) + (BlLength % 4 ? 4 : 0);
                                               if (TapeWindow[WhichWindow].Tape.XMSHandle == -1)                 /* First entry ? */
                                                 XMSAlloc (&(TapeWindow[WhichWindow].Tape.XMSHandle), Entry->AllocLength);
                                               else
                                                 XMSRealloc (TapeWindow[WhichWindow].Tape.XMSHandle,
                                                             TapeWindow[WhichWindow].Tape.Size + Entry->AllocLength, TRUE);
                                               Entry->CalculatedParity = Entry->FlagByte;
                                               if ((Entry->Flags.IsHeader = (Entry->FlagByte == 0x00 && Entry->BlockLength == 17 ?
                                                                  TRUE : FALSE)))
                                                 Entry->Timing.NumPilotPulses = NS_HDPILOT;
                                               XMSIndex = 0;                                            /* Start at the beginning */
                                               LengthToGo = Entry->BlockLength;
                                               while (LengthToGo > 0)                                  /* Read the data in chunks */
                                               {
                                                 ReadLength = (word)(LengthToGo > (dword)49152 ? 49152 : LengthToGo);
                                                 if (read (Fh, WorkBuffer, ReadLength) < ReadLength)           /* Premature EOF ? */
                                                 { free (Entry);
                                                   __INVALIDTAPEFILE: close (Fh);
                                                                      _DestroyPReq ();
                                                                      ErrorRequester (-1, -1, "@Close", NULL, NULL,
                                                                                      _ErrInvalidTapeFile,
                                                                                      TapeWindow[WhichWindow].TapeName);
                                                                      return (FALSE);
                                                 }
                                                 for (Cnt = 0 ; Cnt < ReadLength ; Cnt ++)                    /* Calculate parity */
                                                   Entry->CalculatedParity ^= *(WorkBuffer + Cnt);
                                                 DLengthToGo = (dword)(ReadLength & 0xFFFC) + (ReadLength % 4 ? 4 : 0);
                                                 CopyToXMS (DLengthToGo, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer,
                                                            TapeWindow[WhichWindow].Tape.Size + XMSIndex);
                                                 XMSIndex += ReadLength;
                                                 LengthToGo -= ReadLength;
                                               }
                                             }
                                             else
                                             {
                                               Entry->BlockLength = Entry->AllocLength = 0;
                                               Entry->CalculatedParity = Entry->FlagByte;
                                             }
                                             Entry->Flags.ParityByteUsed = TRUE;
                                             if (read (Fh, &(Entry->ParityByte), 1) <= 0)
                                             { free (Entry); goto __READERROR; }
                                           }
                                         }
                                         if (Entry->Timing.NumBitsInLastByte != 8)                /* Additional bits at the end ? */
                                           if (read (Fh, &(Entry->Timing.CustomLastByte), 1) <= 0)
                                           { free (Entry); goto __READERROR; }
                                         break;
                         }
                         if (InsertBlock)                                                                    /* Update file index */
                         {
                           TapeWindow[WhichWindow].Tape.Size += Entry->AllocLength;
                           Entry->Next = NULL;
                           if (TapeWindow[WhichWindow].Directory == NULL)
                             TapeWindow[WhichWindow].Directory = Entry;
                           else
                           {
                             DirP = TapeWindow[WhichWindow].Directory;
                             while (DirP->Next != NULL)
                               DirP = DirP->Next;
                             DirP->Next = Entry;
                           }
                           TapeWindow[WhichWindow].NumEntries ++;
                         }
                       }
                       break;
      case FT_VOC    : if (ReadingTapeID)
                       {
                         if ((NRead = read (Fh, NameIDBuffer, 20)) <= 0)  goto __READERROR;                 /* Read VOC ID string */
                         if (strncmp (NameIDBuffer, "Creative Voice File\x1A", 20))  goto __INVALIDTAPEFILE;
                         if ((NRead = read (Fh, NameIDBuffer, 6)) <= 0)  goto __INVALIDTAPEFILE;                /* Read some info */
                         TapeWindow[WhichWindow].TapeVersion = (word)NameIDBuffer[2] + (word)NameIDBuffer[3] * 256;
                         WScratch = (word)NameIDBuffer[0] + (word)NameIDBuffer[1] * 256;            /* Offset of first data block */
                         Index = WScratch;
                         if (lseek (Fh, (dword)WScratch, SEEK_SET) == -1)  goto __READERROR;            /* Start at the beginning */
                         ReadingTapeID = FALSE;
                       }
                       PreviousWasVOCID8 = FALSE;
                       while (MoreToGo)                                                                  /* Wait for a data block */
                       {
                         if ((NRead = read (Fh, &BlockTypeID, 1)) == -1)  goto __READERROR;                      /* Read block ID */
                         if (NRead == 0)
                           MoreToGo = FALSE;
                         else if (BlockTypeID == 0x00)                                                            /* (Terminator) */
                         { Index += 1; MoreToGo = FALSE; }
                         else
                         {
                           if ((NRead = read (Fh, &LengthToGo, 3)) == -1)  goto __READERROR;                 /* Read block length */
                           LengthToGo &= 0x00FFFFFF;                                                             /* Only 3 bytes! */
                           Index += 4;
                           switch (BlockTypeID)
                           {
                             case 0x00 : LengthToGo = 0; MoreToGo = FALSE; break;                                 /* (Terminator) */
                             case 0x01 : if ((NRead = read (Fh, &BScratch, 1)) == -1)  goto __READERROR;          /* (Sound data) */
                                         if (!PreviousWasVOCID8)
                                         {
                                           Timing.SampleRate = 1000000 / (256 - BScratch);                         /* Sample rate */
                                           if (VOCHighSpeedSampleRateAdjust && Timing.SampleRate >= 22222)
                                             Timing.SampleRate = (dword)256000000 / ((dword)65536 - (((dword)BScratch << 8) |0x54));
                                           else if (VOCFrequencyShift)
                                             Timing.SampleRate = 1000000 / (257 - BScratch);
                                         }
                                         if ((NRead = read (Fh, &BScratch, 1)) == -1)  goto __READERROR;      /* Compression type */
                                         MoreToGo = FALSE;
                                         Index += 2;
                                         LengthToGo -= 2;
                                         break;
                             case 0x02 : MoreToGo = FALSE; break;                                             /* (Sound Continue) */
                             case 0x03 : if ((NRead = read (Fh, &WScratch, 2)) == -1)  goto __READERROR;             /* (Silence) */
                                         if ((NRead = read (Fh, &BScratch, 1)) == -1)  goto __READERROR;           /* Sample rate */
                                         Timing.SampleRate = 1000000 / (256 - BScratch);
                                         Index += 3;
                                         WScratch = (word)(((double)WScratch * 1000 / Timing.SampleRate) + 0.5);
                                         MoreToGo = TRUE;
                                         if (TapeWindow[WhichWindow].Directory != NULL && WScratch != 0)
                                         {                       /* See if the previous block was a block that uses a PauseLength */
                                           DirP = TapeWindow[WhichWindow].Directory;
                                           while (DirP->Next != NULL)
                                             DirP = DirP->Next;
                                           if ((WScratch != 0 && (DirP->BlockTypeID == 0x10 || DirP->BlockTypeID == 0x11 ||
                                                                  DirP->BlockTypeID == 0x14 || DirP->BlockTypeID == 0x20)) ||
                                               (WScratch == 0 && DirP->BlockTypeID == 0x20 && DirP->PauseLength == 0x00))
                                           {
                                             DirP->PauseLength += WScratch;                      /* Add the silence to that block */
                                             MoreToGo = FALSE;                                                 /* Signal: handled */
                                           }
                                         }
                                         if (MoreToGo)
                                         {
                                           if ((Entry = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
                                             FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
                                           Entry->ComboSelected = 0x00;
                                           Entry->PauseLength = WScratch;
                                           Entry->BlockTypeID = 0x20;                                              /* Pause block */
                                           Entry->BlockLength = 0;
                                           Entry->AllocLength = 0;
                                           GetBlockTypeInfo (0x20, &(Entry->IsTapeData), &(Entry->IsIdentifiedData),
                                                             NULL, NULL, NULL);
                                           Entry->Next = NULL;                                          /* Insert the pause block */
                                           if (TapeWindow[WhichWindow].Directory == NULL)
                                             TapeWindow[WhichWindow].Directory = Entry;
                                           else
                                           {
                                             DirP = TapeWindow[WhichWindow].Directory;
                                             while (DirP->Next != NULL)
                                               DirP = DirP->Next;
                                             DirP->Next = Entry;
                                           }
                                           TapeWindow[WhichWindow].NumEntries ++;
                                         }
                                         MoreToGo = TRUE;                                                 /* (Reset the variable) */
                                         break;
                             case 0x04 : if ((NRead = read (Fh, &WScratch, 2)) == -1)  goto __READERROR;              /* (Marker) */
                                         Index += 2;                                                     /* Discard marker number */
                                         break;
                             case 0x05 : break;                                                                 /* (ASCII String) */
                             case 0x06 : if ((NRead = read (Fh, &WScratch, 2)) == -1)  goto __READERROR;              /* (Repeat) */
                                         Index += 2;
                                         break;
                             case 0x07 : break;                                                                   /* (End Repeat) */
                             case 0x08 : if ((NRead = read (Fh, &WScratch, 2)) == -1)  goto __READERROR;            /* (Extended) */
                                         Timing.SampleRate = (dword)256000000 / ((dword)65536 - ((dword)WScratch));
                                           /* word TimingConstant (2nd formula) */
                                           /* byte Pack */
                                           /* byte Mode (0=mono, 1=stereo) */
                                         if ((NRead = read (Fh, &WScratch, 2)) == -1)  goto __READERROR;
                                         if (WScratch != 0x0000)  goto __INVALIDTAPEFILE;           /* Must be mono, uncompressed */
                                         Index += 4;                                                          /* Discard the info */
                                         PreviousWasVOCID8 = TRUE;                                    /* Lock the samplerate info */
                                         break;
                             case 0x09 : if ((NRead = read (Fh, &DScratch, 4)) == -1)  goto __READERROR;        /* (GoldWave ext) */
                                         Timing.SampleRate = (dword)(DScratch & 0x0000FFFF);                 /* Direct samplerate */
                                           /* byte BitsPerSample (after compression) */
                                           /* byte Channel (1=mono, 2=stereo) */
                                           /* word Format (0x0000=8-bit unsigned PCM, 0x0004=16-bit signed) */
                                           /* byte Reserved[4] */
                                         if ((NRead = read (Fh, &DScratch, 4)) == -1)  goto __READERROR;
                                         if (DScratch != 0x00000108)  goto __INVALIDTAPEFILE;               /* Must be 8-bit mono */
                                         if ((NRead = read (Fh, &DScratch, 4)) == -1)  goto __READERROR;
                                         MoreToGo = FALSE;
                                         Index += 12;                                                         /* Discard the info */
                                         LengthToGo -= 12;
                                         break;
                             default   : goto __INVALIDTAPEFILE;
                           }
                           if (NRead == 0)                                                                       /* Reached EOF ? */
                             MoreToGo = FALSE;
                         }
                       }
                       if (BlockTypeID == 0x01 || BlockTypeID == 0x02 || BlockTypeID == 0x09)            /* A sample data block ? */
                       {
                         if (!UpdatePReq (Index, "Status: Analyzing sample chain ..."))
                           PRCancel = TRUE;
                         else
                         {
                           DBlLength = Index;                                              /* Take a copy of the current location */
                           SkipSoundContinue = FALSE;
                           MoreToGo = TRUE;
                           while (MoreToGo)
                           {
                             if (lseek (Fh, DBlLength + LengthToGo, SEEK_SET) == -1)                    /* Check ID of next block */
                             {                                                                                  /* Pre-mature EOF */
                               if (!QuestionRequester (-1, -1, FALSE, FALSE, "@Stop", "@Decode", NULL, NULL, NULL, NULL,
                                                       "The file ends pre-mature.\n"
                                                       "The samplechain suggests a file size of at least %lu bytes, but the actual "
                                                       "size is %lu bytes.\n"
                                                       "Do you want to decode as much as possible?", DBlLength + LengthToGo + 1,
                                                       _PRFileLength))
                                 goto __CONTINUE;
                               MoreToGo = FALSE;                                                    /* (Go on as far as possible) */
                             }
                             else
                             {
                               Index = DBlLength + LengthToGo;
                               if ((NRead = read (Fh, &BlockTypeID, 1)) == -1)  goto __READERROR;
                               if (BlockTypeID == 0x01 || BlockTypeID == 0x02 || BlockTypeID == 0x09)         /* Sound Continue ? */
                               {
                                 if ((NRead = read (Fh, &DScratch, 3)) == -1)  goto __READERROR;             /* Read block length */
                                 DScratch &= 0x00FFFFFF;
                                 LengthToGo += DScratch + 4;                                                   /* Chain the block */
                                 SkipSoundContinue = TRUE;
                                 VOCHeaderLocations[VOCHeaderLocationIndex ++] = Index;                  /* Remember the location */
                               }
                               else
                                 MoreToGo = FALSE;
                             }
                           }
                           Index = DBlLength;
                           if (lseek (Fh, Index, SEEK_SET) == -1)  goto __READERROR;                      /* Go back to the start */
                           if (InsertDirectRecording)
                           {
                             if (!ReadDirectRecording (WhichWindow, Fh, Timing.SampleRate, LengthToGo, &Index, SkipSoundContinue,
                                                       &PRCancel))
                               goto __READERROR;
                           }
                           else                                                                          /* (Normal VOC decoding) */
                           {
                             if (!DecodeDirectSample (WhichWindow, Fh, Timing.SampleRate, LengthToGo, &Index, SkipSoundContinue,
                                                      &PRCancel, 0))
                               goto __READERROR;
                           }
                         }
                       }
                       VOCHeaderLocationIndex = 0;                                 /* Clear the VOC header location table if used */
                       break;
      case FT_ACORNRAW: if (ReadingTapeID)
                       {
                         if ((NRead = read (Fh, NameIDBuffer, 64)) <= 0)  goto __READERROR;                 /* Read RAW ID string */
                         for (Cnt = 0 ; Cnt < 48 ; Cnt ++)
                           if (NameIDBuffer[Cnt] != '\0')  goto __INVALIDTAPEFILE;                    /* First 48 bytes are zeros */
                         if (strncmp (NameIDBuffer + 48, "Raw tape sample", 15) ||
                             NameIDBuffer[63] != '\0')  goto __INVALIDTAPEFILE;
                         Index = 64;
                         Timing.SampleRate = Z80EMSAMPRATE;
                         LengthToGo = _PRFileLength - Index;
                         ReadingTapeID = FALSE;
                       }
                       if (InsertDirectRecording)
                       {
                         if (!ReadDirectRecording (WhichWindow, Fh, Timing.SampleRate, LengthToGo, &Index, FALSE, &PRCancel))
                           goto __READERROR;
                       }
                       else                                                                              /* (Normal RAW decoding) */
                       {
                         if (!DecodeDirectSample (WhichWindow, Fh, Timing.SampleRate, LengthToGo, &Index, FALSE, &PRCancel, 0))
                           goto __READERROR;
                       }
                       break;
      case FT_CSW    : if (ReadingTapeID)
                       {
                         if ((NRead = read (Fh, NameIDBuffer, 32)) <= 0)  goto __READERROR;                 /* Read RAW ID string */
                         if (strncmp (NameIDBuffer, "Compressed Square Wave\x1A", 23))  goto __INVALIDTAPEFILE;
                         Index = 32;
                         Timing.SampleRate = (byte)NameIDBuffer[25] + (byte)NameIDBuffer[26] * 256;
                         if (NameIDBuffer[27] != 0x01)  goto __INVALIDTAPEFILE;              /* Type other than RLE not supported */
                         LengthToGo = _PRFileLength - Index;
                         ReadingTapeID = FALSE;
                       }
                       if (InsertDirectRecording)
                       {
                         if (!ReadDirectRecording (WhichWindow, Fh, Timing.SampleRate, LengthToGo, &Index, FALSE, &PRCancel))
                           goto __READERROR;
                       }
                       else                                                                              /* (Normal RAW decoding) */
                       {
                         if (!DecodeDirectSample (WhichWindow, Fh, Timing.SampleRate, LengthToGo, &Index, FALSE, &PRCancel, 0))
                           goto __READERROR;
                       }
                       break;
      case FT_TZX    : if (ReadingTapeID)
                       {
                         if ((NRead = read (Fh, NameIDBuffer, 8)) <= 0)  goto __READERROR;                  /* Read TZX ID string */
                         if (strncmp (NameIDBuffer, "ZXTape!\x1A", 8))  goto __INVALIDTAPEFILE;
                         if ((NRead = read (Fh, &WScratch, 2)) <= 0)  goto __INVALIDTAPEFILE;
                         TapeWindow[WhichWindow].TapeVersion = ((WScratch & 0x00FF) << 8) | ((WScratch & 0xFF00) >> 8);
                         Index = 10;
                         ReadingTapeID = FALSE;                                                               /* Read TZX version */
                       }
                       if ((NRead = read (Fh, &BScratch, 1)) == -1)  goto __READERROR;                     /* Can I read a byte ? */
                       Index += NRead;
                       if (NRead == 1)                                                   /* Then there is at least one more block */
                       {
                         if ((Entry = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
                           FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
                         Entry->ComboSelected = 0x00;
                         Entry->Flags.IsNormalSpeed = (BScratch == 0x10 ? TRUE : FALSE);
                         Entry->PauseLength = 0;
                         Entry->AllocLength = Entry->BlockLength = 0;                             /* Assume the block has no body */
                         Entry->BlockTypeID = BScratch;
                         GetBlockTypeInfo (BScratch, &(Entry->IsTapeData), &(Entry->IsIdentifiedData), NULL, NULL, NULL);
                         memcpy (&(Entry->Timing), &NormalTiming, sizeof (struct Timing_s));     /* Prepare for normal speed data */
                         InsertBlock = FALSE;                                  /* Assume the block is fully handled in the 'case' */
                         InsertWorkBuffer = FALSE;
                         InsertLargeBlock = FALSE;
                         AddedSize = 0L;
                         PureData = FALSE;
                         switch (BScratch)
                         {
                           case 0x10 :                                                             /* (Standard Speed Data Block) */
                                       if ((NRead = read (Fh, &(Entry->PauseLength), 2)) <= 0)  goto __READERROR;
                                       if ((NRead = read (Fh, &BlLength, 2)) <= 0)  goto __READERROR;              /* BlockLength */
                                       DBlLength = BlLength;
                                       Index += 4;
                                       InsertBlock = TRUE;                                  /* The data itself must still be done */
                                       break;
                           case 0x11 :                                                              /* (Turbo Loading Data Block) */
                                       if ((NRead = read (Fh, &(Entry->Timing.LengthPilotPulse), 2)) <= 0)  goto __READERROR;
                                       if ((NRead = read (Fh, &(Entry->Timing.LengthSyncPulse[0]), 2)) <= 0)  goto __READERROR;
                                       if ((NRead = read (Fh, &(Entry->Timing.LengthSyncPulse[1]), 2)) <= 0)  goto __READERROR;
                                       if ((NRead = read (Fh, &(Entry->Timing.LengthZeroBitPulse), 2)) <= 0)  goto __READERROR;
                                       if ((NRead = read (Fh, &(Entry->Timing.LengthOneBitPulse), 2)) <= 0)  goto __READERROR;
                                       if ((NRead = read (Fh, &(Entry->Timing.NumPilotPulses), 2)) <= 0)  goto __READERROR;
                                       if ((NRead = read (Fh, &(Entry->Timing.NumBitsInLastByte), 1)) <= 0)  goto __READERROR;
                                       if ((NRead = read (Fh, &(Entry->PauseLength), 2)) <= 0)  goto __READERROR;
                                       if ((NRead = read (Fh, &DBlLength, 3)) <= 0)  goto __READERROR;             /* BlockLength */
                                       Entry->Timing.NumSyncPulses = 2;
                                       DBlLength &= 0x00FFFFFF;                                                  /* Only 3 bytes! */
                                       Index += 18;
                                       InsertBlock = TRUE;                                  /* The data itself must still be done */
                                       break;
                           case 0x12 :                                                                             /* (Pure Tone) */
                                       if ((NRead = read (Fh, &(Entry->Timing.LengthPilotPulse), 2)) <= 0)  goto __READERROR;
                                       if ((NRead = read (Fh, &(Entry->Timing.NumPilotPulses), 2)) <= 0)  goto __READERROR;
                                       Index += 4;
                                       break;                                                         /* Store as pilot tone info */
                           case 0x13 :                                                  /* (Row of Pulses with Different Timings) */
                                       if ((NRead = read (Fh, &BScratch, 1)) <= 0)  goto __READERROR;         /* Number of pulses */
                                       Index += 1;
                                       BlLength = BScratch * 2;                                                        /* (words) */
                                       if (BlLength > 0)
                                       {
                                         if ((NRead = read (Fh, WorkBuffer, BlLength)) <= 0)  goto __READERROR;      /* Read them */
                                         InsertWorkBuffer = TRUE;                                         /* Insert them later on */
                                       }
                                       break;
                           case 0x14 :                                                                       /* (Pure Data Block) */
                                       if ((NRead = read (Fh, &(Entry->Timing.LengthZeroBitPulse), 2)) <= 0)  goto __READERROR;
                                       if ((NRead = read (Fh, &(Entry->Timing.LengthOneBitPulse), 2)) <= 0)  goto __READERROR;
                                       if ((NRead = read (Fh, &(Entry->Timing.NumBitsInLastByte), 1)) <= 0)  goto __READERROR;
                                       if ((NRead = read (Fh, &(Entry->PauseLength), 2)) <= 0)  goto __READERROR;
                                       if ((NRead = read (Fh, &DBlLength, 3)) <= 0)  goto __READERROR;             /* BlockLength */
                                       DBlLength &= 0x00FFFFFF;                                                  /* Only 3 bytes! */
                                       Entry->Timing.NumPilotPulses = 0;                                         /* No pilot tone */
                                       Entry->Timing.NumSyncPulses = 0;                                         /* No sync pulses */
                                       Index += 10;
                                       InsertBlock = TRUE;                                  /* The data itself must still be done */
                                       PureData = TRUE;                                            /* Don't split off sync/parity */
                                       break;
                           case 0x15 :                                                                      /* (Direct Recording) */
                                       if ((NRead = read (Fh, &(Entry->Timing.SampleRate), 2)) <= 0)  goto __READERROR;   /* In T */
                                       if ((NRead = read (Fh, &(Entry->PauseLength), 2)) <= 0)  goto __READERROR;
                                       if ((NRead = read (Fh, &(Entry->Timing.NumBitsInLastByte), 1)) <= 0)  goto __READERROR;
                                       if ((NRead = read (Fh, &DBlLength, 3)) <= 0)  goto __READERROR;             /* BlockLength */
                                       DBlLength &= 0x00FFFFFF;                                                  /* Only 3 bytes! */
                                       Index += 8;
                                       InsertLargeBlock = TRUE;                             /* The data itself must still be done */
                                       break;
                           case 0x20 :                                              /* (Pause (Silence) or Stop the Tape Command) */
                                       if ((NRead = read (Fh, &(Entry->PauseLength), 2)) <= 0)  goto __READERROR;
                                       Index += 2;
                                       break;
                           case 0x21 :                                                                           /* (Group Start) */
                                       if ((NRead = read (Fh, &BScratch, 1)) <= 0)  goto __READERROR;     /* Length of Group name */
                                       Index += 1;
                                       BlLength = (word)BScratch;
                                       *WorkBuffer = '\0';
                                       if (BlLength > 0)
                                         if ((NRead = read (Fh, WorkBuffer, BlLength)) <= 0)  goto __READERROR;     /* Group name */
                                       if (!strnicmp (WorkBuffer, "SpeedLock 2 ", 12))        GlobalDecodingScheme = DS_SLOCK2;
                                       else if (!strnicmp (WorkBuffer, "SpeedLock 2.5 ", 14)) GlobalDecodingScheme = DS_SLOCK3;
                                       else if (!strnicmp (WorkBuffer, "SpeedLock 3 ", 12))   GlobalDecodingScheme = DS_SLOCK3;
                                       else if (!strnicmp (WorkBuffer, "SpeedLock 4 ", 12))   GlobalDecodingScheme = DS_SLOCK4;
                                       else if (!strnicmp (WorkBuffer, "SpeedLock 5 ", 12))   GlobalDecodingScheme = DS_SLOCK5;
                                       else if (!strnicmp (WorkBuffer, "SpeedLock 3.5 ", 14)) GlobalDecodingScheme = DS_SLOCK6;
                                       else if (!strnicmp (WorkBuffer, "SpeedLock 6 ", 12))   GlobalDecodingScheme = DS_SLOCK6;
                                       else if (!strnicmp (WorkBuffer, "SpeedLock 7 ", 12))   GlobalDecodingScheme = DS_SLOCK7;
                                       InsertWorkBuffer = TRUE;
                                       break;
                           case 0x22 : break;                                                                      /* (Group End) */
                           case 0x23 :                                                                         /* (Jump To Block) */
                                       if ((NRead = read (Fh, &(Entry->PauseLength), 2)) <= 0)  goto __READERROR;
                                       Index += 2;                                                        /* Read as pause length */
                                       break;
                           case 0x24 : if ((NRead = read (Fh, &(Entry->PauseLength), 2)) <= 0)  goto __READERROR; /* (Loop Start) */
                                       Index += 2;                                                        /* Read as pause length */
                                       break;
                           case 0x25 : break;                                                                       /* (Loop End) */
                           case 0x26 :                                                                         /* (Call Sequence) */
                                       if ((NRead = read (Fh, &WScratch, 2)) <= 0)  goto __READERROR;          /* Number of calls */
                                       Index += 2;
                                       BlLength = WScratch * 2;
                                       if (BlLength > 0)
                                       {
                                         if ((NRead = read (Fh, WorkBuffer, BlLength)) <= 0)  goto __READERROR;      /* Read them */
                                         InsertWorkBuffer = TRUE;                                         /* Insert them later on */
                                         AddedSize = (dword)BlLength;                               /* Need double, for JumpCodes */
                                       }
                                       break;
                           case 0x27 : break;                                                           /* (Return from Sequence) */
                           case 0x28 :                                                                          /* (Select Block) */
                                       if ((NRead = read (Fh, &DBlLength, 2)) <= 0)  goto __READERROR;            /* Block length */
                                       Index += 2;
                                       DBlLength &= 0x0000FFFF;                                                  /* Only 2 bytes! */
                                       InsertLargeBlock = TRUE;                             /* The data itself must still be done */
                                       break;
                           case 0x2A : if ((NRead = read (Fh, &DBlLength, 4)) <= 0)  goto __READERROR;                /* Extended */
                                       Index += 4;
                                       break;                                                            /* (No real body - is 0) */
                           case 0x30 :                                                                      /* (Text Description) */
                                       if ((NRead = read (Fh, &BScratch, 1)) <= 0)  goto __READERROR;           /* Length of Text */
                                       Index += 1;
                                       BlLength = (word)BScratch;
                                       if (BlLength > 0)
                                       {
                                         if ((NRead = read (Fh, WorkBuffer, BlLength)) <= 0)  goto __READERROR;           /* Text */
                                         InsertWorkBuffer = TRUE;
                                       }
                                       break;
                           case 0x31 :                                                                         /* (Message Block) */
                                       if ((NRead = read (Fh, &BScratch, 1)) <= 0)  goto __READERROR;          /* Time to display */
                                       Entry->PauseLength = BScratch;                                    /* Store as pause length */
                                       if ((NRead = read (Fh, &BScratch, 1)) <= 0)  goto __READERROR;           /* Length of Text */
                                       Index += 2;
                                       BlLength = (word)BScratch;
                                       if (BlLength > 0)
                                       {
                                         if ((NRead = read (Fh, WorkBuffer, BlLength)) <= 0)  goto __READERROR;           /* Text */
                                         InsertWorkBuffer = TRUE;
                                       }
                                       break;
                           case 0x32 :                                                                          /* (Archive info) */
                                       if ((NRead = read (Fh, &BlLength, 2)) <= 0)  goto __READERROR;              /* BlockLength */
                                       Index += 2;
                                       if (BlLength > 0)
                                       {
                                         if ((NRead = read (Fh, WorkBuffer, BlLength)) <= 0)  goto __READERROR;   /* Entire block */
                                         InsertWorkBuffer = TRUE;
                                       }
                                       break;
                           case 0x33 :                                                                         /* (Hardware Type) */
                                       if ((NRead = read (Fh, &BScratch, 1)) <= 0)  goto __READERROR;       /* Number of machines */
                                       Index += 1;
                                       BlLength = BScratch * 3;                                                   /* 3 bytes each */
                                       if (BlLength > 0)
                                       {
                                         if ((NRead = read (Fh, WorkBuffer, BlLength)) <= 0)  goto __READERROR;      /* Read them */
                                         InsertWorkBuffer = TRUE;
                                         for (Cnt = 0 ; Cnt < BlLength ; Cnt += 3)
                                           if (*(WorkBuffer + Cnt + 2) == 0x01 &&                         /* "Runs on and uses" ? */
                                               *(WorkBuffer + Cnt) == 0x00)                                  /* Type = Computer ? */
                                             if (*(WorkBuffer + Cnt + 1) == 0x03 ||                              /* Spectrum 128K */
                                                 *(WorkBuffer + Cnt + 1) == 0x04 ||                           /* Spectrum 128K +2 */
                                                 *(WorkBuffer + Cnt + 1) == 0x05 ||                       /* Spectrum 128K +2A/+3 */
                                                 *(WorkBuffer + Cnt + 1) == 0x0E)                        /* Spanish Spectrum 128K */
                                               TapeWindow[WhichWindow].ModeIs128K = TRUE;                        /* Set the flags */
                                       }
                                       break;
                           case 0x34 :                                                                        /* (Emulation Info) */
                                       BlLength = 8;                                                          /* Total of 8 bytes */
                                       if ((NRead = read (Fh, WorkBuffer, BlLength)) <= 0)  goto __READERROR;        /* Read them */
                                       InsertWorkBuffer = TRUE;
                                       break;
                           case 0x35 : if ((NRead = read (Fh, WorkBuffer, 16)) <= 9) goto __READERROR;     /* (Custom Info Block) */
                                       if (!strncmp (WorkBuffer, "POKEs           ", 16))
                                         Entry->FlagByte = INSRTCI_POKES;
                                       else if (!strncmp (WorkBuffer, "Instructions    ", 16))
                                         Entry->FlagByte = INSRTCI_INSTRUCT;
                                       else if (!strncmp (WorkBuffer, "Spectrum Screen ", 16))
                                         Entry->FlagByte = INSRTCI_SCREEN;
                                       else if (!strncmp (WorkBuffer, "ZX-Edit Document", 16))
                                         Entry->FlagByte = INSRTCI_ZXEDIT;
                                       else if (!strncmp (WorkBuffer, "Picture         ", 16))
                                         Entry->FlagByte = INSRTCI_PICTURE;
                                       else
                                         Entry->FlagByte = INSRTCI_NONE;
                                       if ((NRead = read (Fh, &DBlLength, 4)) <= 0)  goto __READERROR;             /* BlockLength */
                                       Index += 4;
                                       BlLength = 16;
                                       InsertWorkBuffer = TRUE;
                                       InsertLargeBlock = TRUE;
                                       break;
                           case 0x40 :                                                                        /* (Snapshot Block) */
                                       if ((NRead = read (Fh, &DBlLength, 3)) <= 0)  goto __READERROR;             /* BlockLength */
                                       DBlLength &= 0x00FFFFFF;                                                  /* Only 3 bytes! */
                                       Index += 3;
                                       InsertLargeBlock = TRUE;
                                       break;
                           case 0x5A :                                                          /* ('Z' - Merging of 2 TZX files) */
                                       Index += 9;                                       /* Simply skip the header - 9 more bytes */
                                       if (lseek (Fh, Index, SEEK_SET) == -1)  goto __READERROR;
                                       break;
                           default   :                                                        /* (Unknown block for this version) */
                                       if (TapeWindow[WhichWindow].TapeVersion > 0x010B)             /* Tape higher TZX version ? */
                                         if (!TZXTooHighReported)
                                         {
                                           ErrorRequester (-1, -1, NULL, NULL, NULL, "{BNOTICE}B: This tape has TZX version %d.%d\n"
                                                           "TAPER knows TZX up to version 1.12\n"
                                                           "Some blocks are not understood",
                                                           (TapeWindow[WhichWindow].TapeVersion & 0xFF00) >> 8,
                                                           TapeWindow[WhichWindow].TapeVersion & 0x00FF);
                                           TZXTooHighReported = TRUE;
                                         }
                                       if ((NRead = read (Fh, &DBlLength, 4)) <= 0)  goto __READERROR;             /* BlockLength */
                                       Index += 4;
                                       InsertLargeBlock = TRUE;
                                       break;
                         }
                         if (InsertBlock)                                                       /* Still a datablock to be done ? */
                         {                                                                /* Input: Length in (dword) `DBlLength' */
                           if (Entry->Timing.NumBitsInLastByte != 8)                              /* Additional bits at the end ? */
                             DBlLength --;
                           if (DBlLength == 0)                                               /* Flag + block + parity = 0 bytes ? */
                           {
                             Entry->Flags.FlagByteUsed = FALSE;
                             Entry->Flags.ParityByteUsed = FALSE;
                             Entry->BlockLength = Entry->AllocLength = 0;
                           }
                           else
                           {
                             if (PureData)
                             {
                               Entry->Flags.FlagByteUsed = FALSE;
                               Entry->FlagByte = 0;
                             }
                             else
                             {
                               Entry->Flags.FlagByteUsed = TRUE;
                               if (read (Fh, &(Entry->FlagByte), 1) <= 0)
                               { free (Entry); goto __READERROR; }
                             }
                             if (DBlLength == 1 && !PureData)
                             {
                               Entry->Flags.ParityByteUsed = FALSE;
                               Entry->BlockLength = Entry->AllocLength = 0;
                             }
                             else
                             {
                               if (DBlLength > 2 || PureData)
                               {
                                 if (PureData)
                                   Entry->BlockLength = DBlLength;
                                 else
                                   Entry->BlockLength = DBlLength - 2;                                 /* (Flag and parity bytes) */
                                 Entry->AllocLength = (DBlLength & 0xFFFFFFFC) + (DBlLength % 4 ? 4 : 0);
                                 if (TapeWindow[WhichWindow].Tape.XMSHandle == -1)                               /* First entry ? */
                                   XMSAlloc (&(TapeWindow[WhichWindow].Tape.XMSHandle), Entry->AllocLength);
                                 else
                                   XMSRealloc (TapeWindow[WhichWindow].Tape.XMSHandle,
                                               TapeWindow[WhichWindow].Tape.Size + Entry->AllocLength, TRUE);
                                 Entry->CalculatedParity = Entry->FlagByte;
                                 if ((Entry->Flags.IsHeader = (Entry->BlockTypeID == 0x10 && Entry->FlagByte == 0x00 &&
                                                         Entry->BlockLength == 17 ? TRUE : FALSE)))
                                   Entry->Timing.NumPilotPulses = NS_HDPILOT;
                                 XMSIndex = 0;                                                          /* Start at the beginning */
                                 LengthToGo = Entry->BlockLength;
                                 while (LengthToGo > 0)                                                /* Read the data in chunks */
                                 {
                                   ReadLength = (word)(LengthToGo > (dword)49152 ? 49152 : LengthToGo);
                                   if ((RealReadLength = read (Fh, WorkBuffer, ReadLength)) < ReadLength)      /* Premature EOF ? */
                                   { if (errno != 0)                                                          /* A 'real' error ? */
                                     { free (Entry); goto __READERROR; }
                                     if (!QuestionRequester (-1, -1, FALSE, FALSE, "@Drop", "@Keep", NULL, NULL, NULL, NULL,
                                                             "The last block in this tape is incomplete.\n"
                                                             "It is defined as %lu bytes, but only %lu bytes were read.\n"
                                                             "Do you want to drop it or keep it?", Entry->BlockLength,
                                                             Entry->BlockLength - LengthToGo + RealReadLength -
                                                             (PureData ? 0 : 1)))
                                     { free (Entry); goto __CONTINUE; }
                                     DBlLength += RealReadLength - LengthToGo;
                                     if (PureData)
                                       Entry->Flags.ParityByteUsed = FALSE;
                                     else
                                     {
                                       Entry->Flags.ParityByteUsed = TRUE;
                                       Entry->ParityByte = *(WorkBuffer + RealReadLength);
                                       RealReadLength --;
                                     }
                                     Entry->BlockLength += RealReadLength - LengthToGo;
                                     if (PureData && LengthToGo == Entry->BlockLength)
                                     {
                                       Entry->FlagByte = *WorkBuffer;
                                       Entry->CalculatedParity = 0;
                                     }
                                     for (Cnt = 0 ; Cnt < RealReadLength ; Cnt ++)
                                       Entry->CalculatedParity ^= *(WorkBuffer + Cnt);
                                     DLengthToGo = (dword)(RealReadLength & 0xFFFC) + (RealReadLength % 4 ? 4 : 0);
                                     CopyToXMS (DLengthToGo, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer,
                                                TapeWindow[WhichWindow].Tape.Size + XMSIndex);
                                     goto __TZXCONT;
                                   }
                                   if (PureData && LengthToGo == Entry->BlockLength)                   /* First Pure Data chunk ? */
                                   {
                                     Entry->FlagByte = *WorkBuffer;                               /* Copy first byte to flag byte */
                                     Entry->CalculatedParity = 0;                                 /* 'Flag' will be the first one */
                                   }
                                   for (Cnt = 0 ; Cnt < ReadLength ; Cnt ++)                                  /* Calculate parity */
                                     Entry->CalculatedParity ^= *(WorkBuffer + Cnt);
                                   DLengthToGo = (dword)(ReadLength & 0xFFFC) + (ReadLength % 4 ? 4 : 0);
                                   CopyToXMS (DLengthToGo, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer,
                                              TapeWindow[WhichWindow].Tape.Size + XMSIndex);
                                   XMSIndex += ReadLength;
                                   LengthToGo -= ReadLength;
                                   if (PureData && LengthToGo == 0)                                     /* Last Pure Data chunk ? */
                                   {
                                     Entry->ParityByte = *(WorkBuffer + ReadLength - 1);         /* Copy last byte to parity byte */
                                     Entry->CalculatedParity ^= Entry->ParityByte;                   /* Fix the calculated parity */
                                   }
                                 }
                               }
                               else
                               {
                                 Entry->BlockLength = Entry->AllocLength = 0;
                                 Entry->CalculatedParity = Entry->FlagByte;
                               }
                               if (PureData)
                                 Entry->Flags.ParityByteUsed = FALSE;
                               else
                               {
                                 Entry->Flags.ParityByteUsed = TRUE;
                                 if (read (Fh, &(Entry->ParityByte), 1) <= 0)
                                 { free (Entry); goto __READERROR; }
                               }
                             }
                           }
                           if (Entry->Timing.NumBitsInLastByte != 8)                              /* Additional bits at the end ? */
                           {
                             if (read (Fh, &(Entry->Timing.CustomLastByte), 1) <= 0)
                             { free (Entry); goto __READERROR; }
                             DBlLength ++;
                           }
                           __TZXCONT:;
                           Index += DBlLength;                                                               /* Update file index */
                         }
                         if (InsertWorkBuffer)
                         {
                           Entry->BlockLength += BlLength;
                           Entry->AllocLength = (Entry->BlockLength & 0xFFFFFFFC) + (Entry->BlockLength % 4 ? 4 : 0);
                           if (TapeWindow[WhichWindow].Tape.XMSHandle == -1)                                     /* First entry ? */
                             XMSAlloc (&(TapeWindow[WhichWindow].Tape.XMSHandle), Entry->AllocLength + AddedSize);
                           else
                             XMSRealloc (TapeWindow[WhichWindow].Tape.XMSHandle,
                                         TapeWindow[WhichWindow].Tape.Size + Entry->AllocLength + AddedSize, TRUE);
                           CopyToXMS (Entry->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer,
                                      TapeWindow[WhichWindow].Tape.Size);
                           Index += BlLength;                                                                /* Update file index */
                         }
                         if (InsertLargeBlock)
                         {
                           DScratch = Entry->BlockLength;                            /* >0 only if InsertWorkBuffer was also TRUE */
                           Entry->BlockLength += DBlLength;
                           Entry->AllocLength = (Entry->BlockLength & 0xFFFFFFFC) + (Entry->BlockLength % 4 ? 4 : 0);
                           if (TapeWindow[WhichWindow].Tape.XMSHandle == -1)                                     /* First entry ? */
                             XMSAlloc (&(TapeWindow[WhichWindow].Tape.XMSHandle), Entry->AllocLength);
                           else
                             XMSRealloc (TapeWindow[WhichWindow].Tape.XMSHandle,
                                         TapeWindow[WhichWindow].Tape.Size + Entry->AllocLength - DScratch, TRUE);
                           XMSIndex = DScratch;                                                         /* Start at the beginning */
                           LengthToGo = Entry->BlockLength - DScratch;
                           FirstPart = TRUE;
                           while (LengthToGo > 0)                                                      /* Read the data in chunks */
                           {
                             ReadLength = (word)(LengthToGo > (dword)49152 ? 49152 : LengthToGo);
                             if ((RealReadLength = read (Fh, WorkBuffer, ReadLength)) < ReadLength)      /* Premature EOF ? */
                             { if (errno != 0)                                                          /* A 'real' error ? */
                               { free (Entry); goto __READERROR; }
                               if (!QuestionRequester (-1, -1, FALSE, FALSE, "@Drop", "@Keep", NULL, NULL, NULL, NULL,
                                                       "The last block in this tape is incomplete.\n"
                                                       "It is defined as %lu bytes, but only %lu bytes were read.\n"
                                                       "Do you want to drop it or keep it?", Entry->BlockLength,
                                                       Entry->BlockLength - LengthToGo + RealReadLength))
                               { free (Entry); goto __CONTINUE; }
                               DBlLength += RealReadLength - LengthToGo;                                     /* Truncate the file */
                               LengthToGo = ReadLength;
                             }
                             DLengthToGo = (dword)(RealReadLength & 0xFFFC) + (RealReadLength % 4 ? 4 : 0);
                             CopyToXMS (DLengthToGo, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer,
                                        TapeWindow[WhichWindow].Tape.Size + XMSIndex);
                             if (FirstPart && BScratch == 0x28)                                    /* Handling a 'Select Block' ? */
                             {
                               AddedSize = (*WorkBuffer) * 2;                            /* Number of selections is in first byte */
                               XMSRealloc (TapeWindow[WhichWindow].Tape.XMSHandle,            /* Add room for the jump code table */
                                           TapeWindow[WhichWindow].Tape.Size + Entry->AllocLength + AddedSize - DScratch, TRUE);
                               FirstPart = FALSE;
                             }
                             XMSIndex += ReadLength;
                             LengthToGo -= ReadLength;
                           }
                           Index += DBlLength;                                                               /* Update file index */
                         }
                         if (AddedSize > 0L)
                           Entry->AllocLength += AddedSize;
                         TapeWindow[WhichWindow].Tape.Size += Entry->AllocLength;
                         Entry->Next = NULL;
                         if (TapeWindow[WhichWindow].Directory == NULL)
                           TapeWindow[WhichWindow].Directory = Entry;
                         else
                         {
                           DirP = TapeWindow[WhichWindow].Directory;
                           while (DirP->Next != NULL)
                             DirP = DirP->Next;
                           DirP->Next = Entry;
                         }
                         TapeWindow[WhichWindow].NumEntries ++;
                       }
                       break;
      case FT_Z80    :                                                                  /* (All snapshot types are treated alike) */
      case FT_SNA    : if (!ConvertSnapshot (_PRFileLength, WhichWindow, Fh))
                       {
                         close (Fh);
                         _DestroyPReq ();
                         return (FALSE);
                       }
                       goto __CONTINUE;                                               /* The snapshot MUST be dealt with in 1 go! */
    }
    if (Index > _PRFileLength)                                      /* Pointing past the end of the file (truncated last block) ? */
      Index = _PRFileLength;                                                                           /* The correct the pointer */
    if (!UpdatePReq (Index, StatusText))
      PRCancel = TRUE;
    MouseStatus ();
    MoreToGo = (bool)(Index < _PRFileLength ? TRUE : FALSE);
  }
  __CONTINUE:
  CleanupTapeBuffers ();                                                                          /* Release any decoding buffers */
  if (close (Fh) == -1)
    FatalError (10, "Could not close tape: %s", StrError (errno));
  _DestroyPReq ();
  _MustInitializeJumpCodes = TRUE;
  if (!_PreReadingTape)
  {
    SignalTapeChanged (WhichWindow, Merge);
    SignalTapeMode (WhichWindow, TapeWindow[WhichWindow].ModeIs128K);
  }
  if (TapeWindow[WhichWindow].FileType == FT_TZX && TapeWindow[WhichWindow].TapeVersion == 0x0A01)   /* Version number reversed ? */
  {                                             /* An older release of MakeTZX used to do this (create version 10.01 i.s.o. 1.10) */
    ErrorRequester (-1, -1, NULL, NULL, NULL, "The TZX version number is invalid (reversed) and has been adjusted.\n"
                    "Please save the file to apply the correction.");
    TapeWindow[WhichWindow].TapeVersion = 0x010A;
    SignalTapeChanged (WhichWindow, TRUE);                                                       /* Next write will correct this! */
  }
  TapeWindow[WhichWindow].NoTape = FALSE;
  TapeWindow[WhichWindow].NewTape = FALSE;
  return (TRUE);
}

void WriteTape (short WhichWindow, bool FatalSave)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `FatalSave' is TRUE only for FatalError handling.                           */
/* Post  : The tape has been written to disk from the intermediate file format to the requested file format.                      */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *DirP;
  char                NameIDBuffer[65];
  struct Timing_s     LastTiming;
  word                LastPauseLength = 1000;
  bool                MaySkipBlocks = FALSE;                          /* TRUE if the user accepted missing blocks for this format */
  int                 Fh;
  bool                Dummy = TRUE;
  bool                Dummy2 = TRUE;
  int     register    Cnt;
  int     register    Cnt2;
  dword               XMSIndex;
  dword               XMSSubIndex;
  word                WriteLength;
  dword               LengthToGo;
  dword               DLengthToGo;
  byte                BScratch;                                                           /* Simple byte/word/dword write buffers */
  word                WScratch;
  dword               DScratch;
  bool                WritingTapeID = TRUE;
  bool                InsertBlock = FALSE;
  bool                InsertLargeBlock = FALSE;
  bool                InsertWorkBuffer = FALSE;
  word                BlLength;
  dword               DBlLength;
  word                WriteSampleRate;
  comboent            BlockNo = 0;
  comboent            OldBlockNo;
  bool                WriteAll = FALSE;                                                      /* TAP write turbo blocks loose data */
  bool                WriteThis = FALSE;
  bool                OldSnapButton;
  bool                OldSaved = FALSE;
  bool                Cancelled = FALSE;
  bool                TZXLowVer = 10;                                           /* Use v1.10 unless new BlockTypeID 0x2A is found */

  SimulatePlayBlock = FALSE;                                                                    /* Signal: this is the real thing */
  CountPlayTimeLength = FALSE;
  UsePlayTimeClock = FALSE;
  ClearTapeLoopDetection (WhichWindow);
  if (TapeWindow[WhichWindow].FileType == FT_UNKNOWN ||                                                     /* Final sanity check */
      TapeWindow[WhichWindow].FileType == FT_BADSOUND)
    FatalError (6, "%s has unknown Spectrum tape file type", TapeWindow[WhichWindow].TapeName);
  if (MaxFloppyIdChanged)
    WriteConfig ();
  if (TapeWindow[WhichWindow].NewTape)
    if (!ForceNewTapeName (WhichWindow))
      return;
  if (FatalSave)
    TapeWindow[WhichWindow].FileType = FT_TZX;                                                     /* We want to save everything! */
  if (TapeWindow[WhichWindow].TapeName[1] == ':' && IsFloppy (TapeWindow[WhichWindow].TapeName[0]))
    if (!RequestFloppy (TapeWindow[WhichWindow].TapeName[0], TapeWindow[WhichWindow].TapeFloppyId))
      return;
  if (TapeWindow[WhichWindow].FileType != FT_TZX)                            /* Check if all blocks can be written in this format */
  {
    DirP = TapeWindow[WhichWindow].Directory;
    while (Dummy && Dummy2 && DirP)
    {
      if (DirP->IsIdentifiedData &&
          (TapeWindow[WhichWindow].FileType == FT_TAPZ80 && DirP->BlockLength > 65534L) ||
          (TapeWindow[WhichWindow].FileType == FT_LTP && DirP->BlockLength > 65536L))
      {
        Dummy2 = FALSE;
        if (QuestionRequester (-1, -1, TRUE, FALSE, NULL, NULL, NULL, NULL, NULL, NULL, "One or more data blocks in the tape is "
                               "larger than %u bytes, which this format cannot handle.\nThese will not be saved.\n"
                               "Do you want to use this format anyway?",
                               TapeWindow[WhichWindow].FileType == FT_TAPZ80 ? (unsigned)65534 : (unsigned)65536))
          return;
        else
          MaySkipBlocks = TRUE;
      }
      if (((TapeWindow[WhichWindow].FileType == FT_LTP || TapeWindow[WhichWindow].FileType == FT_TAPZ80) &&
           (!DirP->Flags.IsNormalSpeed || !DirP->IsIdentifiedData)) ||
          (TapeWindow[WhichWindow].FileType == FT_VOC && !DirP->IsTapeData))
      {
        Dummy = FALSE;                                                                           /* Unknown block for this format */
        if (QuestionRequester (-1, -1, TRUE, FALSE, NULL, NULL, NULL, NULL, NULL, NULL, "The selected tape format cannot hold all "
                               "blocks in this tape.\nDo you want to use this format anyway?"))
          return;
        else
          MaySkipBlocks = TRUE;
      }
      DirP = DirP->Next;
    }
  }
  else                                                                                          /* TZX - Determine needed version */
  {
    DirP = TapeWindow[WhichWindow].Directory;
    while (DirP && Dummy)
    {
      if (DirP->BlockTypeID == 0x35 &&                                                     /* One of the new Custom Info blocks ? */
          (DirP->FlagByte == INSRTCI_ZXEDIT || DirP->FlagByte == INSRTCI_PICTURE))
      {
        Dummy = FALSE;                                                                           /* Unknown block for this format */
        TZXLowVer = 12;                                                                          /* Then we need to save as v1.12 */
      }
      else if (DirP->BlockTypeID == 0x32)                                                 /* One of the new Archive Info fields ? */
      {
        XMSIndex = GetTapeBlockXMSIndex (WhichWindow, DirP);
        CopyFromXMS (DirP->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);
        for (Cnt = 0, Cnt2 = 1 ; Dummy && Cnt < *WorkBuffer ; Cnt ++)                                   /* Check all text strings */
        {
          if (*(WorkBuffer + Cnt2) >= 0x05 && *(WorkBuffer + Cnt2) <= 0x08)
          {
            Dummy = FALSE;
            TZXLowVer = 12;
          }
          Cnt2 += *WorkBuffer + 1;
        }
      }
      else if (DirP->BlockTypeID == 0x2A)                                                                 /* The new block type ? */
      {
        if (TZXLowVer < 11)                                                                      /* Unknown block for this format */
          TZXLowVer = 11;                                                                        /* Then we need to save as v1.11 */
      }                                                                   /* (But continue checking, as we might even need v1.12) */
      DirP = DirP->Next;
    }
  }
  if (!TapeWindow[WhichWindow].NonexistingTape && !FatalSave && TapeWindow[WhichWindow].BackupTape)
  {
    TapeWindow[WhichWindow].BackupTape = FALSE;
    Dummy = TRUE;
    if (!CopyFile (TapeWindow[WhichWindow].TapeName, TapeWindow[WhichWindow].BackupName, &TapeWindow[WhichWindow].BackupTape,
                   &Dummy, "Making a tape backup..."))
    {
      unlink (TapeWindow[WhichWindow].BackupName);
      return;
    }
  }
  if (!FatalSave && (TapeWindow[WhichWindow].FileType == FT_VOC ||                                 /* About to write a VOC file ? */
                     TapeWindow[WhichWindow].FileType == FT_CSW))                                  /* About to write a CSW file ? */
    if (!SetupVOCWriteTimers (-1, &WriteSampleRate))                                 /* Allow the users to select the sample rate */
      return;
    else
    {
      CountPlayTimeLength = FALSE;                                                                     /* Signal: the real thing! */
      UsePlayTimeClock = FALSE;
    }
  if (!FatalSave)
    MakePRequester (&_PReq, -1, -1, TRUE, NULL, "Writing Tape", "Writing %s tape ...",
                    WhichWindow == TWLEFT ? "source" : "destination");
  _UsingFatalSave = FatalSave;
  if ((Fh = open (FatalSave ? (WhichWindow == TWLEFT ? TAPELSAVE : TAPERSAVE) : TapeWindow[WhichWindow].TempName,
                  O_BINARY | O_RDWR | O_CREAT | O_TRUNC, 0777)) == -1)
    FatalError (10, "Cannot create new tape file: [%s] %d %s", TapeWindow[WhichWindow].TempName, errno, StrError (errno));
  LastTiming.LengthPilotPulse = NS_LENPILOT;                                                     /* Prepare for normal speed data */
  LastTiming.LengthSyncPulse[0] = NS_LENSYNC1;
  LastTiming.LengthSyncPulse[1] = NS_LENSYNC2;
  LastTiming.LengthZeroBitPulse = NS_LEN0BIT;
  LastTiming.LengthOneBitPulse = NS_LEN1BIT;
  LastTiming.NumPilotPulses = NS_HDPILOT;                                                                      /* (Assume header) */
  LastTiming.NumSyncPulses = 2;
  LastPauseLength = 1000;
  DirP = TapeWindow[WhichWindow].Directory;
  XMSIndex = 0;
  while (!Cancelled && DirP != NULL && !_InfiniteTapeLoop)                                                       /* Do all blocks */
  {
    OldBlockNo = BlockNo;
    switch (TapeWindow[WhichWindow].FileType)
    {
      case FT_TAPZ80 :
      case FT_LTP    : if (!DirP->IsIdentifiedData ||
                           (TapeWindow[WhichWindow].FileType == FT_TAPZ80 && DirP->BlockLength > 65534L) ||
                           (TapeWindow[WhichWindow].FileType == FT_LTP && DirP->BlockLength > 65536L))
                         WriteThis = FALSE;                                                     /* Cannot handle non-tape entries */
                       else if (!DirP->Flags.IsNormalSpeed)
                       {
                         if (!WriteAll)
                           switch (QuestionRequester (-1, -1, TRUE, TRUE, NULL, NULL, "N@one", NULL, NULL, "Confirm Loose Timing",
                                   "Entry %u has custom timing, which will be lost when saving.\n"
                                   "Do you want to save this entry?", BlockNo))
                           {
                             case 0 : WriteThis = TRUE;                                                                  /* (Yes) */
                                      break;
                             case 1 : WriteThis = FALSE;                                                                  /* (No) */
                                      break;
                             case 2 : WriteThis = FALSE;                                                              /* (Cancel) */
                                      Cancelled = TRUE;
                                      break;
                             case 3 : WriteThis = TRUE;                                                                  /* (All) */
                                      WriteAll = TRUE;
                           }
                           else
                             WriteThis = TRUE;
                         if (!OldSaved)                             /* Keep the mouse where it is, it's much too anoying if it is */
                         {                                        /* moved to the default button for every confirmation requester */
                           OldSnapButton = Mouse.SnapToDefaultButton;
                           Mouse.SnapToDefaultButton = FALSE;
                           OldSaved = TRUE;
                         }
                       }
                       else
                         WriteThis = TRUE;                                                                /* (Normal Speed block) */
                       if (WriteThis)
                       {                                                                           /* Save only TAP kind of block */
                         WriteLength = (word)DirP->BlockLength;
                         if (TapeWindow[WhichWindow].FileType == FT_TAPZ80)
                         {
                           if (DirP->Flags.FlagByteUsed)
                             WriteLength ++;
                           if (DirP->Flags.ParityByteUsed)
                             WriteLength ++;
                         }
                         if (write (Fh, &WriteLength, 2) == -1)                                /* Write length of following block */
                         { __WRITEERROR: close (Fh);
                                         _DestroyPReq ();
                                         ErrorRequester (-1, -1, "@Close", NULL, NULL, _FileWriteError,
                                         TapeWindow[WhichWindow].TapeName, StrError (errno));
                                         return;
                         }
                         if (DirP->Flags.FlagByteUsed)                                               /* First comes the flag byte */
                           if (write (Fh, &(DirP->FlagByte), 1) == -1)  goto __WRITEERROR;
                         if (DirP->BlockLength > 0)
                         {
                           XMSSubIndex = 0;                                                             /* Start at the beginning */
                           LengthToGo = DirP->BlockLength;
                           while (LengthToGo > 0)                                                     /* Write the data in chunks */
                           {
                             WriteLength = (word)(LengthToGo > (dword)49152 ? 49152 : LengthToGo);
                             DLengthToGo = (dword)(WriteLength & 0xFFFC) + (WriteLength % 4 ? 4 : 0);
                             CopyFromXMS (DLengthToGo, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex + XMSSubIndex, WorkBuffer);
                             if (write (Fh, WorkBuffer, WriteLength) < WriteLength)  goto __WRITEERROR;
                             XMSSubIndex += WriteLength;
                             LengthToGo -= WriteLength;
                           }
                         }
                         if (DirP->Flags.ParityByteUsed)                                                  /* Then the parity byte */
                           if (write (Fh, &(DirP->ParityByte), 1) == -1)  goto __WRITEERROR;
                       }
                       break;
      case FT_VOC    : if (DirP->IsTapeData)
                       {
                         if (WritingTapeID)
                         {
                           if (write (Fh, "Creative Voice File\x1A\x1A\x00\x0A\x01\x29\x11", 26) != 26)  goto __WRITEERROR;
                           WritingTapeID = FALSE;                                                             /* Write VOC header */
                           LengthToGo = 26;                                                     /* Use `LengthToGo' as file index */
                         }
                         if (!WriteVOCBlock (WhichWindow, Fh, DirP, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, &LengthToGo,
                                             &BlockNo))
                           goto __WRITEERROR;
                       }
                       else
                         BlockNo ++;
                       break;
      case FT_ACORNRAW : if (DirP->IsTapeData)
                       {
                         if (WritingTapeID)
                         {
                           for (Cnt = 0 ; Cnt < 48 ; Cnt ++)
                             NameIDBuffer[Cnt] = '\0';                                                /* First 48 bytes are zeros */
                           strcpy (NameIDBuffer + 48, "Raw tape sample");
                           if (write (Fh, NameIDBuffer, 64) != 64)  goto __WRITEERROR;
                           WritingTapeID = FALSE;                                                             /* Write VOC header */
                           LengthToGo = 64;                                                     /* Use `LengthToGo' as file index */
                         }
                         if (!WriteVOCBlock (WhichWindow, Fh, DirP, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, &LengthToGo,
                                             &BlockNo))
                           goto __WRITEERROR;
                       }
                       else
                         BlockNo ++;
                       break;
      case FT_CSW    : if (DirP->IsTapeData)
                       {
                         if (WritingTapeID)
                         {
                           strcpy (NameIDBuffer, "Compressed Square Wave\x1A");
                           NameIDBuffer[23] = 0x01;                                                              /* (Version 1.0) */
                           NameIDBuffer[24] = 0x00;
                           NameIDBuffer[25] = (byte)(WriteSampleRate & 0x00FF);                                   /* (SampleRate) */
                           NameIDBuffer[26] = (byte)(WriteSampleRate >> 8);
                           NameIDBuffer[27] = 0x01;                                                          /* (RLE compression) */
                           NameIDBuffer[28] = 0x00;                                                                 /* (Reserved) */
                           NameIDBuffer[29] = 0x00;
                           NameIDBuffer[30] = 0x00;
                           NameIDBuffer[31] = 0x00;
                           if (write (Fh, NameIDBuffer, 32) != 32)  goto __WRITEERROR;
                           WritingTapeID = FALSE;                                                             /* Write VOC header */
                           LengthToGo = 32;                                                     /* Use `LengthToGo' as file index */
                         }
                         if (!WriteVOCBlock (WhichWindow, Fh, DirP, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, &LengthToGo,
                                             &BlockNo))
                           goto __WRITEERROR;
                       }
                       else
                         BlockNo ++;
                       break;
      case FT_TZX    : if (WritingTapeID)
                       {
                         memcpy (NameIDBuffer, "ZXTape!\x1A\x01", 9);
                         NameIDBuffer[9] = TZXLowVer;                                                         /* Set file version */
                         if (write (Fh, NameIDBuffer, 10) != 10)  goto __WRITEERROR;
                         WritingTapeID = FALSE;                                                               /* Write TZX header */
                       }
                       InsertBlock = InsertLargeBlock = InsertWorkBuffer = FALSE;
                       if (write (Fh, &(DirP->BlockTypeID), 1) == -1)  goto __WRITEERROR;
                       switch (DirP->BlockTypeID)
                       {
                         case 0x10 : BlLength = 0;
                                     if (DirP->Next == NULL && TPConfig.TZXStripPauseLengths)         /* Last block in the tape ? */
                                     { if (write (Fh, &BlLength, 2) == -1)  goto __WRITEERROR; }       /* Write 'PauseLength = 0' */
                                     else if (DirP->Next->BlockTypeID == 0x20 &&               /* Next block is 'Stop the tape' ? */
                                              DirP->Next->PauseLength == 0 && TPConfig.TZXStripPauseLengths)
                                     { if (write (Fh, &BlLength, 2) == -1)  goto __WRITEERROR; }       /* Write 'PauseLength = 0' */
                                     else if (DirP->Next->BlockTypeID == 0x22 &&                     /* Next block is 'Group End' */
                                              DirP->Next->Next == NULL && TPConfig.TZXStripPauseLengths)   /* which is the last ? */
                                     { if (write (Fh, &BlLength, 2) == -1)  goto __WRITEERROR; }       /* Write 'PauseLength = 0' */
                                     else
                                       if (write (Fh, &(DirP->PauseLength), 2) == -1)  goto __WRITEERROR;
                                     DBlLength = DirP->BlockLength;
                                     //if (DirP->Timing.NumBitsInFirstByte != 8)  DBlLength ++;
                                     if (DirP->Flags.FlagByteUsed)              DBlLength ++;
                                     if (DirP->Flags.ParityByteUsed)            DBlLength ++;
                                     //if (DirP->Timing.NumBitsInLastByte != 8)   DBlLength ++;  Not supported by format!
                                     if (write (Fh, &DBlLength, 2) == -1)  goto __WRITEERROR;
                                     InsertBlock = TRUE;
                                     break;
                         case 0x11 : if (write (Fh, &(DirP->Timing.LengthPilotPulse), 2) == -1)  goto __WRITEERROR;
                                     if (write (Fh, &(DirP->Timing.LengthSyncPulse[0]), 2) == -1)  goto __WRITEERROR;
                                     if (write (Fh, &(DirP->Timing.LengthSyncPulse[1]), 2) == -1)  goto __WRITEERROR;
                                     if (write (Fh, &(DirP->Timing.LengthZeroBitPulse), 2) == -1)  goto __WRITEERROR;
                                     if (write (Fh, &(DirP->Timing.LengthOneBitPulse), 2) == -1)  goto __WRITEERROR;
                                     if (write (Fh, &(DirP->Timing.NumPilotPulses), 2) == -1)  goto __WRITEERROR;
                                     if (write (Fh, &(DirP->Timing.NumBitsInLastByte), 1) == -1)  goto __WRITEERROR;
                                     BlLength = 0;
                                     if (DirP->Next == NULL && TPConfig.TZXStripPauseLengths)         /* Last block in the tape ? */
                                     { if (write (Fh, &BlLength, 2) == -1)  goto __WRITEERROR; }       /* Write 'PauseLength = 0' */
                                     else if (DirP->Next->BlockTypeID == 0x20 &&               /* Next block is 'Stop the tape' ? */
                                              DirP->Next->PauseLength == 0 && TPConfig.TZXStripPauseLengths)
                                     { if (write (Fh, &BlLength, 2) == -1)  goto __WRITEERROR; }       /* Write 'PauseLength = 0' */
                                     else if (DirP->Next->BlockTypeID == 0x22 &&                     /* Next block is 'Group End' */
                                              DirP->Next->Next == NULL && TPConfig.TZXStripPauseLengths)   /* which is the last ? */
                                     { if (write (Fh, &BlLength, 2) == -1)  goto __WRITEERROR; }       /* Write 'PauseLength = 0' */
                                     else
                                       if (write (Fh, &(DirP->PauseLength), 2) == -1)  goto __WRITEERROR;
                                     DBlLength = DirP->BlockLength;
                                     if (DirP->Timing.NumBitsInFirstByte != 8)  DBlLength ++;
                                     if (DirP->Flags.FlagByteUsed)              DBlLength ++;
                                     if (DirP->Flags.ParityByteUsed)            DBlLength ++;
                                     if (DirP->Timing.NumBitsInLastByte != 8)   DBlLength ++;
                                     if (write (Fh, &DBlLength, 3) == -1)  goto __WRITEERROR;
                                     InsertBlock = TRUE;
                                     break;
                         case 0x12 : if (write (Fh, &(DirP->Timing.LengthPilotPulse), 2) == -1)  goto __WRITEERROR;
                                     if (write (Fh, &(DirP->Timing.NumPilotPulses), 2) == -1)  goto __WRITEERROR;
                                     break;
                         case 0x13 : BScratch = DirP->BlockLength / 2;
                                     if (write (Fh, &BScratch, 1) == -1)  goto __WRITEERROR;
                                     InsertWorkBuffer = TRUE;
                                     break;
                         case 0x14 : if (write (Fh, &(DirP->Timing.LengthZeroBitPulse), 2) == -1)  goto __WRITEERROR;
                                     if (write (Fh, &(DirP->Timing.LengthOneBitPulse), 2) == -1)  goto __WRITEERROR;
                                     if (write (Fh, &(DirP->Timing.NumBitsInLastByte), 1) == -1)  goto __WRITEERROR;
                                     BlLength = 0;
                                     if (DirP->Next == NULL && TPConfig.TZXStripPauseLengths)         /* Last block in the tape ? */
                                     { if (write (Fh, &BlLength, 2) == -1)  goto __WRITEERROR; }       /* Write 'PauseLength = 0' */
                                     else if (DirP->Next->BlockTypeID == 0x20 &&               /* Next block is 'Stop the tape' ? */
                                              DirP->Next->PauseLength == 0 && TPConfig.TZXStripPauseLengths)
                                     { if (write (Fh, &BlLength, 2) == -1)  goto __WRITEERROR; }       /* Write 'PauseLength = 0' */
                                     else if (DirP->Next->BlockTypeID == 0x22 &&                     /* Next block is 'Group End' */
                                              DirP->Next->Next == NULL && TPConfig.TZXStripPauseLengths)   /* which is the last ? */
                                     { if (write (Fh, &BlLength, 2) == -1)  goto __WRITEERROR; }       /* Write 'PauseLength = 0' */
                                     else
                                       if (write (Fh, &(DirP->PauseLength), 2) == -1)  goto __WRITEERROR;
                                     DBlLength = DirP->BlockLength;
                                     if (DirP->Timing.NumBitsInFirstByte != 8)  DBlLength ++;
                                     if (DirP->Timing.NumBitsInLastByte != 8)   DBlLength ++;
                                     if (write (Fh, &DBlLength, 3) == -1)  goto __WRITEERROR;
                                     InsertBlock = TRUE;
                                     break;
                         case 0x15 : if (write (Fh, &(DirP->Timing.SampleRate), 2) == -1)  goto __WRITEERROR;
                                     BlLength = 0;
                                     if (DirP->Next == NULL && TPConfig.TZXStripPauseLengths)         /* Last block in the tape ? */
                                     { if (write (Fh, &BlLength, 2) == -1)  goto __WRITEERROR; }       /* Write 'PauseLength = 0' */
                                     else if (DirP->Next->BlockTypeID == 0x20 &&               /* Next block is 'Stop the tape' ? */
                                              DirP->Next->PauseLength == 0 && TPConfig.TZXStripPauseLengths)
                                     { if (write (Fh, &BlLength, 2) == -1)  goto __WRITEERROR; }       /* Write 'PauseLength = 0' */
                                     else
                                       if (write (Fh, &(DirP->PauseLength), 2) == -1)  goto __WRITEERROR;
                                     if (write (Fh, &(DirP->Timing.NumBitsInLastByte), 1) == -1)  goto __WRITEERROR;
                                     DBlLength = DirP->BlockLength;
                                     if (write (Fh, &DBlLength, 3) == -1)  goto __WRITEERROR;
                                     InsertLargeBlock = TRUE;
                                     break;
                         case 0x20 : if (write (Fh, &(DirP->PauseLength), 2) == -1)  goto __WRITEERROR;
                                     break;
                         case 0x21 : BScratch = (byte)DirP->BlockLength;
                                     if (write (Fh, &BScratch, 1) == -1)  goto __WRITEERROR;
                                     InsertWorkBuffer = TRUE;
                                     break;
                         case 0x22 : break;
                         case 0x23 : if (write (Fh, &(DirP->PauseLength), 2) == -1)  goto __WRITEERROR;
                                     break;
                         case 0x24 : if (write (Fh, &(DirP->PauseLength), 2) == -1)  goto __WRITEERROR;
                                     break;
                         case 0x25 : break;
                         case 0x26 : WScratch = DirP->BlockLength / 2;
                                     if (write (Fh, &WScratch, 2) == -1)  goto __WRITEERROR;
                                     InsertWorkBuffer = TRUE;
                                     break;
                         case 0x27 : break;
                         case 0x28 : WScratch = (word)DirP->BlockLength;
                                     if (write (Fh, &WScratch, 2) == -1)  goto __WRITEERROR;
                                     InsertLargeBlock = TRUE;
                                     break;
                         case 0x2A : DScratch = 0;
                                     if (write (Fh, &DScratch, 4) == -1)  goto __WRITEERROR;
                                     break;
                         case 0x30 : BScratch = (byte)DirP->BlockLength;
                                     if (write (Fh, &BScratch, 1) == -1)  goto __WRITEERROR;
                                     InsertWorkBuffer = TRUE;
                                     break;
                         case 0x31 : BScratch = (byte)DirP->BlockLength;
                                     if (write (Fh, &(DirP->PauseLength), 1) == -1)  goto __WRITEERROR;
                                     if (write (Fh, &BScratch, 1) == -1)  goto __WRITEERROR;
                                     InsertWorkBuffer = TRUE;
                                     break;
                         case 0x32 : BlLength = (word)DirP->BlockLength;
                                     if (write (Fh, &BlLength, 2) == -1)  goto __WRITEERROR;
                                     InsertWorkBuffer = TRUE;
                                     break;
                         case 0x33 : BScratch = (byte)(DirP->BlockLength / 3);
                                     if (write (Fh, &BScratch, 1) == -1)  goto __WRITEERROR;
                                     InsertWorkBuffer = TRUE;
                                     break;
                         case 0x34 : InsertWorkBuffer = TRUE;
                                     break;
                         case 0x35 : BScratch = 4;
                                     BlLength = 16;
                                     InsertWorkBuffer = TRUE;
                                     InsertLargeBlock = TRUE;
                                     break;
                         case 0x40 : if (write (Fh, &(DirP->BlockLength), 3) == -1)  goto __WRITEERROR;
                                     InsertLargeBlock = TRUE;
                                     break;
                         default   : if (write (Fh, &(DirP->BlockLength), 4) == -1)  goto __WRITEERROR;
                                     InsertLargeBlock = TRUE;
                                     break;
                       }
                       if (InsertBlock)
                       {
                         if (DirP->Timing.NumBitsInFirstByte != 8)
                           if (write (Fh, &(DirP->Timing.CustomFirstByte), 1) != 1)  goto __WRITEERROR;
                         if (DirP->Flags.FlagByteUsed)
                           if (write (Fh, &(DirP->FlagByte), 1) != 1)  goto __WRITEERROR;
                         if (DirP->BlockLength > 0)
                         {
                           XMSSubIndex = 0;                                                             /* Start at the beginning */
                           LengthToGo = DirP->BlockLength;
                           while (LengthToGo > 0)                                                     /* Write the data in chunks */
                           {
                             WriteLength = (word)(LengthToGo > (dword)49152 ? 49152 : LengthToGo);
                             DLengthToGo = (dword)(WriteLength & 0xFFFC) + (WriteLength % 4 ? 4 : 0);
                             CopyFromXMS (DLengthToGo, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex + XMSSubIndex, WorkBuffer);
                             if (write (Fh, WorkBuffer, WriteLength) < WriteLength)  goto __WRITEERROR;
                             XMSSubIndex += WriteLength;
                             LengthToGo -= WriteLength;
                           }
                         }
                         if (DirP->Flags.ParityByteUsed)
                           if (write (Fh, &(DirP->ParityByte), 1) != 1)  goto __WRITEERROR;
                         if (DirP->BlockTypeID != 0x10 && DirP->Timing.NumBitsInLastByte != 8)      /* Not supported for ID 0x10! */
                           if (write (Fh, &(DirP->Timing.CustomLastByte), 1) != 1)  goto __WRITEERROR;
                       }
                       if (InsertWorkBuffer)
                       {
                         if (InsertLargeBlock)                                    /* Both InsertWorkBuffer AND InsertLargeBlock ? */
                           WriteLength = BlLength;                                       /* Then the length is passed in BlLength */
                         else
                           WriteLength = (word)DirP->BlockLength;                             /* Otherwise write the entire block */
                         if (WriteLength > 0)
                         {
                           DLengthToGo = (WriteLength & 0xFFFFFFFC) + (WriteLength % 4 ? 4 : 0);
                           CopyFromXMS (DLengthToGo, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);
                           if (write (Fh, WorkBuffer, WriteLength) < WriteLength)  goto __WRITEERROR;
                         }
                       }
                       if (InsertLargeBlock)
                       {
                         if (InsertWorkBuffer)                                    /* Both InsertWorkBuffer AND InsertLargeBlock ? */
                         {
                           XMSSubIndex = (BlLength & 0xFFFC) + (BlLength % 4 ? 4 : 0);         /* BlLength was done as WorkBuffer */
                           LengthToGo = DirP->BlockLength - XMSSubIndex;                                /* So this is what's left */
                           if (write (Fh, &LengthToGo, BScratch) < BScratch)  goto __WRITEERROR;            /* Write length first */
                         }
                         else
                         {
                           XMSSubIndex = 0;                                                             /* Start at the beginning */
                           LengthToGo = DirP->BlockLength;
                         }
                         while (LengthToGo > 0)                                                       /* Write the data in chunks */
                         {
                           WriteLength = (word)(LengthToGo > (dword)49152 ? 49152 : LengthToGo);
                           DLengthToGo = (dword)(WriteLength & 0xFFFC) + (WriteLength % 4 ? 4 : 0);
                           CopyFromXMS (DLengthToGo, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex + XMSSubIndex, WorkBuffer);
                           if (write (Fh, WorkBuffer, WriteLength) < WriteLength)  goto __WRITEERROR;
                           XMSSubIndex += WriteLength;
                           LengthToGo -= WriteLength;
                         }
                       }
                       break;
    }
    if (TapeWindow[WhichWindow].FileType == FT_VOC ||
        TapeWindow[WhichWindow].FileType == FT_ACORNRAW ||
        TapeWindow[WhichWindow].FileType == FT_CSW)
    {
      if (BlockNo == OldBlockNo + 1)                                                                      /* Just do next block ? */
      {
        XMSIndex += DirP->AllocLength;
        DirP = DirP->Next;
      }
      else                                                                                                /* Nope, find the block */
      {
        OldBlockNo = BlockNo;
        DirP = TapeWindow[WhichWindow].Directory;
        XMSIndex = 0;
        while (OldBlockNo > 0 && DirP)
        {
          XMSIndex += DirP->AllocLength;
          DirP = DirP->Next;
          OldBlockNo --;
        }
      }
    }
    else
    {
      XMSIndex += DirP->AllocLength;
      BlockNo ++;
      DirP = DirP->Next;
    }
    if (!FatalSave && TapeWindow[WhichWindow].Tape.Size > 0)
      if (!UpdatePRequester (&_PReq, (double)XMSIndex * 100 / TapeWindow[WhichWindow].Tape.Size, NULL))            /* Cancelled ? */
      { close (Fh); unlink (TapeWindow[WhichWindow].TempName); _DestroyPReq (); return; }
    MouseStatus ();
  }
  if (TapeWindow[WhichWindow].FileType == FT_VOC)                                                         /* Writing a VOC file ? */
    if (write (Fh, "\x00\x00\x00\x00", 4) != 4)  goto __WRITEERROR;                                        /* Write the endmarker */
  if (close (Fh) == -1)
    FatalError (10, "Could not close tape: %s", StrError (errno));
  CleanupTapeBuffers ();                                                                          /* Release any encoding buffers */
  if (!FatalSave)
  {
    if (unlink (TapeWindow[WhichWindow].TapeName) == -1 && errno != ENOENT)
    {
      ErrorRequester (-1, -1, "@Close", NULL, NULL, "Cannot update file: %s%s", StrError (errno),
                      errno == 6 ? "\nSharing violation?" : "");
      unlink (TapeWindow[WhichWindow].TempName);
    }
    else
    {
      if (rename (TapeWindow[WhichWindow].TempName, TapeWindow[WhichWindow].TapeName) == -1)
        FatalError (10, "Could not install new tape: %s", StrError (errno));
      SignalTapeChanged (WhichWindow, FALSE);
    }
    _DestroyPReq ();
  }
  TapeWindow[WhichWindow].NonexistingTape = FALSE;                                                   /* Signal: tape exists (now) */
  TapeWindow[WhichWindow].NewTape = FALSE;
  UnghostMenuBackup (WhichWindow);
  if (OldSaved)
    Mouse.SnapToDefaultButton = OldSnapButton;
}

void ReleaseTape (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/* Post  : The tape in that window has been destroyed.                                                                            */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *DirP;
  struct TapeEntry_s *FreeP;

  if (TapeWindow[WhichWindow].Tape.XMSHandle != -1)                                                 /* Cleanup any remaining tape */
    XMSFree (TapeWindow[WhichWindow].Tape.XMSHandle);
  DirP = TapeWindow[WhichWindow].Directory;
  while (DirP != NULL)
  {
    FreeP = DirP;
    DirP = DirP->Next;
    free (FreeP);
  }
  TapeWindow[WhichWindow].Directory = NULL;
  TapeWindow[WhichWindow].NumEntries = -1;
  TapeWindow[WhichWindow].EmptyTape = TRUE;
  TapeWindow[WhichWindow].Tape.XMSHandle = -1;
  TapeWindow[WhichWindow].Tape.Size = 0;
  TapeWindow[WhichWindow].NonexistingTape = TRUE;
  TapeWindow[WhichWindow].TapeChanged = FALSE;
  TapeWindow[WhichWindow].TapeVersion = 0x0000;                                                                     /* (Not used) */
  TapeWindow[WhichWindow].NextJumpCode = 0x0000;
  TapeWindow[WhichWindow].ModeIs128K = FALSE;
  ReInitComboWindow (&(TapeWindow[WhichWindow].IndexWindow), 0);
  CleanupDisassemblerLabels ();                                                              /* Remove any assembler symbol table */
}

dword GetTapeBlockXMSIndex (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the requested entry.                                      */
/* Post  : The XMS starting index of this entry is returned.                                                                      */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *Directory;
  dword               XMSIndex = 0;

  Directory = TapeWindow[WhichWindow].Directory;
  while (Directory != NULL && Directory != Entry)
  {
    XMSIndex += Directory->AllocLength;
    Directory = Directory->Next;
  }
  return (XMSIndex);
}

word GetTapeBlock (short WhichWindow, dword *NextByte, bool *More)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/*         `*NextByte' should be 0 when first called. If `*More' is TRUE, each next call adjusts `*NextByte' automatically.       */
/*         Passing these values again until `*More' is FALSE, the entire block is done in 48K chunks.                             */
/* Post  : The currently selected tapeblock's data has been copied to the WorkBuffer.                                             */
/*         If the data length is > 49152, `More' is set TRUE and only max 49152 bytes are done at a time.                         */
/*         `NextByte' is adjusted so that you know where the next part starts on consequetive calls.                              */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct ComboEntry   Entry;
  struct TapeEntry_s *DirP;
  dword               XMSIndex = 0;
  dword               BlSize;

  *More = FALSE;
  GetComboEntry (&(TapeWindow[WhichWindow].IndexWindow), &Entry, TapeWindow[WhichWindow].SelectedLine);
  DirP = (struct TapeEntry_s *)Entry.Identifier;
  if (DirP)
  {
    if (DirP->BlockLength == 0)
    {
      *More = FALSE;
      return (0);
    }
    XMSIndex = GetTapeBlockXMSIndex (WhichWindow, DirP);
    if (DirP->AllocLength > (dword)49152)
    {
      BlSize = DirP->AllocLength - *NextByte;
      if (BlSize > (dword)49152)
      {
        CopyFromXMS ((dword)49152, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex + *NextByte, WorkBuffer);
        *NextByte += (dword)49152;
        *More = TRUE;
        return (49152);
      }
      else
      {
        CopyFromXMS (BlSize, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex + *NextByte, WorkBuffer);
        return (BlSize);
      }
    }
    else
    {
      CopyFromXMS (DirP->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);
      return (DirP->BlockLength);
    }
  }
  return (0);
}

void DeleteTapeBlock (short WhichWindow, struct TapeEntry_s *Entry, bool RebuildTape, bool Changed)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the directory entry of the block to be removed.           */
/* Post  : The selected block has been removed from the tape (thus making `Entry' invalid).                                       */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *DirPrev  = NULL;
  struct TapeEntry_s *Directory;
  dword               XMSIndex = 0;

  Directory = TapeWindow[WhichWindow].Directory;
  while (Directory != NULL && Directory != Entry)
  {
    XMSIndex += Directory->AllocLength;
    DirPrev = Directory;
    Directory = Directory->Next;
  }
  if (Entry->AllocLength > 0)
  {
    if (XMSIndex + Entry->AllocLength < TapeWindow[WhichWindow].Tape.Size)                                /* Not the last entry ? */
      CopyInXMS (TapeWindow[WhichWindow].Tape.Size - XMSIndex - Entry->AllocLength,                  /* Remove the block from XMS */
                 TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex + Entry->AllocLength,
                 TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex);
    TapeWindow[WhichWindow].Tape.Size -= Entry->AllocLength;
    if (TapeWindow[WhichWindow].Tape.Size > 0)
      XMSRealloc (TapeWindow[WhichWindow].Tape.XMSHandle, TapeWindow[WhichWindow].Tape.Size, FALSE);
  }
  if (DirPrev == NULL)                                                                                   /* Was the first entry ? */
    TapeWindow[WhichWindow].Directory = Entry->Next;
  else
    DirPrev->Next = Entry->Next;
  free (Entry);                                                                                 /* Remove from the tape directory */
  if (-- TapeWindow[WhichWindow].NumEntries < 0)
  {
    if (TapeWindow[WhichWindow].Tape.XMSHandle != -1)
    {
      XMSFree (TapeWindow[WhichWindow].Tape.XMSHandle);
      TapeWindow[WhichWindow].Tape.XMSHandle = -1;
    }
    TapeWindow[WhichWindow].Directory = NULL;
    TapeWindow[WhichWindow].NumEntries = -1;
    TapeWindow[WhichWindow].EmptyTape = TRUE;
    TapeWindow[WhichWindow].Tape.Size = 0;
    TapeWindow[WhichWindow].TapeVersion = 0x0000;
    TapeWindow[WhichWindow].NextJumpCode = 0x0000;
    TapeWindow[WhichWindow].ModeIs128K = FALSE;
    GhostEmptyTape (WhichWindow);
  }
  if (TapeWindow[WhichWindow].SelectedLine > TapeWindow[WhichWindow].IndexWindow.NumEntries)
    TapeWindow[WhichWindow].SelectedLine = TapeWindow[WhichWindow].IndexWindow.NumEntries;
  if (Changed)
    SignalTapeChanged (WhichWindow, TRUE);
  if (RebuildTape)
    SetupNewData (WhichWindow);                                                                     /* Rebuild the tape directory */
}

void CopyTapeBlock (short SourceWindow, struct TapeEntry_s *SDirP, short DestinationWindow, struct TapeEntry_s *DDirP, bool Move,
                    bool CompleteCollapsedGroups)

/**********************************************************************************************************************************/
/* Pre   : All variables are set and valid. The xxxEntry variables are indices in the appropriate combo windows.                  */
/* Post  : The currently selected tapeblock has been copied to the desired location in the destination tape.                      */
/*         If `Move' was TRUE, the block has been removed from the source tape.                                                   */
/* Import: CopyTapeBlock (recursive).                                                                                             */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *SDirPrev  = NULL;
  struct TapeEntry_s *SDirGSub;                                                               /* (When copying a collapsed group) */
  struct TapeEntry_s *DDirGSub;
  struct TapeEntry_s *SDirectory;
  struct TapeEntry_s *DDirectory;
  struct TapeEntry_s *NewDirEntry;
  dword               SXMSIndex = 0;
  dword               DXMSIndex = 0;
  bool                GroupDone = TRUE;
  bool                LoopDone  = TRUE;

  if (SourceWindow == DestinationWindow)                                                                    /* Within same tape ? */
  {
    SDirectory = TapeWindow[SourceWindow].Directory;
    while (SDirectory != NULL && SDirectory != SDirP)
    {
      SXMSIndex += SDirectory->AllocLength;
      SDirPrev = SDirectory;
      SDirectory = SDirectory->Next;
    }
    if (DDirP != NULL)                                                                                 /* Insert before the top ? */
    {                                                                                                                     /* Nope */
      DDirectory = TapeWindow[SourceWindow].Directory;
      while (DDirectory != NULL && DDirectory != DDirP)
      {
        DXMSIndex += DDirectory->AllocLength;
        DDirectory = DDirectory->Next;
      }
      DXMSIndex += DDirectory->AllocLength;                                                 /* Insert AFTER the destination entry */
    }
    if (SDirP->AllocLength > 0)
      XMSRealloc (TapeWindow[SourceWindow].Tape.XMSHandle,                                                    /* Make room anyway */
                  TapeWindow[SourceWindow].Tape.Size + SDirP->AllocLength, TRUE);
    if (Move)
    {
      if (SDirP->AllocLength > 0)
      {
        CopyInXMS (SDirP->AllocLength,                       /* Use the expanded room to copy the requested block to a safe place */
                   TapeWindow[SourceWindow].Tape.XMSHandle, SXMSIndex,
                   TapeWindow[SourceWindow].Tape.XMSHandle, TapeWindow[SourceWindow].Tape.Size);
        if (SXMSIndex < DXMSIndex)                                                                                 /* Moving up ? */
          CopyInXMS (DXMSIndex - SXMSIndex + SDirP->AllocLength,
                     TapeWindow[SourceWindow].Tape.XMSHandle, SXMSIndex + SDirP->AllocLength,
                     TapeWindow[SourceWindow].Tape.XMSHandle, SXMSIndex);
        else                                                                                                     /* (Moving down) */
          CopyInXMS (SXMSIndex - DXMSIndex,
                     TapeWindow[SourceWindow].Tape.XMSHandle, DXMSIndex,
                     TapeWindow[SourceWindow].Tape.XMSHandle, DXMSIndex + SDirP->AllocLength);
        CopyInXMS (SDirP->AllocLength,                                                         /* Copy the saved block into place */
                   TapeWindow[SourceWindow].Tape.XMSHandle, TapeWindow[SourceWindow].Tape.Size,
                   TapeWindow[SourceWindow].Tape.XMSHandle, DXMSIndex);
        XMSRealloc (TapeWindow[SourceWindow].Tape.XMSHandle,                          /* Release the temporary extra buffer space */
                    TapeWindow[SourceWindow].Tape.Size, FALSE);
      }
      if (DDirP == NULL)                                                                       /* Insert before the first entry ? */
      {
        SDirPrev->Next = SDirP->Next;
        SDirP->Next = TapeWindow[SourceWindow].Directory;
        TapeWindow[SourceWindow].Directory = SDirP;
      }
      else if (SDirPrev == NULL)                                                                         /* Was the first entry ? */
      {
        TapeWindow[SourceWindow].Directory = SDirP->Next;
        SDirP->Next = DDirP->Next;
        DDirP->Next = SDirP;
      }
      else
      {
        SDirPrev->Next = SDirP->Next;
        SDirP->Next = DDirP->Next;
        DDirP->Next = SDirP;
      }
    }
    else                                                                                                                /* (Copy) */
    {
      if (SDirP->AllocLength > 0)
      {
        if (DXMSIndex < TapeWindow[SourceWindow].Tape.Size)
          CopyInXMS (TapeWindow[SourceWindow].Tape.Size - DXMSIndex,                   /* 'Move' the room to the desired location */
                     TapeWindow[SourceWindow].Tape.XMSHandle, DXMSIndex,                              /* Unless adding to the end */
                     TapeWindow[SourceWindow].Tape.XMSHandle, DXMSIndex + SDirP->AllocLength);
        if (SXMSIndex > DXMSIndex)                                                     /* Room was created before the SXMSIndex ? */
          CopyInXMS (SDirP->AllocLength,
                     TapeWindow[SourceWindow].Tape.XMSHandle, SXMSIndex + SDirP->AllocLength,           /* (Adjust for the shift) */
                     TapeWindow[SourceWindow].Tape.XMSHandle, DXMSIndex);
        else
          CopyInXMS (SDirP->AllocLength,
                     TapeWindow[SourceWindow].Tape.XMSHandle, SXMSIndex,
                     TapeWindow[SourceWindow].Tape.XMSHandle, DXMSIndex);
      }
      if ((NewDirEntry = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
        FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
      memcpy (NewDirEntry, SDirP, sizeof (struct TapeEntry_s));
      NewDirEntry->ComboSelected &= ~SELECTBIT;                                                /* Don't keep the selection status */
      NewDirEntry->SrcJumpCode = 0xFFFF;                                                     /* Signal: non-initialized jump code */
      if (DDirP == NULL)                                                                       /* Insert before the first entry ? */
      {
        NewDirEntry->Next = TapeWindow[SourceWindow].Directory;
        TapeWindow[SourceWindow].Directory = NewDirEntry;
      }
      else
      {
        NewDirEntry->Next = DDirP->Next;
        DDirP->Next = NewDirEntry;
      }
      TapeWindow[SourceWindow].NumEntries ++;
      TapeWindow[SourceWindow].Tape.Size += SDirP->AllocLength;
    }
    SignalTapeChanged (SourceWindow, TRUE);
  }
  else                                                                                              /* (From 1 tape to the other) */
  {
    if (SDirP != NULL)                                                                                    /* Prepare source entry */
    {
      SDirectory = TapeWindow[SourceWindow].Directory;
      while (SDirectory != NULL && SDirectory != SDirP)
      {
        SXMSIndex += SDirectory->AllocLength;
        SDirPrev = SDirectory;
        SDirectory = SDirectory->Next;
      }
      if (!TapeWindow[DestinationWindow].EmptyTape)                                              /* Prepare destination directory */
      {
        if (DDirP != NULL)                                                                            /* Insert before the top ? */
        {                                                                                                                /* Nope */
          DDirectory = TapeWindow[DestinationWindow].Directory;
          while (DDirectory != NULL && DDirectory != DDirP)
          {
            DXMSIndex += DDirectory->AllocLength;
            DDirectory = DDirectory->Next;
          }
          DXMSIndex += DDirectory->AllocLength;                                             /* Insert AFTER the destination entry */
        }
        if (SDirP->AllocLength > 0)
        {
          XMSRealloc (TapeWindow[DestinationWindow].Tape.XMSHandle,
                      TapeWindow[DestinationWindow].Tape.Size + SDirP->AllocLength, TRUE);
          if (DXMSIndex < TapeWindow[DestinationWindow].Tape.Size)                         /* Make room, unless adding to the end */
            CopyInXMS (TapeWindow[DestinationWindow].Tape.Size - DXMSIndex,
                       TapeWindow[DestinationWindow].Tape.XMSHandle, DXMSIndex,
                       TapeWindow[DestinationWindow].Tape.XMSHandle, DXMSIndex + SDirP->AllocLength);
        }
      }
      else if (SDirP->AllocLength > 0)                                                                            /* (Empty tape) */
        XMSAlloc (&(TapeWindow[DestinationWindow].Tape.XMSHandle), SDirP->AllocLength);
      if (SDirP->AllocLength > 0)
        CopyInXMS (SDirP->AllocLength,                                                   /* Copy tape block between the 2 windows */
                   TapeWindow[SourceWindow].Tape.XMSHandle, SXMSIndex,
                   TapeWindow[DestinationWindow].Tape.XMSHandle, DXMSIndex);
      TapeWindow[DestinationWindow].Tape.Size += SDirP->AllocLength;
      if (Move)                                                                       /* Remove from the source tape if requested */
      {
        if (SDirP->AllocLength > 0)
        {
          if (SXMSIndex + SDirP->AllocLength < TapeWindow[SourceWindow].Tape.Size)     /* Move back, unless removing from the end */
            CopyInXMS (TapeWindow[SourceWindow].Tape.Size - SDirP->AllocLength - SXMSIndex,
                       TapeWindow[SourceWindow].Tape.XMSHandle, SXMSIndex + SDirP->AllocLength,
                       TapeWindow[SourceWindow].Tape.XMSHandle, SXMSIndex);
          TapeWindow[SourceWindow].Tape.Size -= SDirP->AllocLength;
          if (TapeWindow[SourceWindow].Tape.Size != 0)
            XMSRealloc (TapeWindow[SourceWindow].Tape.XMSHandle, TapeWindow[SourceWindow].Tape.Size, FALSE);
        }
        SDirP->ComboSelected &= ~SELECTBIT;                                                    /* Don't keep the selection status */
        SDirP->SrcJumpCode = 0xFFFF;                                                         /* Signal: non-initialized jump code */
        if (SDirPrev == NULL)                                                                            /* Was the first entry ? */
        {
          TapeWindow[SourceWindow].Directory = SDirP->Next;
          if (TapeWindow[SourceWindow].Directory == NULL)                                                  /* Tape is now empty ? */
            TapeWindow[SourceWindow].EmptyTape = TRUE;
        }
        else
          SDirPrev->Next = SDirP->Next;                                                 /* Remove from the source directory chain */
        if (TapeWindow[DestinationWindow].EmptyTape)                                                 /* Will be the first entry ? */
        {
          SDirP->Next = NULL;
          TapeWindow[DestinationWindow].Directory = SDirP;
        }
        else
          if (DDirP == NULL)                                                                   /* Insert before the first entry ? */
          {
            SDirP->Next = TapeWindow[DestinationWindow].Directory;
            TapeWindow[DestinationWindow].Directory = SDirP;
          }
          else
          {
            SDirP->Next = DDirP->Next;                                               /* Insert in the destination directory chain */
            DDirP->Next = SDirP;
          }
        TapeWindow[SourceWindow].NumEntries --;
      }
      else                                                                                            /* Copy the directory entry */
      {
        if ((NewDirEntry = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
          FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
        memcpy (NewDirEntry, SDirP, sizeof (struct TapeEntry_s));
        NewDirEntry->ComboSelected &= ~SELECTBIT;                                              /* Don't keep the selection status */
        NewDirEntry->SrcJumpCode = 0xFFFF;                                                   /* Signal: non-initialized jump code */
        if (TapeWindow[DestinationWindow].EmptyTape)
        {
          NewDirEntry->Next = NULL;
          TapeWindow[DestinationWindow].Directory = NewDirEntry;
        }
        else
          if (DDirP == NULL)                                                                   /* Insert before the first entry ? */
          {
            NewDirEntry->Next = TapeWindow[DestinationWindow].Directory;
            TapeWindow[DestinationWindow].Directory = NewDirEntry;
          }
          else
          {
            NewDirEntry->Next = DDirP->Next;
            DDirP->Next = NewDirEntry;
          }
      }
    }
    TapeWindow[DestinationWindow].NumEntries ++;
    TapeWindow[DestinationWindow].EmptyTape = FALSE;
    UnghostTape (DestinationWindow);
    if (TapeWindow[SourceWindow].NumEntries < 0)
    {
      if (TapeWindow[SourceWindow].Tape.XMSHandle != -1)
      {
        XMSFree (TapeWindow[SourceWindow].Tape.XMSHandle);
        TapeWindow[SourceWindow].Tape.XMSHandle = -1;
      }
      TapeWindow[SourceWindow].Directory = NULL;
      TapeWindow[SourceWindow].NumEntries = -1;
      TapeWindow[SourceWindow].EmptyTape = TRUE;
      TapeWindow[SourceWindow].Tape.Size = 0;
      TapeWindow[SourceWindow].TapeVersion = 0x0000;
      TapeWindow[SourceWindow].NextJumpCode = 0x0000;
      TapeWindow[SourceWindow].ModeIs128K = FALSE;
      GhostEmptyTape (SourceWindow);
    }
    SignalTapeChanged (DestinationWindow, TRUE);
    if (Move)
      SignalTapeChanged (SourceWindow, TRUE);
  }
  _CheckJumpCodes (SourceWindow);                                           /* Adjust the jumping tape entries that are unchanged */
  if (SourceWindow != DestinationWindow)
    _CheckJumpCodes (DestinationWindow);
  if (CompleteCollapsedGroups && (SDirP->BlockTypeID == 0x21 || SDirP->BlockTypeID == 0x24) && (SDirP->ComboSelected & EXPANDBIT))
  {                                                                                                     /* Collapsed Group/Loop ? */
    if (DDirP == NULL)
      DDirGSub = TapeWindow[DestinationWindow].Directory;
    else
      DDirGSub = DDirP->Next;
    SDirGSub = SDirP->Next;
    if (SDirP->BlockTypeID == 0x21)
      GroupDone = FALSE;
    else
      LoopDone = FALSE;
    while (!GroupDone || !LoopDone)
      if (SDirGSub == NULL)
        GroupDone = LoopDone = TRUE;
      else
      {
        CopyTapeBlock (SourceWindow, SDirGSub, DestinationWindow, DDirGSub, Move, FALSE);                      /* (Don't re-loop) */
        if (SDirGSub->BlockTypeID == 0x22)                                                  /* Just done the matching Group End ? */
          GroupDone = TRUE;                                                                                    /* Then we're done */
        else if (SDirGSub->BlockTypeID == 0x25)                                              /* Just done the matching Loop End ? */
          LoopDone = TRUE;                                                                                     /* Then we're done */
        SDirGSub = SDirGSub->Next;
        DDirGSub = DDirGSub->Next;
      }
  }
  _MustInitializeJumpCodes = TRUE;                              /* Signal: jump codes of the copied entries are invalid - rebuild */
}

short DragTapeBlock (void)

/**********************************************************************************************************************************/
/* Pre   : A combo window has received the WINH_DRAGDROPPED message, so the global `DraggedComboEntry' is set and valid.          */
/* Post  : All given tape blocks have been copied/moved to the destination window.                                                */
/*         The return value is TWLEFT or TWRIGHT, which is the destination window.                                                */
/* Import: CopyTapeBlock.                                                                                                         */
/**********************************************************************************************************************************/

{
  struct ComboEntry   Entry;
  struct TapeEntry_s *SDirP;
  struct TapeEntry_s *NextSDirP;
  struct TapeEntry_s *DDirP;
  short               SWindow = TWLEFT;
  short               DWindow = TWLEFT;
  comboent            SAdjust;

  SetMouseShape (DefaultWaitingMouseShape);
  if (DraggedComboEntry.SourceWindow == &(TapeWindow[TWRIGHT].IndexWindow))                                 /* Get window indices */
    SWindow = TWRIGHT;
  if (DraggedComboEntry.DestinationWindow == &(TapeWindow[TWRIGHT].IndexWindow))
    DWindow = TWRIGHT;
  AutoInsertNewTape (DWindow);
  if (DraggedComboEntry.DestinationEntry == -1)
    DDirP = NULL;
  else
  {
    GetComboEntry (&(TapeWindow[DWindow].IndexWindow), &Entry, DraggedComboEntry.DestinationEntry);
    DDirP = (struct TapeEntry_s *)Entry.Identifier;
    if (DDirP->BlockTypeID == 0x21 && (DDirP->ComboSelected & EXPANDBIT))                       /* Drag after a collapsed Group ? */
      while (DDirP->BlockTypeID != 0x22)                                                           /* Find the matching Group End */
      {
        DDirP = DDirP->Next;
        if (!DDirP)                                                                                                  /* Ooops.... */
        {
          SetMouseShape (DefaultNormalMouseShape);
          return (DWindow);
        }
      }
    else if (DDirP->BlockTypeID == 0x24 && (DDirP->ComboSelected & EXPANDBIT))                   /* Drag after a collapsed Loop ? */
      while (DDirP->BlockTypeID != 0x25)                                                            /* Find the matching Loop End */
      {
        DDirP = DDirP->Next;
        if (!DDirP)                                                                                                  /* Ooops.... */
        {
          SetMouseShape (DefaultNormalMouseShape);
          return (DWindow);
        }
      }
  }
  SAdjust = TapeWindow[SWindow].SelectedLine;
  if (SWindow != DWindow && DraggedComboEntry.MoveEntries &&                             /* Moving from one window to the other ? */
      SAdjust > 0)                                                                      /* Poiting to at least the second entry ? */
  {
    GetComboEntry (&(TapeWindow[SWindow].IndexWindow), &Entry, SAdjust - 1);                        /* Peek at the previous entry */
    if ((Entry.Selected & ENT_NOSELECT) || Entry.Identifier == 0)                      /* The previous entry was non-selectable ? */
      SAdjust --;                                  /* Then we're now at the second line of a multi-line entry. Point to the first */
  }
  if (DraggedComboEntry.MultipleEntries)
  {
    SDirP = TapeWindow[SWindow].Directory;
    while (SDirP != NULL)
    {
      NextSDirP = SDirP->Next;
      if (SDirP->ComboSelected & SELECTBIT)
      {
        CopyTapeBlock (SWindow, SDirP, DWindow, DDirP, DraggedComboEntry.MoveEntries, FALSE);
        if (DDirP == NULL)
          DDirP = TapeWindow[DWindow].Directory;
        else
          DDirP = DDirP->Next;
      }
      SDirP = NextSDirP;
    }
  }
  else
  {
    GetComboEntry (&(TapeWindow[SWindow].IndexWindow), &Entry, DraggedComboEntry.SourceEntry);
    SDirP = (struct TapeEntry_s *)Entry.Identifier;
    CopyTapeBlock (SWindow, SDirP, DWindow, DDirP, DraggedComboEntry.MoveEntries, TRUE);
  }
  TapeWindow[DWindow].SelectedLine = DraggedComboEntry.DestinationEntry + 1;              /* Make sure the selection bar is valid */
  SetupNewData (DWindow);                                                                            /* Display the new directory */
  SelectNewTapeEntry (DWindow, TapeWindow[DWindow].SelectedLine);
  if (!(SWindow != DWindow && !DraggedComboEntry.MoveEntries))                                            /* Both tapes changed ? */
  {
    _MustInitializeJumpCodes = TRUE;                                    /* Fully rebuild the jump code table of this tape as well */
    SetupNewData (SWindow);
  }
  SelectNewTapeEntry (SWindow, SAdjust);
  SetMouseShape (DefaultNormalMouseShape);
  return (DWindow);
}

void SetTapeBlockSelectFlag (struct TapeEntry_s *Entry, short NewValue)

/**********************************************************************************************************************************/
/* Pre   : `Entry' points to the tape block directory entry, `NewValue' is the new block select value.                            */
/* Post  : The block select value has been updated. Only the SELECTBIT is kept.                                                   */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *DirP;
  bool                GroupDone = TRUE;
  bool                LoopDone = TRUE;

  Entry->ComboSelected = (Entry->ComboSelected & ~SELECTBIT) | (NewValue & SELECTBIT);
  if ((Entry->BlockTypeID == 0x21 || Entry->BlockTypeID == 0x24) && (Entry->ComboSelected & EXPANDBIT))
  {                                                                                    /* Entry is a collapsed Group/Loop block ? */
    if (Entry->BlockTypeID == 0x21)
      GroupDone = FALSE;
    else
      LoopDone = FALSE;
    DirP = Entry->Next;
    while ((!GroupDone || !LoopDone) && DirP != NULL)
    {
      DirP->ComboSelected = (DirP->ComboSelected & ~SELECTBIT) | (NewValue & SELECTBIT);                   /* Handle entire group */
      if (DirP->BlockTypeID == 0x21)                                                                             /* Group Start ? */
        GroupDone = FALSE;                                                                            /* Marks beginning of group */
      if (DirP->BlockTypeID == 0x22)                                                                               /* Group End ? */
        GroupDone = TRUE;                                                                                   /* Marks end of group */
      if (DirP->BlockTypeID == 0x24)                                                                              /* Loop Start ? */
        LoopDone = FALSE;                                                                              /* Marks beginning of loop */
      if (DirP->BlockTypeID == 0x25)                                                                                /* Loop End ? */
        LoopDone = TRUE;                                                                                     /* Marks end of loop */
      DirP = DirP->Next;
    }
  }
}

void SetTapeBlockExpandFlag (struct TapeEntry_s *Entry, short NewValue)

/**********************************************************************************************************************************/
/* Pre   : `Entry' points to the tape block directory entry, `NewValue' is the new group expand value.                            */
/* Post  : The group expand value has been updated. Only the EXPANDBIT is kept.                                                   */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  Entry->ComboSelected = (Entry->ComboSelected & ~EXPANDBIT) | (NewValue & EXPANDBIT);
}

void SetupNewData (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/* Post  : A catalog of the file has been inserted in the entry table of the combo window. A line is built up as follows:         */
/*         1. (Expanded) Spectrum file name;                                                                                      */
/*         2. Type of file: PROG, CODE, A-AR, N-AR, HEAD, ????;                                                                   */
/*         3. Depending on the type:                                                                                              */
/*            PROG (Program)                 : Auto-start line or ---- if none;                                                   */
/*            CODE (Code)                    : Startaddress of the code;                                                          */
/*            A-AR (Alphanumerical Array)    : -                                                                                  */
/*            N-AR (Numerical Array)         : -                                                                                  */
/*            HEAD (Headerless block)        : The Flag byte;                                                                     */
/*            ???? (Headerless block)        : The Type byte;                                                                     */
/*            DATA (Following a header)      : -                                                                                  */
/*         4. Length of the block.                                                                                                */
/*         Headerless blocks with total length < 2 are reported as being noise (multiple blocks are reported once).               */
/*         PROG, CODE and headerless blocks are selectable, all others are not.                                                   */
/*         - Entries with an error (parity error) are printed in red.                                                             */
/*         - Entries with losing data (ID 0x10 with incomplete last byte) are printed in blue.                                    */
/* Import: GetBlockTypeInfo, _CheckGroupBlocks, _CheckJumpBlocks, _CheckJumpCodes.                                                */
/**********************************************************************************************************************************/

{
  struct MouseShape   *OldShape;
  struct TapeRecord_s *TRecord;
  struct TapeEntry_s  *DirEntry;
  struct ComboEntry    NewEntry;
  struct ComboEntry    ErrEntry;
  word       register  M;
  dword                BlLength;
  byte                 LastFlag;
  byte                 LastBlockTypeID = 0x00;
  byte                 FlagByte;
  char                 FType[49];
  short                TokLen;
  dword                XMSIndex     = 0;
  bool                 HasBadParity;
  char                 ParCol[4];
  char                 SpeedInfo[3];
  bool                 IsTapeData;
  bool                 IsIdentifiedData;
  char                *FullName;
  char                *TapeWindowName;
  bool                 InsertDescription;
  comboent             BlockNumber = 0;
  short                FirstCol;
  bool                 Collapsed   = FALSE;
  bool                 ColGroup    = FALSE;
  bool                 ColLoop     = FALSE;

  _CheckGroupBlocks (WhichWindow);                                              /* Test and repair possible Group inconsistencies */
  _CheckJumpBlocks (WhichWindow);                                                /* Test and repair possible Jump inconsistencies */
  if (_MustInitializeJumpCodes)                                                             /* Are all jump codes valid already ? */
    _InitializeJumpCodes (WhichWindow);
  else
    _CheckJumpCodes (WhichWindow);
  ReInitComboWindow (&(TapeWindow[WhichWindow].IndexWindow), TapeWindow[WhichWindow].NumEntries + 1);
  if (TapeWindow[WhichWindow].EmptyTape)
  {
    NewEntry.Selected = 0x01 | ENT_NOSELECT;
    NewEntry.Identifier = 0;
    strcpy (NewEntry.Text, "{+1<Empty tape>}N");
    if (!InsertComboEntry (&(TapeWindow[WhichWindow].IndexWindow), &NewEntry, NULL, -1))
      FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
    PrintComboEntries (&(TapeWindow[WhichWindow].IndexWindow));
    UpdateTapeSize (WhichWindow, -1);
    return;
  }
  DirEntry = TapeWindow[WhichWindow].Directory;
  NewEntry.Selected = 0;
  NewEntry.Identifier = 0;
  strcpy (ErrEntry.Text, "      >>>>> Missing Datablock <<<<<");
  ErrEntry.Selected = ENT_NOSELECT | ENT_NODROP;
  ErrEntry.Identifier = 0;
  TRecord = (struct TapeRecord_s *)WorkBuffer;
  OldShape = CurrentMouseShape;
  SetMouseShape (DefaultWaitingMouseShape);
  LastFlag = 0xFF;                                                                      /* Signal: last block was of unknown type */
  while (DirEntry != NULL)
  {
    NewEntry.Selected = DirEntry->ComboSelected & SELECTBIT;                                        /* Fetch old selection status */
    InsertDescription = FALSE;
    if (!GetBlockTypeInfo (DirEntry->BlockTypeID, &IsTapeData, &IsIdentifiedData, &FullName, &TapeWindowName, NULL))
      InsertDescription = TRUE;                                                                                 /* Unknown type ? */
    else if (!IsIdentifiedData)                                                              /* Normal/Turbo, but identified data */
      InsertDescription = TRUE;                                                                 /* Nope, insert ID string instead */
    if (InsertDescription)
    {
      if (!Collapsed)
      {
        if (LastFlag == 0x00 && (LastBlockTypeID == 0x10 || LastBlockTypeID == 0x11))         /* Last block was not a datablock ? */
        {
          if (!InsertComboEntry (&(TapeWindow[WhichWindow].IndexWindow), &ErrEntry, NULL, -1))
            FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
        }
        if (DirEntry->BlockTypeID == 0x21 ||                                                              /* Special: Group Start */
            DirEntry->BlockTypeID == 0x30)                                                           /* Special: Text Description */
        {
          CopyFromXMS (DirEntry->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);
          *(WorkBuffer + DirEntry->BlockLength) = '\0';
          *(WorkBuffer + 50) = '\0';                                                                     /* Force some max length */
          if (DirEntry->BlockTypeID == 0x21)                                                                     /* (Group Start) */
            sprintf (NewEntry.Text, "%4d%c{+1GROUP \"%s\"", BlockNumber, DirEntry->ComboSelected & EXPANDBIT ? 0x05 : 0x06,
                     WorkBuffer);
          else                                                                                              /* (Text Description) */
            sprintf (NewEntry.Text, "%4d {+1  {{Text \"%s\"}}", BlockNumber, WorkBuffer);
        }
        else if (DirEntry->BlockTypeID == 0x35)                                                           /* Special: Custom Info */
        {
          M = 16;
          CopyFromXMS (M, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);                       /* Fetch ID string */
          while (-- M > 0 && *(WorkBuffer + M) == ' ')                                                    /* Skip trailing spaces */
            ;
          *(WorkBuffer + M + 1) = '\0';
          memset (NewEntry.Text, ' ', 255);
          sprintf (NewEntry.Text, "%4d {+1  {{Custom Info \"%s\"}}", BlockNumber, WorkBuffer);
          NewEntry.Text[strlen (NewEntry.Text)] = ' ';
          sprintf (WorkBuffer, " %lu", DirEntry->BlockLength);
          strcpy (NewEntry.Text + 47 + (M < 18 ? 1 : 0) - strlen (WorkBuffer), WorkBuffer);
        }
        else if (DirEntry->BlockTypeID == 0x24)                                                            /* Special: Loop Start */
          sprintf (NewEntry.Text, "%4d%c{+1LOOP %d TIMES", BlockNumber, DirEntry->ComboSelected & EXPANDBIT ? 0x05 : 0x06,
                   DirEntry->PauseLength);
        else if (DirEntry->BlockTypeID == 0x23)                                                         /* Special: Jump to Block */
          sprintf (NewEntry.Text, "%4d {+1  {{Jump to Entry %d}}", BlockNumber, BlockNumber + (short)DirEntry->PauseLength);
        else if (DirEntry->BlockTypeID == 0x14)                                                             /* Special: Pure Data */
        {
          if (DirEntry->Timing.LengthZeroBitPulse == NS_LEN0BIT && DirEntry->Timing.LengthOneBitPulse == NS_LEN1BIT)
            sprintf (NewEntry.Text, "%4d   PURE DATA                    %7lu", BlockNumber, DirEntry->BlockLength);
          else
            sprintf (NewEntry.Text, "%4d * PURE DATA                    %7lu", BlockNumber, DirEntry->BlockLength);
        }
        else
        {
          FirstCol = GetBlockTypeIndex (DirEntry->BlockTypeID);
          if (BlockType[FirstCol].ReportSize)                                          /* Report size of the data block as well ? */
            sprintf (NewEntry.Text, "%4d {+1%-33s%7lu", BlockNumber, TapeWindowName, DirEntry->BlockLength);
          else
            sprintf (NewEntry.Text, "%4d {+1%s", BlockNumber, TapeWindowName);
        }
        NewEntry.Identifier = (dword)DirEntry;
        if (!InsertComboEntry (&(TapeWindow[WhichWindow].IndexWindow), &NewEntry, NULL, -1))
          FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
      }
      LastFlag = 0xFF;
      BlockNumber ++;
    }
    else
    {
      strcpy (SpeedInfo, DirEntry->BlockTypeID == 0x10 ? "  " : "* ");                               /* Mark turbo loading blocks */
      FlagByte = DirEntry->Flags.FlagByteUsed ? DirEntry->FlagByte : 0xFF;
      if ((HasBadParity = (DirEntry->Flags.ParityByteUsed && DirEntry->ParityByte != DirEntry->CalculatedParity)))
        strcpy (ParCol, "{+4");
      else if (DirEntry->BlockTypeID == 0x10 && DirEntry->Timing.NumBitsInLastByte != 8)
        strcpy (ParCol, "{+C");
      else
        strcpy (ParCol, "{++");
      BlLength = DirEntry->BlockLength;
      sprintf (NewEntry.Text, "%4d %s%s", BlockNumber, ParCol, SpeedInfo);
      if (FlagByte != 0x00 || (FlagByte == 0x00 && BlLength != 17))                                         /* It is a data block */
      {
        if (LastFlag != 0x00)                                                            /* Last read block was also a data block */
        {
          if (!Collapsed)
          {
            if (DirEntry->Flags.FlagByteUsed)
              if (BlLength >= 100000)
                sprintf (NewEntry.Text + strlen (NewEntry.Text), "HEAD                    %5u%7lu", FlagByte, BlLength);
              else
                sprintf (NewEntry.Text + strlen (NewEntry.Text), "HEAD                     %5u%6lu", FlagByte, BlLength);
            else
              sprintf (NewEntry.Text + strlen (NewEntry.Text), "HEAD                         %7lu", BlLength);
            NewEntry.Identifier = (dword)DirEntry;                                                    /* It is a headerless block */
            if (!InsertComboEntry (&(TapeWindow[WhichWindow].IndexWindow), &NewEntry, NULL, -1))
              FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
          }
        }
        else                                                                          /* It belongs to the previously read header */
        {
          if (!Collapsed)
          {
            sprintf (NewEntry.Text + strlen (NewEntry.Text), "DATA                         %7lu", BlLength);
            NewEntry.Identifier = (dword)DirEntry;
            if (!InsertComboEntry (&(TapeWindow[WhichWindow].IndexWindow), &NewEntry, NULL, -1))
              FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
          }
          LastFlag = (FlagByte == 0x00 ? 0xFF : FlagByte);
        }
      }
      else                                                                                                /* It is a header block */
      {
        if (LastFlag == 0x00)                                                           /* Previous block was also a header block */
        {
          if (!Collapsed)
          {
            if (!InsertComboEntry (&(TapeWindow[WhichWindow].IndexWindow), &ErrEntry, NULL, -1))             /* This is weird.... */
              FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
            NewEntry.Selected = 0;
          }
        }
        else
          LastFlag = 0x00;
        strcat (NewEntry.Text, "      \"");
        FirstCol = strlen (NewEntry.Text);
        CopyFromXMS (DirEntry->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);
        M = RemapSpectrumString (TRecord->HName, 10, NewEntry.Text + FirstCol, 80, FALSE, FALSE, &TokLen, TRUE, FALSE, TRUE);
        NewEntry.Text[M + FirstCol] = '\0';                                                        /* Insert filename from header */
        if (TokLen > 10)                                                                                    /* Over 10 expanded ? */
        {
          while (TokLen > 10 && NewEntry.Text[M + FirstCol - 1] == ' ')                      /* Try eliminating spaces at the end */
          {
            M --;
            TokLen --;
          }
          NewEntry.Text[M + FirstCol] = '\0';
          sprintf (NewEntry.Text + strlen (NewEntry.Text), "%s}N\"", ParCol);
        }
        else if (TokLen < 10)                                                     /* Unexpandable tokens were removed (AT, TAB) ? */
        {
          sprintf (NewEntry.Text + strlen (NewEntry.Text), "%s}N\"", ParCol);            /* Ensure there's an attribute endmarker */
          strcat (NewEntry.Text, "          ");                                                     /* Pad the result with spaces */
          NewEntry.Text[M + FirstCol + 16 - TokLen] = '\0';
          TokLen = 10;
        }
        else
          sprintf (NewEntry.Text + strlen (NewEntry.Text), "%s}N\"", ParCol);
        if (TokLen > 10)
        {
          if (!Collapsed)
          {
            NewEntry.Identifier = 0;
            NewEntry.Selected = ENT_NOSELECT | ENT_NODROP;
            if (!InsertComboEntry (&(TapeWindow[WhichWindow].IndexWindow), &NewEntry, NULL, -1))
              FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
            sprintf (NewEntry.Text, "     %s%s                  ", ParCol, SpeedInfo);
            NewEntry.Selected = 0;
          }
          M = 10;
        }
        switch (TRecord->HType)                                                                         /* Block type from header */
        {
          case 0 : if (TRecord->HStart > 32767)
                     strcpy (FType, "  PROG  ---- ");
                   else
                     sprintf (FType, "  PROG %5u ", TRecord->HStart);
                   sprintf (FType + strlen (FType), "%5u", TRecord->HLength);
                   break;
          case 1 : sprintf (FType, "  N-AR       %5u", TRecord->HLength);
                   break;
          case 2 : sprintf (FType, "  A-AR       %5u", TRecord->HLength);
                   break;
          case 3 : sprintf (FType, "  CODE %5u %5u", TRecord->HStart, TRecord->HLength);
                   break;
          default: sprintf (FType, "  ???? %5u%6lu", TRecord->HType, BlLength);               /* Usually in M1 blocks; just 17 bt */
                   LastFlag = 0xFF;                                                   /* Signal: also print info on the datablock */
        }
        DirEntry->Flags.IsHeader = TRUE;                                           /* Ensure the directory entry has correct info */
        if (!Collapsed)
        {
          strcat (NewEntry.Text, FType);
          NewEntry.Selected |= HEADERBLOCK;                                                              /* Selection prints info */
          NewEntry.Identifier = (dword)DirEntry;
          if (!InsertComboEntry (&(TapeWindow[WhichWindow].IndexWindow), &NewEntry, NULL, -1))
            FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
        }
      }
      BlockNumber ++;
    }
    if (DirEntry->BlockTypeID == 0x21 && (DirEntry->ComboSelected & EXPANDBIT))                                  /* Group Start ? */
      ColGroup = TRUE;
    else if (DirEntry->BlockTypeID == 0x22 && ColGroup)                                                            /* Group End ? */
      ColGroup = FALSE;
    else if (DirEntry->BlockTypeID == 0x24 && (DirEntry->ComboSelected & EXPANDBIT))                              /* Loop Start ? */
      ColLoop = TRUE;
    else if (DirEntry->BlockTypeID == 0x25 && ColLoop)                                                              /* Loop End ? */
      ColLoop = FALSE;
    Collapsed = (ColGroup | ColLoop);                                                             /* Are we done collapsing now ? */
    XMSIndex += DirEntry->AllocLength;
    DirEntry = DirEntry->Next;
    LastBlockTypeID = DirEntry->BlockTypeID;
  }
  if (LastFlag == 0x00 && (LastBlockTypeID == 0x10 || LastBlockTypeID == 0x11))               /* Last block was not a datablock ? */
  {
    if (!InsertComboEntry (&(TapeWindow[WhichWindow].IndexWindow), &ErrEntry, NULL, -1))                      /* This is weird... */
      FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
  }
  if (TapeWindow[WhichWindow].SelectedLine >= 0)                                                  /* Keep the same entry selected */
    SelectNewTapeEntry (WhichWindow, TapeWindow[WhichWindow].SelectedLine);
  else
    PrintComboEntries (&(TapeWindow[WhichWindow].IndexWindow));
  SetMouseShape (OldShape);
  UpdateTapeSize (WhichWindow, (long)TapeWindow[WhichWindow].Tape.Size +
                  (long)(TapeWindow[WhichWindow].NumEntries + 1) * sizeof (struct TapeEntry_s));
}

comboent GetBlockNumber (char *ComboTextLine)

/**********************************************************************************************************************************/
/* Pre   : `ComboTextEntry' holds the combo entry Text field to be considered.                                                    */
/* Post  : The block number (in the first 4 columns) has been read and returned, or -1 if there was no valid number.              */
/*         Any (match) coloring in front is skipped.                                                                              */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  short register Cnt;
  comboent       BlockNumber = 0;
  short          Start = 0;
  short          End = 3;                                                                   /* (A block number holds 4 positions) */

  while (ComboTextLine[Start] == '{')                                                                /* Block number is colored ? */
  { Start += 3; End += 3; }                                                                                  /* Skip all coloring */
  for (Cnt = Start ; Cnt <= End ; Cnt ++)
    if (ComboTextLine[Cnt] == '\0')                                                                              /* End of text ? */
      return (-1);                                                                                         /* Now that's funny... */
    else if (ComboTextLine[Cnt] >= '0' && ComboTextLine[Cnt] <= '9')
      BlockNumber = BlockNumber * 10 + ComboTextLine[Cnt] - '0';                                               /* Read the number */
  return (BlockNumber);
}

comboent FindBlockNumber (short WhichWindow, comboent BlockNumber)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `BlockNumber' is what we're looking for.                                    */
/* Post  : The combo entry number that contains this block number has been returned, or -1 is it does not exist.                  */
/* Import: GetBlockNumber.                                                                                                        */
/**********************************************************************************************************************************/

{
  struct ComboEntry CEntry;
  comboent register Cnt;
  comboent          ReadNumber;

  for (Cnt = 0 ; Cnt <= TapeWindow[WhichWindow].IndexWindow.NumEntries ; Cnt ++)
  {
    GetComboEntry (&(TapeWindow[WhichWindow].IndexWindow), &CEntry, Cnt);
    if (!(CEntry.Selected & ENT_NOSELECT))                                                /* Does it have a block number anyway ? */
    {
      ReadNumber = GetBlockNumber (CEntry.Text);                                                         /* Read the block number */
      if (ReadNumber == -1)
        return (-1);
      if (ReadNumber == BlockNumber)                                                                          /* What we wanted ? */
        return (Cnt);
    }
  }
  return (-1);
}

void ViewTapeBlockInfo (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/* Post  : The directory entry information of the currently selected tape block has been shown.                                   */
/* Import: GetBlockTypeInfo, SetupNewData.                                                                                        */
/**********************************************************************************************************************************/

{
  struct WindowInfo CWindow  = {186, 130, 270, 81, 270, 81, 270, 81, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER| WINF_TCLEAR, NULL,
                                "Tape Block Information"};
  struct ButtonInfo POk      = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 109, 55, 8, DBLACK, DWHITE, NULL, 1, "@Ok", NULL};
  struct ButtonInfo PEdit    = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 57, 55, 8, DBLACK, DWHITE, NULL, 2, "@Edit", NULL};

  struct ComboInfo    TWindow;
  struct ComboEntry   TEntry;
  struct ButtonInfo  *ButtonP;
  struct ComboEntry   Entry;
  struct TapeEntry_s *DirP;
  struct TapeEntry_s *Directory;
  bool                IsTapeData;
  bool                IsIdentifiedData;
  char               *FullName;
  char               *TapeWindowName;
  char               *HWIndex;                                                                         /* Used only for type 0x33 */
  char               *HWName;
  dword              *DTable;
  short              *WTable;
  short               NumInfoLines;                                                                   /* Total number is this + 3 */
  short               Line         = 4;                                                     /* Custom block info starts at line 4 */
  comboent            AddLines;
  comboent            BlockNumber  = 0;
  short      register Cnt;
  word       register Cnt2;
  word       register Cnt3;
  word       register Cnt4;
  byte                Bt1;
  byte                Bt2;
  byte                Bt3;
  byte                LastValue;
  byte                LastType;
  byte                NumHWTypes;
  dword               XMSIndex     = 0;
  word                WScratch;
  char                TextBuffer[257];
  bool                Again        = TRUE;
  bool                Stop         = FALSE;
  bool                EntryChanged = FALSE;
  char                TBits[5];

  TEntry.Selected = 0;
  TEntry.Identifier = 0;
  GetComboEntry (&(TapeWindow[WhichWindow].IndexWindow), &Entry, TapeWindow[WhichWindow].SelectedLine);
  DirP = (struct TapeEntry_s *)Entry.Identifier;
  if (DirP != NULL)                                                                        /* Determine the start of the XMS part */
  {
    Directory = TapeWindow[WhichWindow].Directory;
    while (Directory != NULL && Directory != DirP)
    {
      XMSIndex += Directory->AllocLength;
      Directory = Directory->Next;
      BlockNumber ++;
    }
  }
  if (DirP->BlockTypeID == 0x35 && DirP->FlagByte == INSRTCI_SCREEN)                               /* Standardized Screen block ? */
  {
    if (BlockType[GetBlockTypeIndex (DirP->BlockTypeID)].EditHandler (WhichWindow, DirP))                               /* Edit ! */
    {
      SignalTapeChanged (WhichWindow, TRUE);
      EntryChanged = TRUE;
      SetupNewData (WhichWindow);                                                                   /* Rebuild the tape directory */
    }
    return;
  }
  if (DirP->BlockTypeID == 0x35 && DirP->FlagByte == INSRTCI_INSTRUCT)                       /* Standardized Instructions block ? */
  {
    if (ViewInstructionsBlock (WhichWindow, DirP))                                                         /* Use a custom viewer */
    {
      SignalTapeChanged (WhichWindow, TRUE);
      EntryChanged = TRUE;
      SetupNewData (WhichWindow);                                                                   /* Rebuild the tape directory */
    }
    return;
  }
  GetBlockTypeInfo (DirP->BlockTypeID, &IsTapeData, &IsIdentifiedData, &FullName, &TapeWindowName, &NumInfoLines);
  if (NumInfoLines >= 0)
    AddLines = NumInfoLines;
  else
    AddLines = 16;
  CWindow.Height = CWindow.MinHeight = CWindow.MaxHeight = CWindow.Height + 10 * AddLines;
  POk.StartY += (10 * AddLines);
  if ((Cnt = GetBlockTypeIndex (DirP->BlockTypeID)) >= 0)
    if (BlockType[Cnt].EditHandler != NULL)                                                       /* A block that can be edited ? */
    {
      POk.StartX += 46;
      PEdit.StartY = POk.StartY;
      AddButton (&CWindow, &POk, TRUE);
      AddButton (&CWindow, &PEdit, FALSE);
    }
    else
      AddButton (&CWindow, &POk, TRUE);
  else
    AddButton (&CWindow, &POk, TRUE);
  DrawWindow (&CWindow, NULL, WING_CLOSE, TRUE);
  while (Again)
  {
    Line = 4;
    if (NumInfoLines < 0)
      MakeComboWindow (&TWindow, &CWindow, COMT_VIEW, COMF_NOBUFFER | COMF_VSLIDER | COMF_SYNCSLIDER, CWindow.StartX + 8,
                       CWindow.StartY + 40, 0, 0, 46, 40, 16, NULL, FALSE, 0, NULL, DBLACK, DWHITE, 0, 0);
    GetBlockTypeInfo (DirP->BlockTypeID, NULL, NULL, &FullName, NULL, NULL);
    Wprintf (&CWindow, 1, 2, " {+1Block type       }N %s (%02X)", FullName, DirP->BlockTypeID);
    Wprintf (&CWindow, 2, 2, " {+1Block length     }N %lu", DirP->BlockLength);
    strcpy (TBits, DirP->Timing.NumBitsInLastByte == 1 ? "bit" : "bits");
    switch (DirP->BlockTypeID)
    {
      case 0x10 : if (DirP->Flags.FlagByteUsed)
                    Wprintf (&CWindow, Line ++, 2, " {+1Flag byte        }N %4u", DirP->FlagByte);
                  else
                    Wprintf (&CWindow, Line ++, 2, " {+1Flag byte        }N NONE");
                  if (DirP->Flags.ParityByteUsed)
                    if (DirP->ParityByte == DirP->CalculatedParity)
                      Wprintf (&CWindow, Line ++, 2, " {+1Parity byte      }N   %02X", DirP->ParityByte);
                    else
                      Wprintf (&CWindow, Line ++, 2, " {+4Parity byte         %02X (should be %02X)}N",
                               DirP->ParityByte, DirP->CalculatedParity);
                  else
                    Wprintf (&CWindow, Line ++, 2, " {+1Parity byte      }N NONE");
                  Wprintf (&CWindow, Line ++, 2, " {+1Pause length     }N%5u ms.",    DirP->PauseLength);
                  Wprintf (&CWindow, Line ++, 2, " {+1PILOT pulse      }N %4u T",      NS_LENPILOT);
                  Wprintf (&CWindow, Line ++, 2, " {+1PILOT length     }N %4u pulses",
                           DirP->Flags.IsHeader ? NS_HDPILOT : NS_DTPILOT);
                  Wprintf (&CWindow, Line ++, 2, " {+1SYNC pulse 1     }N %4u T",      NS_LENSYNC1);
                  Wprintf (&CWindow, Line ++, 2, " {+1SYNC pulse 2     }N %4u T",      NS_LENSYNC2);
                  Wprintf (&CWindow, Line ++, 2, " {+1ZERO bit pulse   }N %4u T",      NS_LEN0BIT);
                  Wprintf (&CWindow, Line ++, 2, " {+1ONE bit pulse    }N %4u T",      NS_LEN1BIT);
                  if (DirP->Timing.NumBitsInLastByte != 8)
                  {
                    Wprintf (&CWindow, Line ++, 2, " {+1Last byte length {+C %4u %s}N",   DirP->Timing.NumBitsInLastByte, TBits);
                    Wprintf (&CWindow, Line ++, 2, " {+1Last byte        {+C   %02X}N",   DirP->Timing.CustomLastByte);
                  }
                  else
                  {
                    WClearLine (&CWindow, Line ++);
                    WClearLine (&CWindow, Line ++);
                  }
                  break;
      case 0x11 : if (DirP->Flags.FlagByteUsed)
                    Wprintf (&CWindow, Line ++, 2, " {+1Flag byte        }N %4u", DirP->FlagByte);
                  else if (DirP->BlockLength > 0)
                    Wprintf (&CWindow, Line ++, 2, " {+1Flag byte        }N NONE (%3u)", DirP->FlagByte);
                  else
                    Wprintf (&CWindow, Line ++, 2, " {+1Flag byte        }N NONE");
                  if (DirP->Flags.ParityByteUsed)
                    if (DirP->ParityByte == DirP->CalculatedParity)
                      Wprintf (&CWindow, Line ++, 2, " {+1Parity byte      }N   %02X", DirP->ParityByte);
                    else
                      Wprintf (&CWindow, Line ++, 2, " {+4Parity byte         %02X (should be %02X)}N",
                               DirP->ParityByte, DirP->CalculatedParity);
                  else if (DirP->BlockLength >= 2)
                    Wprintf (&CWindow, Line ++, 2, " {+1Parity byte      }N NONE (%3u|%3u)",
                             DirP->ParityByte, DirP->CalculatedParity);
                  else
                    Wprintf (&CWindow, Line ++, 2, " {+1Parity byte      }N NONE");
                  Wprintf (&CWindow, Line ++, 2, " {+1Pause length     }N%5u ms.",     DirP->PauseLength);
                  Wprintf (&CWindow, Line ++, 2, " {+1PILOT pulse      }N %4u T",      DirP->Timing.LengthPilotPulse);
                  Wprintf (&CWindow, Line ++, 2, " {+1PILOT length     }N %4u pulses", DirP->Timing.NumPilotPulses);
                  if (DirP->Timing.NumSyncPulses == 0)
                    Wprintf (&CWindow, Line ++, 2, " {+1SYNC pulses      }N NONE");
                  else
                    Wprintf (&CWindow, Line ++, 2, " {+1SYNC pulse 1     }N %4u T",    DirP->Timing.LengthSyncPulse[0]);
                  if (DirP->Timing.NumSyncPulses > 1)
                    Wprintf (&CWindow, Line ++, 2, " {+1SYNC pulse 2     }N %4u T",    DirP->Timing.LengthSyncPulse[1]);
                  if (DirP->Timing.NumSyncPulses > 2)
                    Wprintf (&CWindow, Line ++, 2, " {+1SYNC pulse 3     }N %4u T",    DirP->Timing.LengthSyncPulse[2]);
                  Wprintf (&CWindow, Line ++, 2, " {+1ZERO bit pulse   }N %4u T",      DirP->Timing.LengthZeroBitPulse);
                  Wprintf (&CWindow, Line ++, 2, " {+1ONE bit pulse    }N %4u T",      DirP->Timing.LengthOneBitPulse);
                  if (DirP->Timing.NumBitsInLastByte != 8)
                  {
                    Wprintf (&CWindow, Line ++, 2, " {+1Last byte length }N %4u %s",   DirP->Timing.NumBitsInLastByte, TBits);
                    Wprintf (&CWindow, Line ++, 2, " {+1Last byte        }N   %02X",   DirP->Timing.CustomLastByte);
                  }
                  else
                  {
                    WClearLine (&CWindow, Line ++);
                    WClearLine (&CWindow, Line ++);
                  }
                  break;
      case 0x12 : Wprintf (&CWindow, Line ++, 2, " {+1Pulse length     }N %4u T", DirP->Timing.LengthPilotPulse);
                  Wprintf (&CWindow, Line ++, 2, " {+1Tone length      }N%5u pulses", DirP->Timing.NumPilotPulses);
                  break;
      case 0x13 : CopyFromXMS (DirP->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);
                  for (AddLines = 0 ; AddLines < DirP->BlockLength / 2 ; AddLines ++)
                  {
                    WScratch = *((word *)WorkBuffer + AddLines);                                            /* Type-cast to words */
                    sprintf (TEntry.Text, "%3d - %5u T", AddLines, WScratch);
                    if (!InsertComboEntry (&TWindow, &TEntry, NULL, -1))
                      FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
                  }
                  break;
      case 0x14 : Wprintf (&CWindow, Line ++, 2, " {+1Flag byte        }N NONE (%3u)", DirP->FlagByte);
                  Wprintf (&CWindow, Line ++, 2, " {+1Parity byte      }N NONE (%3u|%3u)",
                           DirP->ParityByte, DirP->CalculatedParity);
                  Wprintf (&CWindow, Line ++, 2, " {+1ZERO bit pulse   }N %4u T",    DirP->Timing.LengthZeroBitPulse);
                  Wprintf (&CWindow, Line ++, 2, " {+1ONE bit pulse    }N %4u T",    DirP->Timing.LengthOneBitPulse);
                  Wprintf (&CWindow, Line ++, 2, " {+1Last byte length }N %4u %s",   DirP->Timing.NumBitsInLastByte, TBits);
                  if (DirP->Timing.NumBitsInLastByte != 8)
                    Wprintf (&CWindow, Line ++, 2, " {+1Last byte        }N   %02X", DirP->Timing.CustomLastByte);
                  Wprintf (&CWindow, Line ++, 2, " {+1Pause length     }N%5u ms.",  DirP->PauseLength);
                  WClearLine (&CWindow, Line ++);
                  WClearLine (&CWindow, Line ++);
                  WClearLine (&CWindow, Line ++);
                  WClearLine (&CWindow, Line ++);
                  break;
      case 0x15 : Wprintf (&CWindow, Line ++, 2, " {+1Sample rate      }N %4u T (%5u Hz)",  DirP->Timing.SampleRate,
                           (word)(3500000L / DirP->Timing.SampleRate));
                  Wprintf (&CWindow, Line ++, 2, " {+1Pause length     }N%4u ms.",         DirP->PauseLength);
                  Wprintf (&CWindow, Line ++, 2, " {+1Last byte length }N %4u samples",     DirP->Timing.NumBitsInLastByte);
                  break;
      case 0x20 : if (DirP->PauseLength > 0)                                                                 /* Inifinite pause ? */
                    Wprintf (&CWindow, Line ++, 2, " {+1Pause length     }N%5u ms.", DirP->PauseLength);
                  else
                    Wprintf (&CWindow, Line ++, 2, " {+1Pause length     }N INFINITE");
                  break;
      case 0x21 : CopyFromXMS (DirP->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);
                  *(WorkBuffer + DirP->BlockLength) = '\0';
                  Wprintf (&CWindow, Line ++, 2, " {+1Name}N \"%s\"", WorkBuffer);
                  break;
      case 0x22 : break;
      case 0x23 : Wprintf (&CWindow, Line ++, 2, " {+1Jump to entry    }N %d", BlockNumber + (short)DirP->PauseLength);
                  break;
      case 0x24 : Wprintf (&CWindow, Line ++, 2, " {+1Repetitions      }N %d", DirP->PauseLength);
                  break;
      case 0x25 : break;
      case 0x26 : CopyFromXMS (DirP->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);
                  for (Cnt = 0, WTable = (short *)WorkBuffer ; Cnt < (DirP->BlockLength / 2) ; Cnt ++, WTable ++)
                  {
                    sprintf (TEntry.Text, "Call entry %3d", BlockNumber + *WTable);
                    if (!InsertComboEntry (&TWindow, &TEntry, NULL, -1))
                      FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
                  }
                  break;
      case 0x27 : break;
      case 0x28 : CopyFromXMS (DirP->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);
                  for (Cnt = 0, Cnt2 = 1 ; Cnt < *WorkBuffer ; Cnt ++)
                  {
                    Bt1 = *(WorkBuffer + Cnt2 + 2);                                               /* (Length of description text) */
                    sprintf (TEntry.Text, "%3d {+1%3d}N ", Cnt + 1, BlockNumber + *((short *)(WorkBuffer + Cnt2)));
                    strncpy (TEntry.Text + 13, WorkBuffer + Cnt2 + 3, Bt1);
                    TEntry.Text[Bt1 + 13] = '\0';
                    if (!InsertComboEntry (&TWindow, &TEntry, NULL, -1))
                      FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
                    Cnt2 += (word)Bt1 + 3;
                  }
                  break;
      case 0x2A : break;
      case 0x30 : CopyFromXMS (DirP->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);
                  for (Cnt = 0, Cnt2 = 256 ; Cnt < DirP->BlockLength ; Cnt ++, Cnt2 ++)
                  {
                    *(WorkBuffer + Cnt2) = *(WorkBuffer + Cnt);
                    if (*(WorkBuffer + Cnt) == '{' || *(WorkBuffer + Cnt) == '}')                    /* Dismap special characters */
                      *(WorkBuffer + (++ Cnt2)) = *(WorkBuffer + Cnt);
                  }
                  *(WorkBuffer + Cnt2) = '\0';
                  if (!FitComboString (&TWindow, NULL, WorkBuffer + 256, 38, 0, 0, FALSE))
                    FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
                  break;
      case 0x31 : CopyFromXMS (DirP->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);
                  if (DirP->PauseLength == 0)
                    sprintf (TEntry.Text, " {+1Display time}N INFINITE");
                  else
                    sprintf (TEntry.Text, " {+1Display time}N %u seconds", DirP->PauseLength);
                  if (!InsertComboEntry (&TWindow, &TEntry, NULL, -1))
                    FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
                  TEntry.Text[0] = '\0';
                  if (!InsertComboEntry (&TWindow, &TEntry, NULL, -1))
                    FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
                  strcpy (TEntry.Text, " {+1Message}N");
                  if (!InsertComboEntry (&TWindow, &TEntry, NULL, -1))
                    FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
                  for (Cnt = 0, Cnt2 = 256 ; Cnt < DirP->BlockLength ; Cnt ++, Cnt2 ++)
                    if (*(WorkBuffer + Cnt) != 0x0A)                                                /* Remove linefeed characters */
                    {
                      if (*(WorkBuffer + Cnt) == 0x0D)                                   /* Convert carriages return to linefeeds */
                        *(WorkBuffer + Cnt2) = 0x0A;
                      else
                        *(WorkBuffer + Cnt2) = *(WorkBuffer + Cnt);
                      if (*(WorkBuffer + Cnt) == '{' || *(WorkBuffer + Cnt) == '}')                  /* Dismap special characters */
                        *(WorkBuffer + (++ Cnt2)) = *(WorkBuffer + Cnt);
                    }
                  *(WorkBuffer + Cnt2) = '\0';
                  if (!FitComboString (&TWindow, NULL, WorkBuffer + 256, 30, 0, 0, TRUE))
                    FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
                  break;
      case 0x32 : CopyFromXMS (DirP->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);
                  for (Cnt = 0, Cnt2 = 1 ; Cnt < *WorkBuffer ; Cnt ++)                                     /* Do all text strings */
                  {
                    switch (*(WorkBuffer + (Cnt2 ++)))                                                   /* (Use 'Cnt2' as index) */
                    {
                      case 0x00 : strcpy (TEntry.Text, "{+1Full Title}N"); break;
                      case 0x01 : strcpy (TEntry.Text, "{+1Software House / Publisher}N"); break;
                      case 0x02 : strcpy (TEntry.Text, "{+1Author(s)}N"); break;
                      case 0x03 : strcpy (TEntry.Text, "{+1Year of Publication}N"); break;
                      case 0x04 : strcpy (TEntry.Text, "{+1Language}N"); break;
                      case 0x05 : strcpy (TEntry.Text, "{+1Game / Utility Type}N"); break;
                      case 0x06 : strcpy (TEntry.Text, "{+1Price}N"); break;
                      case 0x07 : strcpy (TEntry.Text, "{+1Protection Scheme / Loader}N"); break;
                      case 0x08 : strcpy (TEntry.Text, "{+1Origin}N"); break;
                      default   : strcpy (TEntry.Text, "{+1Comments}N"); break;
                    }
                    if (!InsertComboEntry (&TWindow, &TEntry, NULL, -1))
                      FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
                    for (Cnt3 = 0, Cnt4 = 0 ; Cnt3 < *(WorkBuffer + Cnt2) ; Cnt3 ++, Cnt4 ++)
                    {
                      if (*(WorkBuffer + Cnt2 + Cnt3 + 1) == 0x0D)                       /* Convert carriage returns to linefeeds */
                        *(TextBuffer + Cnt4) = 0x0A;
                      else
                        *(TextBuffer + Cnt4) = *(WorkBuffer + Cnt2 + Cnt3 + 1);
                      if (*(WorkBuffer + Cnt2 + Cnt3 + 1) == '{' ||
                          *(WorkBuffer + Cnt2 + Cnt3 + 1) == '}')                                    /* Dismap special characters */
                        *(TextBuffer + (++ Cnt4)) = *(WorkBuffer + Cnt2 + Cnt3 + 1);
                    }
                    *(TextBuffer + Cnt4) = '\0';
                    if (!FitComboString (&TWindow, NULL, TextBuffer, 38, 0, 0, FALSE))
                      FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
                    Cnt2 += Cnt3 + 1;                                                        /* Cnt2 now points after this string */
                  }
                  break;
      case 0x33 : HWIndex = WorkBuffer + (dword)49152 - HWTableLength;                           /* Use the end of the WorkBuffer */
                  CopyFromXMS (HWTableLength, HWTableXMSHandle, 0, HWIndex);                          /* Read hardware type table */
                  DTable = (dword *)WorkBuffer;
                  GetHardwareTable (WhichWindow, DirP);                                                        /* Fetch tape data */
                  NumHWTypes = (byte)(DirP->BlockLength / 3);
                  LastValue = LastType = 0xFF;
                  if (NumHWTypes > 0)
                    for (Cnt = 0 ; Cnt < NumHWTypes ; Cnt ++)                                                     /* Do all types */
                    {
                      Bt1 = (byte)((*(DTable + Cnt) & 0x0000FF00) >> 8);                                       /* (Hardware Type) */
                      Bt2 = (byte)(*(DTable + Cnt) & 0x000000FF);                                                /* (Hardware ID) */
                      Bt3 = (byte)((*(DTable + Cnt) & 0x00FF0000) >> 16);                                              /* (Value) */
                      if (LastValue == 0xFF || Bt3 != LastValue)
                      {
                        LastValue = Bt3;
                        LastType = 0xFF;                                                     /* Force hardware type to be printed */
                        switch (Bt3)
                        {
                          case 0 : strcpy (TEntry.Text, "{BRuns & could use  }B"); break;
                          case 1 : strcpy (TEntry.Text, "{BRuns & uses       }B"); break;
                          case 2 : strcpy (TEntry.Text, "{BRuns & doesn\'t use}B"); break;
                          case 3 : strcpy (TEntry.Text, "{BDoesn\'t run on    }B"); break;
                          default: strcpy (TEntry.Text, "{B<UNDEFINED>       }B"); break;
                        }
                        if (!InsertComboEntry (&TWindow, &TEntry, NULL, -1))
                          FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
                      }
                      if (Bt1 < *HWIndex)
                      {
                        if (LastType == 0xFF || Bt1 != LastType)
                        {
                          LastType = Bt1;
                          sprintf (TEntry.Text, " {+1%s}N", HWIndex + 4 + (dword)*(HWIndex + 2) * Bt1);
                          if (!InsertComboEntry (&TWindow, &TEntry, NULL, -1))
                            FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
                        }
                        HWName = HWIndex + 4 + ((dword)*(HWIndex + 2) * *HWIndex) +
                                               ((dword)*(HWIndex + 2) * *(HWIndex + 1) * Bt1) +
                                               ((dword)*(HWIndex + 2) * Bt2);
                        if (*HWName == '\0')                                                                   /* Not filled in ? */
                          strcpy (TEntry.Text, "  <Undefined>");
                        else
                          sprintf (TEntry.Text, "  %s", HWName);
                      }
                      else
                        strcat (TEntry.Text, "<Undefined>)}N");
                      if (!InsertComboEntry (&TWindow, &TEntry, NULL, -1))
                        FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
                    }
                  break;
      case 0x34 : CopyFromXMS (DirP->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);     /* (8 bytes) */
                  Wprintf (&CWindow, Line ++, 2, " {+1R register emul  }N %s", *WorkBuffer & 0x01 ? "On" : "Off");
                  Wprintf (&CWindow, Line ++, 2, " {+1LDIR emul        }N %s", *WorkBuffer & 0x02 ? "On" : "Off");
                  Wprintf (&CWindow, Line ++, 2, " {+1High res color   }N %s", *WorkBuffer & 0x04 ? "On" : "Off");
                  if ((*WorkBuffer & 0x18) == 0x08)       Wprintf (&CWindow, Line ++, 2, " {+1Video sync       }N High");
                  else if ((*WorkBuffer & 0x18) == 0x18)  Wprintf (&CWindow, Line ++, 2, " {+1Video sync       }N Low");
                  else                                    Wprintf (&CWindow, Line ++, 2, " {+1Video sync       }N Normal");
                  Wprintf (&CWindow, Line ++, 2, " {+1Fast ROM loading }N %s", *WorkBuffer & 0x20 ? "On" : "Off");
                  Wprintf (&CWindow, Line ++, 2, " {+1Border emulation }N %s", *WorkBuffer & 0x40 ? "On" : "Off");
                  Wprintf (&CWindow, Line ++, 2, " {+1Screen refresh   }N %s", *WorkBuffer & 0x80 ? "On" : "Off");
                  Wprintf (&CWindow, Line ++, 2, " {+1Refresh delay    }N %u int", *(WorkBuffer + 2));
                  Wprintf (&CWindow, Line ++, 2, " {+1Start tape immed.}N %s", *(WorkBuffer + 1) & 0x01 ? "On" : "Off");
                  Wprintf (&CWindow, Line ++, 2, " {+1Autoload 128K    }N %s", *(WorkBuffer + 1) & 0x02 ? "On" : "Off");
                  Wprintf (&CWindow, Line ++, 2, " {+1Interrupt freq   }N %u", (word)*(WorkBuffer + 3) + 256 * *(WorkBuffer + 4));
                  break;
      case 0x35 : CopyFromXMS (20, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);               /* Fetch Block ID */
                  Bt1 = *(WorkBuffer + 0x0010);                                                                  /* (Keep a copy) */
                  *(WorkBuffer + 0x0010) = '\0';                                                              /* Make it a string */
                  sprintf (TEntry.Text, "{+1ID String        }N \"%s\"", WorkBuffer);
                  if (!InsertComboEntry (&TWindow, &TEntry, NULL, -1))
                    FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
                  if (DirP->FlagByte == INSRTCI_POKES ||                                            /* Standardized POKEs block ? */
                      DirP->FlagByte == INSRTCI_ZXEDIT ||                                /* Standardized ZX-Edit Document block ? */
                      DirP->FlagByte == INSRTCI_PICTURE)                                          /* Standardized Picture block ? */
                  {
                    TEntry.Text[0] = '\0';
                    if (!InsertComboEntry (&TWindow, &TEntry, NULL, -1))
                      FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
                    strcpy (TEntry.Text, "{+1General Description}N");
                    if (!InsertComboEntry (&TWindow, &TEntry, NULL, -1))
                      FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
                    Cnt2 = 0x0011;                                                        /* Set start of the General Description */
                    if (DirP->FlagByte == INSRTCI_PICTURE)
                    {                                         /* The Picture block's General Description is further in the header */
                      Bt1 = *(WorkBuffer + 0x0011);
                      Cnt2 = 0x0012;
                    }
                    if (Bt1 > 0)                                                            /* General description has a length ? */
                    {
                      if (DirP->AllocLength > 280)
                        CopyFromXMS (280, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);
                      else
                        CopyFromXMS (DirP->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);
                      *(WorkBuffer + Bt1 + Cnt2) = '\0';                                           /* Prepare General Description */
                      for (Cnt = 0 ; Cnt < Bt1 ; Cnt ++)
                        if (*(WorkBuffer + Cnt2 + Cnt) == 0x0D)                          /* Convert carriages return to linefeeds */
                          *(WorkBuffer + Cnt2 + Cnt) = 0x0A;
                      if (!FitComboString (&TWindow, NULL, WorkBuffer + Cnt2, 30, 0, 0, FALSE))      /* Insert at 30-column width */
                        FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
                    }
                    else if (DirP->FlagByte != INSRTCI_POKES)
                    {
                      switch (DirP->FlagByte)
                      {
                        case INSRTCI_SCREEN  : strcpy (TEntry.Text, "Loading Screen"); break;
                        case INSRTCI_ZXEDIT  : strcpy (TEntry.Text, "Instructions"); break;
                        case INSRTCI_PICTURE : strcpy (TEntry.Text, "Inlay Card"); break;
                      }
                      if (!InsertComboEntry (&TWindow, &TEntry, NULL, -1))
                        FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
                    }
                  }
                  if (DirP->FlagByte == INSRTCI_ZXEDIT || DirP->FlagByte == INSRTCI_PICTURE)
                  {
                    TEntry.Text[0] = '\0';
                    if (!InsertComboEntry (&TWindow, &TEntry, NULL, -1))
                      FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
                    strcpy (TEntry.Text, "{+1<Use the Edit button to view this block>}N");
                    if (!InsertComboEntry (&TWindow, &TEntry, NULL, -1))
                      FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
                  }
                  break;
      case 0x40 : CopyFromXMS (1, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);         /* (Info part is 1 byte) */
                  Wprintf (&CWindow, Line ++, 2, " {+1Snapshot type    }N %s", *WorkBuffer ? "SNA" : "Z80");
                  Wprintf (&CWindow, Line ++, 2, " {+1Snapshot length  }N %lu", DirP->BlockLength - 8);
                  break;
      default   : break;
    }
    if (NumInfoLines < 0)
      PrintComboEntries (&TWindow);
    Stop = FALSE;
    while (!Stop)
    {
      switch (HandleWindow (&CWindow, &ButtonP))
      {
        case WINH_RNOHAND : Stop = TRUE;
                            if (ButtonP->ReactNum == 2)                                                                 /* Edit ? */
                            {
                              if (BlockType[GetBlockTypeIndex (DirP->BlockTypeID)].EditHandler (WhichWindow, DirP))     /* Edit ! */
                              {
                                SignalTapeChanged (WhichWindow, TRUE);
                                EntryChanged = TRUE;
                              }
                            }
                            else                                                                                          /* (Ok) */
                              Again = FALSE;
                            break;
        case WINH_CLOSE   : Stop = TRUE; Again = FALSE; break;
      }
      if (NumInfoLines < 0)
        HandleComboWindow (&TWindow, &AddLines);
      MouseStatus ();
      if (GlobalKey == K_ESC)
      { Stop = TRUE; Again = FALSE; }
    }
    if (NumInfoLines < 0)
      DestroyComboWindow (&TWindow, 0);
  }
  DestroyWindow (&CWindow);
  if (EntryChanged)                                                                                /* The entry has been edited ? */
    SetupNewData (WhichWindow);                                                                     /* Rebuild the tape directory */
  GlobalKey = 0x0000;
}

bool GetBlockTypeInfo (byte BlockTypeID, bool *IsTapeData, bool *IsIdentifiedData, char **FullName, char **TapeWindowName,
                       short *NumInfoLines)

/**********************************************************************************************************************************/
/* Pre   : `BlockTypeID' holds the ID to be tested; any of the parameters may be NULL.                                            */
/* Post  : TRUE is returned if this ID is identified and named.                                                                   */
/*         `IsTapeData', `IsIdentifiedData, `FullName', `TapeWindowName' and `NumInfoLines' are set to point to the appropriate   */
/*         values in the named fields, unless the provided pointer was NULL.                                                      */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  short register Cnt;

  for (Cnt = 0 ; Cnt < NumBlockTypes ; Cnt ++)
    if (BlockType[Cnt].TypeID == BlockTypeID)
    {
      if (IsTapeData != NULL)        *IsTapeData = BlockType[Cnt].IsTapeData;
      if (IsIdentifiedData != NULL)  *IsIdentifiedData = BlockType[Cnt].IsIdentifiedData;
      if (FullName != NULL)          *FullName = BlockType[Cnt].FullName;
      if (TapeWindowName != NULL)    *TapeWindowName = BlockType[Cnt].TapeWindowName;
      if (NumInfoLines != NULL)      *NumInfoLines = BlockType[Cnt].NumInfoLines;
      return (TRUE);
    }
  if (IsTapeData != NULL)        *IsTapeData = BlockType[Cnt].IsTapeData;
  if (IsIdentifiedData != NULL)  *IsIdentifiedData = BlockType[Cnt].IsIdentifiedData;
  if (FullName != NULL)          *FullName = BlockType[Cnt].FullName;
  if (TapeWindowName != NULL)    *TapeWindowName = BlockType[Cnt].TapeWindowName;
  if (NumInfoLines != NULL)      *NumInfoLines = BlockType[Cnt].NumInfoLines;
  return (FALSE);
}

short GetBlockTypeIndex (byte BlockTypeID)

/**********************************************************************************************************************************/
/* Pre   : `BlockTypeID' holds the ID to be found.                                                                                */
/* Post  : The index into the BlockType named structure has been returned.                                                        */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  short register Cnt;

  for (Cnt = 0 ; Cnt < NumBlockTypes ; Cnt ++)
    if (BlockType[Cnt].TypeID == BlockTypeID)
      return (Cnt);
  return (-1);
}

void ExpandCollapse (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : The user clicked on a Group Start block.                                                                               */
/* Post  : The Group body is expand/collapse state is toggled if the mouse was on the 'page' character.                           */
/* Import: SetupNewData.                                                                                                          */
/**********************************************************************************************************************************/

{
  if (Entry->BlockTypeID != 0x21 && Entry->BlockTypeID != 0x24)                     /* Only applicable to Group/Loop Start blocks */
    return;
  if (Mouse.Xx < TapeWindow[WhichWindow].IndexWindow.ComboWindow.StartX + 24 ||             /* (Is the 5th character in the line) */
      Mouse.Xx >= TapeWindow[WhichWindow].IndexWindow.ComboWindow.StartX + 30)
    return;
  if (Entry->ComboSelected & EXPANDBIT)                                                               /* Was currently expanded ? */
    Entry->ComboSelected &= ~EXPANDBIT;                                                                /* Then collapse the group */
  else                                                                                                      /* (Nope, expand now) */
    Entry->ComboSelected |= EXPANDBIT;
  SetupNewData (WhichWindow);                                                                       /* Rebuild the tape directory */
}

void DecodeDRB (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the directory entry of the Direct Recording Block.        */
/* Post  : The block has been decoded and appended to the same tape.                                                              */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  dword Index = 0;
  dword XMSIndex;
  bool  PRCancel = FALSE;

  DecodeDirectRecording = TRUE;
  if (!SetupVOCReadTimers (FT_UNKNOWN))                                            /* Allow the users to select the timing values */
    return;
  MakePRequester (&_PReq, -1, -1, TRUE, NULL, "Decoding Direct Recording", "");
  XMSIndex = GetTapeBlockXMSIndex (WhichWindow, Entry);
  _PRFileLength = (Entry->BlockLength - 1) * 8 + Entry->Timing.NumBitsInLastByte;                 /* Sst number of samples (bits) */
  DecodeDirectSample (WhichWindow, -1, (word)(3500000L / Entry->Timing.SampleRate), _PRFileLength, &Index, FALSE, &PRCancel,
                      XMSIndex);
  CleanupTapeBuffers ();                                                                          /* Release any decoding buffers */
  _DestroyPReq ();
  SetupNewData (WhichWindow);                                                                       /* Rebuild the tape directory */
  SignalTapeChanged (WhichWindow, TRUE);
}

void TmpExpandTape (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/* Post  : All Group blocks are expanded. The current expansion state is saved and must be restored with `TmpRestoreTape'.        */
/* Import: SetupNewData.                                                                                                          */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *DirP;
  bool                Changed = FALSE;

  DirP = TapeWindow[WhichWindow].Directory;
  while (DirP != NULL)
  {
    if ((DirP->BlockTypeID == 0x21 || DirP->BlockTypeID == 0x24) && (DirP->ComboSelected & EXPANDBIT))       /* Collapsed group ? */
    {
      DirP->ComboSelected = (DirP->ComboSelected & ~EXPANDBIT) | BUEXPANDBIT;                  /* Remember and drop collapse flag */
      Changed = TRUE;
    }
    DirP = DirP->Next;
  }
  if (Changed)
    SetupNewData (WhichWindow);                                                                     /* Rebuild the tape directory */
}

void TmpRestoreTape (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/* Post  : The tape was fully expanded with a `TmpExpandTape' call. The previous expand/collapse state has now been restored.     */
/* Import: SetupNewData.                                                                                                          */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *DirP;
  bool                Changed = FALSE;

  DirP = TapeWindow[WhichWindow].Directory;
  while (DirP != NULL)
  {
    if ((DirP->BlockTypeID == 0x21 || DirP->BlockTypeID == 0x24) && (DirP->ComboSelected & BUEXPANDBIT)) /* Temp expanded group ? */
    {
      DirP->ComboSelected = (DirP->ComboSelected & ~BUEXPANDBIT) | EXPANDBIT;                          /* Restore collapsed state */
      Changed = TRUE;
    }
    DirP = DirP->Next;
  }
  if (Changed)
    SetupNewData (WhichWindow);                                                                     /* Rebuild the tape directory */
}

void ClearTapeLoopDetection (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/* Post  : The flags to detect an infinite tape write loop have been reset.                                                       */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *DirP;

  DirP = TapeWindow[WhichWindow].Directory;
  while (DirP != NULL)
  {
    DirP->Flags.IsWritten = FALSE;
    DirP = DirP->Next;
  }
  _InfiniteTapeLoop = FALSE;
}
